/*  kwaterfall.c

    Main file for  kwaterfall  (X11 waterfall display tool for Karma).

    Copyright (C) 1998-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will display a running waterfall.
    This module runs on an X11 server.


    Written by      Richard Gooch   6-JUL-1998: Copied from kview/main_xt.c

    Updated by      Richard Gooch   10-JUL-1998: Changed from K_FLOAT to
  K_UBYTE. Added clever scrolling by fiddling vertical offset array.

    Updated by      Richard Gooch   21-MAR-2003: Added #include <k_version.h>

    Last updated by Richard Gooch   11-APR-2004: Added -slow option.


*/
#include <stdio.h>
#include <stdlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <karma.h>
#include <k_version.h>
#define NEW_WIN_SCALE
#define X11
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_im.h>
#include <karma_ds.h>
#include <karma_ic.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_n.h>
#include <Xkw/ImageDisplay.h>

#define VERSION          "1.0.3"
#define SCROLL_INTERVAL  10  /*  milliseconds  */
#define IMAGE_SIZE      512
#define MAX_VALUE       100.0


/*  Private functions  */
STATIC_FUNCTION (void timer_cbk_bulk_copy,
		 (XtPointer client_data, XtIntervalId *id) );
STATIC_FUNCTION (void timer_cbk_fiddle_offsets,
		 (XtPointer client_data, XtIntervalId *id) );


/*  Private data  */
String fallback_resources[] =
{
    "Kwaterfall*pseudoColourCanvas*background:              black",
    "Kwaterfall*trueColourCanvas*background:                black",
    "Kwaterfall*pseudoColourStereoCanvas*background:        black",
    "Kwaterfall*trueColourStereoCanvas*background:          black",
    "Kwaterfall*pseudoColourCanvas*foreground:              white",
    "Kwaterfall*Command*background:                         grey70",
    "Kwaterfall*Repeater*background:                        grey70",
    "Kwaterfall*Ktoggle*background:                         grey80",
    "Kwaterfall*closeButton*background:                     grey90",
    "Kwaterfall*ChoiceMenu.background:                      turquoise",
    "Kwaterfall*ExclusiveMenu.background:                   turquoise",
    "Kwaterfall*Value*background:                           #d0a0a0",
    "Kwaterfall*ImageDisplay*quit*background:               orange",
    "Kwaterfall*SimpleSlider.foreground:                    Sea Green",
    "Kwaterfall*ImageDisplay*intensityMenu*font:            10x20",
    "Kwaterfall*ImageDisplay*zoomMenu*font:                 10x20",
    "Kwaterfall*ImageDisplay*controlMenu*font:              10x20",
    "Kwaterfall*ImageDisplay*theMenu*font:                  10x20",
    "Kwaterfall*ImageDisplay.Command.font:                  7x13bold",
    "Kwaterfall*ImageDisplay.Ktoggle.font:                  7x13bold",
    "Kwaterfall*ImageDisplay.ChoiceMenu.font:               7x13bold",
    "Kwaterfall*ImageDisplay.exportMenu.menuButton.font:    7x13bold",
    "Kwaterfall*ImageDisplay.overlayMenu.menuButton.font:   7x13bold",
    "Kwaterfall*ImageDisplay*zoomMenu*Unzoom*foreground:    red",
    "Kwaterfall*SimpleSlider.borderWidth:                   0",
    "Kwaterfall*IntensityPolicy*showIntensityReset:         True",
    "Kwaterfall*IntensityPolicy*resetIntensityToggle.state: True",
    "Kwaterfall*IntensityPolicy*autoIntensityScale:         False",
    "Kwaterfall*ZoomPolicy*showMiddleZoom:                  True",
    "Kwaterfall*ZoomPolicy*middleZoomToggle.state:          True",
    "Kwaterfall*ZoomPolicy*showRightZoom:                   True",
    "Kwaterfall*ZoomPolicy*rightZoomToggle.state:           True",
    "Kwaterfall*font:                                       9x15bold",
    "Kwaterfall*borderColor:                                black",
    "Kwaterfall*background:                                 aquamarine",
    "Kwaterfall*foreground:                                 black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-cmap_master", ".topForm.cmapMaster", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-create_at_init", "*createAtInit", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static XtAppContext app_context = NULL;


int main (int argc, char **argv)
{
    KWorldCanvas wc, wc_true;
    flag controlled;
    Widget main_shell, image_display;
    char title_name[STRING_LENGTH];
    char hostname[STRING_LENGTH];
    void (*timer_cbk) (XtPointer client_data, XtIntervalId *id);
    extern char module_name[STRING_LENGTH + 1];
    extern char module_version_date[STRING_LENGTH + 1];
    extern char karma_library_version[STRING_LENGTH + 1];
    /*static char function_name[] = "main";*/

    /*  Initialise module  */
    im_register_module_name ("kwaterfall");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kwaterfall",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 100,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_CONTROLLED,
					     &controlled,
					     XTMISC_INIT_ATT_END,
					     NULL);
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    r_gethostname (hostname, STRING_LENGTH);
    sprintf (title_name, "%s v%s/Karma v%s @%s",
	     module_name, module_version_date, karma_library_version,
	     hostname);
    XtVaSetValues (main_shell,
		   XtNtitle, title_name,
		   NULL);
    image_display = XtVaCreateManagedWidget ("topForm",
					     imageDisplayWidgetClass,
					     main_shell,
					     XtNborderWidth, 0,
					     XkwNnumDatasets, 0,
					     XkwNcreateFilepopup, FALSE,
					     NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (image_display,
		   XkwNpseudoColourCanvas, &wc,
		   XkwNtrueColourCanvas, &wc_true,
		   NULL);
    if (wc == NULL) wc = wc_true;
    if (wc == NULL)
    {
	fprintf (stderr, "No PseudoColour or TrueColour canvas found\n");
	exit (RV_UNDEF_ERROR);
    }
    timer_cbk = timer_cbk_fiddle_offsets;
    for (--argc, ++argv; argc > 0; --argc, ++argv)
    {
	if (strcmp (*argv, "-slow") == 0) timer_cbk = timer_cbk_bulk_copy;
    }
    if (timer_cbk == timer_cbk_bulk_copy)
	fprintf (stderr, "Scrolling using bulk copy\n");
    else if (timer_cbk == timer_cbk_fiddle_offsets)
	fprintf (stderr, "Scrolling using offset indices manipulation\n");
    XtAppAddTimeOut (app_context, SCROLL_INTERVAL, timer_cbk, wc);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main  */

static void timer_cbk_bulk_copy (XtPointer client_data, XtIntervalId *id)
/*  This is the interval timer callback.
*/
{
    KWorldCanvas canvas = client_data;
    unsigned int x, y;
    static iarray array = NULL;
    static ViewableImage vimage = NULL;
    static char function_name[] = "timer_cbk";

    if (array == NULL)
    {
	/*  First time  */
	if ( ( array = iarray_create_2D (IMAGE_SIZE, IMAGE_SIZE, K_UBYTE) )
	     == NULL ) m_abort (function_name, "array");
	/*  Fill the array  */
	for (y = 0; y < IMAGE_SIZE; ++y) for (x = 0; x < IMAGE_SIZE; ++x)
	    UB2 (array, y, x) = n_uniform () * MAX_VALUE;
	if ( ( vimage = viewimg_create_from_iarray (canvas, array, FALSE) )
	     == NULL ) m_abort (function_name, "viewable image");
	viewimg_set_attributes (vimage,
				VIEWIMG_VATT_VALUE_MIN, 0.0,
				VIEWIMG_VATT_VALUE_MAX, MAX_VALUE,
				VIEWIMG_VATT_END);
	viewimg_make_active (vimage);
    }
    else
    {
	/*  Scroll  */
	unsigned char line_buffer[IMAGE_SIZE];

	for (x = 0; x < IMAGE_SIZE; ++x) line_buffer[x] = UB2 (array, 0, x);
	for (y = 1; y < IMAGE_SIZE; ++y) for (x = 0; x < IMAGE_SIZE; ++x)
	    UB2 (array, y - 1, x) = UB2 (array, y, x);
	/*  Copy saved line to the top  */
	for (x = 0; x < IMAGE_SIZE; ++x)
	    UB2 (array, IMAGE_SIZE - 1, x) = line_buffer[x];
	iarray_register_data_change (array, NULL);
    }
    XtAppAddTimeOut (app_context, SCROLL_INTERVAL, timer_cbk_bulk_copy,
		     canvas);
}   /*  End Function timer_cbk_bulk_copy  */

static void timer_cbk_fiddle_offsets (XtPointer client_data, XtIntervalId *id)
/*  This is the interval timer callback.
*/
{
    KWorldCanvas canvas = client_data;
    unsigned int x, y;
    static array_desc *arr_desc = NULL;
    static ViewableImage vimage = NULL;
    static char *elem_name = "Data Value";
    static char function_name[] = "timer_cbk_fiddle_offsets";

    if (arr_desc == NULL)
    {
	/*  First time  */
	unsigned int type = K_UBYTE;
	unsigned char *ub_ptr;
	uaddr lengths[2];
	static array_pointer arrayp;

	lengths[0] = IMAGE_SIZE;
	lengths[1] = IMAGE_SIZE;
	if ( ( arr_desc =
	       ds_easy_alloc_array_desc (2, lengths, NULL, NULL, NULL, NULL,
					 1, &type,
					 (CONST char **) &elem_name) )
	     == NULL )
	    m_abort (function_name, "array descriptor");
	if ( !ds_compute_array_offsets (arr_desc) )
	    m_abort (function_name, "array offsets");
	if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, TRUE) )
	    m_abort (function_name, "array");
	/*  Fill the array  */
	for (y = 0; y < IMAGE_SIZE; ++y)
	{
	    ub_ptr = (unsigned char *) arrayp.array + arr_desc->offsets[0][y];
	    for (x = 0; x < IMAGE_SIZE; ++x)
		ub_ptr[ arr_desc->offsets[1][x] ] = n_uniform () * MAX_VALUE;
	}
	if ( ( vimage = viewimg_create (canvas, NULL, arr_desc, arrayp.array,
					1, 0, 0) )
	     == NULL ) m_abort (function_name, "viewable image");
	viewimg_set_attributes (vimage,
				VIEWIMG_VATT_VALUE_MIN, 0.0,
				VIEWIMG_VATT_VALUE_MAX, MAX_VALUE,
				VIEWIMG_VATT_END);
	viewimg_make_active (vimage);
    }
    else
    {
	/*  Scroll: be clever and just scroll the vertical offset array. On my
	    dual PPro 180 this makes the per-frame scroll and display time take
	    0.75 times the time for bulk scrolling  */
	uaddr offset;

	offset = arr_desc->offsets[0][0];
	m_copy (arr_desc->offsets[0], arr_desc->offsets[0] + 1,
		(IMAGE_SIZE - 1) * sizeof *arr_desc->offsets[0]);
	arr_desc->offsets[0][IMAGE_SIZE - 1] = offset;
	viewimg_register_data_change (vimage);
    }
    XtAppAddTimeOut (app_context, SCROLL_INTERVAL, timer_cbk_fiddle_offsets,
		     canvas);
}   /*  End Function timer_cbk_fiddle_offsets  */
