/*  generate_summary.c

    Source file for  generate_summary  (generate summary of packages).

    Copyright (C) 1997-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma utility programme will generate a summary of packages.


    Written by      Richard Gooch   2-APR-1997

    Updated by      Richard Gooch   7-APR-1997: Generate README.

    Updated by      Richard Gooch   11-MAY-1997: Functions directory is now
  optional: README is simply not updated.

    Last updated by Richard Gooch   20-AUG-2002: Use fixed-space font for
  package names in HTML output.


*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <karma.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_st.h>


#define BUF_SIZE 1048576


/*  Private functions  */
STATIC_FUNCTION (flag process_package,
		 (Channel readme_ch, Channel WWW_ch, CONST char *package_name,
		  Channel libline_ch, Channel func_ch, Channel memory_ch) );


int main (int argc, char **argv)
{
    Channel readme_ch, WWW_ch, func_ch, libline_ch, memory_ch;
    char *libline_dir, *func_dir, *readme_filename, *WWW_filename;
    char libline_filename[STRING_LENGTH], func_filename[STRING_LENGTH];
    static char usage[] =
      "generate_summary libline_dir func_dir readme_file WWW_file packages...";

    if (argc < 6)
    {
	fprintf (stderr, "%s\n", usage);
	exit (RV_MISSING_PARAM);
    }
    libline_dir = argv[1];
    func_dir = argv[2];
    readme_filename = argv[3];
    WWW_filename = argv[4];
    argc -= 5;
    argv += 5;
    if ( ( memory_ch = ch_open_memory (NULL, BUF_SIZE) ) == NULL )
    {
	exit (RV_CANNOT_OPEN);
    }
    if (strcmp (func_dir, "-") != 0)
    {
	if ( ( readme_ch = ch_open_file (readme_filename, "a+") ) == NULL )
	{
	    fprintf (stderr, "Error opening: %s\t%s\n",
		     readme_filename, ERRSTRING);
	    exit (RV_CANNOT_OPEN);
	}
    }
    else readme_ch = NULL;
    if ( ( WWW_ch = ch_open_file (WWW_filename, "a+") ) == NULL )
    {
	fprintf (stderr, "Error opening: %s\t%s\n",
		 WWW_filename, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    ch_printf (WWW_ch, "<table>\n");
    for (; argc > 0; --argc, ++argv)
    {
	sprintf (libline_filename, "%s/%s", libline_dir, *argv);
	if ( ( libline_ch = ch_open_file (libline_filename, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening: %s\t%s\n",
		     libline_filename, ERRSTRING);
	    exit (RV_CANNOT_OPEN);
	}
	if (strcmp (func_dir, "-") != 0)
	{
	    sprintf (func_filename, "%s/%s", func_dir, *argv);
	    if ( ( func_ch = ch_open_file (func_filename, "r") ) == NULL )
	    {
		fprintf (stderr, "Error opening: %s\t%s\n",
			 func_filename, ERRSTRING);
		exit (RV_CANNOT_OPEN);
	    }
	}
	else func_ch = NULL;
	if ( !process_package (readme_ch, WWW_ch, *argv, libline_ch, func_ch,
			       memory_ch) )
	{
	    fprintf (stderr, "Error processing package: \"%s\"\n", *argv);
	    exit (RV_READ_ERROR);
	}
	ch_close (libline_ch);
	if (func_ch != NULL) ch_close (func_ch);
    }
    if (readme_ch != NULL) ch_close (readme_ch);
    ch_printf (WWW_ch, "</table>\n");
    ch_close (WWW_ch);
    exit (RV_OK);
}   /*  End Function main  */


/*  Private functions follow  */

static flag process_package (Channel readme_ch, Channel WWW_ch,
			     CONST char *package_name,
			     Channel libline_ch, Channel func_ch,
			     Channel memory_ch)
/*  [SUMMARY] Process a package.
    <readme_ch> The channel to which README information is written.
    <WWW_ch> The channel to which WWW information is written.
    <package_name> The name of the package.
    <libline_ch> The channel from which liblines are read.
    <func_ch> The channel from which functions are read.
    <memory_ch> A temporary channel.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int num_funcs, count;
    char txt[STRING_LENGTH];

    /*  Read first line which has summary of package  */
    if ( !chs_get_line (libline_ch, txt, STRING_LENGTH) ) return (FALSE);
    ch_printf(WWW_ch,
	      "<tr><td><a href=%s.html><tt>%s</tt></a></td><td>%s</td></tr>\n",
	      package_name, package_name, txt);
    if (readme_ch == NULL) return (TRUE);
    ch_seek (memory_ch, 0);
    ch_printf (readme_ch, "\n");
    ch_puts (readme_ch,
	     "-----------------------------------------------------------------------------",
	     TRUE);
    ch_puts (readme_ch, txt, TRUE);
    /*  Copy package summary  */
    ch_copy (readme_ch, libline_ch, 0);
    ch_printf (readme_ch, "\n");
    /*  Count number of functions  */
    for (num_funcs = 0; chs_get_line (func_ch, txt, STRING_LENGTH);
	 ++num_funcs);
    ch_seek (func_ch, 0);
    /*  Write functions to README in two columns  */
    for (count = 0; chs_get_line (func_ch, txt, STRING_LENGTH); ++count)
    {
	if (count % 2 == 0) ch_printf (readme_ch, "%-40s", txt);
	else ch_puts (readme_ch, txt, TRUE);
    }
    if (count % 2 != 0) ch_printf (readme_ch, "\n");
    return (TRUE);
}   /*  End Function process_package  */
