/*  karma_doc2man.c

    Source file for  karma_doc2man  (convert documentation to manual pages).

    Copyright (C) 1997-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma utility programme will read Karma documentation file and will
    write out manual pages.


    Written by      Richard Gooch   23-MAR-1997

    Updated by      Richard Gooch   24-MAR-1997

    Updated by      Richard Gooch   19-SEP-1997: Corrected URL for ftp.

    Updated by      Richard Gooch   13-JAN-1999: Added support for tables.

    Updated by      Richard Gooch   19-MAY-1999: Added support for $NAME and
  $SYNOPSIS tags.

    Last updated by Richard Gooch   28-MAY-1999: Added default synopsis.


*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <karma.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_st.h>


#define BUF_SIZE 1048576


/*  Private functions  */
STATIC_FUNCTION (flag get_tag, (Channel channel, CONST char *tag,
				Channel memory_ch, unsigned int *num_notes) );
STATIC_FUNCTION (flag get_value_for_tag,
		 (Channel channel, CONST char *tag,
		  char value[STRING_LENGTH], flag warn) );
STATIC_FUNCTION (flag process_function,
		 (Channel docfile_ch, Channel package_ch, Channel function_ch,
		  Channel memory_ch, int section, CONST char *package_name,
		  CONST char *tag, CONST char *function_name,
		  CONST char *formatted_time, CONST char *synopsis,
		  CONST char *software_name) );
STATIC_FUNCTION (flag process_table,
		 (Channel docfile_ch, Channel package_ch,
		  CONST char *table_name) );
STATIC_FUNCTION (flag process_tag_with_lines,
		 (Channel docfile_ch, Channel out_ch,
		  char line[STRING_LENGTH], flag space) );
STATIC_FUNCTION (flag process_parameter,
		 (Channel docfile_ch, Channel function_ch,
		  char line[STRING_LENGTH]) );
STATIC_FUNCTION (flag write_tail,
		 (Channel channel, CONST char *software_name) );


int main (int argc, char **argv)
{
    Channel docfile_ch, function_ch, memory_ch;
    Channel package_ch = NULL;
    int section = -1;
    CONST char *docfilename, *mandir;
    struct tm *tm;
    time_t tv;
    char line[STRING_LENGTH], txt[STRING_LENGTH];
    char package_name[STRING_LENGTH], PACKAGE_NAME[STRING_LENGTH];
    char formatted_time[STRING_LENGTH], synopsis[STRING_LENGTH];
    char software_name[STRING_LENGTH];
    char p1[STRING_LENGTH], p2[STRING_LENGTH];
    static char usage_string[] = "karma_doc2man [-section num] docfile mandir";

    if ( (argc > 1) && (strcmp (argv[1], "-section") == 0) )
    {
	if (argc < 3)
	{
	    fprintf (stderr, "Usage:\t%s\n", usage_string);
	    exit (RV_MISSING_PARAM);
	}
	section = atoi (argv[2]);
	argc -= 2;
	argv += 2;
    }
    if (argc != 3)
    {
	fprintf (stderr, "Usage:\t%s\n", usage_string);
	exit (RV_MISSING_PARAM);
    }
    docfilename = argv[1];
    mandir = argv[2];
    tv = time (NULL);
    tm = localtime (&tv);
    strftime (formatted_time, STRING_LENGTH, "%d %b %Y", tm);
    if ( ( memory_ch = ch_open_memory (NULL, BUF_SIZE) ) == NULL )
    {
	fprintf (stderr, "Error opening memory buffer\n");
	exit (RV_CANNOT_OPEN);
    }
    if ( ( docfile_ch = ch_open_file (docfilename, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening: \"%s\"\t%s\n",
		 docfilename, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    if (section < 0) section = 3;
    strcpy (software_name, "Karma");
    /*  $PACKAGE  */
    if ( !get_value_for_tag (docfile_ch, "$PACKAGE", package_name, TRUE) )
	exit(RV_BAD_DATA);
    strcpy (PACKAGE_NAME, package_name);
    st_upr (PACKAGE_NAME);
    sprintf (synopsis, "#include <karma_%s.h>", package_name);  /*  Default  */
    sprintf (txt, "%s/%s.%d", mandir, package_name, section);
    if ( ( package_ch = ch_open_file (txt, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 txt, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    while ( chs_get_line (docfile_ch, line, STRING_LENGTH) )
    {
	char *ptr;
	char tag[STRING_LENGTH];
	static flag write_title = TRUE;

	if (line[0] != '$')
	{
	    ch_puts (package_ch, line, TRUE);
	    continue;
	}
	strcpy (tag, line);
	for (ptr = tag; !isspace (*ptr) && (*ptr != '\0'); ++ptr);
	if (*ptr == '\0') ptr = NULL;
	else
	{
	    *ptr = '\0';
	    for (ptr++; isspace (*ptr); ++ptr)
	    {
		if (*ptr == '\0')
		{
		    fprintf (stderr, "Error parsing line\n");
		    return (FALSE);
		}
	    }
	}
	if (strcmp (tag, "$END") == 0) break;
	if (strcmp (tag, "$NAME") == 0) strcpy (software_name, ptr);
	if (write_title)
	{
	    ch_printf (package_ch, ".TH %s %d \"%s\" \"%s Distribution\"\n",
		       PACKAGE_NAME, section, formatted_time, software_name);
	    write_title = FALSE;
	}
	if (strcmp (tag, "$NAME") == 0) continue;
	if (strcmp (tag, "$SYNOPSIS") == 0) strcpy (synopsis, ptr);
	else if (strcmp (tag, "$SUMMARY") == 0)
	    ch_printf (package_ch, ".SH PACKAGE\n%s \\- %s\n", package_name,
		       ptr);
	else if (strcmp (tag, "$PURPOSE") == 0)
	    ch_puts (package_ch, ".SH PURPOSE", TRUE);
	else if (strcmp (tag, "$LIBRARY") == 0)
	    ch_printf (package_ch, ".SH LIBRARY\n%s\n", ptr);
	else if (strcmp (tag, "$LINKWITH") == 0)
	    ch_printf (package_ch, ".SH LINK WITH\n%s\n", ptr);
	else ch_puts (package_ch, line, TRUE);
    }
    ch_printf (package_ch, ".SH FUNCTIONS\n");
    while ( chs_get_line (docfile_ch, line, STRING_LENGTH) )
    {
	sscanf (line, "%s %s", p1, p2);
	if ( (strcmp (p1, "$FUNCTION") == 0) ||
	     (strcmp (p1, "$EXPERIMENTAL_FUNC") == 0) ||
	     (strcmp (p1, "$PROTO_FUNC") == 0) )
	{
	    sprintf (txt, "%s/%s.%d", mandir, p2, section);
	    if ( ( function_ch = ch_open_file (txt, "w") ) == NULL )
		exit (RV_CANNOT_OPEN);
	    if ( !process_function (docfile_ch, package_ch, function_ch,
				    memory_ch, section,
				    package_name, p1 + 1, p2,
				    formatted_time, synopsis, software_name) )
	    {
		ch_close (docfile_ch);
		ch_close (package_ch);
		ch_close (function_ch);
		exit (RV_BAD_DATA);
	    }
	    ch_close (function_ch);
	}
	else if (strcmp (p1, "$TABLE") == 0)
	{
	    if ( !process_table (docfile_ch, package_ch, p2) )
	    {
		ch_close (docfile_ch);
		ch_close (package_ch);
		exit (RV_BAD_DATA);
	    }
	}
	else
	{
	    fprintf (stderr, "Bad line: \"%s\"\n", line);
	    exit (RV_BAD_DATA);
	}
    }
    ch_close (docfile_ch);
    if ( !write_tail (package_ch, software_name) ) exit (RV_WRITE_ERROR);
    ch_close (package_ch);
    return (RV_OK);
}   /*  End Function main  */


/*  Private functions follow  */

static flag get_tag (Channel channel, CONST char *tag, Channel memory_ch,
		     unsigned int *num_notes)
/*  [SUMMARY] Get a tag.
    <channel> The channel to read from.
    <tag> The tag name.
    <memory_ch> The memory channel to write notes to.
    <num_notes> This is incremented for each note.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char line[STRING_LENGTH];

    while ( chs_get_line (channel, line, STRING_LENGTH) )
    {
	if (strcmp (line, "$NOTE") == 0)
	{
	    if ( (memory_ch == NULL) || (num_notes == NULL) )
	    {
		fprintf (stderr, "$NOTE tag not handled\n");
		return (FALSE);
	    }
	    ++*num_notes;
	    if ( !process_tag_with_lines (channel, memory_ch, line, TRUE) )
		return (FALSE);
	}
	if (strcmp (line, tag) != 0)
	{
	    fprintf (stderr, "%s tag not found in: \"%s\"\n", tag, line);
	    return (FALSE);
	}
	return (TRUE);
    }
    fprintf (stderr, "Error reading %s tag\n", tag);
    return (FALSE);
}   /*  End Function get_tag  */

static flag get_value_for_tag (Channel channel, CONST char *tag,
			       char value[STRING_LENGTH], flag warn)
/*  [SUMMARY] Get the value for a tag.
    <channel> The channel to read from.
    <tag> The tag name.
    <value> The value is written here.
    <warn> If TRUE, warning messages may be displayed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char line[STRING_LENGTH];
    char *ptr;

    if ( !chs_get_line (channel, line, STRING_LENGTH) )
    {
	fprintf (stderr, "Error reading %s tag\n", tag);
	return (FALSE);
    }
    for (ptr = line; !isspace (*ptr); ++ptr);
    *ptr = '\0';
    for (ptr++; isspace (*ptr); ++ptr)
    {
	if (*ptr == '\0')
	{
	    fprintf (stderr, "Error parsing line\n");
	    return (FALSE);
	}
    }
    if (strcmp (line, tag) != 0)
    {
	if (warn) fprintf (stderr, "%s tag not found in: \"%s\"\n", tag, line);
	return (FALSE);
    }
    strcpy (value, ptr);
    return (TRUE);
}   /*  End Function get_value_for_tag  */

static flag process_function (Channel docfile_ch, Channel package_ch,
			      Channel function_ch, Channel memory_ch,
			      int section,
			      CONST char *package_name, CONST char *tag,
			      CONST char *function_name,
			      CONST char *formatted_time, CONST char *synopsis,
			      CONST char *software_name)
/*  [SUMMARY] Process a function.
    <docfile_ch> The channel to read from.
    <package_ch> The channel to write package information to.
    <function_ch> The channel to write function information to.
    <memory_ch> The channel to write temporary information to.
    <section> The section number.
    <package_name> The name of the package.
    <tag> The tag for this function.
    <function_name> The name of the function.
    <formatted_time> The formatted time string.
    <synopsis> The synopsis formatting string.
    <software_name> The name of the software.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag keep_going;
    unsigned int num_notes = 0;
    unsigned long read_pos, write_pos;
    char *ptr;
    char line[STRING_LENGTH], FUNCTION_NAME[STRING_LENGTH];
    char return_type[STRING_LENGTH], mt_level[STRING_LENGTH];

    strcpy (FUNCTION_NAME, function_name);
    st_upr (FUNCTION_NAME);
    strcpy (mt_level, "Unsafe.");
    ch_printf (function_ch,".TH %s %d \"%s\" \"%s Distribution\"\n",
	       FUNCTION_NAME, section, formatted_time, software_name);
    /*  $RETURN_TYPE  */
    if ( !get_value_for_tag (docfile_ch, "$RETURN_TYPE", return_type, TRUE) )
	return (FALSE);
    /*  $PARAMETER_LIST  */
    if ( !get_tag (docfile_ch, "$PARAMETER_LIST", memory_ch, &num_notes) )
	return (FALSE);
    ch_seek (memory_ch, 0);
    for (keep_going = TRUE; keep_going; )
    {
	if ( !chs_get_line (docfile_ch, line, STRING_LENGTH) ) return (FALSE);
	if (line[0] == '$')
	{
	    keep_going = FALSE;
	    continue;
	}
	if ( !ch_puts (memory_ch, line, TRUE) ) return (FALSE);
    }
    /*  $SUMMARY  */
    if (strncmp (line, "$SUMMARY", 8) != 0)
    {
	fprintf (stderr, "$SUMMARY tag not found in line: \"%s\"\n", line);
	return (FALSE);
    }
    for (ptr = line; !isspace (*ptr); ++ptr);
    for (ptr++; isspace (*ptr); ++ptr)
    {
	if (*ptr == '\0')
	{
	    fprintf (stderr, "Error parsing line\n");
	    return (FALSE);
	}
    }
    ch_printf (package_ch, ".IP \\fI%s\\fP 1i\n%s\n", function_name, ptr);
    ch_printf (function_ch, ".SH %s\n%s \\- %s\n", tag, function_name, ptr);
    ch_printf (function_ch, ".SH SYNOPSIS\n.B ");
    ch_printf (function_ch, synopsis, package_name);
    ch_printf (function_ch, "\n.sp\n");
    ch_printf (function_ch, ".B %s %s\n.sp\n", return_type, function_name);
    ch_tell (memory_ch, &read_pos, &write_pos);
    if (ch_copy (function_ch, memory_ch, write_pos) < write_pos) return FALSE;
    /*  Get ready for notes  */
    ch_seek (memory_ch, 0);
    /*  $PURPOSE  */
    if ( !get_tag (docfile_ch, "$PURPOSE", memory_ch, &num_notes) )
	return (FALSE);
    ch_printf (function_ch, ".SH DESCRIPTION\n");
    for (keep_going = TRUE; keep_going; )
    {
	if ( !chs_get_line (docfile_ch, line, STRING_LENGTH) ) return (FALSE);
	if (line[0] == '$')
	{
	    keep_going = FALSE;
	    continue;
	}
	if ( !ch_puts (function_ch, line, TRUE) ) return (FALSE);
    }
    while (TRUE)
    {
	if (strcmp (line, "$NOTE") == 0)
	{
	    ++num_notes;
	    if ( !process_tag_with_lines (docfile_ch, memory_ch, line, TRUE) )
		return (FALSE);
	    continue;
	}
	if (strcmp (line, "$END") == 0)
	{
	    if (num_notes > 0)
	    {
		/*  Write notes  */
		if (num_notes > 1) ch_printf (function_ch, ".SH NOTES\n");
		else ch_printf (function_ch, ".SH NOTE\n");
		ch_tell (memory_ch, &read_pos, &write_pos);
		if (ch_copy (function_ch, memory_ch, write_pos) < write_pos)
		    return (FALSE);
	    }
	    return write_tail (function_ch, software_name);
	}
	if (strncmp (line, "$PARAMETER", 10) == 0)
	{
	    if ( !process_parameter (docfile_ch, function_ch, line) )
		return (FALSE);
	    continue;
	}
	if (strcmp (line, "$RETURNS") == 0)
	{
	    ch_printf (function_ch, ".SH RETURN VALUE\n");
	    if ( !process_tag_with_lines (docfile_ch, function_ch, line,
					  FALSE) ) return (FALSE);
	    ch_printf (function_ch, ".SH Multi-threading Level\n%s\n",
		       mt_level);
	    continue;
	}
	if (strncmp (line, "$MT-LEVEL", 9) == 0)
	{
	    for (ptr = line; !isspace (*ptr); ++ptr);
	    for (ptr++; isspace (*ptr); ++ptr)
	    {
		if (*ptr == '\0')
		{
		    fprintf (stderr, "Error parsing line\n");
		    return (FALSE);
		}
	    }
	    strcpy (mt_level, ptr);
	    if ( !chs_get_line (docfile_ch, line, STRING_LENGTH) )
		return (FALSE);
	    continue;
	}
	if (strcmp (line, "$SEE_ALSO") == 0)
	{
	    ch_printf (function_ch, ".SH SEE ALSO\n");
	    if ( !process_tag_with_lines (docfile_ch, function_ch, line,
					  FALSE) ) return (FALSE);
	    continue;
	}
	fprintf (stderr, "Bad line: \"%s\"\n", line);
	return (FALSE);
    }
    return (FALSE);
}   /*  End Function process_function  */

static flag process_table (Channel docfile_ch, Channel package_ch,
			   CONST char *table_name)
/*  [SUMMARY] Process a table.
    <docfile_ch> The channel to read from.
    <package_ch> The channel to write package information to.
    <table_name> The name of the table.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char ch;
    unsigned int num_columns, count;
    char line[STRING_LENGTH], columns[STRING_LENGTH];
    char summary[STRING_LENGTH];

    ch_printf (package_ch, ".SH TABLE %s\n", table_name);
    /*  $COLUMNS  */
    if ( !get_value_for_tag (docfile_ch, "$COLUMNS", columns, TRUE) )
	return (FALSE);
    num_columns = atoi (columns);
    /*  $SUMMARY  */
    if ( !get_value_for_tag (docfile_ch, "$SUMMARY", summary, TRUE) )
	return (FALSE);
    ch_puts (package_ch, summary, TRUE);
    ch_puts (package_ch, "\n.TS", TRUE);
    for (count = 1; count < num_columns; ++count)
	ch_puts (package_ch, "l ", FALSE);
    ch_puts (package_ch, "l", TRUE);
    for (count = 1; count < num_columns; ++count)
	ch_puts (package_ch, "_ ", FALSE);
    ch_puts (package_ch, "_", TRUE);
    for (count = 1; count < num_columns; ++count)
	ch_puts (package_ch, "l ", FALSE);
    ch_puts (package_ch, "l.", TRUE);
    /*  $TABLE_DATA  */
    if ( !get_tag (docfile_ch, "$TABLE_DATA", NULL, NULL) ) return (FALSE);
    while (TRUE)
    {
	if ( !chs_get_line (docfile_ch, line, STRING_LENGTH) ) return (FALSE);
	if (line[0] == '$') break;
	for (count = 0, ch = line[0]; ch != '\0'; ch = line[++count])
	{
	    if (ch == '|') ++count;
	    else ch_write (package_ch, &ch, 1);
	}
	ch = '\n';
	ch_write (package_ch, &ch, 1);
    }
    /*  $END  */
    if (strcmp (line, "$END") != 0)
    {
	fprintf (stderr, "$END tag not found in line: \"%s\"\n", line);
	return (FALSE);
    }
    ch_puts (package_ch, ".TE", TRUE);
    return (TRUE);
}   /*  End Function process_table  */

static flag process_tag_with_lines (Channel docfile_ch, Channel out_ch,
				    char line[STRING_LENGTH], flag space)
/*  [SUMMARY] Process a tag with one or more lines following it.
    <docfile_ch> The channel to read from.
    <out_ch> The channel to write information to.
    <space> If TRUE, a spacer line is placed after the tag lines.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    while ( chs_get_line (docfile_ch, line, STRING_LENGTH) )
    {
	if (line[0] == '$')
	{
	    if (space) ch_puts (out_ch, ".sp", TRUE);
	    return (TRUE);
	}
	else
	{
	    ch_puts (out_ch, line, TRUE);
	}
    }
    return (FALSE);
}   /*  End Function process_tag_with_lines  */

static flag process_parameter (Channel docfile_ch, Channel function_ch,
			       char line[STRING_LENGTH])
/*  [SUMMARY] Process a note.
    <docfile_ch> The channel to read from.
    <function_ch> The channel to write function information to.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char *ptr, *p1;

    while (TRUE)
    {
	if (strncmp (line, "$PARAMETER", 10) == 0)
	{
	    for (ptr = line; !isspace (*ptr); ++ptr);
	    for (; isspace (*ptr); ++ptr);
	    for (p1 = ptr; !isspace (*p1); ++p1);
	    *p1 = '\0';
	    ch_printf (function_ch, ".IP \\fI%s\\fP 1i\n", ptr);
	}
	else if (line[0] == '$') return (TRUE);
	else ch_puts (function_ch, line, TRUE);
	if ( !chs_get_line (docfile_ch, line, STRING_LENGTH) ) return (FALSE);
    }
    return (TRUE);
}   /*  End Function process_parameter  */

static flag write_tail (Channel channel, CONST char *software_name)
/*  [SUMMARY] Write the tail (author and availability).
    <channel> The channel to write to.
    <software_name> The name of the software.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    if ( !ch_puts (channel, ".SH AUTHOR\nRichard Gooch (rgooch@atnf.csiro.au)",
		   TRUE) ) return (FALSE);
    if (strcmp (software_name, "Karma") != 0) return (TRUE);
    if ( !ch_puts (channel, ".SH AVAILABLITY", TRUE) ) return (FALSE);
    if (!ch_puts(channel,
		 "The Karma Distribution is available for anonymous ftp from:",
		 TRUE) ) return (FALSE);
    if ( !ch_puts (channel, "\nftp://ftp.atnf.csiro.au/pub/software/karma/",
		   TRUE) ) return (FALSE);
    if (!ch_puts(channel,
		 "ftp://wuarchive.wustl.edu/graphics/graphics/packages/karma/",
		 TRUE) ) return (FALSE);
    if ( !ch_puts (channel, "ftp://ftp.pwr.wroc.pl/pub/karma/",
		   TRUE) ) return (FALSE);
    if ( !ch_puts (channel, "\nThe Karma Home Page is:\n", TRUE) )
	return (FALSE);
    if ( !ch_puts (channel, "http://www.atnf.csiro.au/karma/", TRUE) )
	return (FALSE);
    return (TRUE);
}   /*  End Function write_tail  */
