/*  shader2c.c

    Source file for  shader2c  (shader to c pre-compiler).

    Copyright (C) 1994-1998  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This programme will convert source code in the Karma shader language to
    C code.


    Written by      Richard Gooch   9-MAY-1994

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   17-JUN-1995: Stipped smooth raycast code
  generation and added  -codeset  option.

    Updated by      Richard Gooch   24-DEC-1995: Tested for no parameters.

    Updated by      Richard Gooch   26-MAR-1996: Added typedef for
  RotatedKcoord_3d.

    Updated by      Richard Gooch   29-JUL-1996: Added #include <karma_m.h>

    Updated by      Richard Gooch   29-JAN-1997: Minor cosmetic changes.

    Last updated by Richard Gooch   3-NOV-1998: Added some CONST declarations.

    $Id: shader2c.c,v 2.1 2011/07/27 02:44:10 cal103 Exp $

*/
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>

STATIC_FUNCTION (void copy_file, (Channel dest, char *sourcefile) );

int main (argc, argv)
int argc;
char *argv[];
{
    flag some_shaders = FALSE;
    flag usage_error = FALSE;
    Channel inp;
    Channel out;
    CONST char *p, *includedir;
    char *keyword, *func;
    char *codeset, *sourcefile = NULL;
    char *destfile = NULL;
    char buffer[STRING_LENGTH];
    char txt[STRING_LENGTH];
    static char usage_string[] = "Usage:\tshader2c [-codeset target] sourcefile destfile";
    static char *default_codeset = "generic";
    static char function_name[] = "main";

    if (argc < 3)
    {
	fprintf (stderr, "%s\n", usage_string);
	exit (RV_MISSING_PARAM);
    }
    codeset = default_codeset;
    argc--;
    ++argv;
    while ( (argc > 0) && !usage_error )
    {
	if (strcmp (argv[0], "-codeset") == 0)
	{
	    if (argc < 4)
	    {
		usage_error = TRUE;
		continue;
	    }
	    codeset = argv[1];
	    argc -= 2;
	    argv += 2;
	}
	else
	{
	    if (argc < 2)
	    {
		usage_error = TRUE;
		continue;
	    }
	    sourcefile = argv[0];
	    destfile = argv[1];
	    argc -= 2;
	    argv += 2;
	}
    }
    if ( (sourcefile == NULL) || (destfile == NULL) ) usage_error = TRUE;
    if (usage_error)
    {
	fprintf (stderr, "%s\n", usage_string);
	exit (RV_MISSING_PARAM);
    }
    if ( ( includedir = r_getenv ("KARMAINCLUDEPATH") ) == NULL )
    {
	fprintf (stderr,
		 "Environment variable: \"KARMAINCLUDEPATH\" not set\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( ( inp = ch_open_file (sourcefile, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 sourcefile, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    if ( ( out = ch_open_file (destfile, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 destfile, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    if ( !ch_puts (out, "#include <stdio.h>", TRUE) )
    {
	fprintf (stderr, "Write error\n");
	exit (RV_WRITE_ERROR);
    }
    if (!ch_puts(out,
		 "typedef struct {float h; float v; float d;} RotatedKcoord_3d;"
		 , TRUE) )
    {
	fprintf (stderr, "Write error\n");
	exit (RV_WRITE_ERROR);
    }
    /*  Start reading the source file  */
    while ( ch_getl (inp, buffer, STRING_LENGTH) )
    {
	if (strlen (buffer) < 1)
	{
	    if ( !ch_puts (out, "", TRUE) )
	    {
		fprintf (stderr, "Write error\n");
		exit (RV_WRITE_ERROR);
	    }
	    continue;
	}
	if ( ( keyword = ex_word (buffer, &p) ) == NULL )
	{
	    m_abort (function_name, "keyword");
	}
	if (strcmp (keyword, "SHADER") == 0)
	{
	    if (some_shaders)
	    {
		if ( !ch_puts (out, "}\n\n", TRUE) )
		{
		    fprintf (stderr, "Write error\n");
		    exit (RV_WRITE_ERROR);
		}
	    }
	    some_shaders = TRUE;
	    if ( ( func = ex_word (p, &p) ) == NULL )
	    {
		fprintf (stderr, "No function name\n");
		exit (RV_BAD_DATA);
	    }
	    if (p != NULL)
	    {
		fprintf (stderr, "Bad line: \"%s\"\n", buffer);
		exit (RV_BAD_DATA);
	    }
	    sprintf (txt, "void shader_%s ", func);
	    if ( !ch_puts (out, txt, FALSE) )
	    {
		fprintf (stderr, "Write error\n");
		exit (RV_WRITE_ERROR);
	    }
	    /*  Copy function parameters and initialisation  */
	    sprintf (txt, "%s/shader/%s.setup",
			    includedir, codeset);
	    copy_file (out, txt);
	    sprintf (txt, "    static char function_name[] = \"%s\";\n",
			    func);
	    if ( !ch_puts (out, txt, TRUE) )
	    {
		fprintf (stderr, "Write error\n");
		exit (RV_WRITE_ERROR);
	    }
	    m_free (func);
	}
	else if (strcmp (keyword, "CODE") == 0)
	{
	    /*  Write raycasting code  */
	    sprintf (txt, "%s/shader/%s.cast",
			    includedir, codeset);
	    copy_file (out, txt);
	}
	else if (strcmp (keyword, "CLEANUP") == 0)
	{
	    /*  Write some comments code  */
	    strcpy (txt,
		    "\t/*  END Shader specific computations  */\n    }\n    /*  Shader specific clean-up  */");
	    if ( !ch_puts (out, txt, TRUE) )
	    {
		fprintf (stderr, "Write error\n");
		exit (RV_WRITE_ERROR);
	    }
	}
	else
	{
	    /*  Just copy the line  */
	    if ( !ch_puts (out, buffer, TRUE) )
	    {
		fprintf (stderr, "Write error\n");
		exit (RV_WRITE_ERROR);
	    }
	}
	m_free (keyword);
    }
    /*  End of source  */
    ch_close (inp);
    if (some_shaders)
    {
	if ( !ch_puts (out, "}", TRUE) )
	{
	    fprintf (stderr, "Write error\n");
	    exit (RV_WRITE_ERROR);
	}
    }
    ch_close (out);
    return (RV_OK);
}   /*  End Function main  */

static void copy_file (Channel dest, char *sourcefile)
{
    char ch;
    Channel source;

    if ( ( source = ch_open_file (sourcefile, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 sourcefile, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    while (ch_read (source, &ch, 1) > 0)
    {
	if (ch_write (dest, &ch, 1) < 1)
	{
	    fprintf (stderr, "Error writing\t%s\n", ERRSTRING);
	}
    }
    ch_close (source);
}   /*  End Function copy_file  */
