/*LINTLIBRARY*/
/*  iraf_read.c

    This code provides an IRAF read facility.

    Copyright (C) 2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  the IRAF format.


    Written by      Richard Gooch   6-JUN-2000: Copied from aips1_read.c

    Last updated by Richard Gooch   6-JUN-2000


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_dir.h>
#include <karma_pio.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_p.h>
#include <karma_r.h>
#include <karma_a.h>

#define GET_VALUE(out, offset) get_value ((out), (char *) slot + (offset), \
					  sizeof *(out), swap_bytes)
#define MAGIC_NUMBER 327674912

#define VERIFY_CONTEXT(ctx) if (ctx == NULL) \
{fprintf (stderr, "NULL context passed\n"); \
 a_prog_bug (function_name); } \
if (ctx->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid context object\n"); \
 a_prog_bug (function_name); }



/*  Private structures  */
struct iraf_data_context_type
{
    unsigned int magic_number;
    Channel channel;
    flag swap_bytes;
    flag have_blanking;
    int blank_value;
    unsigned int num_dim;
    unsigned long blank_count;
    uaddr read_pos;
    uaddr dim_lengths[MAX_DIMENSIONS];
    char datafile[STRING_LENGTH];
};


/*  Private functions  */
STATIC_FUNCTION (void strip_trailing_whitespace, (char *string) );
STATIC_FUNCTION (void get_value,
		 (void *dest, CONST void *source, uaddr size, flag swap) );
STATIC_FUNCTION (uaddr read_blocks,
		 (Channel channel, char *buffer, uaddr num, uaddr size,
		  flag swap_bytes) );


/*  Private data  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_iraf_read (CONST char *setname)
/*  [SUMMARY] Read an IRAF dataset.
    <setname> The IRAF dataset name.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    KIRAFDataContext context;
    uaddr arr_size;
    array_pointer arrayp;
    multi_array *multi_desc;
    /*static char function_name[] = "foreign_iraf_read";*/

    if ( ( multi_desc = foreign_iraf_read_header (setname, TRUE, &context) )
	 == NULL ) return (NULL);
    /*  Now read data  */
    arr_size =
	ds_get_array_size ( (array_desc *)
			    multi_desc->headers[0]->element_desc[0] );
    m_copy (&arrayp, multi_desc->data[0], sizeof arrayp);
    if ( !foreign_iraf_read_data (context, arrayp.array, arr_size) )
    {
	ds_dealloc_multi (multi_desc);
	foreign_iraf_close_data_context (context);
	return (NULL);
    }
    if (context->blank_count > arr_size >> 4)
    {
	fprintf (stderr,
		 "WARNING: %lu out of %lu values are blank: bad data?\n",
		 context->blank_count, arr_size);
    }
    foreign_iraf_close_data_context (context);
    return (multi_desc);
}   /*  End Function foreign_iraf_read  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_iraf_read_header (CONST char *setname, flag data_alloc,
				       KIRAFDataContext *context)
/*  [SUMMARY] Read an IRAF dataset header.
    <setname> The IRAF dataset name.
    <data_alloc> If TRUE, the data space is allocated.
    <context> The data reading context is written here, to be used in
    subsequent calls to [<foreign_iraf_read_data>]. This may be NULL.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    Channel channel;
    flag swap_bytes;
    int len, slot_count, slot_offset, user_id;
    unsigned long num_slots;
    char *ptr, *record, *slot;
    double drecord[RECORD_LEN];
    char dirname[STRING_LENGTH], catfile[STRING_LENGTH];
    char sname[STRING_LENGTH], id_string[4];
    static char function_name[] = "foreign_iraf_read_header";

    {
	array_pointer arrayp;
	int axis_length, seq, naxis, ival;
	unsigned int dim_count, axis_count;
	unsigned int ktype = NONE;
	float bmin, bmaj, fval;
	struct iraf_data_context_type  ctx;
	multi_array *multi_desc;
	array_desc *arr_desc;
	char *array;
	CONST char *dim_names[MAX_DIMENSIONS];
	char irafname[STRING_LENGTH], txt[STRING_LENGTH];
	char headername[STRING_LENGTH];
	char bunit[STRING_LENGTH], coord_names[MAX_DIMENSIONS][STRING_LENGTH];
	char slot_string[4], seq_string[4];
	double value[2];

	if ( (channel = ch_open_file (headername, "r") ) == NULL )
	{
	    fprintf (stderr, "%s: error opening header file: \"%s\"\t%s\n",
		     function_name, headername, ERRSTRING);
	    return (NULL);
	}
	if (ch_read (channel, record, RECORD_BYTES) < RECORD_BYTES)
	{
	    fprintf (stderr, "%s: error reading map header\t%s\n",
		     function_name, ERRSTRING);
	    ch_close (channel);
	    return (NULL);
	}
	/*  At this point assume any further failures result in the function to
	    return rather than continue the loop. NOTE: header file is
	    closed here  */
	ch_close (channel);
	slot = record;
	GET_VALUE (&naxis, KIDIM * 4);
	if (naxis < 2)
	{
	    fprintf (stderr, "%s: one dimensional data not supported\n",
		     function_name);
	    return (NULL);
	}
	m_copy (txt, record + KHCTP * 4, 4 * 2 * KICTPN);
	txt[4 * 2 * KICTPN] = '\0';
	for (axis_count = 0, ctx.num_dim = 0; axis_count < naxis; ++axis_count)
	{
	    GET_VALUE (&axis_length, (KINAX + axis_count) * 4);
	    if (axis_length < 2) continue;
	    m_copy (coord_names[axis_count],
		    record + KHCTP * 4 + axis_count * 8, 8);
	    coord_names[axis_count][8] = '\0';
	    strip_trailing_whitespace (coord_names[axis_count]);
	    ++ctx.num_dim;
	}
	for (axis_count = 0, dim_count = ctx.num_dim - 1; axis_count < naxis;
	     ++axis_count)
	{
	    GET_VALUE (&axis_length, (KINAX + axis_count) * 4);
	    if (axis_length < 2) continue;
	    ctx.dim_lengths[dim_count] = axis_length;
	    dim_names[dim_count] = coord_names[axis_count];
	    --dim_count;
	}
	m_copy (bunit, record + KHBUN * 4, 8);
	bunit[8] = '\0';
	strip_trailing_whitespace (bunit);
	ktype = K_FLOAT;
	ptr = bunit;
	arr_desc = ds_easy_alloc_array_desc (ctx.num_dim, ctx.dim_lengths,
					     NULL, NULL, NULL, dim_names,
					     1, &ktype, (CONST char **) &ptr);
	if (arr_desc == NULL) return (NULL);
	if (data_alloc)
	{
	    if ( !ds_alloc_vm_array (&arrayp, arr_desc, FALSE, FALSE) &&
		 !ds_alloc_mmap_array (-1, 0, 0, TRUE, &arrayp, arr_desc,
				       FALSE, FALSE) )
	    {
		m_error_notify (function_name, "array data");
		ds_dealloc_array_desc (arr_desc);
		return (NULL);
	    }
	}
	else m_clear (&arrayp, sizeof arrayp);
	array = arrayp.array;
	if ( ( multi_desc =
	       ds_easy_alloc_array_from_array_desc (arr_desc, &arrayp, TRUE) )
	     == NULL )
	{
	    ds_dealloc_array_desc (arr_desc);
	    if (arrayp.free != NULL) (*arrayp.free) (arrayp.info, &arrayp);
	}
	/*  Attach the axis information  */
	value[1] = 0.0;
	value[0] = naxis;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "NAXIS", K_INT, value, FALSE);
	for (axis_count = 0; axis_count < naxis; ++axis_count)
	{
	    GET_VALUE (&axis_length, (KINAX + axis_count) * 4);
	    sprintf (txt, "NAXIS%d", axis_count + 1);
	    value[0] = axis_length;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_INT, value, FALSE);
	}
	/*  Attach the co-ordinate information  */
	for (axis_count = 0; axis_count < naxis; ++axis_count)
	{
	    /*  CTYPEn  */
	    sprintf (txt, "CTYPE%d", axis_count + 1);
	    ds_put_unique_named_string (multi_desc->headers[0],
					multi_desc->data,
					txt, coord_names[axis_count],
					FALSE);
	    /*  CRVALn  */
	    sprintf (txt, "CRVAL%d", axis_count + 1);
	    GET_VALUE (value, KDCRV * 4 + axis_count * 8);
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	    /*  CRPIXn  */
	    sprintf (txt, "CRPIX%d", axis_count + 1);
	    GET_VALUE (&fval, KRCRP * 4 + axis_count * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	    /*  CDELTn  */
	    sprintf (txt, "CDELT%d", axis_count + 1);
	    GET_VALUE (&fval, KRCIC * 4 + axis_count * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	    /*  CROTAn  */
	    sprintf (txt, "CROTA%d", axis_count + 1);
	    GET_VALUE (&fval, KRCRT * 4 + axis_count * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       txt, K_DOUBLE, value, FALSE);
	}
	/*  Attach other keywords from first record  */
	/*  BUNIT  */
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "BUNIT", bunit, FALSE);
	/*  OBJECT  */
	m_copy (txt, record + KHOBJ * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "OBJECT", txt, FALSE);
	/*  TELESCOP  */
	m_copy (txt, record + KHTEL * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "TELESCOP", txt, FALSE);
	/*  INSTRUME  */
	m_copy (txt, record + KHINS * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "INSTRUME", txt, FALSE);
	/*  OBSERVER  */
	m_copy (txt, record + KHOBS * 4, 8);
	txt[8] = '\0';
	strip_trailing_whitespace (txt);
	ds_put_unique_named_string (multi_desc->headers[0], multi_desc->data,
				    "OBSERVER", txt, FALSE);
	/*  EPOCH  */
	GET_VALUE (&fval, KREPO * 4);
	value[0] = fval;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "EPOCH", K_DOUBLE, value, FALSE);
	/*  DATAMIN  */
	GET_VALUE (&fval, KRDMN * 4);
	value[0] = fval;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "DATAMIN", K_DOUBLE, value, FALSE);
	/*  DATAMAX  */
	GET_VALUE (&fval, KRDMX * 4);
	value[0] = fval;
	ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				   "DATAMAX", K_DOUBLE, value, FALSE);
	GET_VALUE (&ival, KRBLK * 4);
	if (ival != 0)
	{
	    ctx.have_blanking = TRUE;
	    ctx.blank_value = ival;
	}
	GET_VALUE (&bmin, KRBMN * 4);
	GET_VALUE (&bmaj, KRBMJ * 4);
	if ( (bmin != 0.0) && (bmaj != 0.0) )
	{
	    /*  BMIN  */
	    value[0] = bmin;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       "BMIN", K_DOUBLE, value, FALSE);
	    /*  BMAJ  */
	    value[0] = bmaj;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       "BMAJ", K_DOUBLE, value, FALSE);
	    /*  BPA  */
	    GET_VALUE (&fval, KRBPA * 4);
	    value[0] = fval;
	    ds_put_unique_named_value (multi_desc->headers[0],multi_desc->data,
				       "BPA", K_DOUBLE, value, FALSE);
	}
	if (context == NULL) return (multi_desc);
	if ( ( *context = m_dup (&ctx, sizeof ctx) ) == NULL )
	{
	    m_error_notify (function_name, "data context");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
	(*context)->magic_number = MAGIC_NUMBER;
	(*context)->channel = NULL;
	(*context)->swap_bytes = swap_bytes;
	(*context)->blank_count = 0;
	(*context)->read_pos = 0;
	return (multi_desc);
    }
    ch_close (channel);
    fprintf (stderr, "%s: dataset \"%s\" not found in catalogue\n",
	     function_name, sname);
    return (NULL);
}   /*  End Function foreign_iraf_read_header  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_iraf_read_data (KIRAFDataContext context, char *data,
			      uaddr num_values)
/*  [SUMMARY] Read data in an IRAF file.
    [PURPOSE] This routine will read the data of an IRAF dataset from a
    KIRAFDataContext object. The header section is NOT read.
    <context> The context to read from.
    <data> The array to write the data into. If this is NULL, input data is
    skipped and nothing is written.
    <num_values> The number of values to read.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int rows_per_block;
    unsigned int ktype = K_FLOAT;
    uaddr values_per_row, bytes_per_row, values_per_image;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "foreign_iraf_read_data";

    VERIFY_CONTEXT (context);
    if (context->channel == NULL)
    {
	if ( ( context->channel = ch_open_file (context->datafile, "rz") )
	     == NULL )
	{
	    fprintf (stderr, "%s: error opening file: \"%s\"\t%s\n",
		     function_name, context->datafile, ERRSTRING);
	    return (FALSE);
	}
    }
    values_per_row = context->dim_lengths[context->num_dim - 1];
    values_per_image = values_per_row*context->dim_lengths[context->num_dim-2];
    bytes_per_row = values_per_row * host_type_sizes[ktype];
    if (bytes_per_row > RECORD_BYTES) rows_per_block = 1;
    else rows_per_block = RECORD_BYTES / bytes_per_row;
    while (num_values > 0)
    {
	flag drain = FALSE;
	uaddr pos_in_image = context->read_pos % values_per_image;
	uaddr pos_in_row = pos_in_image % values_per_row;
	uaddr values_to_read, bytes_read;

	values_to_read = num_values;
	if (values_to_read + pos_in_row > values_per_row)
	    values_to_read = values_per_row - pos_in_row;
	if ( ( bytes_read =
	       read_blocks (context->channel, data, values_to_read,
			    host_type_sizes[ktype], context->swap_bytes) )
	     < host_type_sizes[ktype] * values_to_read )
	{
	    fprintf (stderr, "%s: error reading: %lu bytes, got: %lu\t%s\n",
		     function_name, host_type_sizes[ktype] * values_to_read,
		     bytes_read, ERRSTRING);
	    return (FALSE);
	}
	if ( context->have_blanking && (data != NULL) )
	{
	    int val_count;
	    int *iptr;

	    for (val_count = 0, iptr = (int *) data;
		 val_count < values_to_read;
		 ++val_count, ++iptr)
	    {
		if (*iptr == context->blank_value)
		{
		    *(float *) iptr = TOOBIG;
		    ++context->blank_count;
		}
	    }
	}
	if ( (pos_in_row + values_to_read == values_per_row) &&
	     (rows_per_block < 2) ) drain = TRUE;
	if (pos_in_image + values_to_read == values_per_image) drain = TRUE;
	if (drain)
	{
	    if ( !ch_skip_to_boundary (context->channel, RECORD_BYTES) )
	    {
		fprintf (stderr, "%s: error reading to record boundary\n",
			 function_name);
		return (FALSE);
	    }
	}
	num_values -= values_to_read;
	context->read_pos += values_to_read;
	if (data != NULL) data += bytes_read;
    }
    return (TRUE);
}   /*  End Function foreign_iraf_read_data  */

/*EXPERIMENTAL_FUNCTION*/
void foreign_iraf_close_data_context (KIRAFDataContext context)
/*  [SUMMARY] Close a KIRAFDataContext object.
    <context> The context.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "foreign_iraf_close_data_context";

    VERIFY_CONTEXT (context);
    if (context->channel != NULL) ch_close (context->channel);
    context->magic_number = 0;
    m_free (context);
}   /*  End Function foreign_iraf_close_data_context  */


/*  Private functions follow  */

static void strip_trailing_whitespace (char *string)
/*  [SUMMARY] Strip trailing whitespace.
    <string> The string to strip.
    [RETURNS] Nothing.
*/
{
    char *ptr;

    for (ptr = string + strlen (string) - 1; (ptr > string) && isspace (*ptr);
	 --ptr) *ptr = '\0';
}   /*  End Function strip_trailing_whitespace  */

static void get_value (void *dest, CONST void *source, uaddr size, flag swap)
{
    if (swap) m_copy_and_swap_blocks (dest, source, size, size, size, 1);
    else m_copy (dest, source, size);
}   /*  End Function get_value  */

static uaddr read_blocks (Channel channel, char *buffer, uaddr num, uaddr size,
			  flag swap_bytes)
/*  [SUMMARY] Read blocks of data from a channel.
    <channel> The channel object.
    <buffer> The buffer to write the data into. If NULL, data are skipped.
    <num> The number of data blocks to read.
    <size> The size of the data blocks.
    <swap_bytes> If TRUE, data is byte-swapped after reading.
    [RETURNS] The number of bytes read.
*/
{
    if (buffer == NULL) return ch_skip (channel, num *size) ? (num * size) : 0;
    if (swap_bytes) return ch_read_and_swap_blocks (channel, buffer, num,size);
    else return ch_read (channel, buffer, num * size);
}   /*  End Function read_blocks  */
