/*LINTLIBRARY*/
/*  main.c

    This code provides KVectorField objects.

    Copyright (C) 2004  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating vector fields.


    Written by      Richard Gooch   19-AUG-2004

    Last updated by Richard Gooch   19-AUG-2004


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <karma.h>
#include <k_list.h>
#define NEW_WIN_SCALE
#include <karma_vectmap.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define VECFIELD_USED_MAGIC_NUMBER    1920918331
#define VECFIELD_FREE_MAGIC_NUMBER    3110496127
#define HOLDER_MAGIC_NUMBER           3667415821
#define VECFIELD_GROUP_MAGIC_NUMBER   2862936115
#define DEFAULT_COLOUR                   "green"

#define VERIFY_HOLDER(holder) if (holder == NULL) \
{fprintf (stderr, "NULL canvas holder passed\n"); \
 a_prog_bug (function_name); } \
if (holder->magic_number != HOLDER_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas holder object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_USED_VECFIELD(vfield) if (vfield == NULL) \
{fprintf (stderr, "NULL vector field passed\n"); \
 a_prog_bug (function_name); } \
if (vfield->magic_number != VECFIELD_USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid vector field object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_FREE_VECFIELD(vfield) if (vfield == NULL) \
{fprintf (stderr, "NULL vector field passed\n"); \
 a_prog_bug (function_name); } \
if (vfield->magic_number != VECFIELD_FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed vector field object\n"); \
 a_prog_bug (function_name); }

#define VERIFY_VECFIELD_GROUP(group) if (group == NULL) \
{fprintf (stderr, "NULL KVectorFieldGroup passed\n"); \
 a_prog_bug (function_name); } \
if (group->magic_number != VECFIELD_GROUP_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid KVectorFieldGroup object\n"); \
 a_prog_bug (function_name); }


typedef struct canvas_holder_type * CanvasHolder;

struct canvas_holder_type
{
    unsigned int magic_number;
    KWorldCanvas canvas;
    flag needs_refresh;
    KVectorField first;
    KVectorField last;
    CanvasHolder prev;
    CanvasHolder next;
    char *colourname;
};

struct vector_field_group
{
    unsigned int magic_number;
    KVectorField first;
    KVectorField last;
};

struct vector_field_chain_type
{
    KVectorField prev;           /*  Previous member of list for this chain  */
    KVectorField next;           /*  Next member of list for this chain      */
};

struct data_container
{
    multi_array *multi_desc;
    array_desc *arr_desc;
    char *slice;
    unsigned int hdim;
    unsigned int vdim;
    unsigned int elem_index;
};

struct vector_field_type
{
    unsigned int magic_number;
    KVectorFieldGroup group;
    CanvasHolder canvas_holder;
    /*  Image data  */
    struct data_container set0;
    struct data_container set1;
    struct data_container set2;
    KwcsAstro astro_projection;
    /*  Display attributes  */
    /*  Other information  */
    unsigned int num_restrictions;
    char **restriction_names;
    double *restriction_values;
    char *inline_restr_names[1];
    double inline_restr_values[1];
    KCallbackFunc data_change_callback;
    struct vector_field_chain_type holder_chain;
    struct vector_field_chain_type group_chain;
};


/*  Private data  */
static CanvasHolder first_canvas_holder = NULL;
static CanvasHolder last_canvas_holder = NULL;
static KVectorField first_free_vfield = NULL;


/*  Private functions  */
STATIC_FUNCTION (KVectorField create_restr,
		 (multi_array *multi_desc, flag try_ap,
		  array_desc *arr_desc, char *slice,
		  unsigned int hdim, unsigned int vdim,
		  unsigned int elem_index,
		  unsigned int num_restr, char **restr_names,
		  double *restr_values, KVectorFieldGroup group) );
STATIC_FUNCTION (unsigned int _vectmap_get_gen_index,
		 (CONST multi_array *multi_desc, CONST array_desc *arr_desc) );
STATIC_FUNCTION (CanvasHolder _vectmap_get_canvas_holder,
		 (KWorldCanvas canvas, char *func_name) );
STATIC_FUNCTION (CanvasHolder _vectmap_alloc_canvas_holder,
		 (KWorldCanvas canvas) );
STATIC_FUNCTION (void _vectmap_free_canvas_holder, (CanvasHolder holder) );
STATIC_FUNCTION (void _vectmap_process_canvas_attributes,
		 (CanvasHolder holder, va_list argp) );
STATIC_FUNCTION (void _vectmap_worldcanvas_precompute_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _vectmap_precompute_image,
		 (KViewableVectmapImage vfield,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void precompute_level,
		 (KViewableVectmapImage vfield, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _vectmap_compute_lin_limits,
		 (KViewableVectmapImage vfield, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  double win_x_min, double win_y_min,
		  double win_x_max, double win_y_max) );
STATIC_FUNCTION (void _vectmap_compute_pix_segs,
		 (KViewableVectmapImage vfield, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  double win_x_min, double win_y_min,
		  double win_x_max, double win_y_max) );
STATIC_FUNCTION (void compute_pixel_coords,
		 (KWorldCanvas canvas, struct viewable_levelinfo_type *level,
		  CONST struct linearsegments_level_struct *linsegs_level,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_world_segs, unsigned int num_pix_segs) );
STATIC_FUNCTION (void _vectmap_worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  CONST struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void refresh_image,
		 (KViewableVectmapImage vfield,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void draw_level,
		 (KViewableVectmapImage vfield, unsigned int level_index,
		  CONST struct win_scale_type *win_scale,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _vectmap_dealloc_levels, (KVectorField vfield) );
STATIC_FUNCTION (void _vectmap_dealloc_viewable_levels,
		 (KViewableVectmapImage vfield) );
STATIC_FUNCTION (void _vectmap_find_linsegs_cache,
		 (KViewableVectmapImage vfield,
		  CONST struct win_scale_type *win_scale) );
STATIC_FUNCTION (void _vectmap_release_linsegs_cache,
		 (struct linearsegments_cache_struct *cache) );


/* Public functions follow */

/*EXPERIMENTAL_FUNCTION*/
void vectmap_init (KWorldCanvas canvas, ...)
/*  [SUMMARY] Initialise the package for a particular canvas.
    [PURPOSE] This routine will initialise the [<vectmap>] package for a
    particular world canvas. Calling this routine causes a number of callback
    routines internal to the package to be registered with the canvas (such
    as refresh and position event callbacks). This routine must be called
    before creating vectmap images.
    <canvas> The world canvas object.
    [VARARGS] The list of parameter attribute-key attribute-value-ptr pairs
    must follow. This list must be terminated with the VECTMAP_CANVAS_ATT_END.
    See [<vectmap_CANVAS_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    CanvasHolder holder;
    static char function_name[] = "vectmap_init";

    va_start (argp, canvas);
    if (canvas == NULL)
    {
	fprintf (stderr, "NULL world canvas passed\n");
	a_prog_bug (function_name);
    }
    if (_vectmap_get_canvas_holder (canvas, NULL) != NULL)
    {
	fprintf (stderr, "Canvas already initialised\n");
	a_prog_bug (function_name);
    }
    if ( ( holder = _vectmap_alloc_canvas_holder (canvas) ) == NULL )
	m_abort (function_name, "canvas holder");
    _vectmap_process_canvas_attributes (holder, argp);
    va_end (argp);
}   /*  End Function vectmap_init  */

/*EXPERIMENTAL_FUNCTION*/
KVectorFieldGroup vectmap_create_group ()
/*  [SUMMARY] Create a container to hold a group of KVectorField objects.
    [RETURNS] The group container on success, else NULL.
*/
{
    KVectorFieldGroup group;
    static char function_name[] = "vectmap_create_group";

    if ( ( group = m_calloc (sizeof *group) ) == NULL )
    {
	m_error_notify (function_name, "vfield group");
	return (NULL);
    }
    group->magic_number = VECFIELD_GROUP_MAGIC_NUMBER;
    return (group);
}   /*  End Function vectmap_create_group  */

/*EXPERIMENTAL_FUNCTION*/
void vectmap_destroy_group (KVectorFieldGroup group)
/*  [SUMMARY] Destroy a group of KVectorField objects.
    <group> The group to destroy. The group container is also destroyed.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "vectmap_destroy_group";

    VERIFY_VECFIELD_GROUP (group);
    while (group->first) vectmap_destroy (group->first);
    group->magic_number = 0;
    m_free (group);
}   /*  End Function vectmap_destroy_group  */

/*EXPERIMENTAL_FUNCTION*/
KVectorField vectmap_create_restr (KWorldCanvas canvas,
				   multi_array *multi_desc0,
				   array_desc *arr_desc0, char *slice0,
				   unsigned int hdim0, unsigned int vdim0,
				   unsigned int elem_index0,
				   unsigned int num_restr0,
				   char **restr_names0, double *restr_values0,
				   multi_array *multi_desc1,
				   array_desc *arr_desc1, char *slice1,
				   unsigned int hdim1, unsigned int vdim1,
				   unsigned int elem_index1,
				   unsigned int num_restr1,
				   char **restr_names1, double *restr_values1,
				   multi_array *multi_desc2,
				   array_desc *arr_desc2, char *slice2,
				   unsigned int hdim2, unsigned int vdim2,
				   unsigned int elem_index2,
				   unsigned int num_restr2,
				   char **restr_names2, double *restr_values2,
				   KVectorFieldGroup group)
/*  [SUMMARY] Create vectmap image from 2D slice with restrictions.
    [PURPOSE] This routine will create a vectmap image object from a
    2-dimensional slice of a Karma data structure. At a later time, this
    vectmap image may be used to create a viewable vectmap image.
    <multi_desc> The  multi_array  descriptor which contains the Karma data
    structure. The routine increments the attachment count on the descriptor
    on successful completion. This may be NULL.
    <arr_desc> The array descriptor.
    <slice> The start of the slice data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index in the data packets.
    <num_restr> The number of matched restrictions.
    <restr_names> The restriction names.
    <restr_values> The restriction values.
    <group> The KVectorFieldGroup to add the new object too. If NULL, it is
    not added to any group.
    [NOTE] Restriction information is automatically deallocated when
    [<vectmap_destroy>] is called.
    [RETURNS] A KVectorField object on success, NULL.
*/
{
    static char function_name[] = "vectmap_create_restr";

    if (group) VERIFY_CIMAGE_GROUP (group);
    if (arr_desc == NULL)
    {
	fprintf (stderr, "NULL array descriptor pointer passed\n");
	a_prog_bug (function_name);
    }
    if (slice == NULL)
    {
	fprintf (stderr, "NULL slice pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Sanity checks  */
    if (hdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "hdim: %u greater than number of dimensions: %u\n",
		 hdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[hdim]->coordinates)
	fprintf (stderr, "WARNING: hdim: %u not regularly spaced\n", hdim);
    if (vdim >= arr_desc->num_dimensions)
    {
	fprintf (stderr, "vdim: %u greater than number of dimensions: %u\n",
		 vdim, arr_desc->num_dimensions);
	a_prog_bug (function_name);
    }
    if (arr_desc->dimensions[vdim]->coordinates)
	fprintf (stderr, "WARNING: vdim: %u not regularly spaced\n", vdim);
    if (elem_index >= arr_desc->packet->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, arr_desc->packet->num_elements);
	a_prog_bug (function_name);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    return _vectmap_create_restr (multi_desc, TRUE, arr_desc, slice,
				  hdim, vdim, elem_index,
				  num_restr, restr_names, restr_values, group);
}   /*  End Function vectmap_create_restr  */

/*EXPERIMENTAL_FUNCTION*/
KVectorField vectmap_create_from_iarray (iarray array, flag swap,
					  KVectorFieldGroup group)
/*  [SUMMARY] Create a vectmap image from an Intelligent Array.
    [PURPOSE] This routine will create a vectmap image object from a
    2-dimensional Intelligent Array. At a later time, this vectmap image may be
    used to create a viewable vectmap image.
    <array> The Intelligent Array. The underlying <<multi_array>> data
    structure will have its attachment count incremented upon successful
    completion.
    <swap> If TRUE the y axis will be displayed horizontally.
    <group> The KVectorFieldGroup to add the new object too. If NULL, it is
    not added to any group.
    [RETURNS] A KVectorField object on success, else NULL.
*/
{
    KVectorField cimage;
    unsigned int num_restr, xdim, ydim;
    dim_desc *xdim_desc, *ydim_desc;
    char **restr_names;
    double *restr_values;
    static char function_name[] = "vectmap_create_from_iarray";

    if (array == NULL)
    {
	fprintf (stderr, "NULL Intelligent Array passed\n");
	a_prog_bug (function_name);
    }
    FLAG_VERIFY (swap);
    if (group) VERIFY_CIMAGE_GROUP (group);
    if (iarray_num_dim (array) != 2)
    {
	fprintf (stderr, "Intelligent Array must have exactly 2 dimensions\n");
	return (NULL);
    }
    if (swap) xdim = 0, ydim = 1;
    else xdim = 1, ydim = 0;
    xdim_desc = iarray_get_dim_desc (array, xdim);
    ydim_desc = iarray_get_dim_desc (array, ydim);
    if ( (iarray_dim_length (array, xdim) != xdim_desc->length)
	 || (iarray_dim_length (array, ydim) != ydim_desc->length) )
    {
	fprintf (stderr, "%s: Intelligent Array must not be a sub-array\n",
		 function_name);
	return (NULL);
    }
    num_restr = iarray_get_restrictions (array, &restr_names, &restr_values);
    cimage = _vectmap_create_restr (array->multi_desc, TRUE,
				    array->arr_desc, array->data,
				    array->orig_dim_indices[xdim],
				    array->orig_dim_indices[ydim],
				    array->elem_index,
				    num_restr, restr_names, restr_values,
				    group);
    if (!cimage) return (NULL);
    cimage->data_change_callback =
	iarray_register_data_change_func (array, vectmap_register_data_change,
					  cimage);
    return (cimage);
}   /*  End Function vectmap_create_from_iarray  */


/*  Private functions follow  */

static CanvasHolder _vectmap_get_canvas_holder (KWorldCanvas canvas,
						char *func_name)
/*  [SUMMARY] Get the canvas holder for a world canvas.
    <canvas> The canvas.
    <func_name> The name of the calling function. If this is not NULL, the
    process will abort if a canvas holder is not found.
    [RETURNS] The canvas holder on success, else NULL.
*/
{
    CanvasHolder canvas_holder;

    /*  First search for existing canvas holder  */
    for (canvas_holder = first_canvas_holder; canvas_holder;
	 canvas_holder = canvas_holder->next)
    {
	if (canvas == canvas_holder->canvas) return (canvas_holder);
    }
    if (func_name)
    {
	fprintf (stderr, "CanvasHolder not found!\n");
	a_prog_bug (func_name);
    }
    return (NULL);
}   /*  End Function _vectmap_get_canvas_holder  */

static CanvasHolder _vectmap_alloc_canvas_holder (KWorldCanvas canvas)
/*  This routine will allocate a canvas holder for a world canvas.
    The canvas must be given by  canvas  .
    The routine returns the canvas holder on success, else it returns NULL.
*/
{
    CanvasHolder holder;
    static char function_name[] = "_vectmap_alloc_canvas_holder";

    /*  Must allocate holder  */
    if ( ( holder = m_calloc (sizeof *holder) ) == NULL )
    {
	m_error_notify (function_name, "canvas holder");
	return (NULL);
    }
    holder->magic_number = HOLDER_MAGIC_NUMBER;
    holder->canvas = canvas;
    KLIST_APPEND2 (first_canvas_holder, last_canvas_holder, holder);
    canvas_register_precompute_func (canvas,
				     _vectmap_worldcanvas_precompute_func,
				     holder);
    canvas_register_refresh_func (canvas, _vectmap_worldcanvas_refresh_func,
				  holder);
    canvas_register_destroy_func (canvas, FALSE, _vectmap_free_canvas_holder,
				  holder);
    return (holder);
}   /*  End Function _vectmap_alloc_canvas_holder  */

static void _vectmap_free_canvas_holder (CanvasHolder holder)
/*  [SUMMARY] Free a canvas holder (and any associated KVectorField objects).
    <holder> The canvas holder.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_vectmap_free_canvas_holder";

    VERIFY_HOLDER (holder);
    while (holder->first) vectmap_destroy (holder->first);
    KLIST_REMOVE2 (first_canvas_holder, last_canvas_holder, holder);
    m_free2 (holder->colourname);
    holder->magic_number = 0;
    m_free (holder);
}   /*  End Function _vectmap_free_canvas_holder  */

static void _vectmap_process_canvas_attributes (CanvasHolder holder,
						va_list argp)
/*  [PURPOSE] This routine will set the attributes for a canvas.
    <holder> The canvas holder.
    <argp> The optional list of parameter attribute-key attribute-value
    pairs.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int att_key;
    char *colour;
    static char function_name[] = "_vectmap_process_canvas_attributes";

    while ( ( att_key = va_arg (argp, unsigned int) ) !=
	    VECTMAP_CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case VECTMAP_CANVAS_ATT_COLOURNAME:
	    colour = va_arg (argp, char *);
	    if (colour == NULL) colour = st_dup (DEFAULT_COLOUR);
	    else colour = st_dup (colour);
	    if (colour == NULL) m_abort (function_name, "colour name");
	    m_free2 (holder->colourname);
	    holder->colourname = colour;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
}   /*  End Function _vectmap_process_canvas_attributes  */

static void _vectmap_worldcanvas_precompute_func
    (KWorldCanvas canvas, int width, int height,
     CONST struct win_scale_type *win_scale, Kcolourmap cmap, flag cmap_resize,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <canvas> The world canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The canvas colourmap.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    CanvasHolder holder = *info;
    KViewableVectmapImage vfield;
    static char function_name[] = "_vectmap_worldcanvas_precompute_func";

    VERIFY_HOLDER (holder);
    if (canvas != holder->canvas)
    {
	fprintf (stderr, "Different canvas in canvas holder object\n");
	a_prog_bug (function_name);
    }
    /*  Show each active image  */
    for (vfield = holder->first;vfield; vfield = vfield->holder_chain.next)
    {
	if (vfield->active) _vectmap_precompute_image (vfield, win_scale,
							num_areas, areas);
    }
}   /*  End Function _vectmap_worldcanvas_precompute_func  */

static void _vectmap_precompute_image (KViewableVectmapImage vfield,
				       CONST struct win_scale_type *win_scale,
				       unsigned int num_areas,
				       KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine processes a precompute event for a world canvas.
    <vfield> The KViewableVectmapImage object.
    <win_scale> A pointer to the window scaling information.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    [RETURNS] Nothing.
*/
{
    KVectorField vfield = vfield->vfield;
    KwcsAstro ap;
    unsigned int count;
    CONST packet_desc *pack_desc;
    CONST uaddr *hoffsets, *voffsets;
    CONST dim_desc *hdim, *vdim;
    /*static char function_name[] = "_vectmap_precompute_image";*/

    if (vfield->num_levels < 1) return;
    pack_desc = vfield->arr_desc->packet;
    hdim = vfield->arr_desc->dimensions[vfield->hdim];
    vdim = vfield->arr_desc->dimensions[vfield->vdim];
    hoffsets = vfield->arr_desc->offsets[vfield->hdim];
    voffsets = vfield->arr_desc->offsets[vfield->vdim];
    ap = vfield->astro_projection;
    if (!vfield->world_segs_valid)
    {
	/*  Compute segments  */
	if ( !ds_vectmap (vfield->slice,
			  pack_desc->element_types[vfield->elem_index],
			  hdim, hoffsets, vdim, voffsets,
			  vfield->num_levels, vfield->vectmap_levels,
			  sizeof *vfield->worldsegs,
			  &vfield->worldsegs[0].coord_buf_size,
			  &vfield->worldsegs[0].x0, &vfield->worldsegs[0].y0,
			  &vfield->worldsegs[0].x1, &vfield->worldsegs[0].y1,
			  &vfield->worldsegs[0].num_segments) ) return;
	/*  Convert from data pixel to world co-ordinates  */
	for (count = 0; count < vfield->num_levels; ++count)
	{
	    if (ap == NULL) break;
	    wcs_astro_transform3 (ap, TRUE,
				  vfield->worldsegs[count].num_segments,
				  hdim->name,vfield->worldsegs[count].x0,FALSE,
				  vdim->name,vfield->worldsegs[count].y0,FALSE,
				  NULL, NULL, FALSE,
				  vfield->num_restrictions,
				  (CONST char **) vfield->restriction_names,
				  vfield->restriction_values);
	    wcs_astro_transform3 (ap, TRUE,
				  vfield->worldsegs[count].num_segments,
				  hdim->name,vfield->worldsegs[count].x1,FALSE,
				  vdim->name,vfield->worldsegs[count].y1,FALSE,
				  NULL, NULL, FALSE,
				  vfield->num_restrictions,
				  (CONST char **) vfield->restriction_names,
				  vfield->restriction_values);
	}
	vfield->world_segs_valid = TRUE;
    }
    if (vfield->pixel_segments_valid)
    {
	if ( (vfield->old_x_offset != win_scale->x_offset)
	     || (vfield->old_y_offset != win_scale->y_offset)
	     || (vfield->old_x_pixels != win_scale->x_pixels)
	     || (vfield->old_y_pixels != win_scale->y_pixels)
	     || (vfield->old_left_x != win_scale->left_x)
	     || (vfield->old_right_x != win_scale->right_x)
	     || (vfield->old_bottom_y != win_scale->bottom_y)
	     || (vfield->old_top_y != win_scale->top_y) )
	{   /*  The linear->pixel transformation has changed  */
	    vfield->pixel_segments_valid = FALSE;
	}
    }
    _vectmap_find_linsegs_cache (vfield, win_scale);
    /*  Now go ahead and process each level separately  */
    for (count = 0; count < vfield->num_levels; ++count)
	precompute_level (vfield, count, win_scale, num_areas, areas);
    vfield->linsegs_cache->valid = TRUE;
    vfield->pixel_segments_valid = TRUE;
    vfield->old_x_offset = win_scale->x_offset;
    vfield->old_y_offset = win_scale->y_offset;
    vfield->old_x_pixels = win_scale->x_pixels;
    vfield->old_y_pixels = win_scale->y_pixels;
    vfield->old_left_x = win_scale->left_x;
    vfield->old_right_x = win_scale->right_x;
    vfield->old_bottom_y = win_scale->bottom_y;
    vfield->old_top_y = win_scale->top_y;
}   /*  End Function _vectmap_precompute_image  */
