/*  gls_main.cc

    This code provides Glish support.

    Copyright (C) 1996-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains all routines needed for interfacing Glish to Xt.


    Written by      Richard Gooch   21-DEC-1996: Copied from glish.cc in
  MultibeamView programme.

    Updated by      Richard Gooch   26-DEC-1996: Made use of
  <ds_event_dispatch>.

    Updated by      Richard Gooch   27-DEC-1996: Completed conversion of
  FITS-style records to Karma data structure.

    Updated by      Richard Gooch   5-JAN-1997: Created
  <DL_gls_register_glish_event_handler> function.

    Updated by      Richard Gooch   12-FEB-1997: Added support for string value
  fields in record.

    Updated by      Richard Gooch   24-MAR-1997: Created
  <DL_gls_send_plain_event> function.

    Updated by      Richard Gooch   17-MAR-1999: Added some #include's.

    Updated by      Richard Gooch   2-JUL-1999: Allow records with no "array"
  field.

    Updated by      Richard Gooch   6-JUL-1999: Moved call to function
  <add_input_handlers> to after the client->NextEvent() method is called, as
  required by Glish. Consume and ignore "*rendezvous-resp*" event.

    Updated by      Richard Gooch   7-JUL-1999: Do extra processing for
  FITS-style "NAXIS*", "CTYPEn" and "BUNIT" record fields.

    Updated by      Richard Gooch   29-NOV-1999: Support vectors in
  <DL_gls_send_plain_event>.

    Updated by      Richard Gooch   9-DEC-1999: Changed <<data>> pointer type
  for <DL_gls_send_plain_event>.

    Updated by      Richard Gooch   15-JUN-2001: Changed to "glish_complex" and
  "glish_dcomplex" data types due to Yet Another Glish change.

    Updated by      Richard Gooch   23-JUL-2001: Trap "terminate" event in
  <handle_event>.

    Updated by      Richard Gooch   2-AUG-2001: Trap "exit" event instead of
  "terminate", because Glish traps "terminate" (undocumented feature).

    Updated by      Richard Gooch   5-SEP-2001: Trap NULL events, which the
  Glish client library generates in response to a "terminate" event. No longer
  trap "exit" event, since it's now redundant.

    Updated by      Richard Gooch   15-OCT-2002: Changed function casts for
  <c_register_callback> from "flag (*) (...)" to "flag (*) ()" for gcc 3.1.1.

    Updated by      Richard Gooch   20-JUL-2003: Trap for NULL string value.

    Last updated by Richard Gooch   17-APR-2004: Improved diagnostic messages.


*/
#include <stdio.h>
#include <Glish/Value.h>
#include <Glish/Client.h>
/*#include <iostream.h>*/  /*  Disabled because gcc 3.1.1 complains, and is
                               apparently not needed anyway  */
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <dlfcn.h>
#include <karma.h>
#include <karma_dmp.h>
#include <karma_dm.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>

#define EVENT_STATUS_CONSUMED      0
#define EVENT_STATUS_NOT_CONSUMED  1
#define EVENT_STATUS_EXIT          2


/*  Private structure types  */
typedef struct
{
    CONST char *name;
    CONST void *value;
} glish_info_type;

typedef struct
{
    CONST char *name;
    CONST void *data;
    unsigned int type;
    unsigned int num_elements;
    unsigned int num_axes;
    CONST int *axis_lengths;
} plain_info_type;

typedef flag (*GlishEventFuncType) (Client *client,
				    void *object, CONST char *name,
				    CONST void *value_ptr);
typedef flag (*PlainEventFuncType) (void *object, CONST char *name,
				    CONST void *data, unsigned int type,
				    unsigned int num_elements,
				    unsigned int num_axes,
				    CONST int *axis_lengths);


/*  Functions to export which need to be marked as C, to prevent C++ name
    mangling  */
EXTERN_FUNCTION (flag DL_gls_initialise, (int *argc, char **argv) );
EXTERN_FUNCTION (KCallbackFunc DL_gls_register_glish_event_handler,
		 (flag (*func) (...), CONST char *name, void *object) );
EXTERN_FUNCTION (KCallbackFunc DL_gls_register_plain_event_handler,
		 (flag (*func) (...), CONST char *name, void *object) );
EXTERN_FUNCTION (void DL_gls_send_plain_event,
		 (CONST char *name, CONST void *data,
		  unsigned int type, unsigned int num_elements,
		  unsigned int num_axes, CONST int *axis_lengths) );


/*  Private functions  */
STATIC_FUNCTION (void add_input_handlers, (Client *client) );
STATIC_FUNCTION (flag input_func, (int fd, void **info) );
STATIC_FUNCTION (unsigned int handle_event, (GlishEvent *event) );
STATIC_FUNCTION (flag glish_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag plain_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag karma_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag get_plain_info, (plain_info_type *info, Value *value) );
STATIC_FUNCTION (multi_array *get_karma_from_plain_info,
		 (plain_info_type info) );


/*  Private data  */
static Client *client = NULL;
static KCallbackList glish_callback_list = NULL;
static KCallbackList plain_callback_list = NULL;


/*  Public functions follow  */

/*UNPUBLISHED_FUNCTION*/
flag DL_gls_initialise (int *argc, char **argv)
/*  [SUMMARY] Initialise Glish support.
    <argc> As passed to the <<main>> function.
    <argv> As passed to the <<main>> function.
    [RETURNS] TRUE if running under Glish, else FALSE.
*/
{
    int argc_local = *argc;
    static char function_name[] = "DL_gls_initialise";

    if ( !dm_test_if_setup () )
    {
	fprintf (stderr, "<dm> package not setup\n");
	a_prog_bug (function_name);
    }
    client = new Client (argc_local, argv);
    *argc = argc_local;
    add_input_handlers (client);
    return (TRUE);
}   /*  End Function DL_gls_initialise  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc DL_gls_register_glish_event_handler (flag (*func) (...),
						   CONST char *name,
						   void *object)
/*  [SUMMARY] Register a Glish event handler for Glish.
    <func> The function that is called.
    <name> The name of the Glish events the handler will receive. If this is
    NULL, all events will be received.
    <object> The object.
    [RETURNS] A KCallbackFunc.
*/
{
    void *info = (void *) func;

    return c_register_callback (&glish_callback_list,
				( flag (*) () ) glish_event_func,
				object, info, FALSE, (void *) name, FALSE,
				TRUE);
}   /*  End Function DL_gls_register_glish_event_handler  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc DL_gls_register_plain_event_handler (flag (*func) (...),
						   CONST char *name,
						   void *object)
/*  [SUMMARY] Register a plain event handler for Glish.
    <func> The function that is called.
    <name> The name of the Glish events the handler will receive. If this is
    NULL, all events will be received.
    <object> The object.
    [RETURNS] A KCallbackFunc.
*/
{
    void *info = (void *) func;

    return c_register_callback (&plain_callback_list,
				( flag (*) () ) plain_event_func,
				object, info, FALSE, (void *) name, FALSE,
				TRUE);
}   /*  End Function DL_gls_register_plain_event_handler  */

/*UNPUBLISHED_FUNCTION*/
void DL_gls_send_plain_event (CONST char *name, CONST void *data,
			      unsigned int type, unsigned int num_elements,
			      unsigned int num_axes, CONST int *axis_lengths)
/*  [SUMMARY] Send a plain Glish event.
    <name> The event name.
    <data> The data to send (this is a vector).
    <type> The type of the data.
    <num_elements> The number of elements in the vector.
    <num_axes> The number of axes. This may be 0.
    <axis_lengths> The length of each axis.
    [RETURNS] Nothing.
*/
{
    Value *value;
    static char function_name[] = "DL_gls_send_plain_event";

    if (num_elements < 1)
    {
	client->PostEvent (name, (const Value *) NULL);
	return;
    }
    if (num_elements < 2)
    {
	glish_complex zf;
	glish_dcomplex zd;

	switch (type)
	{
	  case K_BYTE:
	    value = new Value (*(byte *) data);
	    break;
	  case K_SHORT:
	    value = new Value (*(short *) data);
	    break;
	  case K_INT:
	    value = new Value (*(int *) data);
	    break;
	  case K_FLOAT:
	    value = new Value (*(float *) data);
	    break;
	  case K_DOUBLE:
	    value = new Value (*(double *) data);
	    break;
	  case K_COMPLEX:
	    zf.r = *(float *) data;
	    zf.i = *( (float *) data + 1);
	    value = new Value (zf);
	    break;
	  case K_DCOMPLEX:
	    zd.r = *(double *) data;
	    zd.i = *( (double *) data + 1);
	    value = new Value (zd);
	    break;
	  case K_VSTRING:
	    value = new Value (*(char **) data);
	    break;
	  default:
	    fprintf (stderr, "%s: type: %u not supported\n",
		     function_name, type);
	    return;
	    /*break;*/
	}
	client->PostEvent (name, value);
	delete value;
	return;
    }
    if (num_axes < 1)
    {
	unsigned int count;
	glish_complex *zf;
	glish_dcomplex *zd;

	switch (type)
	{
	  case K_BYTE:
	    value = new Value ((byte *) data, num_elements, PRESERVE_ARRAY);
	    break;
	  case K_SHORT:
	    value = new Value ((short *) data, num_elements, PRESERVE_ARRAY);
	    break;
	  case K_INT:
	    value = new Value ((int *) data, num_elements, PRESERVE_ARRAY);
	    break;
	  case K_FLOAT:
	    value = new Value ((float *) data, num_elements, PRESERVE_ARRAY);
	    break;
	  case K_DOUBLE:
	    value = new Value ((double *) data, num_elements, PRESERVE_ARRAY);
	    break;
	  case K_COMPLEX:
	    zf = new glish_complex[num_elements];
	    for (count = 0; count < num_elements; ++count)
	    {
		zf[count].r = *( (float *) data + count * 2 );
		zf[count].i = *( (float *) data + count * 2 + 1);
	    }
	    value = new Value (zf, num_elements, TAKE_OVER_ARRAY);
	    break;
	  case K_DCOMPLEX:
	    zd = new glish_dcomplex[num_elements];
	    for (count = 0; count < num_elements; ++count)
	    {
		zd[count].r = *( (double *) data + count * 2 );
		zd[count].i = *( (double *) data + count * 2 + 1);
	    }
	    value = new Value (zd, num_elements, TAKE_OVER_ARRAY);
	    break;
	  case K_VSTRING:
	    value = new Value ((const char **) data, num_elements,
			       PRESERVE_ARRAY);
	    break;
	  default:
	    fprintf (stderr, "%s: type: %u not supported\n",
		     function_name, type);
	    return;
	    /*break;*/
	}
	client->PostEvent (name, value);
	delete value;
	return;
    }
    fprintf (stderr, "%s: n-dimensional data not yet supported\n",
	     function_name);
}   /*  End Function DL_gls_send_plain_event  */


/*  Private functions follow  */

static void add_input_handlers (Client *client)
/*  [SUMMARY] Add event handlers for Glish.
    <client> The Glish client.
    [RETURNS] Nothing.
*/
{
    int num_set, count;
    fd_set new_fdset;
    static fd_set glish_fdset;
    static flag first_time = TRUE;
    static char function_name[] = "_gls_add_input_handlers";

    if (first_time)
    {
	first_time = FALSE;
	FD_ZERO (&glish_fdset);
    }
    new_fdset = glish_fdset;
    num_set = client->AddInputMask (&new_fdset);
    for (count = 0; (num_set > 0) && (count < FD_SETSIZE); ++count)
    {
	if ( !FD_ISSET (count, &new_fdset) ) continue;
	/*  Ignore if descriptor already registered  */
	if ( FD_ISSET (count, &glish_fdset) ) continue;
	fprintf (stderr, "Managing Glish FD: %d\n", count);
	if ( !dm_manage (count, client, FALSE,
			 KARMA_FD_EVENT_INPUT, input_func,
			 KARMA_FD_EVENT_END) )
	{
	    fprintf (stderr, "Error managing descriptor\n");
	    a_prog_bug (function_name);
	}
    }
    glish_fdset = new_fdset;
}   /*  End Function add_input_handlers  */

static flag input_func (int fd, void **info)
/*  [SUMMARY] Input event callback.
    [PURPOSE] This routine is called when new input occurs on a descriptor.
    <fd> The file descriptor.
    <info> A pointer to the arbitrary information pointer. The pointer
    written here will persist until the descriptor is unmanaged (or a
    subsequent callback routine changes it).
    [RETURNS] TRUE if the descriptor is to remain managed and open, else it
    returns FALSE (indicating that the descriptor is to be unmanaged and
    closed). This routine MUST NOT unmanage or close the descriptor.
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    fd_set event_fdset;
    Client *client = (Client *) *info;
    GlishEvent *event;

    FD_ZERO (&event_fdset);
    FD_SET (fd, &event_fdset);
    switch ( handle_event ( client->NextEvent (&event_fdset) ) )
    {
      case EVENT_STATUS_CONSUMED:
	break;
      case EVENT_STATUS_NOT_CONSUMED:
	client->Unrecognized ();
	break;
      case EVENT_STATUS_EXIT:
	delete client;
	exit (RV_OK);
	/*break;*/
    }
    add_input_handlers (client);
    return (TRUE);
}   /*  End Function input_func  */

static unsigned int handle_event (GlishEvent *event)
/*  [SUMMARY] Handle a Glish event.
    <event> The Glish event.
    [RETURNS] A status code.
*/
{
    unsigned int num_fields, count, ktype;
    glish_info_type glish_info;
    plain_info_type plain_info;
    multi_array *multi_desc;
    array_desc *arr_desc;
    Value *main_value, *array_value, *field_value;
    const char *main_name;
    double value[2];
    static char function_name[] = "_gls_handle_event";

    if (event == NULL) return (EVENT_STATUS_EXIT);
    main_value = event->Val ();
    main_name = event->Name ();
    glish_info.name = main_name;
    /*  Special events to consume and ignore  */
    if (strcmp (main_name, "*rendezvous-resp*") == 0)
	return (EVENT_STATUS_CONSUMED);
    /*  Generic events: pass to callbacks  */
    glish_info.value = main_value;
    if ( c_call_callbacks (glish_callback_list, &glish_info) )
	return (EVENT_STATUS_CONSUMED);
    if ( main_value->IsNumeric () || (main_value->Type () == TYPE_STRING) )
    {
	/*  Not a record, just a plain value  */
	if ( !get_plain_info (&plain_info, main_value) )
	    return (EVENT_STATUS_NOT_CONSUMED);
	plain_info.name = main_name;
	if ( c_call_callbacks (plain_callback_list, &plain_info) )
	    return (EVENT_STATUS_CONSUMED);
	if ( ( multi_desc = get_karma_from_plain_info (plain_info) ) == NULL )
	{
	    fprintf (stderr, "%s: plain event \"%s\" not consumed\n",
		     function_name, main_name);
	    return (EVENT_STATUS_NOT_CONSUMED);
	}
	if ( ds_event_dispatch (multi_desc, "GLISH", main_name) )
	{
	    ds_dealloc_multi (multi_desc);
	    return (EVENT_STATUS_CONSUMED);
	}
	ds_dealloc_multi (multi_desc);
	fprintf (stderr, "%s: simple Karma event \"%s\" not consumed\n",
		 function_name, main_name);
	return (EVENT_STATUS_NOT_CONSUMED);
    }
    if (main_value->Type () != TYPE_RECORD)
    {
	fprintf (stderr, "%s: value is not a record type\n", function_name);
	return (EVENT_STATUS_NOT_CONSUMED);
    }
    /*  Decode the record  */
    if ( ( array_value = main_value->Field ("array") ) == NULL )
    {
	if ( ( multi_desc = ds_alloc_multi (1) ) == NULL )
	{
	    m_error_notify (function_name, "multi_array");
	    return (EVENT_STATUS_NOT_CONSUMED);
	}
	if ( ( multi_desc->headers[0] = ds_alloc_packet_desc (1) ) == NULL )
	{
	    m_error_notify (function_name, "packet descriptor");
	    ds_dealloc_multi (multi_desc);
	    return (EVENT_STATUS_NOT_CONSUMED);
	}
	--multi_desc->headers[0]->num_elements;
	m_free (multi_desc->headers[0]->element_types);
	multi_desc->headers[0]->element_types = NULL;
	m_free (multi_desc->headers[0]->element_desc);
	multi_desc->headers[0]->element_desc = NULL;
	arr_desc = NULL;
    }
    else
    {
	if ( !get_plain_info (&plain_info, array_value) )
	    return (EVENT_STATUS_NOT_CONSUMED);
	if ( ( multi_desc = get_karma_from_plain_info (plain_info) ) == NULL )
	{
	    fprintf (stderr, "%s: no Karma structure for \"%s\" event\n",
		     function_name, main_name);
	    return (EVENT_STATUS_NOT_CONSUMED);
	}
	arr_desc = (array_desc *) multi_desc->headers[0]->element_desc[0];
    }
    num_fields = main_value->Length ();
    for (count = 1; count <= num_fields; ++count)
    {
	int ival, dim;
	const char *field_name, *string_value;

	if ( ( field_value = main_value->NthField (count) ) == NULL )
	{
	    fprintf (stderr, "%s: error getting field %u\n",
		     function_name, count);
	    ds_dealloc_multi (multi_desc);
	    return (EVENT_STATUS_NOT_CONSUMED);
	}
	field_name = main_value->NthFieldName (count);
	if (strcmp (field_name, "array") == 0) continue;
	value[1] = 0.0;
	switch ( field_value->Type () )
	{
	  case TYPE_BOOL:
	    fprintf (stderr, "%s: boolean type not supported\n",function_name);
	    ds_dealloc_multi (multi_desc);
	    return (EVENT_STATUS_NOT_CONSUMED);
	    /*break;*/
	  case TYPE_INT:
	    ktype = K_INT;
	    ival = *field_value->IntPtr ();
	    value[0] = ival;
	    if (arr_desc == NULL) break;
	    /*  Extra processing for FITS-style "NAXIS*" keywords  */
	    if (strcmp (field_name, "NAXIS") == 0)
	    {
		if (arr_desc->num_dimensions != ival)
		{
		    arr_desc = NULL;
		    break;
		}
	    }
	    else if (strncmp (field_name, "NAXIS", 5) == 0)
	    {
		dim = atoi (field_name + 5);
		if ( (dim < 1) || (dim > arr_desc->num_dimensions) )
		{
		    arr_desc = NULL;
		    break;
		}
		dim = arr_desc->num_dimensions - dim;
		if (arr_desc->dimensions[dim]->length != ival) arr_desc = NULL;
	    }
	    break;
	  case TYPE_FLOAT:
	    ktype = K_FLOAT;
	    value[0] = *field_value->FloatPtr ();
	    break;
	  case TYPE_DOUBLE:
	    ktype = K_DOUBLE;
	    value[0] = *field_value->DoublePtr ();
	    break;
	  case TYPE_COMPLEX:
	    ktype = K_COMPLEX;
	    value[0] = field_value->ComplexPtr ()->r;
	    value[1] = field_value->ComplexPtr ()->i;
	    break;
	  case TYPE_DCOMPLEX:
	    ktype = K_DCOMPLEX;
	    value[0] = field_value->DcomplexPtr ()->r;
	    value[1] = field_value->DcomplexPtr ()->i;
	    break;
	  case TYPE_STRING:
	    ktype = K_VSTRING;
	    string_value = field_value->StringVal (' ');
	    if (!string_value)
	    {
		fprintf (stderr, "%s: NULL string for field: \"%s\" ignored\n",
			 function_name, field_name);
		break;
	    }
	    if (arr_desc == NULL) break;
	    /*  Extra processing for FITS-style "CTYPEn" keywords  */
	    if (strncmp (field_name, "CTYPE", 5) == 0)
	    {
		char *str;

		dim = atoi (field_name + 5);
		if ( (dim < 1) || (dim > arr_desc->num_dimensions) ) break;
		dim = arr_desc->num_dimensions - dim;
		if ( ( str = st_dup (string_value) ) == NULL )
		{
		    m_error_notify (function_name, "CTYPEn string value");
		    break;
		}
		m_free (arr_desc->dimensions[dim]->name);
		arr_desc->dimensions[dim]->name = str;
	    }
	    else if (strcmp (field_name, "BUNIT") == 0)
	    {
		char *str;

		if ( ( str = st_dup (string_value) ) == NULL )
		{
		    m_error_notify (function_name, "BUNIT string value");
		    break;
		}
		m_free (arr_desc->packet->element_desc[0]);
		arr_desc->packet->element_desc[0] = str;
	    }
	    break;
	  case TYPE_RECORD:
	    fprintf (stderr, "%s: record type not supported, name: \"%s\"\n",
		     function_name, field_name);
	    return (EVENT_STATUS_NOT_CONSUMED);
	    /*break;*/
	  default:
	    fprintf (stderr, "%s: unsupported value type, name: \"%s\"\n",
		     function_name, field_name);
	    ds_dealloc_multi (multi_desc);
	    return (EVENT_STATUS_NOT_CONSUMED);
	    /*break;*/
	}
	if (field_value->Length () != 1)
	{
	    fprintf ( stderr, "%s: field: \"%s\" has length: %u, not 1\n",
		      function_name, field_name, field_value->Length () );
	    ds_dealloc_multi (multi_desc);
	    return (EVENT_STATUS_NOT_CONSUMED);
	}
	if ( ds_element_is_atomic (ktype) )
	{
	    if ( !ds_put_unique_named_value (multi_desc->headers[0],
					     multi_desc->data,
					     field_name, ktype, value, FALSE) )
	    {
		fprintf (stderr, "%s: error adding field: \"%s\"\n",
			 function_name, field_name);
		ds_dealloc_multi (multi_desc);
		return (EVENT_STATUS_NOT_CONSUMED);
	    }
	}
	else if (ktype == K_VSTRING)
	{

	    if ( !ds_put_unique_named_string (multi_desc->headers[0],
					      multi_desc->data,
					      field_name, string_value,FALSE) )
	    {
		fprintf (stderr, "%s: error adding field: \"%s\"\n",
			 function_name, field_name);
		ds_dealloc_multi (multi_desc);
		return (EVENT_STATUS_NOT_CONSUMED);
	    }
	}
    }
    if ( ds_event_dispatch (multi_desc, "GLISH", main_name) )
    {
	ds_dealloc_multi (multi_desc);
	return (EVENT_STATUS_CONSUMED);
    }
    ds_dealloc_multi (multi_desc);
    fprintf (stderr, "%s: Karma event \"%s\" not consumed\n",
	     function_name, main_name);
    return (EVENT_STATUS_NOT_CONSUMED);
}   /*  End Function handle_event  */

static flag glish_event_func (void *object, void *client1_data,
			      void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    glish_info_type *info = (glish_info_type *) call_data;
    CONST char *name = (CONST char *) client2_data;
    GlishEventFuncType func = (GlishEventFuncType) client1_data;

    if ( name && (strcmp (name, info->name) != 0) ) return (FALSE);
    return (*func) (client, object, info->name, info->value);
}   /*  End Function glish_event_func  */

static flag plain_event_func (void *object, void *client1_data,
			      void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    plain_info_type *info = (plain_info_type *) call_data;
    CONST char *name = (CONST char *) client2_data;
    PlainEventFuncType func = (PlainEventFuncType) client1_data;

    if ( name && (strcmp (name, info->name) != 0) ) return (FALSE);
    return (*func) (object, info->name, info->data, info->type,
		    info->num_elements, info->num_axes, info->axis_lengths);
}   /*  End Function plain_event_func  */

static flag get_plain_info (plain_info_type *info, Value *value)
/*  [SUMMARY] Get the plain_info from a Glish value.
    <info> The plain info is written here.
    <value> The Glish value.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int ktype;
    unsigned int num_axes = 0;
    const int *axis_lengths = NULL;
    const void *data = NULL;
    const Value *shape_value;
    static char function_name[] = "_gls_get_plain_info";

    shape_value = value->HasAttribute ("shape");
    if (shape_value)
    {
	if (shape_value->Type () != TYPE_INT)
	{
	    fprintf (stderr, "%s: shape value is not type integer\n",
		     function_name);
	    return (FALSE);
	}
	num_axes = shape_value->Length ();
	if (num_axes < 1)
	{
	    fprintf (stderr, "%s: less than one element in shape value\n",
		     function_name);
	    return (FALSE);
	}
	axis_lengths = shape_value->IntPtr ();
    }
    /*  Switch on it's type  */
    switch ( value->Type () )
    {
      case TYPE_BOOL:
	fprintf (stderr, "%s: boolean type not supported\n", function_name);
	return (FALSE);
	/*break;*/
      case TYPE_INT:
	ktype = K_INT;
	data = value->IntPtr ();
	break;
      case TYPE_FLOAT:
	ktype = K_FLOAT;
	data = value->FloatPtr ();
	break;
      case TYPE_DOUBLE:
	ktype = K_DOUBLE;
	data = value->DoublePtr ();
	break;
      case TYPE_COMPLEX:
	ktype = K_COMPLEX;
	data = value->ComplexPtr ();
	break;
      case TYPE_DCOMPLEX:
	ktype = K_DCOMPLEX;
	data = value->DcomplexPtr ();
	break;
      case TYPE_STRING:
	if (value->Length () > 1)
	{
	    fprintf (stderr, "%s: vector of strings not supported\n",
		     function_name);
	    return (FALSE);
	    /*break;*/
	}
	ktype = K_VSTRING;
	data = value->StringPtr ();
	break;
      case TYPE_RECORD:
	fprintf (stderr, "%s: record type not supported\n", function_name);
	return (FALSE);
	/*break;*/
      default:
	fprintf (stderr, "%s: unsupported value type\n", function_name);
	return (FALSE);
	/*break;*/
    }
    if (data == NULL)
    {
	fprintf (stderr, "%s: no data!\n", function_name);
	return (FALSE);
    }
    /*  Not a record: handle now  */
    info->data = (CONST char *) data;
    info->type = ktype;
    info->num_elements = value->Length ();
    info->num_axes = num_axes;
    info->axis_lengths = axis_lengths;
    return (TRUE);
}   /*  End Function get_plain_info  */

static multi_array *get_karma_from_plain_info (plain_info_type info)
/*  [SUMMARY] Get Karma data structure from plain info.
    <info> The plain_info.
    [RETURNS] A multi_array descriptor on success, else NULL.
*/
{
    unsigned int count;
    multi_array *multi_desc;
    char *array;
    uaddr *dim_lengths;
    extern char host_type_sizes[NUMTYPES];
    static char function_name[] = "_gls_get_karma_from_plain_info";

    if (info.num_elements < 2) return (NULL);
    if (info.num_axes < 1)
    {
	if ( ( dim_lengths =
	       (uaddr *) m_alloc (sizeof *dim_lengths) ) == NULL )
	{
	    m_error_notify (function_name, "array of dimension lengths");
	    return (NULL);
	}
	dim_lengths[0] = info.num_elements;
	info.num_axes = 1;
    }
    else
    {
	if ( ( dim_lengths =
	       (uaddr *) m_alloc (sizeof *dim_lengths * info.num_axes) )
	     == NULL )
	{
	    m_error_notify (function_name, "array of dimension lengths");
	    return (NULL);
	}
	for (count = 0; count < info.num_axes; ++count)
	    dim_lengths[info.num_axes - count - 1] = info.axis_lengths[count];
    }
    array = ds_easy_alloc_array (&multi_desc, info.num_axes,
				 dim_lengths, NULL, NULL, NULL,
				 info.type, NULL);
    m_free (dim_lengths);
    if (array == NULL)
    {
	m_error_notify (function_name, "Karma data structure");
	return (NULL);
    }
    m_copy (array, info.data, info.num_elements * host_type_sizes[info.type]);
    return (multi_desc);
}   /*  End Function get_karma_from_plain_info  */
