/*  raw_grey2karma.c

    Source file for  raw_grey2karma  (convert raw data to Karma format).

    Copyright (C) 1993-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will generate a multi-dimensional array of data from a
    raw data file containing greyscale (or intensity) values and will write the
    data to a Karma data file.
    The raw data file must be in host natural format.


    Written by      Richard Gooch   1-NOV-1992

    Updated by      Richard Gooch   4-DEC-1992

    Updated by      Richard Gooch   13-OCT-1993: Moved  main  into this file.

    Updated by      Richard Gooch   22-MAY-1994: Changed over to  panel_
  package.

    Updated by      Richard Gooch   30-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   3-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   19-MAR-2002: Switched to Channels.

    Last updated by Richard Gooch   6-SEP-2003: Added support for network byte
  ordering format.


*/
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_dsrw.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_m.h>

#define VERSION "1.4"

#define MAX_DIMENSIONS (unsigned int) 100


STATIC_FUNCTION (flag raw_grey2karma, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void generate_file,
		 (char *inp_filename, char *arrayname,
		  unsigned int elem_type) );
STATIC_FUNCTION (flag read_values,
		 (Channel channel, char *array, unsigned int type,
		  unsigned long num_values, unsigned int format) );


/*  Private data  */
static unsigned int num_dimensions = 0;
static char *names[MAX_DIMENSIONS];
static unsigned long lengths[MAX_DIMENSIONS];
static double minima[MAX_DIMENSIONS];
static double maxima[MAX_DIMENSIONS];
static unsigned int elem_type = K_FLOAT;
static char *elem_name = "Intensity";
static flag invert_flag = TRUE;

#define FORMAT_NATIVE       0
#define FORMAT_NETWORK      1
#define NUM_FORMAT_CHOICES  2
static char *format_option_alternatives[NUM_FORMAT_CHOICES] =
{
    "native",
    "network",
};
static unsigned int format = FORMAT_NATIVE;


int main (int argc, char **argv)
{
    KControlPanel panel, group;
    extern char *data_type_names[];
    static char function_name[] = "main";

    if ( ( group = panel_create_group () ) == NULL )
    {
	m_abort (function_name, "group panel");
    }
    panel_add_item (group, "maximum", "", K_DOUBLE, maxima, PIA_END);
    panel_add_item (group, "minimum", "", K_DOUBLE, minima, PIA_END);
    panel_add_item (group, "length", "", K_ULONG, lengths, PIA_END);
    panel_add_item (group, "name", "", K_VSTRING, names, PIA_END);
    if ( ( panel = panel_create (FALSE) ) == NULL )
	m_abort (function_name, "control panel");
    panel_add_item (panel, "dimensions", "", PIT_GROUP, group,
		    PIA_ARRAY_MAX_LENGTH, MAX_DIMENSIONS,
		    PIA_ARRAY_LENGTH, &num_dimensions,
		    PIA_END);
    panel_add_item (panel, "elem_name", "", K_VSTRING, &elem_name, PIA_END);
    panel_add_item (panel, "elem_type", "", PIT_CHOICE_INDEX, &elem_type,
		    PIA_NUM_CHOICE_STRINGS, NUMTYPES,
		    PIA_CHOICE_STRINGS, data_type_names,
		    PIA_END);
    panel_add_item (panel, "format", "", PIT_CHOICE_INDEX, &format,
		    PIA_NUM_CHOICE_STRINGS, NUM_FORMAT_CHOICES,
		    PIA_CHOICE_STRINGS, format_option_alternatives,
		    PIA_END);
    panel_add_item (panel, "invert_image", "flag", PIT_FLAG, &invert_flag,
		    PIA_END);
    panel_push_onto_stack (panel);
    module_run (argc, argv, "raw_grey2karma", VERSION, raw_grey2karma, -1, 0,
		FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag raw_grey2karma (CONST char *command, FILE *fp)
{
    char *arrayfile;
    char *input_filename;
    extern unsigned int elem_type;
    extern unsigned int num_dimensions;

    if (num_dimensions < 1)
    {
	fprintf (fp, "Must have at least one dimension to create\n");
	return (TRUE);
    }
    if (ds_element_is_atomic ( (unsigned int) elem_type ) == FALSE)
    {
        fprintf (fp, "elem_type must be atomic: defaulting to K_FLOAT\n");
        elem_type = K_FLOAT;
        return (TRUE);
    }
    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( input_filename = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting input filename\n");
	    return (TRUE);
	}
	if ( ( arrayfile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting arrayfile name\n");
	    m_free (input_filename);
	    return (TRUE);
	}
	generate_file (input_filename, arrayfile, elem_type);
	m_free (input_filename);
	m_free (arrayfile);
    }
    return (TRUE);
}   /*  End Function raw_grey2karma  */

static void generate_file (char *inp_filename, char *arrayname,
			   unsigned int elem_type)
/*  This routine will generate a Karma arrayfile of the new format with a
    multi-dimensional array of a single atomic element. The data will be
    converted from the input file.
    The name of the input file must be pointed to by  inp_filename  .
    the name of the Karma arrayfile must be pointed to by  arrayname  .
    The type of the data elements must be in  elem_type  .
    The routine returns nothing.
*/
{
    Channel inp_ch;
    unsigned int array_size;
    unsigned int block_size;
    unsigned int block_count;
    unsigned int dim_count;
    char *array;
    multi_array *multi_desc;
    extern flag invert_flag;
    extern unsigned int num_dimensions;
    extern unsigned long lengths[MAX_DIMENSIONS];
    extern double minima[MAX_DIMENSIONS];
    extern double maxima[MAX_DIMENSIONS];
    extern char *names[MAX_DIMENSIONS];
    extern char *elem_name;
    extern char host_type_sizes[NUMTYPES];
    /*static char function_name[] = "generate_file";*/

    /*  Determine array size (in co-ordinates)  */
    array_size = 1;
    for (dim_count = 0; dim_count < num_dimensions; ++dim_count)
    {
	if (lengths[dim_count] < 1)
	{
	    fprintf (stderr,
		     "Dimension: %u has length: %lu: must be greater than 0\n",
		     dim_count, lengths[dim_count]);
	    return;
	}
	array_size *= (unsigned int) lengths[dim_count];
    }
    /*  Try to open input file  */
    if ( ( inp_ch = ch_open_file (inp_filename, "rz") ) == NULL )
    {
	fprintf (stderr, "\nError opening file: \"%s\"\t%s",
		 inp_filename, ERRSTRING);
	return;
    }
    if ( ( array = ds_easy_alloc_array (&multi_desc,
					(unsigned int) num_dimensions,
					lengths, minima, maxima,
					(CONST char **) names,
					elem_type, elem_name) )
	== NULL )
    {
	fprintf (stderr,"NULL return value from function: easy_alloc_array\n");
	ch_close (inp_ch);
	return;
    }
    /*  Determine if inversion of image is needed  */
    if (invert_flag)
    {
	/*  Must invert image  */
	block_size = array_size / (unsigned int) lengths[0];
	for (block_count = 0; block_count < lengths[0]; ++block_count)
	{
	    errno = 0;
	    if ( !read_values (inp_ch, array + host_type_sizes[elem_type] *
			       block_size * (lengths[0] - block_count - 1),
			       elem_type, block_size, format) )
	    {
		ds_dealloc_multi (multi_desc);
		ch_close (inp_ch);
		return;
	    }
	}
    }
    else
    {
	/*  Do a block read of entire image  */
	errno = 0;
	if ( !read_values (inp_ch, array, elem_type, array_size, format) )
	{
	    fprintf (stderr, "Error reading file: \"%s\"\t%s\n",
		     inp_filename, ERRSTRING);
	    ds_dealloc_multi (multi_desc);
	    ch_close (inp_ch);
	    return;
	}
    }
    ch_close (inp_ch);
    /*  Write to arrayfile  */
    if ( !dsxfr_put_multi (arrayname, multi_desc) )
    {
	fprintf (stderr, "Error putting array\n");
    }
    /*  Write comments to new array  */
#if 0
    ez_description (comp_array, vble);
    sprintf (txt, "Read raw data file: \"%s\"", inp_filename);
    composite_descr (txt, comp_array);
#endif
    ds_dealloc_multi (multi_desc);
}   /*  End Function generate_file  */

static flag read_values (Channel channel, char *array, unsigned int type,
			 unsigned long num_values, unsigned int format)
/*  [SUMMARY] Read contiguous values.
    <channel> The channel to read from.
    <array> The array to write the data to.
    <type> The type of the data.
    <num_values> The number of values to read.
    <format> The format of the input data.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    packet_desc pack_desc;
    extern char host_type_sizes[NUMTYPES];

    switch (format)
    {
      case FORMAT_NATIVE:
	if (ch_read (channel, array, host_type_sizes[type] * num_values)
	    < host_type_sizes[type] * num_values)
	{
	    fprintf (stderr, "Error reading\t%s\n", ERRSTRING);
	    return (FALSE);
	}
	break;
      case FORMAT_NETWORK:
	m_clear (&pack_desc, sizeof pack_desc);
	pack_desc.num_elements = 1;
	pack_desc.element_types = &type;
	if ( !dsrw_read_packets (channel, &pack_desc, array, num_values) )
	    return (FALSE);
	break;
    }
    return (TRUE);
}   /*  End Function read_values  */
