/*  main_xt.c

    Main file for  khuei  (X11 24bit Huei/Intensity display tool for Karma).

    Copyright (C) 1996-2001  Simone Magri
    Incorporated into Karma by permission.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Simone Magri may be reached by email at  smagri@atnf.csiro.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/


/* khuei allows the user to load two files; all the values in each
  file are scaled to the range 0-255; these values are merged into one
  array and used as an index to a colourmap.  The colourmap of 2^16
  values is generated by the user given values of Hue, Intensity, and
  Saturation.
   
  The program by DEFAULT assumes that file1 data contains amplitude
  values and file2 phase values.  However, (real, imaginary) raw data
  files are also supported.  For each of these the file order can also
  be reversed, giving two other modes of operation.
  
  For any given mode you are able to reset the slider values to their
  default values.  */




/* 
   MODIFICATIONS 

   19may96 SM   Use viewable images to create and manage the colour wedge.
   06jun96 SM   Use viewable images to create and manage the image canvas.
   10oct96 REG  Added print button.
   11oct96 REG  Restored forcing of 1:1 aspect ratio for main canvas.
   11oct96 REG  Only create ViewableImage for new 16 bit image.
   12jan97 REG  Made use of new XkwNtrapDirectoryDatasets resource for
                FilewinWidgetClass.
   21jun97 REG  Added axis labelling and intensity scaling support.
                Restructured file loading and combining.
   22jun97 REG  Made use of <xtmisc_init_app_initialise> function.
   31oct97 REG  Took account of change to to <XkwDataclipNewArray>.
   22feb98 REG  Switched to <XkwDressingControlAddCanvas>.
   01mar98 REG  Made use of new filewinWidget resource.
   17sep98 REG  Replaced DressingControl widget with OverlayMenu widget.
   21jul99 REG  Fixed error message when TrueColour not found.
   30aug99 REG  Added dirname font resource.
   15oct99 REG  Took acount of resource name change for MultiCanvas widget.
   11may01 REG  Changed dirname font resource.
   03oct01 REG  Added translations for Filewin.
   22dec02 REG  Added #include <stdlib.h>
   03oct03 REG  Took account of interface change for Dataclip widget.

*/

/*  TODO
    Do not create new ViewableImage for every
*/ 


/* This module contains Xt interface code */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <sys/time.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <karma.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_contour.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_kcmap.h>
#include <karma_dir.h>
#include <karma_col.h>
#include <karma_im.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/MultiCanvas.h>
#include <Xkw/Value.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/Postscript.h>
#include <Xkw/OverlayMenu.h>
#include <Xkw/Dataclip.h>


/* slider initializations */
#define SLIDER_MIN 0
#define SLIDER_MAX 100
#define SLIDER_HUE_MAX 99

/* identifiers for which widget has just been actived */
#define FILE1 1
#define FILE2 2
#define SHSLD 3
#define EHSLD 4
#define MISLD 5
#define SSLD 6

/* Number of Raw Data Forms the  program can deal with; that is, (File1,File2)
   can be one of RDF_NUM forms and be mapped to (Intensity, Hue) */
#define RDF_NUM 4

/* symbolic constants to identify Raw Data Form of FILE1_FILE2 ( That are
   mapped to INT_HUE respectively */

#define AMP_PHS 0
#define PHS_AMP 1
#define RE_IM 2
#define IM_RE 3
	
/* Private functions */


STATIC_FUNCTION (void reset_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (void rdf_cbk, (Widget w, XtPointer client_data,
				XtPointer call_data) );
STATIC_FUNCTION (void quit_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void slider_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void data_to_huei_pixmap, (void) );
STATIC_FUNCTION (void map_re_im_2_amp_phs,
		 (iarray re_chan, double re_min, double re_max,
		  iarray im_chan, double im_min, double im_max) );
STATIC_FUNCTION (void re_im_2_amp_phs,
		 (double re, double im, double *amp, double *phs));
STATIC_FUNCTION (void data_to_16bit,
		 (iarray amp_data, double amp_min, double amp_max,
		  iarray phs_data, double phs_min, double phs_max) );
STATIC_FUNCTION (void compute_colourmap, (void) );
STATIC_FUNCTION (void disp_cwedge_init_imag, (void) );
STATIC_FUNCTION (flag data_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag test_both_good, (iarray arr1, iarray arr2) );
STATIC_FUNCTION (void region1_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void region2_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );

/* STATIC_FUNCTION is portable across different platforms */



/*  Private data  */

String fallback_resources[] =
{
    "Khuei.geometry:				     -0+0",
    "Khuei*pseudoColourCanvas*background:            black",
    "Khuei*directColourCanvas*background:            black",
    "Khuei*trueColourCanvas*background:              black",
    "Khuei*pseudoColourStereoCanvas*background:      black",
    "Khuei*directColourStereoCanvas*background:      black",
    "Khuei*trueColourStereoCanvas*background:        black",
    "Khuei*AnimateControl*positionCanvas.background: black",
    "Khuei*pseudoColourCanvas*foreground:            white",
    "Khuei*Command*background:                       grey70",
    "Khuei*Repeater*background:                      grey70",
    "Khuei*Ktoggle*background:                       grey80",
    "Khuei*closeButton*background:                   grey90",
    "Khuei*ChoiceMenu.background:                    turquoise",
    "Khuei*ExclusiveMenu.background:                 turquoise",
    "Khuei*Value*background:                         #d0a0a0",
    "Khuei*topForm*quit*background:                  orange",
    "Khuei*ScrollableList.highlightPixel:            Sea Green",
    "Khuei*SimpleSlider.foreground:                  Sea Green",
    "Khuei*SimpleProgressMeter.background:           turquoise3",
    "Khuei*SimpleProgressMeter.foreground:           grey60",
    "Khuei*ImageDisplay*trackLabel0*font:            8x13bold",
    "Khuei*ImageDisplay*trackLabel1*font:            8x13bold",
    "Khuei*ImageDisplay*trackLabel2*font:            8x13bold",
    "Khuei*SimpleMenu*font:                          10x20",
    "Khuei*ImageDisplay.Command.font:                7x13bold",
    "Khuei*ImageDisplay.applicationBox.Command.font: 7x13bold",
    "Khuei*ImageDisplay.Ktoggle.font:                7x13bold",
    "Khuei*ImageDisplay.ChoiceMenu.font:             7x13bold",
    "Khuei*ImageDisplay.exportMenu.menuButton.font:  7x13bold",
    "Khuei*ImageDisplay*zoomMenu*Unzoom*foreground:  red",
    "Khuei*SimpleSlider.borderWidth:                 0",
    "Khuei*Filewin.dirname.textSink.font:            6x13",
    "Khuei*font:                                     9x15bold",
    "Khuei*borderColor:                              black",
    "Khuei*background:                               MediumAquamarine",
    "Khuei*foreground:                               black",
    NULL
};

/* define/declare global widgets */
static Widget main_shell = NULL;
static Widget start_hue_sld,end_hue_sld,min_int_sld,sat_sld;
static Widget overlay_menu = NULL;
static Widget izoomwinpopup1 = NULL;
static Widget izoomwinpopup2 = NULL;


/* implicit external definitions (alloc storage)
   and declarations (nature of variable stated)*/
/* statics only available in this file/local */
/* note: explicit dec with def in fn, like:*/
/* extern int curr_file_selected; */

/* Set-up an indicator for when: the first pair of raw data files,
   having an equal number of data values, have been loaded; and all
   iarrays have been created once.  This is useful in a number of
   circumstances of program function */
flag done_once = FALSE;


iarray raw_chan1 = NULL;  /* intelligent array of file1 data values */
iarray raw_chan2 = NULL;  /* intelligent array of file2 data values */

KwcsAstro file1_ap = NULL;  /*  World co-ordinate system for file1  */
KwcsAstro file2_ap = NULL;  /*  World co-ordinate system for file2  */

double raw_min1, raw_max1;
double raw_min2, raw_max2;


/* iarrays that have been created from raw_chan1 and raw_chan2 being
   converted from (real,imaginary) values and vice versa to (amplitude,phase)
   values */
iarray conv_amp_data = NULL;
iarray conv_phs_data = NULL;

/* selected start hue range value selected */
int start_hue_val = SLIDER_MIN;

/* selected end hue range value selected */
int end_hue_val = SLIDER_HUE_MAX;

/* selected minimum intensity range value, maximum intensity is fixed by
   colourmap creation routine */
int min_int_val = SLIDER_MIN;

/* selected saturation slider value */
int sat_val = SLIDER_MAX;

/* rgb masks for the graphics system/ display; array of transformed rgb
   values of colourmap with the given masks */
unsigned long red_mask, green_mask, blue_mask;

/* The pixel canvases of the true canvas widgets of the image and
   colour-wedge */
static KPixCanvas image_pix_true_cnvs, cwedge_pix_true_cnvs;

/* Number of Raw Data Forms that the program can deal with; map to Intensity
 vs Hue */
/* const unsigned int rdf_num = 2; - doesn't work for external/static vars*/

/* works def */
char *rdf_str[RDF_NUM] =
{ "Amplitude, Phase", "Phase, Amplitude",
  "Real, Imaginary", "Imaginary, Real" };

/* Indicates the current user selected form of the RAW data, default is
   (Amplitude, Phase) */
unsigned int rdf_id = 0;

/* cmap_rgb_array points to a 65536 value (2^16) Karma PseudoColourmap,
   cmap. */
Kcolourmap cmap;
unsigned long *cmap_rgb_array;
KWorldCanvas image_world_cnvs;

/* Scaled values of hue, intensity and saturation; scaling is based on the
   current % slider values. */
const float max_int = 1.0;
float start_hue,end_hue,min_int,sat;

    

int main (int argc, char **argv)
{
    XtAppContext app_context;
    Display *dpy;
    Screen *screen;
    Widget topform;
    Widget filepopup1, filepopup2;
    Widget file1_btn, file2_btn;
    Widget reset_btn, print_btn, iscale1_btn, iscale2_btn;
    Widget quit_btn;
    Widget image_canvas;
    Widget cwedge_canvas;
    Widget image_true_cnvs;
    Widget cwedge_true_cnvs;
    Widget filewin, pswinpopup;
    Widget rdf_menu_btn;
    int canvas_types;
    int image_cnvs_init_siz = 512;
    const int cwedge_width = 256;
    const int cwedge_height = 64;

    /*  Start up Xt  */   
    main_shell = xtmisc_init_app_initialise (&app_context, "Khuei",
					     NULL, 0,
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_END,
					     NULL);
    /*  Initialise communications  */ 
    dpy = XtDisplay (main_shell);
    screen = XtScreen (main_shell);
    /*  Create other widgets  */
    /* PARENT Form widget - main shell */
    topform = XtVaCreateManagedWidget ("topForm", formWidgetClass, main_shell,
				       XtNhSpace, 0,
				       XtNborderWidth, 0,
				       NULL);
    /*  Create popups  */
    filepopup1 = XtVaCreatePopupShell
	("filewinPopup", filepopupWidgetClass, main_shell,
	 XtNtitle, "khuei File1 Selector",
	 XtNx, 0,
	 XtNy, 0,
	 XkwNautoPopdown, True,
	 XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
	 XkwNdomainName, "FILE1",
	 XkwNloadFiles, TRUE,
	 NULL);
    XtVaGetValues (filepopup1, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    filepopup2 = XtVaCreatePopupShell
	("filewinPopup", filepopupWidgetClass, main_shell,
	 XtNtitle, "khuei File2 Selector",
	 XtNx, 0,
	 XtNy, 0,
	 XkwNautoPopdown, True,
	 XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
	 XkwNdomainName, "FILE2",
	 XkwNloadFiles, TRUE,
	 NULL);
    XtVaGetValues (filepopup2, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (data_event_func, NULL);
    /*  Create PostScript control widget  */
    pswinpopup = XtVaCreatePopupShell ("postscriptwinpopup",
				       postscriptWidgetClass, main_shell,
				       XtNtitle, "Postscript Window",
				       XkwNautoIncrement, TRUE,
				       NULL);
    izoomwinpopup1 = XtVaCreatePopupShell("izoomwinpopup1",
					  dataclipWidgetClass, main_shell,
					  XtNtitle,"Intensity Zoom for File 1",
					  NULL);
    XtAddCallback (izoomwinpopup1, XkwNrawRegionCallback, region1_cbk, NULL);
    izoomwinpopup2 = XtVaCreatePopupShell("izoomwinpopup2",
					  dataclipWidgetClass, main_shell,
					  XtNtitle,"Intensity Zoom for File 2",
					  NULL);
    XtAddCallback (izoomwinpopup2, XkwNrawRegionCallback, region2_cbk, NULL);
    /*  Command widgets */
    file1_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
					 XtNlabel, "File1",
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNright, XtChainLeft,
					 NULL);
    XtAddCallback (file1_btn, XtNcallback, xtmisc_popup_cbk, filepopup1);
    file2_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
					 XtNlabel, "File2",
					 XtNfromHoriz, file1_btn,
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNright, XtChainLeft,
					 NULL);
    XtAddCallback (file2_btn, XtNcallback, xtmisc_popup_cbk, filepopup2);
    reset_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
					 XtNlabel, "Reset",
					 XtNfromHoriz, file2_btn,
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNright, XtChainLeft,
					 NULL);
    XtAddCallback (reset_btn, XtNcallback, reset_cbk, NULL);
    /*  Create "Print" button and make PostScript control widget pop up when
	button is pressed  */
    print_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
					 XtNlabel, "Print",
					 XtNfromHoriz, reset_btn,
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNright, XtChainLeft,
					 NULL);
    XtAddCallback (print_btn, XtNcallback, xtmisc_popup_cbk, pswinpopup);
    overlay_menu = XtVaCreateManagedWidget ("overlayMenu",
					    overlayMenuWidgetClass, topform,
					    XtNfromHoriz, print_btn,
					    NULL);
    quit_btn = XtVaCreateManagedWidget ("quit", commandWidgetClass, topform,
					XtNlabel, "Quit",
					XtNfromHoriz, overlay_menu,
					XtNtop, XtChainTop,
					XtNbottom, XtChainTop,
					XtNleft, XtChainLeft,
					XtNright, XtChainLeft,
					NULL);
    XtAddCallback (quit_btn, XtNcallback, quit_cbk, NULL);
    iscale1_btn = XtVaCreateManagedWidget ("button", commandWidgetClass,
					   topform,
					   XtNlabel, "Iscale 1",
					   XtNfromVert, quit_btn,
					   XtNtop, XtChainTop,
					   XtNbottom, XtChainTop,
					   XtNleft, XtChainLeft,
					   XtNright, XtChainLeft,
					   NULL);
    XtAddCallback (iscale1_btn, XtNcallback, xtmisc_popup_cbk, izoomwinpopup1);
    iscale2_btn = XtVaCreateManagedWidget ("button", commandWidgetClass,
					   topform,
					   XtNlabel, "Iscale 2",
					   XtNfromVert, quit_btn,
					   XtNfromHoriz, iscale1_btn,
					   XtNtop, XtChainTop,
					   XtNbottom, XtChainTop,
					   XtNleft, XtChainLeft,
					   XtNright, XtChainLeft,
					   NULL);
    XtAddCallback (iscale2_btn, XtNcallback, xtmisc_popup_cbk, izoomwinpopup2);
    /*  ExclusiveMenu Widget  */
    rdf_menu_btn = XtVaCreateManagedWidget ("menuButton",
					    exclusiveMenuWidgetClass, topform,
					    XtNmenuName, "rdfMenu",
					    XtNfromVert, file1_btn,
					    XkwNchoiceName,
					    "RAW DATA FORM (File1, File2) ",
					    XtNfromVert, iscale1_btn,
					    XtNtop, XtChainTop,
					    XtNbottom, XtChainTop,
					    XtNleft, XtChainLeft,
					    XtNright, XtChainLeft,
					    XkwNnumItems, RDF_NUM,
					    XkwNitemStrings, rdf_str,
					    NULL);
    XtAddCallback (rdf_menu_btn, XkwNselectCallback, rdf_cbk, NULL);
    /*  Value widgets  */
    start_hue_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					     topform,
					     XtNlabel, "% Start Hue",
					     XtNfromVert, rdf_menu_btn,
					     XtNborderWidth, 0,
					     XtNorientation, XtorientVertical,
					     XtNtop, XtChainTop,
					     XtNbottom, XtChainTop,
					     XtNleft, XtChainLeft,
					     XtNright, XtChainLeft,
					     XkwNminimum, SLIDER_MIN,
					     XkwNmaximum, SLIDER_HUE_MAX,
					     XkwNmodifier, 1,
					     XtNvalue, SLIDER_MIN,
					     XkwNvaluePtr, &start_hue_val,
					     NULL);
    XtAddCallback (start_hue_sld, XkwNvalueChangeCallback, slider_cbk, NULL);
    end_hue_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					   topform,
					   XtNlabel, "% End Hue",
					   XtNfromVert, rdf_menu_btn,
					   XtNfromHoriz, start_hue_sld,
					   XtNborderWidth, 0,
					   XtNorientation, XtorientVertical,
					   XtNtop, XtChainTop,
					   XtNbottom, XtChainTop,
					   XtNleft, XtChainLeft,
					   XtNright, XtChainLeft,
					   XkwNminimum, 0,
					   XkwNmaximum, SLIDER_HUE_MAX,
					   XkwNmodifier, 1,
					   XtNvalue, SLIDER_HUE_MAX,
					   XkwNvaluePtr, &end_hue_val,
					   NULL);
    XtAddCallback (end_hue_sld, XkwNvalueChangeCallback, slider_cbk, NULL);
    min_int_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, 
					   topform,
					   XtNlabel, "% Min Intensity",
					   XtNfromVert, start_hue_sld,
					   XtNborderWidth, 0,
					   XtNorientation, XtorientVertical,
					   XtNtop, XtChainTop,
					   XtNbottom, XtChainTop,
					   XtNleft, XtChainLeft,
					   XtNright, XtChainLeft,
					   XkwNminimum, SLIDER_MIN,
					   XkwNmaximum, SLIDER_MAX,
					   XkwNmodifier, 1,
					   XtNvalue, SLIDER_MIN,
					   XkwNvaluePtr, &min_int_val,
					   NULL);
    XtAddCallback (min_int_sld, XkwNvalueChangeCallback, slider_cbk, NULL);
    sat_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
				       topform,
				       XtNlabel, "% Saturation",
				       XtNfromVert, start_hue_sld,
				       XtNfromHoriz, min_int_sld,
				       XtNborderWidth, 0,
				       XtNorientation, XtorientVertical,
				       XtNtop, XtChainTop,
				       XtNbottom, XtChainTop,
				       XtNleft, XtChainLeft,
				       XtNright, XtChainLeft,
				       XkwNminimum, SLIDER_MIN,
				       XkwNmaximum, SLIDER_MAX,
				       XkwNmodifier, 1,
				       XtNvalue, SLIDER_MAX,
				       XkwNvaluePtr, &sat_val,
				       NULL);
    XtAddCallback (sat_sld, XkwNvalueChangeCallback, slider_cbk, NULL);
    /*  Canvas widgets  */
    canvas_types = XkwCanvasTypeTrueColour;
    /*  colour-wedge pixel canvas  */
    cwedge_canvas = XtVaCreateManagedWidget ("multiCanvas",
					     multiCanvasWidgetClass ,
					     topform,
					     XtNfromVert, sat_sld,
					     XtNwidth, cwedge_width,
					     XtNheight, cwedge_height,
					     XkwNrequestMask, canvas_types,
					     XtNtop, XtChainTop,
					     XtNbottom, XtChainTop,
					     XtNleft, XtChainLeft,
					     XtNright, XtChainLeft,
					     NULL);
    cwedge_true_cnvs = XtNameToWidget (cwedge_canvas, "trueColourCanvas");
    if (cwedge_true_cnvs == NULL)
    {
	fprintf(stderr,"No TrueColour display found\n");
	exit (1);
    }
    XtVaSetValues (cwedge_true_cnvs,
		   XtNmappedWhenManaged, True,
		   XkwNsilenceUnconsumed, True,
		   NULL);
    /*  image pixel canvas  */
    if (HeightOfScreen(XtScreen(topform)) < 700) image_cnvs_init_siz = 385;
    image_canvas = XtVaCreateManagedWidget ("multiCanvas",
					    multiCanvasWidgetClass ,
					    topform,
					    XtNfromVert, cwedge_canvas,
					    XtNwidth, image_cnvs_init_siz,
					    XtNheight, image_cnvs_init_siz ,
					    XkwNrequestMask, canvas_types,
					    XtNtop, XtChainTop,
					    XtNbottom, XtChainBottom,
					    XtNleft, XtChainLeft,
					    XtNright, XtChainRight,
					    NULL);
    image_true_cnvs = XtNameToWidget (image_canvas, "trueColourCanvas");
    XtVaSetValues (image_true_cnvs,
		   XtNmappedWhenManaged, True,
		   XkwNsilenceUnconsumed, True,
		   NULL);
    XtRealizeWidget (main_shell);

    XtVaGetValues (cwedge_true_cnvs,
		   XkwNmonoPixCanvas, &cwedge_pix_true_cnvs,
		   NULL);
    XtVaGetValues (image_true_cnvs,
		   XkwNmonoPixCanvas, &image_pix_true_cnvs,
		   NULL);
    disp_cwedge_init_imag ();
    /*  Since only one KPixCanvas is worth printing, register it now  */
    XkwPostscriptRegisterImageAndName (pswinpopup, image_pix_true_cnvs, NULL);

    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main  */


/*  PRIVATE ROUTINES follow  */

static void reset_cbk (Widget w, XtPointer client_data, XtPointer call_data)

/* Resets the slider values to their defaults and replots the corresponding
   hue-intensity pixel map. */

{
    XtVaSetValues (start_hue_sld,
		   XtNvalue, SLIDER_MIN,
		   NULL);
    XtVaSetValues (end_hue_sld,
		   XtNvalue, SLIDER_HUE_MAX,
		   NULL);
    XtVaSetValues (min_int_sld,
		   XtNvalue, SLIDER_MIN,
		   NULL);
    XtVaSetValues (sat_sld,
		   XtNvalue, SLIDER_MAX,
		   NULL);
    compute_colourmap ();
}   /*  End Function reset_cbk  */

static void rdf_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/* Determines the user selection of the Raw Data Form (eg Amp,Phs) and
   uses this to convert the data to hue,intensity mapped pixel map. */
{    
    rdf_id = *(int *) call_data;

    if ( test_both_good (raw_chan1, raw_chan2) ) data_to_huei_pixmap ();
}   /*  End Function rdf_cbk  */

static void quit_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the quit button callback, exits from the application. */
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static void slider_cbk (w, client_data, call_data)
/*  This is the slider callback. */
     Widget w;
     XtPointer client_data;	/* application data */
     XtPointer call_data;	/* widget value data */
{
    /* hue/sat/int slider/s have been activated, values changed */
    compute_colourmap ();
}   /*  End Function slider_cbk   */

static void map_re_im_2_amp_phs (iarray re_chan, double re_min, double re_max,
				 iarray im_chan, double im_min, double im_max)
{
    double amp, phs;
    int i,j;		/* counters for accessing array elements */
    int xlen, ylen;

    /* mapping of (Real,Imaginary) raw data iarrays to (Amplitude, Phase)
       iarrays */
    xlen = iarray_dim_length (re_chan, 1);
    ylen = iarray_dim_length (re_chan, 0);
    iarray_copy_data (conv_amp_data, re_chan, FALSE );
    iarray_copy_data (conv_phs_data, im_chan, FALSE );
    for (i = 0; i < ylen; i++) for (j = 0; j < xlen; j++)
    {
	amp = D2 (conv_amp_data, i, j);
	phs = D2 (conv_phs_data, i, j);
	if (amp < re_min) amp = re_min;
	else if (amp > re_max) amp = re_max;
	if (phs < im_min) phs = im_min;
	else if (phs > im_max) phs = im_max;
	re_im_2_amp_phs (amp, phs, &amp, &phs);
	D2 (conv_amp_data, i, j) = amp;
	D2 (conv_phs_data, i, j) = phs;
    }
} /* End Function map_re_im_2_amp_phs */

static void re_im_2_amp_phs (double re, double im, double *amp, double *phs)
{
    const double zero = 0.0;

/* Calculate Amplitude and Phase values fom Real and Imaginary values. */
/* Account for domain error with pow(x, y) function. */
  
  
    if ( (re == zero) && (im == zero) )
    {
	*amp = zero;
	*phs = zero;
    }
    else if ( (re == zero) && (im != zero))
    {
	*amp = im;
	*phs = atan2 (im, re); 
    }
    else if ( (re != zero) && (im == zero) )
    {
	*amp = re;
	*phs = atan2 (im, re);
    }
    else
    {
	*amp = sqrt( (pow (re, 2) + pow (im, 2)) );
	*phs = atan2 (im, re);
    }
    /*  since we have a linear mapping of phase values in the 16bit array  */
    if (*phs < zero) *phs = TWOPI + *phs;
} /* End Function re_im_2_amp_phs */

static void data_to_16bit (iarray amp_data, double amp_min, double amp_max,
			   iarray phs_data, double phs_min, double phs_max)
{
    /* The 16bit array created from: scaling to 8bit arrays the two input
       channel data arrays ; and merging the two.  The upper 8bits (MSByte)
       represent scaled Amplitude and the lower 8bits (LSByte) represent
       scaled Phase.  This 16bit array is used as an index to the RGB colour
       map. */
    /* The input arrays *must* have already been converted to (Amplitude,
       Phase),
       and this is *always* mapped to (Intensity, Hue) in the pixmap. */
    /* scale factor and offset to convert loaded iarrays to
       values in the range 0-255, that is to 8bit values */
    double scale, offset;
    /* minimum and maximum values of channel data */
    /* counters for accessing array elements */
    int i, j, xlen, ylen;
    static ViewableImage image_vimage = NULL;
    static iarray amp_8bit = NULL;
    static iarray phs_8bit = NULL;
    static iarray merge_16bit = NULL;
    static char function_name[] = "data_to_16bit";

    xlen = iarray_dim_length (amp_data, 1);
    ylen = iarray_dim_length (amp_data, 0);
    /*  Ensure merge_16bit array exists and is the right size  */
    if (merge_16bit != NULL)
    {
	/*  Existing array: test if the same size. Assume that the amplitude
	    and phase arrays are the same size  */
	if ( (iarray_dim_length (merge_16bit, 0) != xlen) ||
	     (iarray_dim_length (merge_16bit, 1) != ylen ) )
	{
	    /*  Wrong size  */
	    iarray_dealloc (merge_16bit);
	    merge_16bit = NULL;
	    iarray_dealloc (amp_8bit);
	    amp_8bit = NULL;
	    iarray_dealloc (phs_8bit);
	    phs_8bit = NULL;
	    if (image_vimage != NULL)
	    {
		viewimg_destroy (image_vimage);
		image_vimage = NULL;
	    }
	}
    }
    if (merge_16bit == NULL)
    {
	if ( ( merge_16bit = iarray_create_from_template (raw_chan1, K_USHORT,
							  TRUE, TRUE, TRUE) )
	     == NULL )
	{
	    m_abort (function_name, "merged 16bit array");
	}
	if ( ( amp_8bit = iarray_create_2D (ylen, xlen, K_UBYTE) ) == NULL )
	{
	    m_abort (function_name, "8bit amplitude array");
	}
	if ( ( phs_8bit = iarray_create_2D (ylen, xlen, K_UBYTE) ) == NULL )
	{
	    m_abort (function_name, "8bit phase array");
	}
    }
    /* (note: conv1_real is ignored unless input values are complex) */
    scale = 255 / (amp_max - amp_min);
    offset = (255 * amp_min) / (amp_min - amp_max);
    iarray_clip_scale_and_offset (amp_8bit, amp_data, scale, offset,
				  amp_min, amp_max, FALSE);
  
    scale = 255 / (phs_max - phs_min);
    offset = (255 * phs_min) / (phs_min - phs_max);
    iarray_clip_scale_and_offset (phs_8bit, phs_data, scale, offset,
				  phs_min, phs_max, FALSE);
  
    for (i=0; i < ylen; i++)
	for (j=0; j < xlen; j++)
	    US2(merge_16bit,i,j) =
		(  ( ((unsigned short int) UB2(amp_8bit,i,j))  ) << 8  ) ^
		UB2(phs_8bit,i,j);

    if (image_vimage == NULL)
    {
	image_vimage = viewimg_create_from_iarray (image_world_cnvs,
						   merge_16bit, FALSE);
	viewimg_make_active (image_vimage);
    }
    else viewimg_register_data_change (image_vimage);
} /* End Function data_to_16bit */

static void compute_colourmap (void)
{
/* Plots the pixmap of the raw data.  The merge_16bit array is used as an
   index to the colourmap.  The RBG colour map (PseudoColour type) is
   created based on the range of  hues, intensities, and the saturation
   value, selected by the user */

    /* scale % slider values of hue, intensity and saturation
       for colourmap routine; the sliders allow the user to effectivly
       select a subspace of the full colour space to create the
       colourmap */
    start_hue =  (start_hue_val/100.0) * 6.0;
    end_hue = (end_hue_val/100.0) * 6.0;
    min_int = (min_int_val/100.0) * max_int;
    sat = (sat_val/100.0) * 1.0;
  
/* Refresh COLOUR-WEDGE and IMAGE pixel canvases.  If merge_16bit contains
   non NULL values for the first time the image viewable image is created. */
    /* calculate current colourmap subspace */
    col_hsb_slice_to_rgb_array (cmap_rgb_array, red_mask, green_mask,blue_mask,
				min_int, max_int, start_hue, end_hue, sat);  
    /*  This is all that is needed to tell the graphics system to redraw the
	16 bit image with the new colourtable  */
    kcmap_notify_pixels_changed (cmap);
}   /*  End Function compute_colourmap  */

static void data_to_huei_pixmap (void)
{
    /* Since raw_chan1 and raw_chan2 have been loaded: if necessary, convert
       this raw data and, or swap the channels the data is stored in; the
       data can then be plotted as a hue_intensity pixmap. */
    int xlen, ylen;
    double amp_min, amp_max, phs_min, phs_max;
    static char function_name[] = "data_to_huei_pixmap";

    xlen = iarray_dim_length (raw_chan1, 1);
    ylen = iarray_dim_length (raw_chan1, 0);
    if (conv_amp_data != NULL)
    {
	/*  Existing array: test if the same size. Assume that the amplitude
	    and phase arrays are the same size  */
	if ( (iarray_dim_length (conv_amp_data, 0) != xlen) ||
	     (iarray_dim_length (conv_amp_data, 1) != ylen ) )
	{
	    /*  Wrong size  */
	    iarray_dealloc (conv_amp_data);
	    conv_amp_data = NULL;
	    iarray_dealloc (conv_phs_data);
	    conv_phs_data = NULL;
	}
    }
    if (conv_amp_data == NULL)
    {
	if ( ( conv_amp_data = iarray_create_2D (ylen, xlen, K_DOUBLE) )
	     == NULL )
	{
	    m_abort (function_name, "conversion amplitude array");
	}
	if ( ( conv_phs_data = iarray_create_2D (ylen, xlen, K_DOUBLE) )
	     == NULL )
	{
	    m_abort (function_name, "conversion phase array");
	}
    }
    switch (rdf_id)
    {
      case AMP_PHS:
	/* This is the program assumed default form of data - raw channel data
	   stored as rawFILE1_rawFILE2;  (Amp,Phs) --> (Intensity,Hue). */
	data_to_16bit (raw_chan1, raw_min1, raw_max1,
		       raw_chan2, raw_min2, raw_max2);
	break;
      case PHS_AMP:
	/* Swap raw channel data: rawFILE1 -> FILE2, rawFILE2 -> FILE1;
	   (Amp,Phs) --> (Intensity,Hue). */ 
	data_to_16bit (raw_chan2, raw_min2, raw_max2,
		       raw_chan1, raw_min1, raw_max1);
	break;
      case RE_IM:
	/* Raw Data Form is: FILE1 has Real values and FILE2 has Imaginary
	   values. */
	/* Use this data to transform/map(Real,Imaginary ) -> (Amplitude,Phase)
	   values; then map (Amp,Phs) -> (Intensity,Hue), that is,
	   (amp_data,phs_data) */
        map_re_im_2_amp_phs (raw_chan1, raw_min1, raw_max1,
			     raw_chan2, raw_min2, raw_max2);
	iarray_min_max (conv_amp_data, CONV_CtoR_REAL, &amp_min, &amp_max);
	iarray_min_max (conv_phs_data, CONV_CtoR_REAL, &phs_min, &phs_max);
        data_to_16bit (conv_amp_data, amp_min, amp_max,
		       conv_phs_data, phs_min, phs_max);
        break;
      case IM_RE:
	/* Raw Data Form is: FILE1 has Imaginary values and FILE2 has Real
	   values. */
	/* Use this data to transform/map(Imaginary,Real ) -> (Amplitude,Phase)
	   values;   then map (Amp,Phs) -> (Intensity,Hue), that is,
	   (amp_data,phs_data) */
	map_re_im_2_amp_phs (raw_chan2, raw_min2, raw_max2,
			     raw_chan1, raw_min1, raw_max1);
	iarray_min_max (conv_amp_data, CONV_CtoR_REAL, &amp_min, &amp_max);
	iarray_min_max (conv_phs_data, CONV_CtoR_REAL, &phs_min, &phs_max);
        data_to_16bit (conv_amp_data, amp_min, amp_max,
		       conv_phs_data, phs_min, phs_max);
        break;
      default:
	break;
    }
} /* End Function data_to_huei_pixmap */

void disp_cwedge_init_imag (void)
/* Displays a 256x256 colour-wedge pixel canvas; data values are drawn
   on a world-canvas (gives world co-ordinate layer on top of a pixel
   canvas for drawing) with the viewable image functionality of the
   karma graphics library. The wedge is used to give the user a feel
   for the full range of intensity and hue values.
   
   In addition, for the image pixel canvas: a world canvas is created;
   and it's viewable image initialized, yet not created or activated. */
{
    KWorldCanvas cwedge_world_cnvs;
    iarray cwedge;
    int i, j;

    ViewableImage cwedge_vimage;
    /* can be ignored - will be removed from karma in the future
       (default values are given by initialisation routine) */
    struct win_scale_type win_scale;
    const int cwedge_range = 256;

    /* initializations */

    /* scale % slider values of hue, intensity and saturation
       for colourmap routine; the sliders allow the user to effectivly
       select a subspace of the full colour space to create the
       colourmap */
  
    start_hue =  (start_hue_val/100.0) * 6.0;
    end_hue = (end_hue_val/100.0) * 6.0;
    min_int = (min_int_val/100.0) * max_int;
    sat = (sat_val/100.0) * 1.0;
  
    /* get rgb colour masks for the graphics system/ display
       (different graphics systems will have different masks)
       for use in creating the colourmap */
  
    kwin_get_attributes (image_pix_true_cnvs,
			 KWIN_ATT_PIX_RED_MASK, &red_mask,
			 KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			 KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			 KWIN_ATT_END);

/* Create a viewable image colourmap; to be shared by *all* application
   canvases. (cwedge and imag )*/
  
    /* allocate storage */
    cmap = kcmap_va_create (NULL, 65536, FALSE, NULL,
			    ( unsigned int (*) () ) NULL,
			    ( void (*) () ) NULL,
			    ( void (*) () ) NULL,
			    ( void (*) () ) NULL,
			    KCMAP_ATT_END);
  
    /* Get pointer to cmap storage, to be used for the cwedge in this case. */
    kcmap_get_pixels(cmap, &cmap_rgb_array);

/* create and initialise the cwedge (16bit index to cmap) */

    cwedge = iarray_create_2D(cwedge_range,cwedge_range,K_USHORT);
    for (i=0; i < cwedge_range; i++){
	for (j=0; j < cwedge_range; j++){
	    US2(cwedge, i, j) =
		( ((unsigned short int) i) << 8 ) ^ ( ((unsigned short int) j) );
	}
    }

    /* Create the RGB colours for the colour-map. This colour-map will
       reflect the users CURRENT choise of: min/max hue, min intensity,
       and saturation.  */
    col_hsb_slice_to_rgb_array(cmap_rgb_array,red_mask,green_mask,blue_mask,
			       min_int,max_int,start_hue,end_hue,sat);  

    /* register the cmap with the viewable images package */
    kcmap_notify_pixels_changed (cmap);

/* Create the world canvas from the pixel canvas and initialize the
   structure that handles the specification of colours for out of
   bounds values, win_scale. Both colour-wedge and image world canvases
   are created. */
  
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    cwedge_world_cnvs = canvas_create (cwedge_pix_true_cnvs, cmap,
				       &win_scale);
    image_world_cnvs = canvas_create (image_pix_true_cnvs, cmap,
				      &win_scale);
    /*  Set the intensity scale manually and later turn off automatic intensity
	scaling since it is not useful  */
    canvas_set_attributes (image_world_cnvs,
			   CANVAS_ATT_VALUE_MIN, 0.0,
			   CANVAS_ATT_VALUE_MAX, 65535.0,
			   CANVAS_ATT_END);
    canvas_set_dressing (image_world_cnvs,
			 KCD_ATT_SHOW_LABELS, TRUE,
			 KCD_ATT_TOP_TICKS, TRUE,
			 KCD_ATT_BOTTOM_TICKS, TRUE,
			 KCD_ATT_LEFT_TICKS, TRUE,
			 KCD_ATT_RIGHT_TICKS, TRUE,
			 KCD_ATT_END);
  
/* Create and display viewable image from the colour-wedge array, on a
   world canvas.  The viewable image for the image data is ONLY
   initilized, NOT created or displayed (activatied). */

    viewimg_init (cwedge_world_cnvs);
    viewimg_init (image_world_cnvs);
    /* facilitate automatic zooming functions (drag zoom with mouse &
       unzoom with character u) */
    viewimg_create_drag_and_zoom_interface (image_world_cnvs);
    viewimg_set_canvas_attributes (cwedge_world_cnvs,
				   VIEWIMG_ATT_INT_X, FALSE,
				   VIEWIMG_ATT_INT_Y, FALSE,
				   VIEWIMG_ATT_END);
    viewimg_set_canvas_attributes (image_world_cnvs,
				   VIEWIMG_ATT_INT_X, FALSE,
				   VIEWIMG_ATT_INT_Y, FALSE,
				   VIEWIMG_ATT_MAINTAIN_ASPECT, TRUE,
				   VIEWIMG_ATT_AUTO_V, FALSE,
				   VIEWIMG_ATT_END);
    cwedge_vimage = viewimg_create_from_iarray (cwedge_world_cnvs, cwedge,
						FALSE);
    viewimg_make_active (cwedge_vimage);
    contour_init (image_world_cnvs,
		  CONTOUR_CANVAS_ATT_COLOURNAME, "green",
		  CONTOUR_CANVAS_ATT_END);
    XkwOverlayMenuAddCanvas (overlay_menu, image_world_cnvs);
}  /* End Function disp_cwedge_init_imag */

static flag data_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process an event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    flag ok;

    if (strcmp (domain, "FILE1") == 0)
    {
	ok = iarray_create_and_setup (&raw_chan1, multi_desc, TRUE, 2,
				      NONE, &raw_min1, &raw_max1, TRUE,
				      &file1_ap);
	XkwDataclipNewArray (izoomwinpopup1, raw_chan1, raw_min1, raw_max1,
			     TRUE);
    }
    else if (strcmp (domain, "FILE2") == 0)
    {
	ok = iarray_create_and_setup (&raw_chan2, multi_desc, TRUE, 2,
				      NONE, &raw_min2, &raw_max2, TRUE,
				      &file2_ap);
	XkwDataclipNewArray (izoomwinpopup2, raw_chan2, raw_min2, raw_max2,
			     TRUE);
    }
    else return (FALSE);
    if (!ok) return (TRUE);
    fprintf (stderr, "Loaded %s as %s\n", name, domain);
    if ( !test_both_good (raw_chan1, raw_chan2) )
    {
	/*  Raw arrays are of a different size: do nothing yet  */
	return (TRUE);
    }
    canvas_use_astro_transform (image_world_cnvs, file1_ap);
    data_to_huei_pixmap ();
    return (TRUE);
}   /*  End Function data_event_func  */

static flag test_both_good (iarray arr1, iarray arr2)
/*  [SUMMARY] Test if two arrays are the same size and are loaded.
    <arr1> The first Intelligent Array.
    <arr2> The second Intelligent Array.
    [RETURNS] TRUE if the arrays are the same size, else FALSE.
*/
{
    if ( (arr1 == NULL) || (arr2 == NULL) ) return (FALSE);
    if ( ( iarray_dim_length (raw_chan1, 0) !=
	   iarray_dim_length (raw_chan2, 0) ) ||
	 ( iarray_dim_length (raw_chan1, 1) !=
	   iarray_dim_length (raw_chan2, 1) ) ) return (FALSE);
    return (TRUE);
}   /*  End Function test_both_good  */

static void region1_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the region callback.
*/
{
    DataclipRegions *regions = (DataclipRegions *) call_data;
    static char function_name[] = "region1_cbk";

    if (regions->num_regions != 1)
    {
	fprintf (stderr, "num_regions: %u is not 1\n", regions->num_regions);
	a_prog_bug (function_name);
    }
    raw_min1 = regions->minima[0];
    raw_max1 = regions->maxima[0];
    if ( test_both_good (raw_chan1, raw_chan2) ) data_to_huei_pixmap ();
}   /*  End Function region1_cbk   */

static void region2_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the region callback.
*/
{
    DataclipRegions *regions = (DataclipRegions *) call_data;
    static char function_name[] = "region2_cbk";

    if (regions->num_regions != 1)
    {
	fprintf (stderr, "num_regions: %u is not 1\n", regions->num_regions);
	a_prog_bug (function_name);
    }
    raw_min2 = regions->minima[0];
    raw_max2 = regions->maxima[0];
    if ( test_both_good (raw_chan1, raw_chan2) ) data_to_huei_pixmap ();
}   /*  End Function region2_cbk   */

#ifdef OS_SunOS

static void dummy ()
/*  [PURPOSE] This routine is needed for SunOS, because of a problem with
    shared archive use. The <<_XtToolkitInitialize>> symbol is defined in
    the Xt shared archive, but it appears the shared archive is not linked for
    this application. Other applications (like <<krenzo>>) have no such
    problems. This is a truly bizzare bug.
*/
{
    XtToolkitInitialize ();
}   /*  End Function dummy  */
#endif
