/*  compute.c

    Compute file for  koords  (X11 co-ordinate generator tool for Karma).

    Copyright (C) 1996-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive generation of an astronomical
    co-ordinate system on a target image, using a reference image.
    This module runs on an X11 server.


    Written by      Richard Gooch   15-OCT-1996

    Updated by      Richard Gooch   17-OCT-1996: Created <apply_coordinates>.

    Updated by      Richard Gooch   3-NOV-1996: Modify displayed information if
  target image had an original co-ordinate system.

    Updated by      Richard Gooch   10-JUN-1997: Remove existing co-ordinate
  system keywords.

    Updated by      Richard Gooch   14-FEB-1998: Changed to dataset_t type.

    Updated by      Richard Gooch   31-MAY-1998: Support Galactic co-ordinates.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Last updated by Richard Gooch   26-JUN-2002: Added "EXTEND" to list of
  keywords to delete (it's not required and Karma doesn't support a boolean
  data type, so bad FITS could be written).


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_dsra.h>
#include <karma_wcs.h>
#include <karma_dmp.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_n.h>
#include "koords.h"


/*  Private functions  */
STATIC_FUNCTION (void setup_headers, () );


/*  Private data  */
static char *str_standard_desc[] =
{
    "PACKET",
    "  9",
    "END",
    "  ELEMENT",
    "    VSTRING",
    "    CTYPE1",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CRVAL1",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CRPIX1",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CDELT1",
    "  END",
    "  ELEMENT",
    "    VSTRING",
    "    CTYPE2",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CRVAL2",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CRPIX2",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CDELT2",
    "  END",
    "  ELEMENT",
    "    DOUBLE",
    "    CROTA2",
    "  END",
    NULL
};
static packet_desc *standard_desc = NULL;
static char *standard_packet = NULL;
static char *deletable_keywords[] =
{
    "CTYPE1",
    "CRVAL1",
    "CRPIX1",
    "CDELT1",
    "CROTA1",
    "CUNIT1",
    "CD001001",
    "CD001002",
    "CD1_1",
    "CD1_2",
    "CTYPE2",
    "CRVAL2",
    "CRPIX2",
    "CDELT2",
    "CROTA2",
    "CUNIT2",
    "CD002001",
    "CD002002",
    "CD2_1",
    "CD2_2",
    "EXTEND",
    "PLTRAH",
    "PLTRAM",
    "PLTRAS",
    "PLTDECSN",
    "PLTDECD",
    "PLTDECM",
    "PLTDECS",
    "PLTSCALE",
    "CNPIX1",
    "CNPIX2",
    "DATATYPE",
    "DCHOPPED",
    "DSHEARED",
    "DSCNDNUM",
    "XPIXELSZ",
    "YPIXELSZ",
    "PPO1",
    "PPO2",
    "PPO3",
    "PPO4",
    "PPO5",
    "PPO6",
    "AMDX1",
    "AMDX2",
    "AMDX3",
    "AMDX4",
    "AMDX5",
    "AMDX6",
    "AMDX7",
    "AMDX8",
    "AMDX9",
    "AMDX10",
    "AMDX11",
    "AMDX12",
    "AMDX13",
    "AMDX14",
    "AMDX15",
    "AMDX16",
    "AMDX17",
    "AMDX18",
    "AMDX19",
    "AMDX20",
    "AMDY1",
    "AMDY2",
    "AMDY3",
    "AMDY4",
    "AMDY5",
    "AMDY6",
    "AMDY7",
    "AMDY8",
    "AMDY9",
    "AMDY10",
    "AMDY11",
    "AMDY12",
    "AMDY13",
    "AMDY14",
    "AMDY15",
    "AMDY16",
    "AMDY17",
    "AMDY18",
    "AMDY19",
    "AMDY20",
    "OBJCTX",
    "OBJCTY",
    NULL
};


/*  Public functions follow  */

void compute_and_store ()
/*  [SUMMARY] Compute co-ordinate system, print and store.
    [RETURNS] Nothing.
*/
{
    flag galactic;
    unsigned int count, xlen, ylen;
    double rms;
    char *p1, *p2;
    char txt[STRING_LENGTH], tmp[STRING_LENGTH];
    double crval1[2], crpix1[2], cdelt1[2], crval2[2], crpix2[2], cdelt2[2];
    double crota[2];
    double pred_x[MAX_PAIRS], pred_y[MAX_PAIRS];
    double diff_x[MAX_PAIRS], diff_y[MAX_PAIRS];
    extern KwcsAstro target_original_ap;
    extern dataset_t reference, target;

    setup_headers ();
    if ( (reference.num_points != target.num_points) ||
	 (target.array == NULL) )
    {
	ring_bell ();
	return;
    }
    galactic = wcs_astro_test_if_glon ( reference.ap,
					iarray_dim_name (reference.array, 1) );
    /*  First compute the new co-ordinate system  */
    xlen = iarray_dim_length (target.array, 1);
    ylen = iarray_dim_length (target.array, 0);
    if (compute_coords (crval1, crpix1, cdelt1, crval2, crpix2, cdelt2, crota,
			NULL, reference.num_points,
			reference.ra, reference.dec, target.ra, target.dec,
			xlen, ylen) < 1)
    {
	fprintf (stderr, "Error computing co-ordinate system\n");
	return;
    }
    /*  Now put entries into temporary header and create projection object  */
    if ( !ds_put_unique_named_string (standard_desc, &standard_packet,
				      "CTYPE1",
				      galactic ? "GLON-ARC" : "RA---ARC",
				      TRUE) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CRVAL1",
				crval1) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CRPIX1",
				crpix1) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CDELT1",
				cdelt1) ) return;
    if ( !ds_put_unique_named_string (standard_desc, &standard_packet,
				      "CTYPE2",
				      galactic ? "GLAT-ARC" : "DEC--ARC",
				      TRUE) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CRVAL2",
				crval2) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CRPIX2",
				crpix2) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CDELT2",
				cdelt2) ) return;
    if ( !ds_put_named_element (standard_desc, standard_packet, "CROTA2",
				crota) ) return;
    if (target.ap != NULL) wcs_astro_destroy (target.ap);
    if ( (target.ap = wcs_astro_setup (standard_desc, standard_packet) )
	 == NULL )
    {
	fprintf (stderr, "Could not create standard projection object\n");
	return;
    }
    /*  Copy RA,DEC values from reference and convert to x,y in target  */
    for (count = 0; count < reference.num_points; ++count)
    {
	pred_x[count] = reference.ra[count];
	pred_y[count] = reference.dec[count];
	if (target_original_ap != NULL)
	{
	    diff_x[count] = reference.ra[count];
	    diff_y[count] = reference.dec[count];
	}
    }
    wcs_astro_transform (target.ap, FALSE, reference.num_points,
			 pred_x, TRUE, pred_y, TRUE, NULL, FALSE,
			 0, NULL, NULL);
    if (galactic)
    {
	p1 = "#  Glon            Glat             Tx     Ty      Px     Py      Dx     Dy\n";
	p2 = "#  Glon            Glat             Tx     Ty      PDx    PDy     Dx     Dy\n";
    }
    else
    {
	p1 = "#  Ra              Dec              Tx     Ty      Px     Py      Dx     Dy\n";
	p2 = "#  Ra              Dec              Tx     Ty      PDx    PDy     Dx     Dy\n";
    }
    if (target_original_ap == NULL) fputs (p1, stderr);
    else
    {
	fputs (p2, stderr);
	wcs_astro_transform (target_original_ap, FALSE, reference.num_points,
			     diff_x, TRUE, diff_y, TRUE, NULL, FALSE,
			     0, NULL, NULL);
    }
    for (count = 0, rms = 0.0; count < reference.num_points; ++count)
    {
	fprintf (stderr, "%-3u", count);
	wcs_astro_format (reference.ap,
			  iarray_dim_name (reference.array, 1), txt,
			  reference.ra[count]);
	if ( ( p1 = strchr (txt, ' ') ) == NULL ) p1 = txt;
	else ++p1;
	fprintf (stderr, "%-16s", p1);
	wcs_astro_format (reference.ap,
			  iarray_dim_name (reference.array, 0), txt,
			  reference.dec[count]);
	if ( ( p2 = strchr (txt, ' ') ) == NULL ) p2 = txt;
	else ++p2;
	fprintf (stderr, "%-15s  %-7.1f%-8.1f",
		 p2, target.ra[count], target.dec[count]);
	if (target_original_ap == NULL)
	{
	    fprintf (stderr, "%-7.1f%-8.1f%-7.1f%-.1f\n",
		     pred_x[count], pred_y[count],
		     target.ra[count] - pred_x[count],
		     target.dec[count] - pred_y[count]);
	}
	else fprintf (stderr, "%-7.1f%-8.1f%-7.1f%-.1f\n",
		      diff_x[count] - pred_x[count],
		      diff_y[count] - pred_y[count],
		      target.ra[count] - pred_x[count],
		      target.dec[count] - pred_y[count]);
	rms = (target.ra[count] - pred_x[count]) * (target.ra[count] -
						    pred_x[count]);
	rms += (target.dec[count] - pred_y[count]) * (target.dec[count] -
						      pred_y[count]);
    }
    rms = sqrt (rms / (double) reference.num_points);
    fprintf (stderr, "RMS pixel error: %.1f\n", rms);
    /*  Print header in nice format  */
    fprintf (stderr, "\nHeader information:\nProjection: ARC (rectangular)\n");
    wcs_astro_format (reference.ap, iarray_dim_name (reference.array, 1), txt,
		      crval1[0]);
    wcs_astro_format (reference.ap, iarray_dim_name (reference.array, 0), tmp,
		      crval2[0]);
    fprintf (stderr, "Reference:  %s  %s   pixel: %.1f  %.1f\n",
	     txt, tmp, crpix1[0], crpix2[0]);
    fprintf (stderr,
	     "Co-ordinate increment:  %s %.3f  %s %.3f  (arcsec/pixel)\n",
	     galactic ? "Glon" : "Ra", cdelt1[0] * 3600.0,
	     galactic ? "Glat" : "Dec", cdelt2[0] * 3600.0);
    fprintf (stderr, "Rotation: %.3f degrees\n", crota[0]);
}   /*  End Function compute_and_store  */

void apply_coordinates ()
/*  [SUMMARY] Apply co-ordinate system to target.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    dim_desc *dim;
    extern dataset_t target;
    static char function_name[] = "apply_coordinates";

    if ( (target.ap == NULL) || (target.array == NULL) )
    {
	ring_bell ();
	return;
    }
    if ( !ds_remove_named_elements (target.array->top_pack_desc,
				    target.array->top_packet,
				    (CONST char **) deletable_keywords) )
    {
	ring_bell ();
	return;
    }
    /*  Copy over each element in the header to the output array  */
    for (count = 0; count < standard_desc->num_elements; ++count)
    {
	if ( !ds_copy_unique_named_element (target.array->top_pack_desc,
					    target.array->top_packet,
					    standard_desc, standard_packet,
					    standard_desc->element_desc[count],
					    FALSE, FALSE, TRUE) )
	{
	    fprintf (stderr, "Error adding co-ordinate system to header\n");
	    return;
	}
    }
    /*  Copy over the axis names if not already  */
    dim = iarray_get_dim_desc (target.array, 1);
    m_free (dim->name);
    dim->name = ds_get_unique_named_string (standard_desc, standard_packet,
					    "CTYPE1");
    if (dim->name == NULL) m_abort (function_name, "horizontal axis name");
    dim = iarray_get_dim_desc (target.array, 0);
    dim->name = ds_get_unique_named_string (standard_desc, standard_packet,
					    "CTYPE2");
    if (dim->name == NULL) m_abort (function_name, "vertical axis name");
}   /*  End Function apply_coordinates  */


/*  Private functions follow  */

static void setup_headers ()
/*  [SUMMARY] Create the temporary headers.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    static char function_name[] = "setup_headers";

    if (standard_desc != NULL) return;
    /*  Create descriptor  */
    if ( ( channel = ch_open_and_fill_memory (str_standard_desc) ) == NULL )
    {
	m_abort (function_name, "memory channel");
    }
    if ( ( standard_desc = dsra_packet_desc (channel) ) == NULL )
    {
	ch_close (channel);
	m_abort (function_name, "standard header descriptor");
    }
    ch_close (channel);
    /*  Create data  */
    if ( ( standard_packet = ds_alloc_data (standard_desc, TRUE, TRUE) )
	 == NULL )
    {
	m_abort (function_name, "standard header data");
    }
}   /*  End Function setup_headers  */
