/*  event.c

    Event file for  koords  (X11 co-ordinate generator tool for Karma).

    Copyright (C) 1996-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive generation of an astronomical
    co-ordinate system on a target image, using a reference image.
    This module runs on an X11 server.


    Written by      Richard Gooch   14-OCT-1996: Copied from kshell/main_xt.c

    Updated by      Richard Gooch   17-OCT-1996: Completed centroid hunter.

    Updated by      Richard Gooch   21-OCT-1996: Made <get_colour> public.
  Created <update_magnifier>.

    Updated by      Richard Gooch   3-NOV-1997: Disabled radius display.

    Updated by      Richard Gooch   8-FEB-1997: Took account of new automatic
  addition of non-linear world co-ordinates on magnifier window by drawing
  ellipse in pixel co-ordinates.

    Updated by      Richard Gooch   17-MAY-1997: Use reverse attribute for
  colourmap to determine if image is inverted and adapt centering algorithm.

    Updated by      Richard Gooch   7-AUG-1997: Made use of new routine
  <iarray_find_centroid>.

    Updated by      Richard Gooch   1-DEC-1997: Took account of change to
  <iarray_find_centroid>.

    Updated by      Richard Gooch   7-DEC-1997: Removed code to set pan centre
  for magnifier as it is now done automatically by the MagnifierPopup widget.

    Updated by      Richard Gooch   14-FEB-1998: Removed code to draw crosshair
  for magnifier as it is now done by the MagnifierPopup widget.

    Updated by      Richard Gooch   16-FEB-1998: Use line dashing for outer
  ring. Draw crosshairs for centroids in magnifier windows.

    Updated by      Richard Gooch   3-MAR-1998: Prevent huge radii in function
  <mag_refresh_func>.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   4-JUN-1998: Discarded <find_object>.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Last updated by Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.


*/
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_canvas.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include "koords.h"


#define ELLIPSE_RADIUS 5.0
#define CROSSHAIR_SIZE 5.0


/*  Private functions  */
STATIC_FUNCTION (unsigned long get_colour,
		 (KWorldCanvas canvas, CONST char *colourname) );


/*  Private data  */


/*  Public functions follow  */

flag reference_event_func (ViewableImage vimage, double x, double y,
			   void *value, unsigned int event_code,
			   void *e_info, void **f_info,
			   double x_lin, double y_lin, unsigned int value_type)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas wc;
    unsigned int index;
    extern flag last_click_was_reference;
    extern dataset_t reference;
    /*static char function_name[] = "reference_event_func";*/

    if (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) return (FALSE);
    wc = viewimg_get_worldcanvas (vimage);
    iarray_find_centroid (reference.array, NULL, NULL,
			  &x_lin, reference.radius, &y_lin, reference.radius);
    x = x_lin;
    y = y_lin;
    wcs_astro_transform (reference.ap, FALSE, 1,
			 &x, FALSE, &y, FALSE, NULL, FALSE,
			 0, NULL, NULL);
    index = reference.num_points;
    if (last_click_was_reference) --index;
    reference.ra[index] = x;
    reference.dec[index] = y;
    if (!last_click_was_reference) ++reference.num_points;
    last_click_was_reference = TRUE;
    kwin_resize (canvas_get_pixcanvas (wc), FALSE, 0, 0, 0, 0);
    kwin_refresh_if_visible (reference.mag_canvas, FALSE);
    return (TRUE);
}   /*  End Function reference_event_func  */

flag target_event_func (ViewableImage vimage, double x, double y,
			void *value, unsigned int event_code,
			void *e_info, void **f_info,
			double x_lin, double y_lin, unsigned int value_type)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas wc;
    extern flag last_click_was_reference;
    extern dataset_t reference, target;
    /*static char function_name[] = "target_event_func";*/

    if (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) return (FALSE);
    if (last_click_was_reference)
    {
	last_click_was_reference = FALSE;
    }
    else
    {
	if (reference.num_points < 1)
	{
	    ring_bell ();
	    return (TRUE);
	}
    }
    wc = viewimg_get_worldcanvas (vimage);
    iarray_find_centroid (target.array, NULL, NULL,
			  &x_lin, target.radius, &y_lin, target.radius);
    target.num_points = reference.num_points;
    target.ra[target.num_points - 1] = x_lin;
    target.dec[target.num_points - 1] = y_lin;
    kwin_resize (canvas_get_pixcanvas (wc), FALSE, 0, 0, 0, 0);
    kwin_refresh_if_visible (target.mag_canvas, FALSE);
    return (TRUE);
}   /*  End Function target_event_func  */

void main_refresh_func (KWorldCanvas canvas, int width, int height,
			struct win_scale_type *win_scale,
			Kcolourmap cmap, flag cmap_resize, void **info,
			PostScriptPage pspage,
			unsigned int num_areas,
			KPixCanvasRefreshArea *areas, flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    dataset_t *dataset = (dataset_t *) *info;
    unsigned int count;
    unsigned long pixel_value;
    char txt[STRING_LENGTH];
    double px[MAX_PAIRS], py[MAX_PAIRS];

    pixel_value = get_colour (canvas, "green");
    pixcanvas = canvas_get_pixcanvas (canvas);
    canvas_convert_from_canvas_coords (canvas, FALSE, dataset->target,
				       dataset->num_points,
				       dataset->ra, dataset->dec, px, py);
    for (count = 0; count < dataset->num_points; ++count)
    {
	kwin_draw_ellipse (pixcanvas, px[count], py[count],
			   ELLIPSE_RADIUS, ELLIPSE_RADIUS, pixel_value);
	sprintf (txt, "%u", count);
	kwin_draw_string (pixcanvas, px[count], py[count] - ELLIPSE_RADIUS * 2,
			  txt, pixel_value, FALSE);
    }
}   /*  End Function main_refresh_func  */

flag mag_event_func (ViewableImage vimage, double x, double y,
		     void *value, unsigned int event_code,
		     void *e_info, void **f_info,
		     double x_lin, double y_lin, unsigned int value_type)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see [<canvas_register_transform_func>]) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    unsigned int hdim, vdim;
    double cx, cy, px, py;
    array_desc *arr_desc;
    dataset_t *dataset = (dataset_t *) *f_info;

    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_END);
    cx = dataset->x_lin;
    ds_convert_coordinates (arr_desc->dimensions[hdim], 1, &cx, TRUE, FALSE,
			    FALSE);
    cy = dataset->y_lin;
    ds_convert_coordinates (arr_desc->dimensions[vdim], 1, &cy, TRUE, FALSE,
			    FALSE);
    px = x_lin;
    ds_convert_coordinates (arr_desc->dimensions[hdim], 1, &px, TRUE, FALSE,
			    FALSE);
    py = y_lin;
    ds_convert_coordinates (arr_desc->dimensions[vdim], 1, &py, TRUE, FALSE,
			    FALSE);
    dataset->radius = sqrt ( (cx - px) * (cx - px) + (cy - py) * (cy - py) );
    canvas_resize (viewimg_get_worldcanvas (vimage), NULL, FALSE);
    return (TRUE);
}   /*  End Function mag_event_func  */

void mag_refresh_func (KWorldCanvas canvas, int width, int height,
		       struct win_scale_type *win_scale,
		       Kcolourmap cmap, flag cmap_resize, void **info,
		       PostScriptPage pspage,
		       unsigned int num_areas,
		       KPixCanvasRefreshArea *areas, flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    ViewableImage vimage;
    unsigned int hdim, vdim, count;
    unsigned long pixel_value;
    double cx, cy, rx, ry, scale;
    array_desc *arr_desc;
    dataset_t *dataset = (dataset_t *) *info;

    vimage = viewimg_get_active (canvas);
    if (vimage == NULL) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    pixel_value = get_colour (canvas, "green");
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &dataset->x_lin, &dataset->y_lin,
				       &cx, &cy);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_END);
    rx = dataset->x_lin;
    ds_convert_coordinates (arr_desc->dimensions[hdim], 1, &rx, TRUE, FALSE,
			    FALSE);
    rx += dataset->radius;
    ry = dataset->y_lin;
    ds_convert_coordinates (arr_desc->dimensions[vdim], 1, &ry, TRUE, FALSE,
			    FALSE);
    ry += dataset->radius;
    rx = ds_get_coordinate (arr_desc->dimensions[hdim], rx);
    ry = ds_get_coordinate (arr_desc->dimensions[vdim], ry);
    if ( DOUBLE_IS_BLANK (rx) || DOUBLE_IS_BLANK (ry) ) return;
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &rx, &ry, &rx, &ry);
    rx = fabs (cx - rx);
    ry = fabs (cy - ry);
    kwin_draw_ellipse (pixcanvas, cx, cy, rx, ry, pixel_value);
    /*  Draw the outer ring  */
    scale = (dataset->radius + 2.0) / dataset->radius;
    if (scale < 1.1) scale = 1.1;
    rx = dataset->x_lin;
    ds_convert_coordinates (arr_desc->dimensions[hdim], 1, &rx, TRUE, FALSE,
			    FALSE);
    rx += dataset->radius * scale;
    ry = dataset->y_lin;
    ds_convert_coordinates (arr_desc->dimensions[vdim], 1, &ry, TRUE, FALSE,
			    FALSE);
    ry += dataset->radius * scale;
    rx = ds_get_coordinate (arr_desc->dimensions[hdim], rx);
    ry = ds_get_coordinate (arr_desc->dimensions[vdim], ry);
    if ( DOUBLE_IS_BLANK (rx) || DOUBLE_IS_BLANK (ry) ) return;
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &rx, &ry, &rx, &ry);
    rx = fabs (cx - rx);
    ry = fabs (cy - ry);
    kwin_set_attributes (pixcanvas, KWIN_ATT_LINE_DASH, TRUE, KWIN_ATT_END);
    kwin_draw_ellipse (pixcanvas, cx, cy, rx, ry, pixel_value);
    kwin_set_attributes (pixcanvas, KWIN_ATT_LINE_DASH, FALSE, KWIN_ATT_END);
    /*  Draw a crosshair at the centroid position  */
    pixel_value = get_colour (canvas, "red");
    for (count = 0; count < dataset->num_points; ++count)
    {
	canvas_convert_from_canvas_coords (canvas, FALSE, dataset->target, 1,
					   dataset->ra + count,
					   dataset->dec + count,
					   &cx, &cy);
	if ( (cx < 0.0) || (cy < 0.0) ) continue;
	if ( (cx > width - 1) || (cy > height - 1) ) continue;
	/*  This one is visible  */
	kwin_draw_line (pixcanvas, cx - CROSSHAIR_SIZE, cy - CROSSHAIR_SIZE,
			cx + CROSSHAIR_SIZE, cy + CROSSHAIR_SIZE, pixel_value);
	kwin_draw_line (pixcanvas, cx - CROSSHAIR_SIZE, cy + CROSSHAIR_SIZE,
			cx + CROSSHAIR_SIZE, cy - CROSSHAIR_SIZE, pixel_value);
    }
}   /*  End Function mag_refresh_func  */


/*  Private functions follow  */

static unsigned long get_colour (KWorldCanvas canvas, CONST char *colourname)
/*  [SUMMARY] Get colour.
    <canvas> The world canvas.
    <colourname> The name of the colour.
    [RETURNS] The specified colour if available, else a default colour.
*/
{
    unsigned long pixel_value;

    if ( canvas_get_colour (canvas, colourname, &pixel_value, NULL,NULL,NULL) )
    {
	return (pixel_value);
    }
    if ( canvas_get_colour (canvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "Error allocating colour: \"%s\" defaulting to white\n",
		 colourname);
	return (pixel_value);
    }
    fprintf (stderr, "Error allocating colour: \"%s\" defaulting to 0\n",
	     colourname);
    return (0);
}   /*  End Function get_colour  */
