/*  main_xt.c

    Main file for  kpolar  (X11 cartesian->polar tool for Karma).

    Copyright (C) 1997-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive generation of an astronomical
    co-ordinate system on a polar image, using a cartesian image.
    This module runs on an X11 server.


    Written by      Richard Gooch   5-AUG-1997: Copied from koords/main_xt.c

    Updated by      Richard Gooch   6-AUG-1997

    Updated by      Richard Gooch   31-OCT-1997: Took account of change to
  <XkwDataclipNewArray>.

    Updated by      Richard Gooch   7-DEC-1997: Added font resource for new
  intensity menu and removed code to set pan centre for magnifier as it is now
  done automatically by the MagnifierPopup widget.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   30-MAY-1998: Added track display of ellipse
  parameters.

    Updated by      Richard Gooch   12-AUG-1998: Update cursor position on all
  canvas events.

    Updated by      Richard Gooch   2-SEP-1998: Removed overlay code.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   21-APR-1999: Added Dataclip control for
  polar window.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   31-AUG-1999: Took account of interface
  change to function <viewimg_track _compute>. Added 'l' key binding in
  function <track_car_event>.

    Updated by      Richard Gooch   9-SEP-1999: Removed track display code and
  made use of XkwNupdateTrackLabels resource for ImageDisplay widget.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   22-JUN-2000: Removed unnecessary #include's

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   6-JUN-2001: Select visible canvas for polar
  window. Trap small radii in <geom_func1>.

    Updated by      Richard Gooch   3-OCT-2001: Added translations for Filewin.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Last updated by Richard Gooch   4-OCT-2003: Added #include <Xkw/Dataclip.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_version.h>
#define NEW_WIN_SCALE
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_wcs.h>
#include <karma_dir.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_xc.h>
#include <karma_ic.h>
#include <karma_ex.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Filewin.h>
#include <Xkw/Dataclip.h>
#include "kpolar.h"


#define VERSION "1.1.5"


/*  Private functions  */
STATIC_FUNCTION (void process_and_setup,
		 (multi_array *multi_desc, CONST char *filename,
		  iarray *array, KwcsAstro *ap,
		  ViewableImage *image, ViewableImage *magnified_image,
		  Widget shell, Widget image_display) );
STATIC_FUNCTION (flag data_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag geom_func1,
		 (KPixCanvasEditorContext context,KPixCanvas canvas,
		  unsigned int object_type, unsigned int num_coords,
		  CONST int *x, CONST int *y,
		  unsigned int raw_event_code, void *raw_e_info,
		  void **f_info) );


/*  Private data  */
String fallback_resources[] =
{
    "Kpolar.geometry:                                      +0+0",
    "Kpolar*pseudoColourCanvas*background:                 black",
    "Kpolar*directColourCanvas*background:                 black",
    "Kpolar*trueColourCanvas*background:                   black",
    "Kpolar*pseudoColourStereoCanvas*background:           black",
    "Kpolar*directColourStereoCanvas*background:           black",
    "Kpolar*trueColourStereoCanvas*background:             black",
    "Kpolar*pseudoColourCanvas*foreground:                 white",
    "Kpolar*Command*background:                            grey70",
    "Kpolar*Repeater*background:                           grey70",
    "Kpolar*Ktoggle*background:                            grey80",
    "Kpolar*closeButton*background:                        grey90",
    "Kpolar*ChoiceMenu.background:                         turquoise",
    "Kpolar*ExclusiveMenu.background:                      turquoise",
    "Kpolar*Value*background:                              #d0a0a0",
    "Kpolar*ImageDisplay*quit*background:                  orange",
    "Kpolar*SimpleSlider.foreground:                       Sea Green",
    "Kpolar*ScrollableList.highlightPixel:                 Sea Green",
    "Kpolar*SimpleProgressMeter.background:                turquoise3",
    "Kpolar*SimpleProgressMeter.foreground:                grey60",
    "Kpolar*polShell.geometry:                             -0+0",
    "Kpolar*ImageDisplay*trackLabel0*font:                 8x13bold",
    "Kpolar*ImageDisplay*trackLabel1*font:                 8x13bold",
    "Kpolar*ImageDisplay*trackLabel2*font:                 8x13bold",
    "Kpolar*ImageDisplay*trackLabel3*font:                 8x13bold",
    "Kpolar*ExclusiveMenu.font:                            8x13bold",
    "Kpolar*ImageDisplay*SimpleMenu*font:                  10x20",
    "Kpolar*ImageDisplay.cmapSize:                         100",
    "Kpolar*ImageDisplay.Command.font:                     7x13bold",
    "Kpolar*ImageDisplay.applicationBox.Command.font:      7x13bold",
    "Kpolar*ImageDisplay.applicationBox.Ktoggle.font:      7x13bold",
    "Kpolar*ImageDisplay.ChoiceMenu.font:                  7x13bold",
    "Kpolar*ImageDisplay.exportMenu.menuButton.font:       7x13bold",
    "Kpolar*ImageDisplay.overlayMenu.menuButton.font:      7x13bold",
    "Kpolar*ImageDisplay*zoomMenu*Unzoom*foreground:       red",
    "Kpolar*Dataclip.AutoValueScale:                       False",
    "Kpolar*SimpleSlider.borderWidth:                      0",
    "Kpolar*IntensityPolicy*autoIntensityScale:            False",
    "Kpolar*ZoomPolicy*integerXZoom:                       False",
    "Kpolar*ZoomPolicy*integerYZoom:                       False",
    "Kpolar*Filewin.dirname.textSink.font:                 6x13",
    "Kpolar*font:                                          9x15bold",
    "Kpolar*borderColor:                                   black",
    "Kpolar*background:                                    aquamarine",
    "Kpolar*foreground:                                    black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget car_image_display = NULL;
static KWorldCanvas car_canvas = NULL;
static Widget pol_shell = NULL;
static Widget pol_image_display = NULL;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KWorldCanvas pc_canvas, true_canvas;
    KPixCanvas pixcanvas;
    KPixCanvasEditorContext ctx;
    XtAppContext app_context;
    Widget filewin, filepopup, box;
    extern KWorldCanvas pol_canvas;
    extern unsigned long car_pixel_value;
    extern char title_name[STRING_LENGTH];

    /*  Initialise module  */
    im_register_module_name ("kpolar");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kpolar",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 100,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (1, -1);
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    /*  Setup the cartesian window  */
    car_image_display = XtVaCreateManagedWidget ("topForm",
						 imageDisplayWidgetClass,
						 main_shell,
						 XtNborderWidth, 0,
						 XkwNnumTrackLabels, 4,
						 XkwNupdateTrackLabels, TRUE,
						 NULL);
    filepopup = XtNameToWidget (car_image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
		   XkwNpreferFloat, TRUE,
		   XkwNdomainName, "FILE",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (data_func, NULL);
    /*  Get the application box  */
    box = XtNameToWidget (car_image_display, "applicationBox");
    XtRealizeWidget (main_shell);
    XtVaGetValues (car_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    car_canvas = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    if (car_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (car_image_display,
		   XkwNvisibleCanvas, car_canvas,
		   NULL);
    pixcanvas = canvas_get_pixcanvas (car_canvas);
    if (car_canvas == true_canvas)
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_GREEN_MASK, &car_pixel_value,
			     KWIN_ATT_END);
    }
    else
    {
	if ( !kwin_get_colour (pixcanvas, "green", &car_pixel_value,
			       NULL, NULL, NULL) )
	{
	    /*  Can't get green: try white instead  */
	    if ( !kwin_get_colour (pixcanvas, "white", &car_pixel_value,
				   NULL, NULL, NULL) )
	    {
		/*  Can't even get white (this should never actually happen
		    with X), so just pick 0  */
		car_pixel_value = 0;
	    }
	}
    }
    ctx = kwin_editor_create_context (pixcanvas,
				      KWIN_EDITOR_CONTEXT_ATT_MOUSE,
				      K_CANVAS_EVENT_RIGHT_MOUSE_MASK,
				      KWIN_EDITOR_CONTEXT_ATT_EDIT_OBJECT,
				      KWIN_EDIT_OBJECT_ELLIPSE,
				      KWIN_EDITOR_CONTEXT_ATT_PIXEL_VALUE,
				      car_pixel_value,
				      KWIN_EDITOR_CONTEXT_ATT_CLEAR_OBJECT,
				      FALSE,
				      KWIN_EDITOR_CONTEXT_ATT_END);
    kwin_editor_register_event_func (ctx, geom_func1, car_canvas);
    kwin_editor_register_event_func (ctx, geom_func, car_canvas);
    canvas_register_refresh_func (car_canvas, car_refresh_func, NULL);
    /*  Setup the polar window  */
    pol_shell = XtVaCreatePopupShell ("polShell", topLevelShellWidgetClass,
				      main_shell,
				      XtNtitle, "Polar Window",
				      NULL);
    pol_image_display = XtVaCreateManagedWidget ("topForm",
						 imageDisplayWidgetClass,
						 pol_shell,
						 XtNborderWidth, 0,
						 XkwNcreateFilepopup, FALSE,
                                                 XkwNnumTrackLabels, 2,
						 XkwNupdateTrackLabels, TRUE,
						 XkwNshowQuitButton, FALSE,
						 NULL);
    /*  Get application box of polar window */
    box = XtNameToWidget (pol_image_display, "applicationBox");
    XtRealizeWidget (pol_shell);
    XtVaGetValues (pol_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    pol_canvas = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    if (pol_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (pol_image_display,
		   XkwNvisibleCanvas, pol_canvas,
		   NULL);
    XtPopup (pol_shell, XtGrabNone);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

void update_polar_dataclip (iarray array, double min, double max)
/*  [SUMMARY] Update the polar Dataclip widget.
    <array> The new polar image.
    <min> The minimum image value.
    <max> The maximum image value.
    [RETURNS] Nothing.
*/
{
    Widget izoomwinpopup;

    izoomwinpopup = XtNameToWidget (pol_image_display, "izoomwinpopup");
    XkwDataclipNewArray (izoomwinpopup, array, min, max, TRUE);
}   /*  End Function update_polar_dataclip  */


/*  Private routines follow  */

static void process_and_setup (multi_array *multi_desc, CONST char *filename,
			       iarray *array, KwcsAstro *ap,
			       ViewableImage *image,
			       ViewableImage *magnified_image,
			       Widget shell, Widget image_display)
{
    KWorldCanvas canvas, mag_canvas;
    double min, max;
    Widget izoomwinpopup, magnifier;
    char *ptr;
    char stripped_filename[STRING_LENGTH];
    char title[STRING_LENGTH];
    extern char title_name[STRING_LENGTH];

    strcpy (stripped_filename, filename);
    if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
    {
	if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
    }
    izoomwinpopup = XtNameToWidget (image_display, "izoomwinpopup");
    XtVaGetValues (image_display,
		   XkwNpseudoColourCanvas, &canvas,
		   XkwNmagnifier, &magnifier,
		   NULL);
    XtVaGetValues (magnifier, XkwNpseudoColourCanvas, &mag_canvas, NULL);
    if (canvas == NULL)
    {
	XtVaGetValues (image_display,
		       XkwNtrueColourCanvas, &canvas,
		       NULL);
	XtVaGetValues (magnifier, XkwNtrueColourCanvas, &mag_canvas, NULL);
    }
    if ( !process_image (multi_desc, array, ap, canvas, mag_canvas,
			 image, magnified_image, &min, &max) ) return;
    sprintf (title, "%s  file: %s\n", title_name, filename);
    XtVaSetValues (shell,
		   XtNtitle, title,
		   NULL);
    XtVaSetValues (image_display,
		   XkwNimageName, stripped_filename,
		   NULL);
    XkwDataclipNewArray (izoomwinpopup, *array, min, max, TRUE);
}   /*  End Function process_and_setup  */

static flag data_func (void *object, multi_array *multi_desc,
		       CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new cartesian image event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    char filename[STRING_LENGTH];
    extern iarray car_array;
    extern KwcsAstro cartesian_ap;
    static ViewableImage car_vimage = NULL;
    static ViewableImage car_magnified_vimage = NULL;
    /*static char function_name[] = "cartesian_data_func";*/

    if (strcmp (domain, "FILE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    process_and_setup (multi_desc, filename, &car_array, &cartesian_ap,
		       &car_vimage,
		       &car_magnified_vimage, main_shell, car_image_display);
    return (TRUE);
}   /*  End Function data_func  */

static flag geom_func1 (KPixCanvasEditorContext context, KPixCanvas canvas,
			unsigned int object_type, unsigned int num_coords,
			CONST int *x, CONST int *y,
			unsigned int raw_event_code, void *raw_e_info,
			void **f_info)
/*  [SUMMARY] Process a geometry editor event on a pixel canvas.
    <context> The KPixCanvasEditorContext.
    <canvas> The pixel canvas on which the event occurred.
    <object_type> The type of the geometric object.
    <num_coords> The number of co-ordinates in the geometric object.
    <x> The array of horizontal co-ordinates for the geometric object.
    <y> The array of vertical co-ordinates for the geometric object.
    <raw_event_code> The raw arbitrary event code.
    <raw_e_info> A pointer to the raw arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas worldcanvas = (KWorldCanvas) *f_info;
    double cx, cy, rx, ry, rotation, ratio;
    Widget track_label;
    char txt[STRING_LENGTH];
    extern iarray car_array;
    /*static char function_name[] = "geom_func1";*/

    if (car_array == NULL) return (FALSE);
    cx = x[0];
    cy = y[0];
    rx = x[1];
    ry = y[1];
    if ( (rx < 2.0) || (ry < 2.0) )
    {
	fprintf (stderr, "Radii (%g,%g) too small\n", rx, ry);
	return (TRUE);
    }
    rotation = (double) x[2] / 64.0;
    canvas_convert_ellipse_coords (worldcanvas, TRUE, FALSE,
				   &cx, &cy, &rx, &ry, &rotation);
    ratio = (rx < ry) ? (ry / rx) : (rx / ry);
    sprintf (txt, "Axis ratio: %g  Position Angle: %.2f degrees",
	     ratio, 90.0 - rotation);
    track_label = XtNameToWidget (car_image_display, "trackLabel3");
    XtVaSetValues (track_label, XtNlabel, txt, NULL);
    return (FALSE);
}   /*  End Function geom_func1  */
