/*  pvget.c

    Slice file for  kpvslice  (X11 Position-Velocity slice tool for Karma).

    Copyright (C) 1996-1997  Vincent McIntyre

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Vincent McIntyre may be reached by email at  vjm@davinci.sci.uow.edu.au
    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive selection of a Position-Velocity
    slice through a data cube.
    This module runs on an X11 server.


    Written by      Vincent McIntyre 12-JUN-1996

    Last updated by Vincent McIntyre 12-AUG-1996; minor tidyup; dump obsolete
  functions and optimise multithreading. New antialias routine.

    Updated by      Richard Gooch   26-SEP-1996: Changed to implement the new
  FITS-style co-ordinate handling, where dimension co-ordinates range from 0
  to length - 1.

    Updated by      Richard Gooch   12-OCT-1996: Removed unthreaded code (was
  disabled anyway).

    Updated by      Richard Gooch   28-OCT-1996: Changed from <abs> to <fabs>.

    Updated by      Richard Gooch   2-FEB-1997: Made use of new
  <wcs_astro_get_associated> function.

    Updated by      Richard Gooch   23-FEB-1997: Made slice X axis "LINEAR"
  rather than "ANGLE", since the slice locus may be arbitrary.

    Updated by      Richard Gooch   19-JUN-1997: Went back to slice X axis
  "ANGLE". Worry about arbitrary slice locus later.

    Updated by      Richard Gooch   21-MAR-2001: Deallocated temporary
  Intelligent Arrays in <pvslice>.

    Updated by      Richard Gooch   5-JAN-2002: Moved command-line code into
  kpvget.c.

    Last updated by Richard Gooch   7-JAN-2002: Copy velocity axis co-ordinate
  information to pvslice image.


             **********  THIS IS A WORK IN PROGRESS  **********

*/
/* #define TESTSUITE */        /* diagnostic messages */
/* #define GAUSSWT */          /* a new addition; summing over gaussian beam */
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_mt.h>
#include <karma_m.h>


#define MAX_SAMPLES 40  /*  10 points, sample 4 neighbours each  */


/* have to make a structure to pass to each thread, as I need >4 params and
   mt_launch_job only allows up to four. */
struct nbjobpars
{
    int ypix;
    int xpix;
    int slpix;
    int yp;
    int xp;
    int MAXNB;
};
struct chjobpars
{
    int    lpix;
    int    zpix;
    int    ypix;
    int    xpix;
    int    nch;
    int    MAXNB;
    iarray nb;
    iarray wt;
};

struct sample_point
{
    unsigned int x, y;
    float weight;
};

struct locus_point
{
    unsigned int num_samples;
    struct sample_point samples[MAX_SAMPLES];
};

static struct locus_point *locus_points = NULL;
static unsigned int max_locus_points = 0;


/*----------------------------------------------------------------------------*/
static float weight (float x, float y, float cx, float cy) {
 /*
 [PURPOSE] Calculate weight to give when summing contributions from neighbours.
 [INPUTS]
  <x>, <y>  current neighbour to calculate weight for.
 <cx>,<cy>  centre point.
 These are assumed to be in PIXEL coordinates.
 [NOTES]
 In this version I make the weight proportional to the product distance
 from the projected centre of the current slice pixel. The pixel this point
 lands on will get highest weight, but the near neighbours will get some
 weight too. This version avoids a division but ensuring positive values
 is costly?
  */

    float dx, dy, w;

    dx = 1 - (x-cx); dy = 1 - (y-cy); w = fabs(dx*dy); 
    return(w);
} /* end weight() */
/*---------------------------------------------------------------------------*/

#ifdef DISABLED
static float oldweight (float x, float y, float cx, float cy) {
 /* calculates weight to give when summing contributions from neighbours.
    In this version I make the weight inversely proportional to the distance
    from the projected centre of the current slice pixel. The pixel this point
    lands on will get highest weight, but the near neighbours will get some
    weight too.
  */

  float dx,dy,r,w;

  dx = x-cx; dy = y-cy;
  r = sqrt((dx*dx+dy*dy));
  if (r>0.0) {
    w = 1.0/r;
  }else{
    w = 1.0;
  }
  return(w);
} /* end weight() */
#endif
/*---------------------------------------------------------------------------*/
static flag check_bounds (int *x1,  int *y1,  int *x2,  int *y2,  int xpix,
		         int ypix, int *blx, int *bly, int *trx, int *try)
{
  /* checks bounds & if endpoints are outside the area of cube's front face,
     computes overlapping section of slice & returns new endpoints */

  int ix1,iy1,ix2,iy2,swapped,v[5],w[5];
  float f,x,y;

  swapped=0;
  ix1 = *x1; iy1 = *y1;  ix2 = *x2; iy2 = *y2;
/*fprintf(stderr, "initialpoints; (%d,%d) (%d,%d)\n",ix1,iy1,ix2,iy2);*/

  /* do everything assuming first point is left of second, swap back later */
  if (ix2 < ix1) {
    ix1 = *x2; ix2 = *x1; iy1 = *y2; iy2 = *y1; swapped=1;
  } else {
    if ( ix1 == ix2 && iy1 > iy2 ) {
      iy1 = *y2; iy2 = *y1; swapped=2;
    }
  }
/*  if (swapped != 0) {
fprintf(stderr, "swappedpoints; (%d,%d) (%d,%d)\n",ix1,iy1,ix2,iy2);
  }*/

  *blx = ix1; *bly = iy1; *trx = ix2; *try = iy2;

  /* handle no-overlap cases: the extended slice line may overlap eventually,
   but the piece within the endpoints does not. So stop here. */
  if ( (ix1 < 0 && ix2 < 0) || (ix1 >= xpix && ix2 >= xpix) ) {
    fprintf(stderr, "PVGET: Illegal x-range\n"); return(FALSE);
  }
  if ( (iy1 < 0 && iy2 < 0) || (iy1 >= ypix && iy2 >= ypix) ) {
    fprintf(stderr, "PVGET: Illegal y-range\n"); return(FALSE);
  }

  /* single pixel case */
  if ( ix1 == ix2 && iy1 == iy2 ) {
    fprintf(stderr, "PVGET: Line segment too short\n"); return(FALSE);
  }

  /* handle the case of a vertical slice */
  if ( ix1 == ix2 ) {

      if (ix1 < 0 || ix1 >= xpix) {
	fprintf(stderr, "PVGET: Illegal x-range\n"); return(FALSE);
      }else{
	*blx = ix1; *trx = ix2;
	/* now check if either of the y coords are good */
	if (iy1 < 0 || iy1 >= ypix) { *bly = 0;}
	if (iy2 < 0 || iy2 >= ypix) { *try = ypix-1;}
      }
  }else{
    /* handle the case of a horizontal slice */
    if ( iy1 == iy2 ) {

      if (iy1 < 0 || iy1 >= ypix) {
	fprintf(stderr, "PVGET: Illegal y-range\n"); return(FALSE);
      }else{    
	*bly = iy1; *try = iy2;
	/* now check if either of the x coords are good */
	if (ix1 < 0 || ix1 >= xpix) { *blx = 0;}
	if (ix2 < 0 || ix2 >= xpix) { *trx = xpix-1;}
      }
    }else{
      /* calculate intercepts with x=0, y=0, x=xpix-1, y=ypix-1
	 store x-intercepts in v[] and y-intercepts in w[]       */
      f = ( (float)(iy2) - (float)(iy1) ) / ( (float)(ix2) - (float)(ix1) );
      
      v[1]=0; w[2]=0; v[3]=xpix-1; w[4]=ypix-1;
      x=0.0;    y=(float)(iy1) + (x-(float)(ix1))*f;          w[1]=(int)(y+0.5);
      y=0.0;    x=(float)(ix1) + (y-(float)(iy1))/f;          v[2]=(int)(x+0.5);
      x=(float)(xpix-1); y=(float)(iy1) + (x-(float)(ix1))*f; w[3]=(int)(y+0.5);
      y=(float)(ypix-1); x=(float)(ix1) + (y-(float)(iy1))/f; v[4]=(int)(x+0.5);
/*      fprintf(stderr,"v= %5d %5d %5d %5d\n",v[1],v[2],v[3],v[4]);
      fprintf(stderr,"w= %5d %5d %5d %5d\n",w[1],w[2],w[3],w[4]);*/

      /*  sample case (v[2]>0 && v[4]<xpix && w[3]<ypix)

               |          |
               |          |
      ---------============----------------
	       =          =    B
	       =          =  .
	       =          =.
	       =         .=
      ---------========.===-----------------
               |     .    |
               |    A     |
               |          |
      */

      /* things are different depending on if slope is +ve or -ve */
      /* For either, there are six cases of where the intercepts of the
	 slice line (extended to infinity in both directions) lie w.r.t. the
	 four lines noted above. See diagram. */
      if ( f >= 0.0 ) {
	if (v[2] < 0 ) {
	  if ( w[1] < ypix ) {
	    *blx = 0; *bly = w[1];
	    if ( w[3] < ypix ) {
	      *trx = xpix-1; *try = w[3];
	    }else{
	      *trx = v[4]; *try = ypix-1;
	    }
	  }else{
	    fprintf(stderr, "PVGET: Out of bounds\n"); return(FALSE);
	  }
	}else{
	  if ( v[2] < xpix ) {
	    *bly = 0; *blx = v[2];
	    if ( v[4] < xpix ) {
	      *try = ypix-1; *trx = v[4];
	    }else{
	      *try = w[3]; *trx = xpix-1;
	    }
	  }else{
	    fprintf(stderr, "PVGET: Out of bounds\n"); return(FALSE);
	  }
	}
	/* end +ve slope cases */
      }else{
	if ( v[4] < 0 ) {
	  if ( w[1] > 0 ) {
	    *blx = 0; *bly = w[1];
	    if ( w[3] > 0 ) {
	      *trx = xpix-1; *try = w[3];
	    }else{
	      *trx = v[2]; *try = 0;
	    }
	  }else{
	    fprintf(stderr, "PVGET: Out of bounds\n"); return(FALSE);
	  }
	}else{
	  if ( v[4] < xpix ) {
	    *blx = v[4]; *bly = ypix-1;
	    if ( v[2] < xpix ) {
	      *trx = v[2]; *try = 0;
	    }else{
	      *trx = xpix-1; *try = w[3];
	    }
	  }else{
	    fprintf(stderr, "PVGET: Out of bounds\n"); return(FALSE);
	  }
	}
      } /* end -ve slope cases */
    } /* end finte-slope cases */
  } /* end normal (nonvertical) cases */

  /* now check to see if either endpoint NEEDS to be replaced:
     if they are legal return the input value */
  if ( (ix1 >= 0 && ix1 < xpix) && (iy1 >= 0 && iy1 < ypix) ) {
    *blx = ix1; *bly = iy1;
  }
  if ( (ix2 >= 0 && ix2 < xpix) && (iy2 >= 0 && iy2 < ypix) ) {
    *trx = ix2; *try = iy2;
  }

  /* swap ends back if necessary */
  if (swapped > 0) {
    ix1 = *blx; *blx = *trx; *trx = ix1;
    iy1 = *bly; *bly = *try; *try = iy1;
  }

  return(TRUE);
  
} /* end check_bounds */
/*-----------------------------------------------------------------------------*/
unsigned int get_slice(iarray cube, int x1, int y1, int x2, int y2,
			double **x_locus, double **y_locus) {

  /* [PURPOSE] determine a locus of pixels in a slice line, from two endpoints */
  /* [NOTES] 
     * The coordinates are returned via pointers to two dynamically allocated
     arrays, written to *x_locus and *y_locus; these must be externally freed.
     * The values have to be reals rather than integers, so that
     antialiasing can be done. Double-precision is used because this is the
     `least common denominator for C.
     * To improve antialiasing, calculate pixels that are sqrt(0.5) apart.
     This changes the angular scale but doesn't matter if I write proper headers

     * This function is designed to extract a straight-line slice.
     Once we allow arbitrary extraction loci, the approach needs to change.
     Then, one has to check on every pixel. If a locus vertex is outside the cube
     extent, just blank it.
     This could be used in implementing arbitrary loci, by chaining together line
     segments but would have to trap the case of going off the edge; blank
     the off-edge pixels in the trace, but keep the vertex points.
     [RETRUNS] Number of pixels in slice.
  */

  int    xpix, ypix, zpix, blx, bly, trx, try;
  int    i, slicepixels;

  float  dx, dy, theta, ctheta, stheta, slicelength;
  double *xloc, *yloc;

  zpix = iarray_dim_length (cube, 0);
  ypix = iarray_dim_length (cube, 1);
  xpix = iarray_dim_length (cube, 2);
  if ( !(xpix*ypix*zpix > 0) ) {
    fprintf (stderr, "GETSLICE: problem with input cube dimensions\n");
  }

  if ( !check_bounds(&x1,&y1,&x2,&y2,xpix,ypix,&blx,&bly,&trx,&try) ) return (0);
#ifdef CMDLINE
  fprintf(stdout, "endpoints are %d %d %d %d\n",blx,bly,trx,try);
#endif
  /* now calculate slice length */
  dx    = (float) (trx-blx+1);
  dy    = (float) (try-bly+1);
  if (blx==trx) { 
    theta = PI/2.0;       ctheta=0.0;          stheta=1.0;
    if ( bly>try ) {theta = 3.0*PI/2.0;        stheta=-1.0; }
  }else{
    theta = atan2(dy,dx); ctheta = cos(theta); stheta = sin(theta);
  }

  slicelength = sqrt(dx*dx+dy*dy);
  slicepixels = (int)(slicelength)+1;

/* allocate enough memory. This has to be deallocated OUTSIDE, once
     the array has been passed to pvslice, and so is finished with */
  xloc = (double *) m_alloc (slicepixels * sizeof *xloc);
  yloc = (double *) m_alloc (slicepixels * sizeof *yloc);

#ifdef TESTSUITE
  fprintf (stderr, " x1: %d  y1: %d  x2: %d  y2: %d Theta: %6.2f\n", x1,y1,x2,y2,theta*180.0/PI);
  fprintf (stderr, "blx: %d bly: %d trx: %d try: %d Theta: %6.2f\n", x1,y1,x2,y2,theta*180.0/PI);
  fprintf (stderr, "dx: %8.3f dy: %8.3f sl: %10.3f sp: %d\n",dx,dy,slicelength,slicepixels);
  fprintf (stderr, "xpix: %3d  ypix: %3d zpix: %3d\n",xpix,ypix,zpix);
#endif

  /* loop over length of slice:
     NB: c-convention array sizing --- index always < size */
  for (i=0; i<slicepixels; i++) {
    /* these are for a straight-line locus; we may want to do other shapes */
    xloc[i] = (double)(i) * ctheta + (double)(blx);
    yloc[i] = (double)(i) * stheta + (double)(bly);
  }

  /* remember x,y_locus are passed in as pointers to pointers; need one deref */
  *x_locus = xloc;
  *y_locus = yloc;
  return(slicepixels);
} /* end get_slice() */
/*-----------------------------------------------------------------------------*/
  void get_n_chan(void *poolinfo,
		  void *callinfo1, void *callinfo2,
		  void *callinfo3, void *callinfo4,
		  void *threadinfo)
  /* [PURPOSE] Execute a number of channel-map slices on the same thread.
     [RETURNS] Nothing.
  */
{
  iarray cube               = (iarray)             callinfo1;
  int    z                  = (iaddr)              callinfo2;
  struct chjobpars *chstuff = (struct chjobpars *) callinfo3;
  iarray pvmap              = (iarray)             callinfo4;

  int    i,j,xp,yp,xn,yn,MAXNB;
  float  sw,f;
  iarray nb,wt;
  int    lpix,xpix,ypix,zpix;
  int    zz, nch;

  lpix =  chstuff->lpix;
  zpix =  chstuff->zpix;
  ypix =  chstuff->ypix;
  xpix =  chstuff->xpix;
  nch  =  chstuff->nch;
  wt   =  chstuff->wt;
  nb   =  chstuff->nb;
  MAXNB=  chstuff->MAXNB;

  /* fprintf(stderr,"  l=%d x=%d y=%d z=%d",lpix,xpix,ypix,z);*/
  if ( (z + nch) > zpix ) nch = zpix - z;
  for ( zz = 0; zz < nch; ++zz, ++z) {   /* do nch channels on this thread */
    for (i=0; i<lpix; i++) { /* start slice loop */
      f = 0.0; sw = 0.0;
      xp = I3(nb, 0,0,i);
      yp = I3(nb, 1,0,i);
      
      for (j=1; j<MAXNB; j++) {
	xn = I3(nb, 0,j,i);
	yn = I3(nb, 1,j,i);
	/* The first test protects the second; it is possible for xn and yn
	   to go out of range (neighbours of a pixel at edge of cube) */
	if ( F2(wt, j,i) >0.0 && F3(cube, z,yn,xn)!=TOOBIG ) {
	  sw = sw + F2(wt, j,i);
	  f  = f  + F3(cube, z,yn,xn) * F2(wt, j,i);
	}
      } /* end neighbours loop */
      
      if ( sw>0.0 ) {
	F2(pvmap, z,i) = f/sw;
      } else {
	if ( xp >= 0 && xp<xpix && yp>=0 && yp<ypix ) {
	  F2(pvmap, z,i) = F3(cube, z,yp,xp);
	}else {
	  F2(pvmap, z,i) = 0.0;
	}
      }
    } /* end slice loop */
  } /* end channels loop */
} /* end get_n_chan */
/*-----------------------------------------------------------------------------*/
iarray pvslice (iarray cube, unsigned int num_points, 
		double *x_locus, double *y_locus)
/*  threaded version */
/*  [PURPOSE] Extract pixels from a cube, along an input locus.
    Result is antialiased.
    [NOTES]
    The cube is assumed to have axes in x,y,v order.
    The output array has the same spatial scale as input cube, ie
    the scale is assumed to be the same for both spatial axes,
    and velocity pixels are given the same width as the channel spacing.
    Output is a weighted sum over all pixels nearby the point where the
    slice locus passes, to reduce aliasing effects.


    There are four cases to consider for the neighbour pixels, depending
    where within a pixel the hit occurs:

             |-----------|-----------|     |-----------|-----------|
             |           |           |	   |           |           |
        2.0  -     3     |     2     | 	   |     3     |     2     |
             |           |           |	   |       xy  |           |
             |-----------|-----------|	   |-----------|-----------|
             |           | xy        |	   |           |           |
        1.0  -     4     |     1     |	   |     4     |     1     |
             |           |           |	   |           |           |
             |-----|-----|-----|-----|	   |-----------|-----------|
                  1.0         2.0

             |-----------|-----------|     |-----------|-----------|
             |           |           |	   |           |           |
             |     3     |     2     | 	   |     3     |     2     |
             |           | xy        |	   |           |           |
             |-----------|-----------|	   |-----------|-----------|
             |           |           |	   |       xy  |           |
             |     4     |     1     |	   |     4     |     1     |
             |           |           |	   |           |           |
             |-----------|-----------|	   |-----------|-----------|

    All these can be handled by int(x+/-0.5), int(y+/-0.5)
    Pixel coordinates are associated with centres of pixels.
    Some pixels don't have four neighbours e.g. at the edges of the channel
    maps. If one of the 4 is missing, I 'mask' it with a zero weight.
    [RETURNS] "position-velocity" iarray, of length <num_points>.
*/
{
    iarray pvmap;
    iarray nb;
    int    xpix, ypix, zpix;
    int    i, j, z, xp, yp, MAXNB;
    int    zstep;
    unsigned int vel_axis;
    float  xc, yc, xx, yy;
    iarray wt;
    KThreadPool pool = NULL; /* need this for thread pool */
    struct chjobpars chstuff;
    double crval, crpix, cdelt;
    struct locus_point *lp;
    char txt[STRING_LENGTH];
    unsigned long dim_lengths[2];
    CONST char *dim_names[2];
    static char function_name[] = "pvslice";

    if (num_points < 2) return (NULL);
    if (max_locus_points < num_points)
    {
	if (locus_points != NULL) m_free (locus_points);
	if ( ( locus_points = m_alloc (num_points * sizeof *locus_points) )
	     == NULL )
	{
	    m_abort (function_name, "local array");
	}
	max_locus_points = num_points;
    }
    /* set number of `neighbour' pixels to use for anitaliasing */
    MAXNB = 5;

    /* NB for iarrays, and Karma genrally; 0th is most significant dimension */
    zpix = iarray_dim_length (cube, 0);
    ypix = iarray_dim_length (cube, 1);
    xpix = iarray_dim_length (cube, 2);
    if ( !(xpix*ypix*zpix > 0) )
    {
	fprintf (stderr, "PVGET: problem with input cube dimensions\n");
    }

#ifdef TESTSUITE
    fprintf (stderr, "locus is %d pixels long\n", num_points);
#endif
    dim_lengths[0] = zpix;
    dim_lengths[1] = num_points;
    dim_names[0] = iarray_dim_name (cube, 0);
    dim_names[1] = "ANGLE";
    pvmap = iarray_create (K_FLOAT, 2, dim_names, dim_lengths,
			   iarray_value_name (cube), NULL);
    wt     = iarray_create_2D (MAXNB, num_points, K_FLOAT);
    nb     = iarray_create_3D (2, MAXNB, num_points, K_INT);

    /* find the neighbour pixels & antialiasing weights.                    */
    /* this is done for one channel only, then list applied to all channels */

    for (i = 0, lp = locus_points; i < num_points; i++, ++lp)
    {
	xc = x_locus[i];              yc = y_locus[i];
	xp = (int)( xc+0.5 );         yp = (int)( yc+0.5 );
	/* here I only do the 4 nearest pixels */
	I3(nb,0,0,i) = xp;            I3(nb,1,0,i) = yp; /* pixel we landed on */
	I3(nb,0,1,i) = (int)(xc+0.5); I3(nb,1,1,i) = (int)(yc-0.5); /* neigbours */
	I3(nb,0,2,i) = (int)(xc+0.5); I3(nb,1,2,i) = (int)(yc+0.5);
	I3(nb,0,3,i) = (int)(xc-0.5); I3(nb,1,3,i) = (int)(yc+0.5);
	I3(nb,0,4,i) = (int)(xc-0.5); I3(nb,1,4,i) = (int)(yc-0.5);
	/* calculate the weight for each neighbour */
	for (j = 1; j < MAXNB; j++)
	{
	    xp = I3 (nb, 0, j, i); yp = I3 (nb, 1, j, i);
	    xx = (float) xp;   yy = (float) yp;
	    if ( (xp >= 0) && (xp < xpix) && (yp >= 0) && (yp < ypix) )
		F2 (wt, j, i) = weight (xx, yy, xc, yc);
	    else F2 (wt, j, i) = 0.0;
	}   /*  end neighbours loop  */
    }   /*  end slice loop  */

    /* apply neighbour list & weights to each x,y plane in turn  --------------- */
    /* do this in threaded code, to take advantage of multiple cpus if available */
    pool = mt_get_shared_pool ();
    zstep = zpix / mt_num_threads (pool);
    /* these are read-only values, so use only one (module-global) instance of the
       structure */
    chstuff.lpix = num_points;
    chstuff.zpix = zpix;
    chstuff.xpix = xpix;
    chstuff.ypix = ypix;
    chstuff.nch  = zstep;
    chstuff.MAXNB= MAXNB;
    chstuff.nb   = nb;
    chstuff.wt   = wt;
    for ( z = 0; z < zpix; z += zstep) {
	mt_launch_job (pool, get_n_chan, (void *) cube, (void *) (iaddr) z,
		       (void *) &chstuff, (void *) pvmap);
    } /* end channels loop */
    mt_wait_for_all_jobs (pool);

    iarray_dealloc (nb);
    iarray_dealloc (wt);
    vel_axis = iarray_get_fits_axis (cube, 0);
    if (vel_axis > 0)
    {
	iarray_put_named_string (pvmap, "CTYPE2", dim_names[0]);
	/* Copy velocity coords from input cube. Assumed to be zeroth axis of
	   iarray.
	   This will be wrong if pvslice doesn't extract all velocity channels,
	   but the way it is done above ensures this so. NB axes numbered from
	   0 */
	sprintf (txt, "CRVAL%u", vel_axis);
	if ( iarray_get_double (cube, txt, &crval, FALSE) )
	    iarray_put_double (pvmap, "CRVAL2", crval);
	sprintf (txt, "CRPIX%u", vel_axis);
	if ( iarray_get_double (cube, txt, &crpix, FALSE) )
	    iarray_put_double (pvmap, "CRPIX2", crpix);
	sprintf (txt, "CDELT%u", vel_axis);
	if ( iarray_get_double (cube, txt, &cdelt, FALSE) )
	    iarray_put_double (pvmap, "CDELT2", cdelt);
    }
    return (pvmap);
} /* end pvslice() */
