/*  kstopper.c

    Main file for  kstopper  (X11 stopper tool for Karma).

    Copyright (C) 1999-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will display a list of Karma GUI tools running on the
    local machine (per user), showing which are idle and which are busy. The
    tool can interrupt the other tools.
    This module runs on an X11 server.


    Written by      Richard Gooch   13-SEP-1999

    Updated by      Richard Gooch   26-JAN-2000: Improved responsiveness and
  reduced client event frequency.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <karma.h>
#include <k_version.h>
#define NEW_WIN_SCALE
#define X11
#include <k_event_codes.h>
#include <karma_xtmisc.h>
#include <karma_conn.h>
#include <karma_im.h>
#include <karma_ic.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/Canvas.h>


#define VERSION "1.0.1"


struct client_info
{
    int vpos;
    int height;
    int ascent;
    int descent;
    pid_t pid;
    flag busy;
    struct timeval last_response;
};


/*  Private functions  */
STATIC_FUNCTION (void refresh_func,
		 (KPixCanvas canvas, int width, int height,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (flag position_func,
		 (KPixCanvas canvas, int x, int y,
		  unsigned int event_code, void *e_info, void **f_info) );
STATIC_FUNCTION (flag open_func, (Connection connection, void **info) );
STATIC_FUNCTION (flag read_func,
		 (Connection connection, void **info, Channel channel) );
STATIC_FUNCTION (void close_func, (Connection connection, void *info) );
STATIC_FUNCTION (void timer_proc, (XtPointer client_data, XtIntervalId *id) );


/*  Private data  */
String fallback_resources[] =
{
    "Kstopper*background:                                 black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-cmap_master", ".topForm.cmapMaster", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-create_at_init", "*createAtInit", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static KPixCanvas canvas = NULL;
static KPixCanvasFont font = NULL;
static unsigned long interval = 5000;  /*  Milliseconds  */
static XtIntervalId xt_interval_id = 0;
static XtAppContext app_context = NULL;


int main (int argc, char **argv)
{
    Widget w;
    /*char title_name[STRING_LENGTH];*/
    /*static char function_name[] = "main";*/

    /*  Initialise module  */
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kstopper",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
#if 0
					     XTMISC_INIT_ATT_TITLE, title_name,
#endif
					     XTMISC_INIT_ATT_END,
					     NULL);
    conn_register_server_protocol ("kstopper", 0, 0,
				   open_func, read_func, close_func);
    /*xtmisc_set_icon (main_shell, ic_write_?_icon);*/
    w = XtVaCreateManagedWidget ("canvas", canvasWidgetClass, main_shell,
				 XtNborderWidth, 0,
				 XtNwidth, 100,
				 XtNheight, 1,
				 XkwNsilenceUnconsumed, TRUE,
				 XkwNclipEvents, TRUE,
				 NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (w, XkwNmonoPixCanvas, &canvas, NULL);
    kwin_register_refresh_func (canvas, refresh_func, NULL);
    kwin_register_position_event_func (canvas, position_func, main_shell);
    if ( ( font = kwin_load_font (canvas, "fixed") ) == NULL )
	exit (RV_UNDEF_ERROR);
    kwin_set_attributes (canvas, KWIN_ATT_FONT, font, KWIN_ATT_END);
    xt_interval_id = XtAppAddTimeOut (app_context, interval, timer_proc,
				      app_context);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main  */


/*  Private functions follow  */

static void refresh_func (KPixCanvas canvas, int width, int height,
			  void **info, PostScriptPage pspage,
			  unsigned int num_areas,
			  KPixCanvasRefreshArea *areas,
			  flag *honoured_areas)
/*  [SUMMARY] Process a refresh event for a pixel canvas.
    <canvas> The pixel canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire pixel canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    int vpos = 0;
    unsigned int num_conn, count;

    num_conn = conn_get_num_serv_connections ("kstopper");
    for (count = 0; count < num_conn; ++count)
    {
	Connection connection;
	unsigned long pixel_value;
	struct client_info *cinfo;

	connection = conn_get_serv_connection ("kstopper", count);
	cinfo = conn_get_connection_info (connection);
	if (cinfo->busy)
	    kwin_get_colour (canvas, "white", &pixel_value,
			     NULL, NULL, NULL);
	else kwin_get_colour (canvas, "grey50", &pixel_value,
			      NULL, NULL, NULL);
	cinfo->vpos = vpos + cinfo->ascent;
	kwin_draw_string (canvas, 0, cinfo->vpos,
			  conn_get_connection_module_name (connection),
			  pixel_value, FALSE);
	vpos += cinfo->height;
    }
}   /*  End Function refresh_func  */

static flag position_func (KPixCanvas canvas, int x, int y,
			   unsigned int event_code, void *e_info,void **f_info)
/*  [SUMMARY] Process a position event on a pixel canvas.
    <canvas> The pixel canvas on which the event occurred.
    <x> The horizontal position of the event, relative to the canvas origin
    <y> The vertical position of the event, relative to the canvas origin.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    unsigned int num_conn, count;
    unsigned long client_interval;
    CONST char *string = e_info;

    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
	break;
      case K_CANVAS_EVENT_POINTER_ENTER:
	interval = 500;
	client_interval = 400;
	break;
      case K_CANVAS_EVENT_POINTER_LEAVE:
	interval = 5000;
	client_interval = 4500;
	break;
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	if (strcmp (string, "q") == 0) exit (RV_OK);
	/*  Fall through  */
      default:
	return (FALSE);
	/*break;*/
    }
    num_conn = conn_get_num_serv_connections ("kstopper");
    for (count = 0; count < num_conn; ++count)
    {
	Connection connection;
	Channel channel;
	struct client_info *cinfo;

	connection = conn_get_serv_connection ("kstopper", count);
	channel = conn_get_channel (connection);
	cinfo = conn_get_connection_info (connection);
	switch (event_code)
	{
	  case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
	    if (y < cinfo->vpos - cinfo->ascent) continue;
	    if (y >= cinfo->vpos + cinfo->descent) continue;
	    if (!cinfo->busy) fprintf (stderr, "Module is idle\n");
	    else kill (cinfo->pid, SIGINT);
	    return (TRUE);
	    /*break;*/
	  case K_CANVAS_EVENT_POINTER_ENTER:
	  case K_CANVAS_EVENT_POINTER_LEAVE:
	    ch_write (channel, (char *) &client_interval,
		      sizeof client_interval);
	    ch_flush (channel);
	    break;
	  default:
	    return (FALSE);
	    /*break;*/
	}
    }
    switch (event_code)
    {
      case K_CANVAS_EVENT_POINTER_ENTER:
      case K_CANVAS_EVENT_POINTER_LEAVE:
	XtRemoveTimeOut (xt_interval_id);
	xt_interval_id = XtAppAddTimeOut (app_context, interval, timer_proc,
					  app_context);
	break;
    }
    return (TRUE);
}   /*  End Function position_func  */

static flag open_func (Connection connection, void **info)
/*  [SUMMARY] Connection open event callback.
    [PURPOSE] This routine is called when a connection opens.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    [RETURNS] TRUE on successful registration, else FALSE (indicating the
    connection should be closed).
    [NOTE] The <<close_func>> will not be called if this routine returns
    FALSE.
*/
{
    Channel channel;
    unsigned long client_interval = interval / 2;
    Dimension height;
    struct client_info *cinfo;
    static char function_name[] = "open_func";

    if ( ( cinfo = m_alloc (sizeof *cinfo) ) == NULL )
    {
	m_error_notify (function_name, "client_info structure");
	return (FALSE);
    }
    m_clear (cinfo, sizeof *cinfo);
    channel = conn_get_channel (connection);
    if (ch_read (channel, (char *) &cinfo->pid, sizeof cinfo->pid)
	< sizeof cinfo->pid)
    {
	fprintf (stderr, "%s: error reading PID\t%s\n",
		 function_name, ERRSTRING);
	m_free (cinfo);
	return (FALSE);
    }
    gettimeofday (&cinfo->last_response, NULL);
    if ( (ch_write (channel, (char *) &client_interval, sizeof client_interval)
	  < sizeof client_interval) || !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error writing interval\t%s\n",
		 function_name, ERRSTRING);
	m_free (cinfo);
	return (FALSE);
    }
    *info = cinfo;
    kwin_get_string_size (font, conn_get_connection_module_name (connection),
			  KWIN_STRING_HEIGHT, &cinfo->height,
			  KWIN_STRING_ASCENT, &cinfo->ascent,
			  KWIN_STRING_DESCENT, &cinfo->descent,
			  KWIN_STRING_END);
    XtVaGetValues (main_shell, XtNheight, &height, NULL);
    height += cinfo->height;
    XtVaSetValues (main_shell, XtNheight, height, NULL);
    return (TRUE);
}   /*  End Function open_func  */

static flag read_func (Connection connection, void **info, Channel channel)
/*  [SUMMARY] Connection read event callback.
    [PURPOSE] This routine is called when data is ready to be read from a
    connection.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    <channel> The channel object on which the read event occurred. This may be
    the normal or datagram channel.
    [RETURNS] TRUE on successful reading, else FALSE (indicating the connection
    should be closed).
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
    struct client_info *cinfo = *info;

    ch_drain (channel, 1);
    if (cinfo->busy)
    {
	unsigned long pixel_value;

	kwin_get_colour (canvas, "grey50", &pixel_value, NULL, NULL, NULL);
	kwin_draw_string (canvas, 0, cinfo->vpos,
			  conn_get_connection_module_name (connection),
			  pixel_value, FALSE);
    }
    cinfo->busy = FALSE;
    gettimeofday (&cinfo->last_response, NULL);
    return (TRUE);
}   /*  End Function read_func  */

static void close_func (Connection connection, void *info)
/*  [SUMMARY] Connection close event callback.
    [PURPOSE] This routine is called when a connection closed.
    When this routine is called, this is the last chance to read any
    buffered data from the channel associated with the connection object.
    <connection> The connection object.
    <info> The arbitrary connection information pointer.
    [RETURNS] Nothing.
*/
{
    Dimension height;
    struct client_info *cinfo = info;

    XtVaGetValues (main_shell, XtNheight, &height, NULL);
    height -= cinfo->height;
    m_free (cinfo);
    XtVaSetValues (main_shell, XtNheight, height, NULL);
}   /*  End Function close_func  */

static void timer_proc (XtPointer client_data, XtIntervalId *id)
/*  [SUMMARY] Xt interval procedure.
*/
{
    Connection connection;
    signed long diff;
    unsigned int num_conn, count;
    struct timeval current_time;
    struct client_info *cinfo;
    XtAppContext app_context = client_data;

    gettimeofday (&current_time, NULL);
    num_conn = conn_get_num_serv_connections ("kstopper");
    for (count = 0; count < num_conn; ++count)
    {
	unsigned long pixel_value;

	connection = conn_get_serv_connection ("kstopper", count);
	cinfo = conn_get_connection_info (connection);
	if (cinfo->busy) continue;
	diff = (current_time.tv_sec - cinfo->last_response.tv_sec) * 1000;
	diff += (current_time.tv_usec - cinfo->last_response.tv_usec) / 1000;
	if (diff < interval) continue;
	cinfo->busy = TRUE;
	kwin_get_colour (canvas, "white", &pixel_value,
			 NULL, NULL, NULL);
	kwin_draw_string (canvas, 0, cinfo->vpos,
			  conn_get_connection_module_name (connection),
			  pixel_value, FALSE);
    }
    xt_interval_id = XtAppAddTimeOut (app_context, interval, timer_proc,
				      client_data);
}   /*  End Function timer_proc  */
