/*  kbwtest.c

    Source file for  kbwtest  (network bandwidth testing module).

    Copyright (C) 2002-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read in a Karma data file and transmit the
    multi-array over one or more Karma connections to a Karma server module.
    The module will also allow a list of files to be transmitted, one after the
    other.


    Written by      Richard Gooch   22-DEC-2002: Copied from tx.c and rx.c

    Last updated by Richard Gooch   20-JUL-2003: Added bits/s measure.


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#if !defined(OS_SunOS) && !defined(OS_IRIX6)
#  include <sched.h>
#endif
#include <sys/time.h>
#include <sys/times.h>
#include <sys/file.h>
#include <errno.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_arln.h>
#include <karma_pio.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_dm.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_hi.h>
#include <karma_a.h>
#include <karma_s.h>
#include <karma_m.h>
#include <karma_n.h>
#include <karma_r.h>
#include <k_version.h>


#define VERSION "1.0"
#define BUF_SIZE 262144
#define DEFAULT_DGRAM_SIZE 4096
#define COMMAND_LINE_LENGTH 4096


/*  Structure definitions  */
typedef struct
{
    unsigned long bytes_to_read;  /* Stream: total bytes, Dgram: packet size */
    flag synchronous;
} spray_info;


STATIC_FUNCTION (flag kbwtest, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag process_one_line,
		 ( CONST char *line, flag (*decode_func) () ) );
STATIC_FUNCTION (flag internal_decode_func, (CONST char *line, FILE *fp) );
STATIC_FUNCTION (char *get_spray_buffer, (unsigned int size) );
STATIC_FUNCTION (void do_spray, (CONST char *p) );
STATIC_FUNCTION (void spray_one,
		 (Connection connection, unsigned int num_bytes,
		  unsigned int packet_size) );
STATIC_FUNCTION (void datagram_spray,
		 (Channel channel, Channel datagram_channel,
		  unsigned int packet_size, unsigned int num_packets,
		  char *buffer) );
STATIC_FUNCTION (void delay, (unsigned long delay_usec) );
STATIC_FUNCTION (flag open_spray, (Connection connection, void **info) );
STATIC_FUNCTION (flag read_spray,
		 (Connection connection, void **info, Channel channel) );
STATIC_FUNCTION (void close_spray, (Connection connection, void *info) );


/*  Private data  */
static flag synchronous_spray = TRUE;
static unsigned long transmit_rate = 0;

#define SPRAY_DATA_RANDOM      0
#define SPRAY_DATA_PATTERN     1
#define NUM_SPRAY_DATA_CHOICES 2
static char *spray_data_choices[NUM_SPRAY_DATA_CHOICES] =
{
    "random",
    "pattern",
};
static unsigned int spray_type = SPRAY_DATA_RANDOM;

#define MAX_PATTERN_LENGTH 16
static unsigned char spray_pattern[MAX_PATTERN_LENGTH] = {0};
static unsigned int pattern_length = 1;
static flag verbose = FALSE;
static unsigned long datagram_packets = 0;
static unsigned long datagram_bytes = 0;
static unsigned long partial_packets = 0;


int main (int argc, char **argv)
{
    KControlPanel panel;
    int def_port_number;
    unsigned int server_port_number;
    char line[COMMAND_LINE_LENGTH];
    char prompt[STRING_LENGTH + 3];
    extern char module_name[STRING_LENGTH + 1];
    static char function_name[] = "main";

    /*  Initialise module  */
    im_register_module_name ("kbwtest");
    im_register_module_version_date (VERSION);
    dm_native_setup ();
    conn_initialise ( ( void (*) () ) NULL );
    conn_register_server_protocol ("spray", 3, 0,
				   open_spray, read_spray, close_spray);
    conn_register_client_protocol ("spray", 3, 0,
				   ( flag (*) () ) NULL,
				   open_spray, read_spray, close_spray);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    if ( ( def_port_number = r_get_def_port (module_name, NULL) ) < 0 )
    {
	fprintf (stderr, "Could not get default port number\n");
	exit (RV_UNDEF_ERROR);
    }
    server_port_number = def_port_number;
    if ( !conn_become_server (&server_port_number, CONN_MAX_INSTANCES) )
    {
	fprintf (stderr, "Error becoming server\n");
	exit (RV_UNDEF_ERROR);
    }
    fprintf (stderr, "Port allocated: %u\n", server_port_number);
    if ( !s_install_handlers () )
    {
	fprintf (stderr, "Module: %s quitting\n", module_name);
	exit (RV_SYS_ERROR);
    }
    /*  Setup user interface  */
    panel_add_item (panel, "tx", "bytes [packet_size]",
		    PIT_FUNCTION, (void *) do_spray,
		    PIA_END);
    panel_add_item (panel, "transmit_rate","maximum transmit rate (bytes/sec)",
		    K_ULONG, (char *) &transmit_rate,
		    PIA_END);
    panel_add_item (panel, "synchronous_spray", "wait for sync at end",
		    PIT_FLAG,
		    &synchronous_spray,
		    PIA_END);
    panel_add_item (panel, "spray_data", "choice of spray data",
		    PIT_CHOICE_INDEX, (char *) &spray_type,
		    PIA_NUM_CHOICE_STRINGS, NUM_SPRAY_DATA_CHOICES,
		    PIA_CHOICE_STRINGS, spray_data_choices,
		    PIA_END);
    panel_add_item (panel, "pattern", "spray pattern", K_UBYTE, spray_pattern,
		    PIA_ARRAY_MIN_LENGTH, 1,
		    PIA_ARRAY_MAX_LENGTH, MAX_PATTERN_LENGTH,
		    PIA_ARRAY_LENGTH, (char *) &pattern_length,
		    PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    /*  Read in defaults  */
    hi_read (module_name, internal_decode_func);
    if ( s_check_for_int () )
    {
	fprintf (stderr, "control_c abort\n");
	exit (RV_CONTROL_C);
    }
    if (argc > 1)
    {
	/*  Concatenate command line arguments into a single line  */
	module_process_argvs (argc - 1, argv + 1, kbwtest);
	/*  Save defaults  */
	hi_write (module_name, internal_decode_func);
	exit (RV_OK);
    }
    /*  Prompt user for commands  */
    /*  Generate prompt. Don't use  sprintf(3)  because it's broken in the
	C library for the VX/MVX
	*/
    strcpy (prompt, module_name);
    strcat (prompt, "> ");
    /*  Read lines and processes until eof on input  */
    while ( arln_read_from_stdin (line, COMMAND_LINE_LENGTH, prompt) &&
	   process_one_line (line, kbwtest) );
    /*  Save defaults  */
    hi_write (module_name, internal_decode_func);
    return (RV_OK);
}   /*  End Function main   */

static flag kbwtest (CONST char *command, FILE *fp)
{
    return (TRUE);
}   /*  End Function kbwtest  */

static flag process_one_line ( CONST char *line, flag (*decode_func) () )
{
    return panel_process_command_with_stack (line, decode_func, stderr);
}   /*  End Function process_one_line  */

static flag internal_decode_func (CONST char *line, FILE *fp)
{
    return panel_process_command_with_stack (line, kbwtest, fp);
}   /*  End Function internal_decode_func  */

static char *get_spray_buffer (unsigned int size)
/*  [SUMMARY] Get the current spray buffer.
    <size> The size of the buffer.
    [RETURNS] A pointer to a filled spray buffer.
*/
{
    unsigned int byte_count;
    static unsigned int rbuf_len = 0;
    static unsigned int vbuf_len = 0;
    static char *rbuffer = NULL;
    static char *vbuffer = NULL;
    static char function_name[] = "get_spray_buffer";

    if ( (spray_type == SPRAY_DATA_RANDOM) && (rbuf_len < size) )
    {
	if (rbuffer != NULL) m_free (rbuffer);
	if ( ( rbuffer = m_alloc (size) ) == NULL )
	{
	    m_abort (function_name, "random number buffer");
	}
	fprintf (stderr, "Filling buffer with random numbers...");
	for (byte_count = 0; byte_count < size; ++byte_count)
	{
	    *(unsigned char *) (rbuffer + byte_count) = n_uniform () * 255.0;
	}
	rbuf_len = size;
	fprintf (stderr, "done\n");
    }
    if ( (spray_type != SPRAY_DATA_RANDOM) && (vbuf_len < size) )
    {
	if (vbuffer != NULL) m_free (vbuffer);
	if ( ( vbuffer = m_alloc (size) ) == NULL )
	{
	    m_abort (function_name, "value buffer");
	}
	vbuf_len = size;
    }
    switch (spray_type)
    {
      case SPRAY_DATA_PATTERN:
	byte_count = size / pattern_length;
	m_fill (vbuffer, pattern_length, (CONST char *) spray_pattern,
		pattern_length, byte_count);
	m_copy (vbuffer + byte_count * pattern_length,
		(CONST char *) spray_pattern,
		size - byte_count * pattern_length);
	break;
      case SPRAY_DATA_RANDOM:
	break;
      default:
	fprintf (stderr, "Illegal spray type: %u\n", spray_type);
	a_prog_bug (function_name);
	break;
    }
    return (spray_type == SPRAY_DATA_RANDOM) ? rbuffer : vbuffer;
}   /*  End Function get_spray_buffer  */

static void do_spray (CONST char *p)
/*  [PURPOSE] This routine will spray bytes to all "spray" servers.
    <p> A string containing the number of bytes to spray.
    [RETURNS] Nothing.
*/
{
    Connection connection;
    unsigned int num_bytes, packet_size;
    unsigned int conn_count, num_conn;
    static char function_name[] = "do_spray";

    if (p == NULL)
    {
	fprintf (stderr, "Must specify number of bytes to spray\n");
	return;
    }
    if ( ( num_bytes = ex_uint (p, &p) ) < 1 )
    {
	fprintf (stderr, "Must specify at least 1 byte\n");
	return;
    }
    packet_size = (p == NULL) ? BUF_SIZE : ex_uint (p, &p);
    num_conn = conn_get_num_client_connections ("spray");
    for (conn_count = 0; conn_count < num_conn; ++conn_count)
    {
	if ( ( connection = conn_get_client_connection ("spray", conn_count) )
	    == NULL )
	{
	    fprintf (stderr, "Client connection: %u does not exist!\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	spray_one (connection, num_bytes, packet_size);
    }
    num_conn = conn_get_num_serv_connections ("spray");
    for (conn_count = 0; conn_count < num_conn; ++conn_count)
    {
	if ( ( connection = conn_get_serv_connection ("spray", conn_count) )
	    == NULL )
	{
	    fprintf (stderr, "Server connection: %u does not exist!\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	spray_one (connection, num_bytes, packet_size);
    }
}   /*  End Function do_spray  */

static void spray_one (Connection connection, unsigned int num_bytes,
		       unsigned int packet_size)
{
    Channel channel, datagram_channel;
    flag wrote_all;
    char dummy;
    unsigned int size, byte_count, block_size;
    unsigned long interval, rem_addr;
    struct karma_qos qos;
    struct timeval start_time;
    struct timeval stop_time;
    char *buffer;
    extern flag synchronous_spray;
    static struct timezone tz = {0, 0};
    float time_taken, usec, rate, multiplier;
    char *prefix;
    char hostname[STRING_LENGTH];
    /*static char function_name[] = "spray_one";*/

    if (transmit_rate < 1) interval = 0;
    else interval = ( (float) packet_size / (float) transmit_rate * 1e6 );
    buffer = get_spray_buffer (packet_size);
    channel = conn_get_channel (connection);
    datagram_channel = conn_get_datagram_channel (connection);
    rem_addr = conn_get_connection_remote_addr (connection);
    if ( !r_gethostbyaddr (hostname, STRING_LENGTH, rem_addr) ) return;
    fprintf (stderr, " %-30s: ", hostname);
    if (datagram_channel)
    {
	ch_get_qos (datagram_channel, &qos);
	size = (packet_size < qos.tx.sdu_max) ? packet_size : qos.tx.sdu_max;
	datagram_spray (channel, datagram_channel, size, num_bytes / size,
			buffer);
	return;
    }
    if (gettimeofday (&start_time, &tz) != 0)
    {
	fprintf (stderr, "Error getting time of day\t%s%c\n", ERRSTRING, BEL);
	exit (RV_SYS_ERROR);
    }
    /*  Tell server how many bytes to expect  */
    if ( !pio_write32 (channel, num_bytes) ) return;
    /*  Tell server whether synchronous operation is required  */
    if ( !pio_write32 (channel, synchronous_spray) ) return;
    /*  Write all the bytes  */
    byte_count = num_bytes;
    wrote_all = TRUE;
    while (byte_count > 0)
    {
	block_size = (byte_count > packet_size) ? packet_size : byte_count;
	if (ch_write (channel, buffer, block_size) < block_size)
	{
	    fprintf (stderr, "Error writing data\t%s\n", ERRSTRING);
	    byte_count = 0;
	    wrote_all = FALSE;
	    break;
	}
	byte_count -= block_size;
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "Error flushing\n");
	    wrote_all = FALSE;
	    break;
	}
	delay (interval);
    }
    if (!wrote_all) return;
    if (synchronous_spray)
    {
	/*  Read the return byte  */
	if (ch_read (channel, &dummy, 1) < 1) return;
    }
    if (gettimeofday (&stop_time, &tz) != 0)
    {
	fprintf (stderr, "Error getting time of day\t%s%c\n", ERRSTRING, BEL);
	exit (RV_SYS_ERROR);
    }
    usec = (float) (stop_time.tv_usec - start_time.tv_usec) * 1e-6;
    time_taken = (float) (stop_time.tv_sec - start_time.tv_sec) + usec;
    rate = (float) num_bytes / time_taken;
    if (rate < 1e3) multiplier = 1.0, prefix = "";
    else if (rate < 1e6) multiplier = 1e-3, prefix = "k";
    else if (rate < 1e9) multiplier = 1e-6, prefix = "M";
    else if (rate < 1e12) multiplier = 1e-9, prefix = "G";
    else multiplier = 1e-12, prefix = "T";
    fprintf (stderr, "%10u Bytes/s  %.1f %sB/s  %.1f %sbits/s\n",
	     (unsigned int) rate, rate * multiplier, prefix,
	     8.0 * rate * multiplier, prefix);
}   /*  End Function spray_one  */

static void datagram_spray (Channel channel, Channel datagram_channel,
			    unsigned int packet_size, unsigned int num_packets,
			    char *buffer)
/*  [PURPOSE] This routine will spray bytes to a datagram channel.
    <channel> The stream channel.
    <datagram_channel> The datagram channel.
    <packet_size> The size of the packets.
    <num_packets> The number of packets to send.
    <buffer> The data to send.
    [RETURNS] Nothing.
*/
{
    unsigned long num_packets_received, num_dropped, num_bytes_received;
    unsigned long num_partial_packets, interval;
    unsigned int packet_count;
    unsigned int num_bytes = 0;
    float time_taken, usec, rate;
    struct timeval start_time;
    struct timeval stop_time;
    static struct timezone tz = {0, 0};
    static char function_name[] = "datagram_spray";

    if (transmit_rate < 1) interval = 0;
    else interval = ( (float) packet_size / (float) transmit_rate * 1e6 );
    if (num_packets < 1)
    {
	fprintf (stderr, "no whole packets to send\n");
	return;
    }
    if ( !pio_write32 (channel, packet_size) || !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error writing packet size\n", function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_packets_received) )
    {
	fprintf (stderr, "%s: error reading dummy num_packets_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_bytes_received) )
    {
	fprintf (stderr, "%s: error reading dummy num_bytes_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_partial_packets) )
    {
	fprintf (stderr, "%s: error reading dummy num_partial_packets\n",
		 function_name);
	return;
    }
    if (gettimeofday (&start_time, &tz) != 0)
    {
	    fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
    }
    /*  Write all the packets  */
    for (packet_count = 0; packet_count < num_packets; ++packet_count)
    {
	if (ch_write (datagram_channel, buffer, packet_size) < packet_size)
	{
	    fprintf (stderr, "%s: error writing data\t%s\n",
		     function_name, ERRSTRING);
	    return;
	}
	if ( !ch_flush (datagram_channel) )
	{
	    fprintf (stderr, "%s: error flusing data\t%s\n",
		     function_name, ERRSTRING);
	    return;
	}
	num_bytes += packet_size;
	delay (interval);
    }
    if ( !pio_write32 (channel, 0) ) return;
    if ( !ch_flush (channel) ) return;
    if ( !pio_read32 (channel, &num_packets_received) )
    {
	fprintf (stderr, "%s: error reading num_packets_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_bytes_received) )
    {
	fprintf (stderr, "%s: error reading num_bytes_received\n",
		 function_name);
	return;
    }
    if ( !pio_read32 (channel, &num_partial_packets) )
    {
	fprintf (stderr, "%s: error reading num_partial_packets\n",
		 function_name);
	return;
    }
    if (gettimeofday (&stop_time, &tz) != 0)
    {
	fprintf (stderr, "Error getting time of day\t%s%c\n",
		 ERRSTRING, BEL);
	exit (RV_SYS_ERROR);
    }
    usec = (float) (stop_time.tv_usec - start_time.tv_usec) * 1e-6;
    time_taken = (float) (stop_time.tv_sec - start_time.tv_sec) + usec;
    rate = (float) num_bytes / time_taken;
    fprintf (stderr, "%u bytes/second transmitted\n", (unsigned int) rate);
    num_dropped = num_packets - num_packets_received;
    fprintf (stderr, "  %lu packets dropped, %u sent (%lu%% loss) ",
	     num_dropped, num_packets,
	     num_dropped * 100 / num_packets);
    rate = (float) num_bytes_received / time_taken;
    fprintf (stderr, "%u bytes/second received\n", (unsigned int) rate);
    fprintf (stderr, "  %lu partial packets received\n", num_partial_packets);
}   /*  End Function datagram_spray  */

static void delay (unsigned long delay_usec)
/*  [SUMMARY] Wait a specified amount of time.
    <delay_usec> The number of microseconds to wait. If this is zero the
    routine does not wait.
    [NOTE] This routine depends on the gettimeofday system call.
    [RETURNS] Nothing.
*/
{
    signed long time_taken;
    unsigned long delay_sec, delay_remainder;
    struct timeval start_time;
    struct timeval stop_time;

    if (delay_usec < 1) return;
    if (gettimeofday (&start_time, NULL) != 0)
    {
	fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    delay_sec = delay_usec / 1000000;
    delay_remainder = delay_usec - delay_sec * 1000000;
    while (TRUE)
    {
	if (gettimeofday (&stop_time, NULL) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (stop_time.tv_sec == start_time.tv_sec)
	{
	    if (stop_time.tv_usec >= start_time.tv_usec + delay_usec) return;
#if !defined(OS_Solaris) && !defined(OS_SunOS) && !defined(OS_IRIX6)
	    sched_yield ();
#endif
	    continue;
	}
	/*  In a new second  */
	stop_time.tv_sec -= delay_sec;
	time_taken = (signed long) stop_time.tv_sec;
	time_taken -= (signed long) start_time.tv_sec;
	time_taken *= 1000000;
	time_taken += (signed long) stop_time.tv_usec;
	time_taken -= (signed long) start_time.tv_usec;
	if (time_taken >= (signed long) delay_remainder) return;
#if !defined(OS_Solaris) && !defined(OS_SunOS) && !defined(OS_IRIX6)
	sched_yield ();
#endif
    }
}   /*  End Function delay  */

static flag open_spray (Connection connection, void **info)
/*  [SUMMARY] Connection open event callback.
    [PURPOSE] This routine is called when a connection opens.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    [RETURNS] TRUE on successful registration, else FALSE (indicating the
    connection should be closed).
    [NOTE] The <<close_func>> will not be called if this routine returns
    FALSE.
*/
{
    spray_info *sinfo;
    static char function_name[] = "open_spray";

    if ( ( sinfo = m_alloc (sizeof *sinfo) ) == NULL )
    {
	m_error_notify (function_name, "spray info");
	return (FALSE);
    }
    sinfo->bytes_to_read = 0;
    sinfo->synchronous = FALSE;
    *info = sinfo;
    return (TRUE);
}   /*  End Function open_spray  */

static flag read_spray (Connection connection, void **info, Channel channel)
/*  [SUMMARY] Connection read event callback.
    [PURPOSE] This routine is called when data is ready to be read from a
    connection.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    <channel> The channel object on which the read event occurred. This may be
    the normal or datagram channel.
    [RETURNS] TRUE on successful reading, else FALSE (indicating the connection
    should be closed).
    [NOTE] The <<close_func>> will not be called if this routine returns
    FALSE.
*/
{
    Channel datagram_channel;
    int bytes_readable;
    unsigned int bytes_to_read;
    unsigned long value;
    spray_info *sinfo = (spray_info *) *info;
    extern flag synchronous_spray;
    static char function_name[] = "read_spray";

    datagram_channel = conn_get_datagram_channel (connection);
    if (channel == datagram_channel)
    {
	if (sinfo->bytes_to_read < 1)
	{
	    /*  Still have some packets in transit after transaction completed:
		make best-guess drain  */
	    fprintf (stderr, "Late datagram packet\n");
	    ch_drain ( channel, ch_get_bytes_readable (channel) );
	    return (TRUE);
	}
	bytes_to_read = ch_drain (channel, sinfo->bytes_to_read);
	if (bytes_to_read < sinfo->bytes_to_read)
	{
	    ++partial_packets;
	    fprintf (stderr, "Received partial of %u bytes, wanted: %lu\n",
		     bytes_to_read, sinfo->bytes_to_read);
	}
	else ++datagram_packets;
	datagram_bytes += bytes_to_read;
	return (TRUE);
    }
    /*  We got something from the stream channel  */
    if (datagram_channel != NULL)
    {
	/*  Special protocol for stream channel  */
	if ( !pio_read32 (channel, &sinfo->bytes_to_read) ) return (FALSE);
	if ( !pio_write32 (channel, datagram_packets) ||
	     !pio_write32 (channel, datagram_bytes) ||
	     !pio_write32 (channel, partial_packets) || !ch_flush (channel) )
	{
	    fprintf (stderr, "%s: error writing\t%s\n",
		     function_name, ERRSTRING);
	    return (FALSE);
	}
	datagram_packets = 0;
	datagram_bytes = 0;
	partial_packets = 0;
	return (TRUE);
    }
    /*  Pure stream protocol  */
    if (sinfo->bytes_to_read == 0)
    {
	/*  New transaction: read number of bytes that will need to be read  */
	if ( !pio_read32 (channel, &value) ) return (FALSE);
	sinfo->bytes_to_read = value;
	if ( !pio_read32 (channel, &value) ) return (FALSE);
	sinfo->synchronous = value;
    }
    if (synchronous_spray)
    {
	bytes_readable = sinfo->bytes_to_read;
	if (verbose) fprintf (stderr, "Starting synchronous read...");
    }
    else
    {
	if ( ( bytes_readable = ch_get_bytes_readable (channel) ) < 0 )
	{
	    exit (RV_SYS_ERROR);
	}
	if (bytes_readable < 1)
	{
	    fprintf(stderr,
		    "%s: apparently no bytes readable. This is probably due\n",
		    function_name);
	    fprintf (stderr,
		     "to an Operating System bug. Solaris 2.x is known to have this bug\n");
	    fprintf (stderr, "Assuming one byte is readable\n");
	    bytes_readable = 1;
	}
    }
    bytes_to_read = (bytes_readable > sinfo->bytes_to_read) ? sinfo->bytes_to_read : bytes_readable;
    if (ch_drain (channel, bytes_to_read) < bytes_to_read)
    {
	fprintf (stderr, "Error reading: %u bytes\t%s\n",
		 bytes_to_read, ERRSTRING);
	return (FALSE);
    }
    if (synchronous_spray && verbose) fprintf (stderr, "\tfinished\n");
    sinfo->bytes_to_read -= bytes_to_read;
    if (sinfo->bytes_to_read < 1)
    {
	if (sinfo->synchronous)
	{
	    if (ch_fill (channel, 1, 0) < 1)
	    {
		fprintf (stderr, "Error writing synchronisation byte\t%s\n",
			 ERRSTRING);
		return (FALSE);
	    }
	    if ( !ch_flush (channel) )
	    {
		fprintf (stderr, "Error flushing synchronisation byte\t%s\n",
			 ERRSTRING);
		return (FALSE);
	    }
	}
    }
    return (TRUE);
}   /*  End Function read_spray  */

static void close_spray (Connection connection, void *info)
/*  [SUMMARY] Connection close event callback.
    [PURPOSE] This routine is called when a connection closed.
    When this routine is called, this is the last chance to read any
    buffered data from the channel associated with the connection object.
    <connection> The connection object.
    <info> The arbitrary connection information pointer.
    [RETURNS] Nothing.
*/
{
    m_free (info);
}   /*  End Function close_spray  */
