/*  kprinthead.c

    Source file for  kprinthead  (print header of data file).

    Copyright (C) 1996-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will print the header information contained in a data
    file. Many formats are supported.


    Written by      Richard Gooch   29-MAY-1996

    Updated by      Richard Gooch   15-JUN-1996: Print array descriptor too.

    Updated by      Richard Gooch   6-AUG-1996: Print history information too.

    Updated by      Richard Gooch   8-AUG-1996: Added history display flag.

    Updated by      Richard Gooch   15-AUG-1996: Supported gzipped Karma and
  FITS files.

    Updated by      Richard Gooch   25-NOV-1996: Don't convert to float when
  doing fast read of FITS files.

    Updated by      Richard Gooch   11-MAR-1998: Minor cosmetic changes.

    Updated by      Richard Gooch   12-MAR-1998: Add ".kf" extension if missing
  in <read_karma>.

    Updated by      Richard Gooch   27-MAY-1998: Add support for empty fits
  files.

    Updated by      Richard Gooch   4-SEP-1998: Supported bzip2ed Karma files.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   20-MAY-2000: Made use of new experimental
  function <foreign_drao_read_header>.

    Updated by      Richard Gooch   28-MAY-2000: Made use of new experimental
  function <foreign_aips1_read_header>.

    Updated by      Richard Gooch   13-NOV-2001: Added mmap flag.

    Updated by      Richard Gooch   14-JUN-2002: Comment update.

    Updated by      Richard Gooch   24-DEC-2002: Supported FITS extensions.

    Updated by      Richard Gooch   4-FEB-2003: Made use of new function
  <foreign_aips1_read_history>.

    Last updated by Richard Gooch   22-APR-2003: Took account of API change for
  DRAO data.


*/
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_dsrw.h>
#include <karma_pio.h>
#include <karma_dmp.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>

#define VERSION "1.2"

STATIC_FUNCTION (flag kprinthead, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void process_file, (CONST char *infile) );
STATIC_FUNCTION (void process_file_fast, (CONST char *infile) );
STATIC_FUNCTION (void show_vm_structure, (CONST multi_array *multi_desc) );
STATIC_FUNCTION (multi_array *read_karma, (CONST char *infile) );
STATIC_FUNCTION (flag read_packet,
		 (Channel channel, CONST packet_desc *descriptor,
		  char *packet) );


/*  Private data  */
static flag fast = TRUE;
static flag show_history = FALSE;
static flag mmap = TRUE;


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "show_history", "flag", PIT_FLAG, &show_history,
		    PIA_END);
    panel_add_item (panel, "mmap", "flag", PIT_FLAG, &mmap,
		    PIA_END);
    panel_add_item (panel, "fast", "flag", PIT_FLAG, &fast,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile");
    module_run (argc, argv, "kprinthead", VERSION, kprinthead, -1, -1, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag kprinthead (CONST char *command, FILE *fp)
{
    char *infile;

    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting infile name\n");
	    return (TRUE);
	}
	if (fast) process_file_fast (infile);
	else process_file (infile);
	m_free (infile);
    }
    return (TRUE);
}   /*  End Function kprinthead  */

static void process_file (CONST char *infile)
/*  [SUMMARY] Process an input file.
    <infile> The name of the input file.
    [RETURNS] Nothing.
*/
{
    multi_array *multi_desc;
    /*static char function_name[] = "process_file";*/

    /*  Read file in whatever format  */
    if ( ( multi_desc =
	   foreign_guess_and_read (infile,
				   mmap ?
				   K_CH_MAP_IF_AVAILABLE : K_CH_MAP_NEVER,
				   FALSE, NULL,
				   FA_GUESS_READ_FITS_TO_FLOAT, FALSE,
				   FA_GUESS_READ_END) ) == NULL )
    {
	fprintf (stderr, "Error reading\n");
	return;
    }
    show_vm_structure (multi_desc);
    ds_dealloc_multi (multi_desc);
}   /*  End Function process_file  */

static void process_file_fast (CONST char *infile)
/*  [SUMMARY] Process an input file.
    <infile> The name of the input file.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    KAIPS1DataContext aips1_ctx = NULL;
    unsigned int ftype;
    multi_array *multi_desc;
    char *ptr;
    char fname[STRING_LENGTH];
    /*static char function_name[] = "process_file_fast";*/

    /*  Determine format  */
    ftype = foreign_guess_format_from_filename (infile);
    if (ftype == FOREIGN_FILE_FORMAT_UNKNOWN) return;
    switch (ftype)
    {
      case FOREIGN_FILE_FORMAT_KARMA:
	if ( ( multi_desc = read_karma (infile) ) == NULL ) return;
	break;
      case FOREIGN_FILE_FORMAT_FITS:
	if ( ( multi_desc =
	       foreign_fits_read (infile, FALSE, TRUE,
				  FA_FITS_READ_SKIP_DATA, TRUE,
				  FA_FITS_READ_END) )
	    == NULL )
	{
	    fprintf (stderr, "Error reading FITS file header(s)\n");
	    return;
	}
	break;
      case FOREIGN_FILE_FORMAT_EMPTY_FITS:
	/*  Read empty FITS file  */
	if ( ( channel = ch_open_file (infile, "rz") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     infile, ERRSTRING);
	    return;
	}
	if ( ( multi_desc =
	       foreign_fits_read_header (channel, FALSE, FALSE, TRUE,
					 FA_FITS_READ_HEADER_ALLOW_TRUNC,
					 TRUE,
					 FA_FITS_READ_HEADER_END) )
	    == NULL )
	{
	    fprintf (stderr, "Error reading FITS file header\n");
	    ch_close (channel);
	    return;
	}
	ch_close (channel);
	break;
      case FOREIGN_FILE_FORMAT_MIRIAD:
	/*  Read Miriad Image file  */
	if ( ( multi_desc =
	       foreign_miriad_read_header2 (infile, FALSE, TRUE,
					    FA_MIRIAD_READ_HEADER_END) )
	    == NULL )
	{
	    fprintf (stderr, "Error reading Miriad Image header\n");
	    return;
	}
	if ( !foreign_miriad_read_history (infile, multi_desc) )
	    return;
	break;
      case FOREIGN_FILE_FORMAT_GIPSY:
	/*  Read GIPSY file  */
	strcpy (fname, infile);
	if ( ( ptr = strrchr (fname, '.') ) == NULL ) return;
	strcpy (ptr, ".descr");
	if ( ( channel = ch_open_file (fname, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     fname, ERRSTRING);
	    return;
	}
	if ( ( multi_desc =
	       foreign_gipsy_read_header (channel, FALSE, TRUE,
					  FA_GIPSY_READ_HEADER_END) )
	    == NULL )
	{
	    fprintf (stderr, "Error reading GIPSY header\n");
	    ch_close (channel);
	    return;
	}
	ch_close (channel);
	break;
      case FOREIGN_FILE_FORMAT_DRAO:
	/*  Read DRAO file  */
	if ( ( multi_desc =
	       foreign_drao_read_header (infile, FALSE, FALSE, NULL,
					 FA_DRAO_READ_HEADER_END) )
	    == NULL )
	{
	    fprintf (stderr, "Error reading DRAO header\n");
	    return;
	}
	break;
      case FOREIGN_FILE_FORMAT_AIPS1:
	/*  Read AIPS1 (Classic AIPS) file  */
	if ( ( multi_desc =
	       foreign_aips1_read_header (infile, FALSE, &aips1_ctx) )
	     == NULL )
	{
	    fprintf (stderr, "Error reading AIPS header\n");
	    return;
	}
	if ( !foreign_aips1_read_history (aips1_ctx, multi_desc) )
	    return;
	break;
      default:
	process_file (infile);
	return;
	/*break;*/
    }
    show_vm_structure (multi_desc);
    ds_dealloc_multi (multi_desc);
}   /*  End Function process_file_fast  */

static void show_vm_structure (CONST multi_array *multi_desc)
/*  [SUMMARY] Show the header of a structure in virtual memory.
    <multi_desc> The structure.
    [RETURNS] Nothing.
*/
{
    unsigned int array_count;
    history *hist;

    if (multi_desc->num_arrays > 1)
	printf ("Total of: %u General Data Structures follow\n",
		multi_desc->num_arrays);
    for (array_count = 0; array_count < multi_desc->num_arrays; ++array_count)
    {
	unsigned int count, type;
	packet_desc *top_pack_desc;
	char *top_packet;

	if (multi_desc->num_arrays > 1)
	{
	    printf ("\nGeneral Data Structure[%u] follows. Name: \"%s\"\n",
		    array_count, multi_desc->array_names[array_count]);
	}
	top_pack_desc = multi_desc->headers[array_count];
	top_packet = multi_desc->data[array_count];
	for (count = 0; count < top_pack_desc->num_elements; ++count)
	{
	    type = top_pack_desc->element_types[count];
	    if (type == K_ARRAY)
	    {
		dmp_array_desc(stdout,
			       (array_desc*)top_pack_desc->element_desc[count],
			       FALSE);
	    }
	    else if (type == LISTP)
	    {
		puts ("Linked List header");
	    }
	    else if ( ds_element_is_named (type) )
	    {
		printf ("Element: \"%s\"\t\t",
			top_pack_desc->element_desc[count]);
		dmp_element (stdout, type, top_pack_desc->element_desc[count],
			     top_packet +
			     ds_get_element_offset (top_pack_desc, count),
			     FALSE);
	    }
	    else
	    {
		printf ("Unknown element type: %u\n", type);
	    }
	}
    }
    if (show_history)
    {
	/*  Show history  */
	if (multi_desc->first_hist != NULL) puts ("");
	for (hist = multi_desc->first_hist; hist != NULL; hist = hist->next)
	{
	    printf ("HISTORY: %s\n", hist->string);
	}
    }
}   /*  End Function show_vm_structure  */

static multi_array *read_karma (CONST char *infile)
/*  [SUMMARY] Read a Karma file but skip n-dimensional arrays.
    <infile> The name of the input file.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    int len;
    unsigned int count, num_arrays, length;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    char *string;
    char *data;
    char fname[STRING_LENGTH];
    static char function_name[] = "read_karma";

    len = strlen (infile);
    strcpy (fname, infile);
    if ( (strcmp (fname + len - 3, ".kf") != 0) &&
	 (strcmp (fname + len - 6, ".kf.gz") != 0) &&
	 (strcmp (fname + len - 7, ".kf.bz2") != 0) )
	strcpy (fname + len, ".kf");
    if ( ( channel = ch_open_file (fname, "rz") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n", fname, ERRSTRING);
	return (NULL);
    }
    if ( ( multi_desc = dsrw_read_multi_header (channel) ) == NULL )
	return (NULL);
    num_arrays = multi_desc->num_arrays;
    for (count = 0; count < num_arrays; ++count)
    {
	/*  Load the descriptor for one array   */
        if ( ( pack_desc = dsrw_read_packet_desc (channel) ) == NULL )
        {
	    ds_dealloc_multi (multi_desc);
            return (NULL);
        }
        multi_desc->headers[count] = pack_desc;
	/*  Allocate the required data space   */
	if ( ( data = ds_alloc_data (multi_desc->headers[count], FALSE,FALSE) )
	     == NULL )
        {
	    m_error_notify (function_name, "general data structure");
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
        multi_desc->data[count] = data;
	/*  Read the data   */
        if ( !read_packet (channel, multi_desc->headers[count],
			   multi_desc->data[count]) )
        {
	    fprintf (stderr, "Error reading array number %u\n", count);
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
    }
    /*  May need to read arraynames  */
    for (count = 0; (count < num_arrays) && (num_arrays > 1); ++count)
    {
	/*  Load the names of the arrays    */
	if ( ( multi_desc->array_names[count] =
	      pio_read_string (channel, &length) ) == NULL )
        {
	    a_func_abort (function_name, "Error reading array name");
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
        if (length == 0)
        {
	    a_func_abort (function_name,
			  "Array name must not be a null string");
            ds_dealloc_multi (multi_desc);
            return (NULL);
        }
    }
    /*  Read history strings  */
    if ( ( string = pio_read_string (channel, &length) ) == NULL )
    {
	a_func_abort (function_name, "Error reading first history string");
	ds_dealloc_multi (multi_desc);
	return (NULL);
    }
    while (length > 0)
    {
	if ( !ds_history_append_string (multi_desc, string, FALSE) )
	{
	    m_abort (function_name, "history entry");
	}
	if ( ( string = pio_read_string (channel, &length) )
	    == NULL )
	{
	    a_func_abort (function_name, "Error reading history string");
	    ds_dealloc_multi (multi_desc);
	    return (NULL);
	}
    }
    m_free (string);
    return (multi_desc);
}   /*  End Function read_karma  */

static flag read_packet (Channel channel, CONST packet_desc *descriptor,
			 char *packet)
/*  [SUMMARY] Read the binary representation of a data packet.
    [PURPOSE] This routine will read the binary representation of a data packet
    from a channel object. The routine will recursively read in sub arrays and
    linked lists.
    <channel> The channel object.
    <descriptor> The descriptor for the packet.
    <packet> The packet data will be written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int count;
    unsigned int type;
    unsigned long pad_bytes;
    array_desc *arr_desc;
    extern char host_type_sizes[NUMTYPES];
    /*static char function_name[] = "read_packet";*/

    for (count = 0; count < descriptor->num_elements;
	 ++count, packet += host_type_sizes[type])
    {
	type = descriptor->element_types[count];
	switch (type)
	{
	  case K_ARRAY:
	    /*  Try and skip past the array  */
	    arr_desc = (array_desc *) descriptor->element_desc[count];
	    if ( !ds_packet_all_data (arr_desc->packet) )
	    {
		fprintf (stderr, "Cannot read non-atomic arrays\n");
		return (FALSE);
	    }
	    /*  Read the array padding  */
	    if (arr_desc->padded)
	    {
		if ( !pio_read32 (channel, &pad_bytes) )
		{
		    fprintf (stderr, "Error reading pad size\n");
		    return (FALSE);
		}
		if (ch_drain (channel, pad_bytes) < pad_bytes)
		{
		    fprintf (stderr, "Error reading byte\t%s\n", ERRSTRING);
		    return (FALSE);
		}
	    }
	    /*  Skip past the array  */
	    /*  NOTE: this should be computed using network type sizes!!!  */
	    pad_bytes = ds_get_packet_size (arr_desc->packet);
	    pad_bytes *= ds_get_array_size (arr_desc);
	    if ( !ch_skip (channel, pad_bytes) )
	    {
		fprintf (stderr, "Error skipping\t%s\n", ERRSTRING);
		return (FALSE);
	    }
	    break;
	  case LISTP:
	    fprintf (stderr, "Cannot read linked lists\n");
	    return (FALSE);
	    /*break;*/
	  default:
	    if ( !dsrw_read_element (channel, type,
				     (char *) descriptor->element_desc[count],
				     packet) )
		return (FALSE);
	    break;
	}
    }
    return (TRUE);
}   /*  End Function read_packet  */
