/*  kreadtape.c

    Source file for  kreadtape  (read files from tape).

    Copyright (C) 2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read multiple files from tape and will attempt to
    create meaningful filenames (based on the "OBJECT" FITS keyword).


    Written by      Richard Gooch   22-JUN-2001

    Updated by      Richard Gooch   6-SEP-2001: Added #include's.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_m.h>
#include <karma_s.h>
#include <k_version.h>


#define VERSION "1.0"
#define MBYTE   1048576

/*  Private functions  */
STATIC_FUNCTION (flag kreadtape, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void read_tape, (CONST char *tape_device) );
STATIC_FUNCTION (flag read_file, (int tape_fd) );
STATIC_FUNCTION (int create_unique_file,
		 (CONST char *format, char name[STRING_LENGTH]) );


/*  Private data  */
static unsigned int blocksize = 28800;


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
	m_abort (function_name, "control panel");
    panel_add_item (panel, "blocksize", "bytes", K_UINT, &blocksize,
		    PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    panel_set_extra_comment_string (panel, "\nGeneral usage: tape-device");
    module_run (argc, argv, "kreadtape", VERSION, kreadtape, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */


/*  Private functions follow  */

static flag kreadtape (CONST char *command, FILE *fp)
{
    char *tape_device;

    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( tape_device = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting tape device name\n");
	    return (TRUE);
	}
	read_tape (tape_device);
	m_free (tape_device);
    }
    return (TRUE);
}   /*  End Function kreadtape  */

static void read_tape (CONST char *tape_device)
/*  [SUMMARY] Process a tape.
    <tape_device> The name of the tape device.
    [RETURNS] Nothing.
*/
{
    int tape_fd;
    /*static char function_name[] = "read_tape";*/

    if ( ( tape_fd = open (tape_device, O_RDONLY, 0) ) < 0 )
    {
	fprintf (stderr, "Error opening tape device: \"%s\"\t%s\n",
		 tape_device, ERRSTRING);
	return;
    }
    while ( read_file (tape_fd) );
    close (tape_fd);
}   /*  End Function read_tape  */

static flag read_file (int tape_fd)
/*  [SUMMARY] Read a file from tape.
    <tape_fd> The tape device file descriptor to read from.
    [RETURNS] TRUE on success, else FALSE on error or end of tape.
*/
{
    Channel channel;
    flag first_block = TRUE;
    int bytes_read, fd;
    unsigned int symbol_index = 0;
    unsigned int bytes_for_symbol = 0;
    multi_array *multi_desc;
    char *object, *ptr;
    char tmpname[STRING_LENGTH], filename[STRING_LENGTH];
    unsigned int buffer_size = 0;
    static char *buffer = NULL;
    static char symbols[] = "|/-\\";
    static char function_name[] = "read_file";

    if (buffer_size < blocksize)
    {
	if (buffer != NULL) m_free (buffer);
	if ( ( buffer = m_alloc (blocksize) ) == NULL )
	    m_abort (function_name, "buffer");
	buffer_size = blocksize;
    }
    if ( ( fd = create_unique_file ("file.%d", tmpname) ) < 0 ) return (FALSE);
    while ( ( bytes_read = read (tape_fd, buffer, blocksize) ) >= 0 )
    {
	if ( s_check_for_int () )
	{
	    fprintf (stderr, "Reading interrupted by user\n");
	    close (fd);
	    unlink (tmpname);
	    return (FALSE);
	}
	if (bytes_read < 1)
	{
	    if (first_block)
	    {
		fprintf (stderr,"Zero-length file read: assuming EndOfTape\n");
		return (FALSE);
	    }
	    break;
	}
	first_block = FALSE;
	if (write (fd, buffer, bytes_read) < bytes_read)
	{
	    fprintf (stderr, "Error writing file: \"%s\"\t%s\n",
		     tmpname, ERRSTRING);
	    close (fd);
	    unlink (tmpname);
	    return (FALSE);
	}
	bytes_for_symbol += bytes_read;
	if (bytes_for_symbol >= MBYTE)
	{
	    bytes_for_symbol = 0;
	    putc (symbols[symbol_index], stderr);
	    putc ('\b', stderr);
	    if (symbols[++symbol_index] == '\0') symbol_index = 0;
	}
    }
    if (bytes_read < 0)
    {
	fprintf (stderr, "Error reading tape\t%s\n", ERRSTRING);
	close (fd);
	unlink (tmpname);
	return (FALSE);
    }
    /*  Finished reading file (all is OK)  */
    close (fd);
    if ( ( channel = ch_open_file (tmpname, "rz") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 tmpname, ERRSTRING);
	return (FALSE);
    }
    if ( ( multi_desc = foreign_fits_read_header (channel, FALSE, FALSE, TRUE,
						  FA_FITS_READ_HEADER_END) )
	 == NULL )
    {
	ch_close (channel);
	fprintf (stderr, "Wrote file: \"%s\"\n", tmpname);
	return (TRUE);
    }
    if ( s_check_for_int () )
    {
	fprintf (stderr, "Reading interrupted by user. Wrote: \"%s\"\n",
		 tmpname);
	return (FALSE);
    }
    object = ds_get_unique_named_string (multi_desc->headers[0],
					 multi_desc->data[0], "OBJECT");
    if (object == NULL)
    {
	object = ds_get_unique_named_string (multi_desc->headers[0],
					     multi_desc->data[0], "IRAFNAME");
	if (object != NULL)
	{
	    if ( ( ptr = strrchr (object, '.') ) != NULL ) *ptr = '\0';
	}
    }
    ds_dealloc_multi (multi_desc);
    if (object == NULL)
    {
	fd = 0;
	do sprintf (filename, "file.%d.fits", fd++);
	while (access (filename, F_OK) == 0);
    }
    else
    {
	/*  Got an OBJECT name! Rename  */
	for (ptr = object; *ptr != '\0'; ++ptr)
	    if (*ptr == '/') *ptr = '=';
	fd = 0;
	do sprintf (filename, "%s.%d.fits", object, fd++);
	while (access (filename, F_OK) == 0);
    }
    if (object != NULL) m_free (object);
    if (rename (tmpname, filename) == 0)
	fprintf (stderr, "Wrote file: \"%s\"\n", filename);
    else fprintf (stderr, "Wrote file: \"%s\"\n", tmpname);
    return (TRUE);
}   /*  End Function read_file  */

static int create_unique_file (CONST char *format, char name[STRING_LENGTH])
{
    int fd;
    int count = 0;

    do sprintf (name, format, count++);
    while ( ( fd = open (name, O_WRONLY | O_CREAT | O_EXCL,
			 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) ) < 0 );
    return (fd);
}   /*  End Function create_unique_file  */
