/*  ksend.c

    Source file for  ksend  (command transmitter module).

    Copyright (C) 1998-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will send commands (encoded in the Karma data format) to
    other programmes (usually GUIs).


    Written by      Richard Gooch   23-OCT-1998

    Updated by      Richard Gooch   24-OCT-1998

    Updated by      Richard Gooch   27-MAY-2001: Added "ppm" command.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/file.h>
#include <errno.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_m.h>
#include <k_version.h>


#define VERSION "1.1"


/*  Private functions  */
STATIC_FUNCTION (flag ksend, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (multi_array *alloc_desc,
		 (CONST char **elem_names, unsigned int num_elem) );
STATIC_FUNCTION (void send_export_command,
		 (CONST char *event_name, CONST char *p) );
STATIC_FUNCTION (void do_print, (CONST char *p) );
STATIC_FUNCTION (void do_ppm, (CONST char *p) );
STATIC_FUNCTION (void do_loadfile, (CONST char *p) );
STATIC_FUNCTION (void do_rclip, (CONST char *p) );
STATIC_FUNCTION (void do_hclip, (CONST char *p) );


/*  Private data  */


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "rclip", "dataset minimum maximum",
		    PIT_FUNCTION, (void *) do_rclip,
		    PIA_END);
    panel_add_item (panel, "print", "filename",
		    PIT_FUNCTION, (void *) do_print,
		    PIA_END);
    panel_add_item (panel, "ppm", "filename",
		    PIT_FUNCTION, (void *) do_ppm,
		    PIA_END);
    panel_add_item (panel, "load", "domain filename",
		    PIT_FUNCTION, (void *) do_loadfile,
		    PIA_END);
    panel_add_item (panel, "hclip", "dataset proportion",
		    PIT_FUNCTION, (void *) do_hclip,
		    PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    /*panel_set_extra_comment_string (panel, "\nGeneral usage: infile");*/
    module_run (argc, argv, "ksend", VERSION, ksend, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */


/*  Private functions follow  */

static multi_array *alloc_desc (CONST char **elem_names, unsigned int num_elem)
/*  [SUMMARY] Allocate a descriptor.
    <elem_names> The element names.
    <num_elem> The number of elements in the top level packet descriptor.
    [RETURNS] The newly allocated descriptor. On failure the process abort.
*/
{
    unsigned int count;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    static char function_name[] = "alloc_desc";

    if ( ( multi_desc = ds_alloc_multi (1) ) == NULL )
	m_abort (function_name, "multi_array");
    if ( ( pack_desc = ds_alloc_packet_desc (num_elem) ) == NULL )
	m_abort (function_name, "packet descriptor");
    multi_desc->headers[0] = pack_desc;
    for (count = 0; count < num_elem; ++count)
    {
	if ( ( pack_desc->element_desc[count] = st_dup (elem_names[count]) )
	     == NULL ) m_abort (function_name, "element name");
    }
    return (multi_desc);
}   /*  End Function alloc_desc   */

static flag ksend (CONST char *command, FILE *fp)
{
    fprintf (stderr, "Ignored: \"%s\"\n", command);
    return (TRUE);
}   /*  End Function ksend  */

static void send_export_command (CONST char *event_name, CONST char *p)
/*  [SUMMARY] Send a print command.
    <event_name> The event name to send.
    <p> A string containing the filename to send for the print command.
    [RETURNS] Nothing.
*/
{
    unsigned int ui_tmp = 0;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    char packet[sizeof ui_tmp + sizeof p];
    static char *names[2] = {"instance_number", "filename"};

    if ( (p == NULL) || (*p == '\0') )
    {
	fprintf (stderr, "Must specify filename\n");
	return;
    }
    multi_desc = alloc_desc ( (CONST char **) names, 2 );
    pack_desc = multi_desc->headers[0];
    pack_desc->element_types[0] = K_UINT;
    pack_desc->element_types[1] = K_VSTRING;
    m_copy (packet, &ui_tmp, sizeof ui_tmp);
    m_copy (packet + ds_get_element_offset (pack_desc, 1), &p, sizeof p);
    multi_desc->data[0] = packet;
    dsxfr_send_multi (NULL, multi_desc, event_name);
    multi_desc->data[0] = NULL;
    ds_dealloc_multi (multi_desc);
}   /*  End Function send_export_command  */

static void do_print (CONST char *p)
/*  [SUMMARY] Send a print command.
    <p> A string containing the filename to send for the print command.
    [RETURNS] Nothing.
*/
{
    send_export_command ("PostscriptWidget::WritePS", p);
}   /*  End Function do_print  */

static void do_ppm (CONST char *p)
/*  [SUMMARY] Send a PPM command.
    <p> A string containing the filename to send for the PPM command.
    [RETURNS] Nothing.
*/
{
    send_export_command ("ExportMenuWidget::WritePPMImageWindow", p);
}   /*  End Function do_ppm  */

static void do_loadfile (CONST char *p)
/*  [SUMMARY] Send a loadfile command.
    <p> A string containing the domain and filenames to send.
    [RETURNS] Nothing.
*/
{
    multi_array *multi_desc;
    packet_desc *pack_desc;
    char *domain;
    char packet[sizeof p + sizeof p];
    static char *names[2] = {"domain", "filename"};

    if ( (p == NULL) || (*p == '\0') )
    {
	fprintf (stderr, "Must specify domain name and filename\n");
	return;
    }
    if ( ( domain = ex_str (p, &p) ) == NULL )
    {
	fprintf (stderr, "Must specify filename\n");
	return;
    }
    multi_desc = alloc_desc ( (CONST char **) names, 2 );
    pack_desc = multi_desc->headers[0];
    pack_desc->element_types[0] = K_VSTRING;
    pack_desc->element_types[1] = K_VSTRING;
    m_copy (packet, &domain, sizeof domain);
    m_copy (packet + ds_get_element_offset (pack_desc, 1), &p, sizeof p);
    multi_desc->data[0] = packet;
    dsxfr_send_multi (NULL, multi_desc, "FilepopupWidget::Load");
    multi_desc->data[0] = NULL;
    m_free (domain);
    ds_dealloc_multi (multi_desc);
}   /*  End Function do_loadfile  */

static void do_rclip (CONST char *p)
/*  [SUMMARY] Send a range clip command.
    <p> A string containing the minimum and maximum.
    [RETURNS] Nothing.
*/
{
    double min, max;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    char *ptr;
    char title[STRING_LENGTH];
    char packet[sizeof &title + sizeof min + sizeof max];
    static char *names[3] = {"title", "minimum", "maximum"};

    if ( (p == NULL) || (*p == '\0') )
    {
	fprintf (stderr, "Must specify dataset, minimum and maximum\n");
	return;
    }
    if ( ( ptr = ex_str (p, &p) ) == NULL )
    {
	fprintf (stderr, "Must specify dataset\n");
	return;
    }
    sprintf (title, "Intensity Zoom for %s", ptr);
    m_free (ptr);
    min = ex_float (p, &p);
    max = ex_float (p, &p);
    multi_desc = alloc_desc ( (CONST char **) names, 3 );
    pack_desc = multi_desc->headers[0];
    pack_desc->element_types[0] = K_VSTRING;
    pack_desc->element_types[1] = K_DOUBLE;
    pack_desc->element_types[2] = K_DOUBLE;
    ptr = title;
    m_copy (packet, &ptr, sizeof ptr);
    m_copy (packet + ds_get_element_offset (pack_desc, 1), &min, sizeof min);
    m_copy (packet + ds_get_element_offset (pack_desc, 2), &max, sizeof max);
    multi_desc->data[0] = packet;
    dsxfr_send_multi (NULL, multi_desc, "DataclipWidget::Clip");
    multi_desc->data[0] = NULL;
    ds_dealloc_multi (multi_desc);
}   /*  End Function do_rclip  */

static void do_hclip (CONST char *p)
/*  [SUMMARY] Send a histogram clip command.
    <p> A string containing the minimum and maximum.
    [RETURNS] Nothing.
*/
{
    double val;
    multi_array *multi_desc;
    packet_desc *pack_desc;
    char *ptr;
    char title[STRING_LENGTH];
    char packet[sizeof &title + sizeof val];
    static char *names[2] = {"title", "hist_clip"};

    if ( (p == NULL) || (*p == '\0') )
    {
	fprintf (stderr, "Must specify dataset and proportion\n");
	return;
    }
    if ( ( ptr = ex_str (p, &p) ) == NULL )
    {
	fprintf (stderr, "Must specify dataset\n");
	return;
    }
    sprintf (title, "Intensity Zoom for %s", ptr);
    m_free (ptr);
    val = ex_float (p, &p);
    multi_desc = alloc_desc ( (CONST char **) names, 2 );
    pack_desc = multi_desc->headers[0];
    pack_desc->element_types[0] = K_VSTRING;
    pack_desc->element_types[1] = K_DOUBLE;
    ptr = title;
    m_copy (packet, &ptr, sizeof ptr);
    m_copy (packet + ds_get_element_offset (pack_desc, 1), &val, sizeof val);
    multi_desc->data[0] = packet;
    dsxfr_send_multi (NULL, multi_desc, "DataclipWidget::Clip");
    multi_desc->data[0] = NULL;
    ds_dealloc_multi (multi_desc);
}   /*  End Function do_hclip  */
