/*  tx.c

    Source file for  tx  (data structure transmitter module).

    Copyright (C) 1993-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will read in a Karma data file and transmit the
    multi-array over one or more Karma connections to a Karma server module.
    The module will also allow a list of files to be transmitted, one after the
    other.


    Written by      Richard Gooch   4-OCT-1992

    Updated by      Richard Gooch   19-DEC-1992

    Updated by      Richard Gooch   1-JAN-1993: Took account of change to
  ch_gets  .

    Updated by      Richard Gooch   27-MAR-1993: Took account of change to
  dsxfr_get_multi  .

    Updated by      Richard Gooch   17-MAY-1993: Took account of change to
  dsxfr_get_multi  .

    Updated by      Richard Gooch   24-JUL-1993: Changed to  ch_open_file  .

    Updated by      Richard Gooch   8-AUG-1993: Tidied up by deallocating
  multi_array data structure when finished.

    Updated by      Richard Gooch   5-OCT-1993: Changed over to  panel_
  package for command line user interface.

    Updated by      Richard Gooch   6-OCT-1993: Moved  main  into this file.

    Updated by      Richard Gooch   23-NOV-1993: Changed to use of
  ex_word_skip  .

    Updated by      Richard Gooch   10-MAR-1995: Added support for "spray"
  protocol.

    Updated by      Richard Gooch   14-JUN-1995: Made use of <ex_uint>.

    Updated by      Richard Gooch   9-AUG-1995: Added support for "hog_request"
  protocol.

    Updated by      Richard Gooch   29-JAN-1996: Added call to
  <im_register_lib_version>.

    Updated by      Richard Gooch   1-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-AUG-1996: Upgraded "spray" protocol to
  support synchronisation.

    Updated by      Richard Gooch   23-MAY-1997: Added "ping_server" protocol.

    Updated by      Richard Gooch   3-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   1-OCT-1997: Added "server_exit" protocol.

    Updated by      Richard Gooch   27-DEC-1997: Added "udp_spray" protocol.

    Updated by      Richard Gooch   28-DEC-1997: Added spray pattern and
  maximum transmit rate.

    Updated by      Richard Gooch   4-FEB-1998: Bug fix in <udp_spray> when
  insufficient arguments given.

    Updated by      Richard Gooch   18-FEB-1997: Removed "udp_spray" protocol
  and added datagram support for "spray" protocol.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   13-DEC-1998: Added call to <sched_yield>
  in <delay>.

    Updated by      Richard Gooch   2-FEB-1999: Disabled call to <sched_yield>
  for OS_Solaris: dumbfucks require -lposix4.

    Updated by      Richard Gooch   19-FEB-1999: Switched back from <write> to
  <ch_write> in <datagram_spray> as XTI interface doesn't support <write>.

    Updated by      Richard Gooch   2-MAR-1999: Added call to <ch_flush> in
  <datagram_spray> to make packets < sdu_max work.

    Updated by      Richard Gooch   25-AUG-1999: Added #ifdef for IRIX 6.2.

    Updated by      Richard Gooch   7-OCT-2000: Took account of interface
  change to function <dm_native_wait_input>.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   17-APR-2002: Added speed multiplier prefix.

    Last updated by Richard Gooch   22-DEC-2002: Moved "spray" protocol to new
  <kbwtest> module.


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#if !defined(OS_SunOS) && !defined(OS_IRIX6)
#  include <sched.h>
#endif
#include <sys/time.h>
#include <sys/times.h>
#include <sys/file.h>
#include <errno.h>
#include <karma.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_pio.h>
#include <karma_ex.h>
#include <karma_ds.h>
#include <karma_dm.h>
#include <karma_im.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_s.h>
#include <karma_m.h>
#include <karma_n.h>
#include <karma_r.h>
#include <k_version.h>


#define VERSION "1.7"
#define BUF_SIZE 262144
#define DEFAULT_DGRAM_SIZE 4096


STATIC_FUNCTION (flag tx, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag transmit_arrayfile, (char *arrayfile) );
STATIC_FUNCTION (void run_animate_file, (char *animate_file) );
STATIC_FUNCTION (void ping, (CONST char *p) );


/*  Private data  */
static flag animate_flag = FALSE;
static int num_iterations = 1;


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    conn_register_client_protocol ("hog_request", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("ping_server", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("server_exit", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "ping", "ping specified number of times",
		    PIT_FUNCTION, (void *) ping,
		    PIA_END);
    panel_add_item (panel, "num_iterations", "number", K_INT, &num_iterations,
		    PIA_END);
    panel_add_item (panel, "animate", "flag", PIT_FLAG, &animate_flag,
		    PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile");
    module_run (argc, argv, "tx", VERSION, tx, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag tx (CONST char *command, FILE *fp)
{
    char filename[STRING_LENGTH+1];
    extern int animate_flag;

    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	sscanf (command, "%s", filename);
	if (animate_flag) run_animate_file (filename);
	else transmit_arrayfile (filename);
    }
    return (TRUE);
}   /*  End Function tx  */

static flag transmit_arrayfile (char *arrayfile)
/*  This routine will transmit an arrayfile over any Karma "multi_array"
    connections.
    The name of the arrayfile must be pointed to by  arrayfile  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    multi_array *multi_desc;
    static char function_name[] = "operate_file";

    if (arrayfile == NULL)
    {
	fprintf (stderr, "NULL array pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Load input array  */
    if ( ( multi_desc = dsxfr_get_multi (arrayfile, TRUE, K_CH_MAP_LOCAL,
					 FALSE) )
	== NULL )
    {
	return (FALSE);
    }
    /*  Transmit multi-array over connection  */
    if (dsxfr_put_multi ("connections", multi_desc) != TRUE)
    {
	fprintf (stderr, "Error transmitting data over connection\n");
	/*  Decrement attachment count  */
	ds_dealloc_multi (multi_desc);
	return (FALSE);
    }
    /*  Decrement attachment count  */
    ds_dealloc_multi (multi_desc);
    return (TRUE);
}   /*  End Function transmit_arrayfile  */

static void run_animate_file (char *animate_file)
/*  This routine will read and transmit a number of arrayfiles over any Karma
    "multi_array" connections.
    The names of the arrayfiles to transmit must be contained in the file
    with name pointed to by  animate_file  .
    The routine returns nothing.
*/
{
    Channel channel;
    int iter_count;
    char line[STRING_LENGTH + 1];
    extern int num_iterations;

    for (iter_count = 0; (iter_count < num_iterations) || (num_iterations < 0);
	 ++iter_count)
    {
	/*  Load animate file  */
	if ( ( channel = ch_open_file (animate_file, "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		     animate_file, ERRSTRING);
	    return;
	}
	/*  Read lines  */
	while ( ch_gets (channel, line, STRING_LENGTH) )
	{
	    /*  Not end of file  */
	    if ( (line[0] == '#') || (line[0] == '\n') )
	    {
		/*  Comment or blank line: skip  */
		continue;
	    }
	    if ( !transmit_arrayfile (line) ) return;
	    if ( s_check_for_int () )
	    {
		fprintf (stderr,
			 "Control C: quitting animate file: \"%s\" after: %d iterations\n",
			 animate_file, iter_count);
		ch_close (channel);
		return;
	    }
	}
	ch_close (channel);
    }
}   /*  End Function run_animate_file  */

static void ping (CONST char *p)
/*  [PURPOSE] This routine will ping all "ping_server" servers.
    <p> A string containing the number of times to ping.
    [RETURNS] Nothing.
*/
{
    Connection connection;
    unsigned int num_pings, size;
    unsigned int conn_count, num_conn, ping_count;
    signed long tx_usec, rtt_usec, rx_usec;
    signed long min_rtt_time, av_rtt_time, max_rtt_time;
    signed long min_tx_time, av_tx_time, max_tx_time;
    signed long min_rx_time, av_rx_time, max_rx_time;
    struct timeval tx_start_time, tx_done_time, rx_start_time, rx_done_time;
    char *name;
    char buffer[BUF_SIZE];
    static char function_name[] = "ping";

    if (p == NULL)
    {
	fprintf (stderr, "Must specify number of times to ping\n");
	return;
    }
    if ( ( num_pings = ex_uint (p, &p) ) < 1 )
    {
	fprintf (stderr, "Must specify at least 1 ping\n");
	return;
    }
    if (p == NULL) size = 1;
    else size = ex_uint (p, &p);
    if (size > BUF_SIZE)
    {
	fprintf (stderr, "Size: %u too large, %u bytes maximum\n",
		 size, BUF_SIZE);
	return;
    }
    m_clear (buffer, size);
    if ( ( num_conn = conn_get_num_client_connections ("ping_server") ) < 1 )
    {
	fprintf (stderr, "No connections to ping servers\n");
	return;
    }
    fprintf (stderr, "Pinging each ping server %d times with size %u bytes\n",
	     num_pings, size);
    for (conn_count = 0; conn_count < num_conn; ++conn_count)
    {
	Channel channel;
	int fd;

	if ( ( connection = conn_get_client_connection ("ping_server",
							conn_count) ) == NULL )
	{
	    fprintf (stderr, "Connection: %u does not exist!\n", conn_count);
	    a_prog_bug (function_name);
	}
	if ( ( name = conn_get_connection_module_name (connection) ) == NULL )
	{
	    fprintf (stderr, "NULL module name for connection: %u\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_datagram_channel (connection);
	if (channel == NULL) channel = conn_get_channel (connection);
	fd = ch_get_descriptor (channel);
	min_rtt_time = 1000000;
	max_rtt_time = 0.0;
	av_rtt_time = 0.0;
	min_tx_time = 1000000;
	max_tx_time = 0.0;
	av_tx_time = 0.0;
	min_rx_time = 1000000;
	max_rx_time = 0.0;
	av_rx_time = 0.0;
	fprintf (stderr, "Pinging connection: %u (module: \"%s\")...\n",
		 conn_count, name);
	fprintf (stderr, "Iter\ttxt (us)    rtt (us)    rxt (us)\n");
	for (ping_count = 0; ping_count < num_pings; ++ping_count)
	{
	    if (gettimeofday (&tx_start_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    if (ch_write (channel, buffer, size) < size)
	    {
		
		fprintf (stderr, "Error writing\t%s\n", ERRSTRING);
		continue;
	    }
	    if ( !ch_flush (channel) )
	    {
		
		fprintf (stderr, "Error flushing\t%s\n", ERRSTRING);
		continue;
	    }
	    if (gettimeofday (&tx_done_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    dm_native_wait_input (fd, -1);
	    if (gettimeofday (&rx_start_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    if (ch_read (channel, buffer, size) < size)
	    {
		
		fprintf (stderr, "Error reading\t%s\n", ERRSTRING);
		continue;
	    }
	    if (gettimeofday (&rx_done_time, NULL) != 0)
	    {
		fprintf (stderr, "Error getting time of day\t%s\n", ERRSTRING);
		exit (RV_SYS_ERROR);
	    }
	    tx_usec = tx_done_time.tv_usec - tx_start_time.tv_usec;
	    tx_usec += (tx_done_time.tv_sec - tx_start_time.tv_sec) * 1000000;
	    if (tx_usec < min_tx_time) min_tx_time = tx_usec;
	    if (tx_usec > max_tx_time) max_tx_time = tx_usec;
	    av_tx_time += tx_usec;
	    rtt_usec = rx_start_time.tv_usec - tx_done_time.tv_usec;
	    rtt_usec += (rx_start_time.tv_sec - tx_done_time.tv_sec) * 1000000;
	    if (rtt_usec < min_rtt_time) min_rtt_time = rtt_usec;
	    if (rtt_usec > max_rtt_time) max_rtt_time = rtt_usec;
	    av_rtt_time += rtt_usec;
	    rx_usec = rx_done_time.tv_usec - rx_start_time.tv_usec;
	    rx_usec += (rx_done_time.tv_sec - rx_start_time.tv_sec) * 1000000;
	    if (rx_usec < min_rx_time) min_rx_time = rx_usec;
	    if (rx_usec > max_rx_time) max_rx_time = rx_usec;
	    av_rx_time += rx_usec;
#if 0
	    printf ("%u\t%-10ld  %-10ld  %ld\n",
		    ping_count, tx_usec, rtt_usec, rx_usec);
#endif
	    usleep (50000);
	}
	fflush (stdout);
	av_tx_time /= num_pings;
	av_rtt_time /= num_pings;
	av_rx_time /= num_pings;
	fprintf (stderr, "Min:\t%-10ld  %-10ld  %ld\n",
		 min_tx_time, min_rtt_time, min_rx_time);
	fprintf (stderr, "Av:\t%-10ld  %-10ld  %ld\n",
		 av_tx_time, av_rtt_time, av_rx_time);
	fprintf (stderr, "Max:\t%-10ld  %-10ld  %ld\n",
		 max_tx_time, max_rtt_time, max_rx_time);
    }
}   /*  End Function ping  */
