#include <stdio.h>
#include <math.h>
#include "karma.h"
#include <karma_ds.h>
#include <karma_a.h>
#include "persp_3d.h"


#define FP_TYPE float

#ifdef i860
#define VXASM
#endif

#ifndef VXASM
#define project_line_loop project_line_loop_l
#endif

EXTERN_FUNCTION (void project_line_loop,
		 (char *line, int length, int inp_stride, float *image,
		  unsigned int *yoffsets,
		  int lower_threshold, float one_on_sq1, float one_on_sq2,
		  float one_on_sq3, float *prep_arrays, float distance_2d_prep,
		  float coord_h_prep, float coord_v_prep,
		  float scale_h, float scale_v,
		  float sq_distance_3d_prep, float minimum_h, float maximum_h,
		  float minimum_v, float maximum_v, float toobig) );

flag project_line_real_byte (char *line, float coord_z, float coord_y,
			     unsigned int length, unsigned int inp_stride,
			     float *image, dim_desc **out_dim_desc,
			     unsigned int *yoffsets, double position[3],
			     double vertical[3], double horizontal[3],
			     double direction[3], C_DATA_TYPE lower_threshold,
			     C_DATA_TYPE upper_threshold, float one_on_sq1,
			     float one_on_sq2, float one_on_sq3,
			     float *prep_arrays)
/*  This routine will project a line in 3D space into 2D space.
    The line values must be pointed to by  line  .
    The co-ordinate of the plane must be given by  coord_z  .
    The co-orinate of the line must be given by  coord_y  .
    The length of the line (in co-ordinates) must be given by  length  .
    The stride (in bytes) between consecutive data points in the cube must be
    given by  inp_stride  .
    The image data must be pointed to by  image  .
    The dimension descriptors for the image must be pointed to by by the array
    of pointers pointed to by  out_dim_desc  .
    The offset of each row in the output image must be in the array pointed to
    by  yoffsets  .
    The eye position must be pointed to by  position  .
    The vertical direction vector must be pointed to by  vertical  .
    The horizontal direction vector must be pointed to by  horizontal  .
    The focus direction vector must be pointed to by  direction  .
    If a data point is less than or equal to  lower_threshold  then it is
    ignored.
    If a data point is greater than or equal to  upper_threshold  then it is
    ignored.
    Some precomputed square roots are passed as  one_on_sq1  ,  one_on_sq2  and
    one_on_sq3  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    float c1;
    float distance_2d_prep;
    float coord_h_prep;
    float coord_v_prep;
    float scale_h;
    float scale_v;
    float sq_distance_3d_prep;
    float minimum_h;
    float maximum_h;
    float minimum_v;
    float maximum_v;
    float point_v[3];
    static char function_name[] = "project_line_real_byte";

    if ( (line == NULL) || (image == NULL) || (out_dim_desc == NULL) )
    {
	(void) fprintf (stderr, "\nNULL pointer(s) passed");
	a_prog_bug (function_name);
    }
    /*  Compute some components of point vector  */
    point_v[1] = coord_y - position[1];
    point_v[2] = coord_z - position[2];
    /*  Compute eye plane constant  */
    c1 = (direction[0] * position[0] + direction[1] * position[1] +
	  direction[2] * position[2]);
    minimum_h = (*out_dim_desc[1]).minimum;
    maximum_h = (*out_dim_desc[1]).maximum;
    minimum_v = (*out_dim_desc[0]).minimum;
    maximum_v = (*out_dim_desc[0]).maximum;
    /*  Compute horizontal and vertical dimension scaling  */
    scale_h = ( (float) ( (*out_dim_desc[1]).length - 1 ) /
	       (maximum_h - minimum_h) );
    scale_v = ( (float) ( (*out_dim_desc[0]).length - 1 ) /
	       (maximum_v - minimum_v) );
    /*  Precompute some more values  */
    distance_2d_prep = direction[1] * coord_y + direction[2] * coord_z - c1;
    coord_h_prep = horizontal[1] * point_v[1] + horizontal[2] * point_v[2];
    coord_v_prep = vertical[1] * point_v[1] + vertical[2] * point_v[2];
    sq_distance_3d_prep = point_v[1] * point_v[1] + point_v[2] * point_v[2];
    project_line_loop (line, length, inp_stride, image, yoffsets,
		       lower_threshold, one_on_sq1, one_on_sq2, one_on_sq3,
		       prep_arrays,
		       distance_2d_prep, coord_h_prep, coord_v_prep,
		       scale_h, scale_v, sq_distance_3d_prep,
		       minimum_h, maximum_h, minimum_v, maximum_v,
		       TOOBIG);
    return (TRUE);
}   /*  End Function project_line_real_byte  */

void project_line_loop_l (char *line, int length, int inp_stride, float *image,
			  unsigned int *yoffsets, int lower_threshold,
			  float one_on_sq1, float one_on_sq2, float one_on_sq3,
			  float *prep_arrays, float distance_2d_prep,
			  float coord_h_prep, float coord_v_prep,
			  float scale_h, float scale_v,
			  float sq_distance_3d_prep, float minimum_h,
			  float maximum_h, float minimum_v, float maximum_v,
			  float toobig)
/*  <prep_arrays> Used by macros.
*/
{
    int inp_value;
    float distance_2d;
    float coord_h;
    float coord_v;
    float sq_distance_3d;
    float *image_ptr;

    /*  Process each point  */
    for (; length-- > 0; line += inp_stride, prep_arrays += 4)
    {
	/*  Convert input value  */
	inp_value = *(C_DATA_TYPE *) line;
	/*  Skip zero value  */
	if (inp_value <= lower_threshold) continue;
	/*  Calculate distance from point to position in eye plane  */
	distance_2d = curr_distance_2d + distance_2d_prep;
	/*  Skip if point is behind eye plane  */
	if (distance_2d <= 0.0) continue;
	distance_2d = 1.0 / (distance_2d * one_on_sq3);
	coord_h = (curr_coord_h + coord_h_prep) * one_on_sq1 * distance_2d;
	/*  Check if point is in field of view  */
	if (coord_h < minimum_h) continue;
	if (coord_h > maximum_h) continue;
	coord_v = (curr_coord_v + coord_v_prep) * one_on_sq2 * distance_2d;
	/*  Check if point is in field of view  */
	if (coord_v < minimum_v) continue;
	if (coord_v > maximum_v) continue;
	/*  Compute true distance to point  */
	sq_distance_3d = curr_sq_point_v0 + sq_distance_3d_prep;
	image_ptr = image;
	image_ptr += yoffsets[(int) (0.5 + (coord_v - minimum_v) * scale_v)];
	image_ptr += (int) (0.5 + (coord_h - minimum_h) * scale_h);
	if (*image_ptr >= toobig)
	{
	    /*  Previously untouched point  */
	    *image_ptr = (float) inp_value / sq_distance_3d;
	}
	else
	{
	    /*  Point has already been written to  */
	    *image_ptr += (float) inp_value / sq_distance_3d;
	}
	/*  Ready for next point  */
    }
}
