/*  This Karma module will render a 3 dimensional data set into a 2 dimensional
    data set, using an external algorithm.
    The output is written to a new Karma arrayfile.


    Written by      Richard Gooch   24-MAY-1992

    Updated by      Richard Gooch   25-DEC-1992

    Updated by      Richard Gooch   3-JAN-1993: Took acccount of change to
  ds_alloc_data

    Updated by      Richard Gooch   28-MAR-1993: Explicitly disabled tiled
  arrays in  generate_desc  .

    Updated by      Richard Gooch   15-SEP-1993: Fixed initialisation of  vble
  to make GCC happy.

    Updated by      Richard Gooch   27-JUN-1999: Switched to <panel> package.

    Last updated by Richard Gooch   30-JUN-1999: Switch to <dsproc_object>.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_dsproc.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include "persp_3d.h"


flag pre_process ();
flag process_array ();
packet_desc *generate_desc ();
flag process_occurrence ();
flag reorder_one_array ();

static int save_unproc = TRUE;

#define VERTICAL_DIMENSION_NAME "y"
#define HORIZONTAL_DIMENSION_NAME "x"
#define DEFAULT_LENGTH 256
#define DEFAULT_MIN -1.0
#define DEFAULT_MAX 1.0

#define NUMARRAYS 100
static char *array_names[NUMARRAYS];
static int num_arrays = 0;

/*  Global variables   */

static double focus[3] = {0.0, 0.0, 0.0};
static int num_focii = 3;

static int h_length = DEFAULT_LENGTH;
static double h_max = DEFAULT_MAX;
static double h_min = DEFAULT_MIN;

static double position[3] = {10.0, 0.0, 0.0};
static int num_positions = 3;

static int v_length = DEFAULT_LENGTH;
static double v_max = DEFAULT_MAX;
static double v_min = DEFAULT_MIN;

static double lower_threshold = -TOOBIG;

static double upper_threshold = TOOBIG;

static double vertical[3] = {0.0, 0.0, 1.0};
static int num_verticals = 3;

static char *dim_names[3] = {NULL, NULL, NULL};

static unsigned int *reorder_list = NULL;

void init_panel ()
{
    KControlPanel panel;
    static char function_name[] = "init_panel";

    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
        m_abort (function_name, "control panel");
    }
    panel_add_item (panel, "array_names", "", K_VSTRING, array_names,
                    PIA_ARRAY_LENGTH, &num_arrays,
		    PIA_ARRAY_MIN_LENGTH, 0,
                    PIA_ARRAY_MAX_LENGTH, NUMARRAYS,
                    PIA_END);
    panel_add_item (panel, "focus", "(3D co-ordinate)", K_DOUBLE, focus,
		    PIA_ARRAY_LENGTH, &num_focii,
		    PIA_ARRAY_MAX_LENGTH, 3,
                    PIA_END);
    panel_add_item (panel, "h_length", "(number of pixels)", K_INT, &h_length,
		    PIA_END);
    panel_add_item (panel, "h_max", "(co-ordinate)", K_DOUBLE, &h_max,
		    PIA_END);
    panel_add_item (panel, "h_min", "(co-ordinate)", K_DOUBLE, &h_min,
		    PIA_END);
    panel_add_item (panel, "lower_threshold", "(value)", K_DOUBLE,
		    &lower_threshold,
		    PIA_END);
    panel_add_item (panel, "position", "(3D co-ordinate)", K_DOUBLE, position,
		    PIA_ARRAY_LENGTH, &num_positions,
		    PIA_ARRAY_MAX_LENGTH, 3,
		    PIA_END);
    panel_add_item (panel, "save_unproc_data", "", PIT_FLAG, &save_unproc,
                    PIA_END);
    
    panel_add_item (panel, "upper_threshold", "(value)", K_DOUBLE,
		    &upper_threshold,
		    PIA_END);
    panel_add_item (panel, "v_length", "(number of pixels)", K_INT, &v_length,
		    PIA_END);
    panel_add_item (panel, "v_max", "(co-ordinate)", K_DOUBLE, &v_max,
		    PIA_END);
    panel_add_item (panel, "v_min", "(co-ordinate)", K_DOUBLE, &v_min,
		    PIA_END);
    panel_add_item (panel, "vertical", "(3D co-ordinate)", K_DOUBLE, vertical,
		    PIA_ARRAY_LENGTH, &num_verticals,
		    PIA_ARRAY_MAX_LENGTH, 3,
		    PIA_END);
    panel_push_onto_stack (panel);
}   /*  End Function init_panel  */

flag render (CONST char *p, FILE *fp)
{
    double mag;
    double direction[3];
    char *arrayfile;
    extern int save_unproc;
    extern int num_arrays;
    extern int num_focii;
    extern int num_positions;
    extern int num_verticals;
    extern double lower_threshold;
    extern double upper_threshold;
    extern double focus[3];
    extern double position[3];
    extern double vertical[3];
    extern char *dim_names[3];
    extern char *array_names[NUMARRAYS];
    /*static char function_name[] = "render";*/

    /*  Check lower and upper thresholds  */
    if (lower_threshold > upper_threshold)
    {
	fprintf (stderr,
		 "WARNING: lower_threshold: %e is greater than upper_threshold: %e\n",
			lower_threshold, upper_threshold);
    }
    /*  Check if focus, position and vertical are all 3D co-ordinates  */
    if (num_focii != 3)
    {
	fprintf (stderr,
			"Must have exactly 3 co-ordinates for focus\n");
	return (TRUE);
    }
    if (num_positions != 3)
    {
	fprintf (stderr,
			"Must have exactly 3 co-ordinates for position\n");
	return (TRUE);
    }
    if (num_verticals != 3)
    {
	fprintf (stderr,
			"Must have exactly 3 co-ordinates for vertical\n");
	return (TRUE);
    }
    /*  Compute vision direction vector  */
    direction[0] = focus[0] - position[0];
    direction[1] = focus[1] - position[1];
    direction[2] = focus[2] - position[2];
    /*  Check if vertical vector is perpendicular to PF (direction vector)  */
    mag = (direction[0] * direction[0] + direction[1] * direction[1] +
	   direction[2] * direction[2]);
    mag *= (vertical[0] * vertical[0] + vertical[1] * vertical[1] +
	    vertical[2] * vertical[2]);
    mag = sqrt (mag);
    if (fabs (direction[0] * vertical[0] + direction[1] * vertical[1] +
	     direction[2] * vertical[2]) > 1.0 / mag)
    {
	fprintf (stderr,
			"Vertical and direction vectors not perpendicular\n");
	fprintf ( stderr, "Dot product: %e\n",
			fabs (direction[0] * vertical[0] +
			      direction[1] * vertical[1] +
			      direction[2] * vertical[2]) );
	return (TRUE);
    }
    for ( ; p; p = ex_command_skip (p) )
    {
	if ( ( arrayfile = ex_str (p, &p) ) == NULL )
	{
	    fprintf (stderr, "Error extracting arrayfile name\n");
	    return (TRUE);
	}
	if ( ( dim_names[0] = ex_str (p, &p) ) == NULL )
	{
	    fprintf (stderr, "Must specify three dimension names to render\n");
	    m_free (arrayfile);
	    return (TRUE);
	}
	if ( ( dim_names[1] = ex_str (p, &p) ) == NULL )
	{
	    fprintf (stderr,
		     "Must specify two more dimension names to render\n");
	    m_free (arrayfile);
	    m_free (dim_names[0]);
	    return (TRUE);
	}
	if ( ( dim_names[2] = ex_str (p, &p) ) == NULL )
	{
	    fprintf (stderr,
		     "Must specify another dimension names to render\n");
	    m_free (arrayfile);
	    m_free (dim_names[0]);
	    m_free (dim_names[1]);
	    return (TRUE);
	}
	dsproc_object (arrayfile, array_names, num_arrays, save_unproc,
		       pre_process, process_array, post_process,
		       K_CH_MAP_LOCAL);
	m_free (arrayfile);
	m_free (dim_names[0]);
	m_free (dim_names[1]);
	m_free (dim_names[2]);
    }
    return (TRUE);
}   /*  End Function render  */

flag pre_process (multi_desc)
/*  This routine will pre process the multi array general data structure
    pointed to by  multi_desc  .
    The routine returns TRUE if processing is to continue, else it
    returns FALSE.
*/
multi_array *multi_desc;
{
    char txt[STRING_LENGTH];
    static char function_name[] = "pre_process";

    if (ds_f_array_name (multi_desc, dim_names[0], (char **) NULL,
			 (unsigned int *) NULL)
        != IDENT_NOT_FOUND)
    {
	sprintf (txt, "array name must not be: \"%s\"", dim_names[0]);
        a_func_abort (function_name, txt);
        return (FALSE);
    }
    if (ds_f_array_name (multi_desc, dim_names[1], (char **) NULL,
			 (unsigned int *) NULL)
        != IDENT_NOT_FOUND)
    {
	sprintf (txt, "array name must not be: \"%s\"", dim_names[1]);
        a_func_abort (function_name, txt);
        return (FALSE);
    }
    if (ds_f_array_name (multi_desc, dim_names[2], (char **) NULL,
			 (unsigned int *) NULL)
        != IDENT_NOT_FOUND)
    {
	sprintf (txt, "array name must not be: \"%s\"", dim_names[2]);
        a_func_abort (function_name, txt);
        return (FALSE);
    }
    return (TRUE);
}   /*  End Function pre_process    */

flag process_array (packet_desc *inp_desc, char *inp_data,
		    packet_desc **out_desc,char **out_data)
/*  This routine will process a general data structure, rendering each
    occurrence of the dimension names in the external strings  dim_names  .
    The input data descriptor and data must be pointed to by  inp_desc  and
    inp_data  ,respectively.
    The output data descriptor pointer will be written to the storage pointed
    to by  out_desc  and the output data pointer will be written to the storage
    pointed to by  out_data  .
    The routine returns TRUE on successful processing, else it displays an
    error message and returns FALSE.
*/
{
    array_desc *array_desc_inp;
    static char function_name[] = "process_array";

    if ( (inp_desc == NULL) || (out_desc == NULL) )
    {
	fprintf (stderr, "NULL descriptor pointer(s)\n");
        a_prog_bug (function_name);
    }
    if ( (inp_data == NULL) || (out_data == NULL) )
    {
	fprintf (stderr, "NULL data pointer(s)\n");
	a_prog_bug (function_name);
    }
    if ( (dim_names[0] == NULL) || (dim_names[1] == NULL) ||
	(dim_names[2] == NULL) )
    {
	fprintf (stderr,
			"NULL external variable dimension name pointer(s)\n");
	a_prog_bug (function_name);
    }
    /*  Generate output descriptor  */
    if ( ( *out_desc = generate_desc (inp_desc) ) == NULL )
    {
	a_func_abort (function_name, "no output descriptor");
        return (FALSE);
    }
    if (reorder_list != NULL)
    {
	/*  Must re-order the input array  */
	fprintf (stderr, "Re-ordering input array\n");
	/*  Re-order the data  */
	if (ds_foreach_occurrence (inp_desc, inp_data, dim_names[0], TRUE,
				   reorder_one_array) == FALSE)
	{
	    a_func_abort (function_name, "error re-ordering input array");
	    return (FALSE);
	}
	/*  Fiddle the array descriptor  */
	switch ( ds_f_name_in_packet (inp_desc, dim_names[0],
				      (char **) &array_desc_inp,
				      (unsigned int *) NULL) )
	{
	  case IDENT_NOT_FOUND:
	    fprintf (stderr, "Dimension: \"%s\" not found\n", dim_names[0]);
	    return (FALSE);
	    /*break;*/
	  case IDENT_ELEMENT:
	    fprintf (stderr,
		     "Item: \"%s\" must be a dimension name, not an element\n",
		     dim_names[0]);
	    return (FALSE);
	    /*break;*/
	  case IDENT_MULTIPLE:
	    fprintf (stderr,
		     "Item \"%s\" found more than once\n", dim_names[0]);
	    return (FALSE);
	    /*break;*/
	  case IDENT_DIMENSION:
	    break;
	  case IDENT_GEN_STRUCT:
	  default:
	    fprintf (stderr,
		     "Bad return value from function: ds_f_name_in_packet\n");
	    a_prog_bug (function_name);
	    break;
	}
	if ( !ds_reorder_array (array_desc_inp, reorder_list, (char *) NULL,
				TRUE) )
	{
	    a_func_abort (function_name,"could not re-order array descriptor");
	    return (FALSE);
	}
    }
    if ( ( *out_data = ds_alloc_data (*out_desc, TRUE, TRUE) ) == NULL )
    {
	a_func_abort (function_name, "could not allocate output data");
        return (FALSE);
    }
    ds_copy_data (inp_desc, inp_data, *out_desc, *out_data);
    if (ds_traverse_and_process (inp_desc, inp_data, *out_desc, *out_data,
				 FALSE, process_occurrence) == FALSE)
    {
	a_func_abort (function_name, "error traversing data structures");
	return (FALSE);
    }
    ds_remove_unique_named_element (*out_desc, out_data, "DATAMIN", TRUE);
    ds_remove_unique_named_element (*out_desc, out_data, "DATAMAX", TRUE);
    ds_remove_unique_named_element (*out_desc, out_data, "BSCALE", TRUE);
    ds_remove_unique_named_element (*out_desc, out_data, "BZERO", TRUE);
    return (TRUE);
}   /*  End Function process_array  */

packet_desc *generate_desc (inp_desc)
/*  This routine will generate a general data structure descriptor for
    the output of a 3d rendering programme.
    The input descriptor, which is partly copied, must be pointed to by
    inp_desc  .
    The input descriptor must contain (at any depth) dimensions with names
    pointed to by the external variable  dim_names  .
    These dimensions will not be copied.
    The routine returns a pointer to the output descriptor on success, else
    it returns NULL.
*/
packet_desc *inp_desc;
{
    flag reorder_needed = FALSE;
    unsigned int dim_count;
    unsigned int dim_num;
    unsigned int dim_index;
    unsigned int elem_count;
    packet_desc *return_value;
    packet_desc *array_pack_desc;
    array_desc *array_desc_inp;
    array_desc *array_desc_out;
    dim_desc *dimension;
    extern int h_length;
    extern int v_length;
    extern double h_min;
    extern double h_max;
    extern double v_min;
    extern double v_max;
    extern char *dim_names[3];
    static char function_name[] = "generate_desc";

    /*  Do tests on input descriptor    */
    switch ( ds_f_name_in_packet (inp_desc, dim_names[2],
				  (char **) &array_desc_inp, &dim_index) )
    {
      case IDENT_NOT_FOUND:
	fprintf (stderr, "Dimension: \"%s\" not found\n", dim_names[2]);
	return (NULL);
	/*break;*/
      case IDENT_ELEMENT:
	fprintf (stderr,
		 "Item: \"%s\" must be a dimension name, not an element\n",
		 dim_names[2]);
	return (NULL);
	/*break;*/
      case IDENT_MULTIPLE:
	fprintf (stderr, "Item \"%s\" found more than once\n", dim_names[2]);
	return (NULL);
	/*break;*/
      case IDENT_DIMENSION:
	break;
      case IDENT_GEN_STRUCT:
      default:
	fprintf (stderr,
		 "Bad return value from function: ds_f_name_in_packet\n");
	a_prog_bug (function_name);
	break;
    }
    if (array_desc_inp->num_levels > 0)
    {
	fprintf (stderr, "Tiling not supported yet\n");
	return (NULL);
    }
    if (array_desc_inp->num_dimensions < 3)
    {
	fprintf (stderr,
		 "Array with dimension: \"%s\" must have at least two other dimensions\n",
		 dim_names[2]);
	return (NULL);
    }
    /*  Deallocate and initialiase dimension re-ordering list  */
    if (reorder_list != NULL)
    {
	m_free (reorder_list);
	reorder_list = NULL;
    }
    if ( ( reorder_list = m_alloc (sizeof *reorder_list *
				   array_desc_inp->num_dimensions) ) == NULL )
    {
	m_error_notify (function_name, "dimension reordering list");
	return (NULL);
    }
    /*  Initialise re-order list  */
    for (dim_count = 0; dim_count < array_desc_inp->num_dimensions;
	 ++dim_count)
    {
	reorder_list[dim_count] = dim_count;
    }
    if (dim_index != array_desc_inp->num_dimensions - 3)
    {
	/*  Array will need to be re-ordered: push dimension to the bottom  */
	reorder_needed = TRUE;
	/*  Move dimension down  */
	for (dim_count = 0, dim_num = 0;
	     dim_count < array_desc_inp->num_dimensions;
	     ++dim_count)
	{
	    if (reorder_list[dim_count] != dim_index)
	    {
		/*  Not the dimension to be lowered  */
		reorder_list[dim_num++] = reorder_list[dim_count];
	    }
	}
	reorder_list[dim_num] = dim_index;
    }
    if ( ( dim_index = ds_f_dim_in_array (array_desc_inp, dim_names[1]) ) >=
	 array_desc_inp->num_dimensions )
    {
	fprintf (stderr,
		 "Dimension: \"%s\" must be in same array as dimension: \"%s\"\n",
		 dim_names[1], dim_names[2]);
	return (NULL);
    }
    if (dim_index != array_desc_inp->num_dimensions - 2)
    {
	/*  Array will need to be re-ordered: push dimension to the bottom  */
	reorder_needed = TRUE;
	/*  Move dimension down  */
	for (dim_count = 0, dim_num = 0;
	     dim_count < array_desc_inp->num_dimensions;
	     ++dim_count)
	{
	    if (reorder_list[dim_count] != dim_index)
	    {
		/*  Not the dimension to be lowered  */
		reorder_list[dim_num++] = reorder_list[dim_count];
	    }
	}
	reorder_list[dim_num] = dim_index;
    }
    if ( ( dim_index = ds_f_dim_in_array (array_desc_inp, dim_names[0]) ) >=
	 array_desc_inp->num_dimensions )
    {
	fprintf (stderr,
		 "Dimension: \"%s\" must be in same array as dimension: \"%s\"\n",
		 dim_names[0], dim_names[2]);
	return (NULL);
    }
    if (dim_index != array_desc_inp->num_dimensions - 1)
    {
	/*  Array will need to be re-ordered: push dimension to the bottom  */
	reorder_needed = TRUE;
	/*  Move dimension down  */
	for (dim_count = 0, dim_num = 0;
	     dim_count < array_desc_inp->num_dimensions;
	     ++dim_count)
	{
	    if (reorder_list[dim_count] != dim_index)
	    {
		/*  Not the dimension to be lowered  */
		reorder_list[dim_num++] = reorder_list[dim_count];
	    }
	}
	reorder_list[dim_num] = dim_index;
    }
    if (reorder_needed == FALSE)
    {
	/*  No re-ordering needed: deallocate re-ordering list  */
	m_free (reorder_list);
	reorder_list = NULL;
    }
    if ( !ds_packet_all_data (array_desc_inp->packet) )
    {
	fprintf (stderr, "Array must have only atomic elements\n");
	return (NULL);
    }
    /*  Copy data structure descriptor  */
    if ( ( return_value = ds_copy_desc_until (inp_desc, (char *) NULL) )
	 == NULL )
    {
	a_func_abort (function_name, "error copying descriptors");
	return (NULL);
    }
    /*  Do tests on output descriptor    */
    switch ( ds_f_name_in_packet (return_value, dim_names[0],
				  (char **) &array_desc_out, &dim_num) )
    {
      case IDENT_NOT_FOUND:
	fprintf (stderr, "Output Dimension: \"%s\" not found\n", dim_names[0]);
	a_prog_bug (function_name);
	break;
      case IDENT_ELEMENT:
	fprintf (stderr,
		 "Output item: \"%s\" must be a dimension name, not an element\n",
		 dim_names[0]);
	a_prog_bug (function_name);
	break;
      case IDENT_MULTIPLE:
	fprintf (stderr, "Output item \"%s\" found more than once\n",
		 dim_names[0]);
	a_prog_bug (function_name);
	break;
      case IDENT_DIMENSION:
	break;
      case IDENT_GEN_STRUCT:
      default:
	fprintf (stderr,
		 "Bad return value from function: ds_f_name_in_packet\n");
	a_prog_bug (function_name);
	break;
    }
    if (array_desc_inp->num_dimensions != array_desc_out->num_dimensions)
    {
	fprintf (stderr,
		 "Number of dimensions in array has changed from: %u to: %u\n",
		 array_desc_inp->num_dimensions,
		 array_desc_out->num_dimensions);
	a_prog_bug (function_name);
    }
    if (dim_num != dim_index)
    {
	fprintf (stderr, "Dimension index has changed from: %u to: %u\n",
		 dim_index, dim_num);
	a_prog_bug (function_name);
    }
    /*  Get array packet descriptor  */
    array_pack_desc = array_desc_out->packet;
    /*  Remove named dimensions  */
    if ( !ds_remove_dim_desc (array_desc_out, dim_names[0]) )
    {
	fprintf (stderr, "Error removing dimension descriptor: \"%s\"\n",
		 dim_names[0]);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    if ( !ds_remove_dim_desc (array_desc_out, dim_names[1]) )
    {
	fprintf (stderr, "Error removing dimension descriptor: \"%s\"\n",
		 dim_names[1]);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    if ( !ds_remove_dim_desc (array_desc_out, dim_names[2]) )
    {
	fprintf (stderr, "Error removing dimension descriptor: \"%s\"\n",
		 dim_names[2]);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    /*  Create (append) image plane dimensions  */
    if ( ( dimension = ds_alloc_dim_desc (VERTICAL_DIMENSION_NAME,
					  h_length, h_min, h_max, TRUE) )
	 == NULL )
    {
	fprintf (stderr, "Error creating dimension descriptor: \"%s\"\n",
		 VERTICAL_DIMENSION_NAME);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    if ( !ds_append_dim_desc (array_desc_out, dimension) )
    {
	fprintf (stderr, "Error appending dimension descriptor: \"%s\"\n",
		 VERTICAL_DIMENSION_NAME);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    if ( ( dimension = ds_alloc_dim_desc (HORIZONTAL_DIMENSION_NAME,
					  v_length, v_min, v_max, TRUE) )
	 == NULL )
    {
	fprintf (stderr, "Error creating dimension descriptor: \"%s\"\n",
		 HORIZONTAL_DIMENSION_NAME);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    if ( !ds_append_dim_desc (array_desc_out, dimension) )
    {
	fprintf (stderr, "Error appending dimension descriptor: \"%s\"\n",
		 HORIZONTAL_DIMENSION_NAME);
	ds_dealloc_packet (return_value, (char *) NULL);
	return (NULL);
    }
    /*  Convert all elements to single precision floating point  */
    for (elem_count = 0; elem_count < array_pack_desc->num_elements;
	 ++elem_count)
    {
	if ( ds_element_is_complex (array_pack_desc->element_types[elem_count]) )
	    array_pack_desc->element_types[elem_count] = K_COMPLEX;
	else
	    array_pack_desc->element_types[elem_count] = K_FLOAT;
    }
    return (return_value);
}   /*  End Function generate_desc  */

flag process_occurrence (inp_desc, inp_type, inp_data,
			 out_desc, out_type, out_data)
/*  This routine will process an occurrence of a difference in the input and
    output data structures.
    The input data structure descriptor is pointed to by  inp_desc  ,the input
    descriptor type is in  inp_type  and the input data is pointed to by
    inp_data  .
    The output data structure descriptor is pointed to by  out_desc  ,the
    output descriptor type in in  out_type  and the output data is pointed to
    by  out_data  .
    The routine is meant to be called from the  ds_traverse_and_process
    family of routines.
    The routine returns TRUE on success, else it returns FALSE.
*/
char *inp_desc;
unsigned int inp_type;
char *inp_data;
char *out_desc;
unsigned int out_type;
char *out_data;
{
    unsigned int num_dim;
    unsigned int dim_count;
    unsigned int inp_stride;
    unsigned int out_stride;
    unsigned int iter_above_count;
    unsigned int pack_size_inp;
    unsigned int pack_size_out;
    unsigned int iter_above;
    unsigned int dim_index;
    unsigned int elem_count;
    unsigned int elem_type;
    unsigned int lengths[3];
    double *coordinates[3];
    array_desc *array_desc_inp;
    array_desc *array_desc_out;
    packet_desc *pack_desc_inp;
    packet_desc *pack_desc_out;
    extern double lower_threshold;
    extern double upper_threshold;
    extern double focus[3];
    extern double position[3];
    extern double vertical[3];
    extern char *data_type_names[NUMTYPES];
    static char function_name[] = "process_occurrence";

    if ( (inp_desc == NULL) || (inp_data == NULL)
	 || (out_desc == NULL) || (out_data == NULL) )
    {
	fprintf (stderr, "NULL descriptor(s) passed\n");
	a_prog_bug (function_name);
    }
    if ( (inp_type != K_ARRAY) || (out_type != K_ARRAY) )
    {
	/*  Descriptors are not array descriptors  */
	fprintf (stderr,
		 "Descriptors: input: %s output: %s are not both K_ARRAY\n",
		 data_type_names[inp_type], data_type_names[out_type]);
	a_prog_bug (function_name);
    }
    array_desc_inp = (array_desc *) inp_desc;
    array_desc_out = (array_desc *) out_desc;
    num_dim = array_desc_inp->num_dimensions;
    /*  Compute index of top cube/image dimension  */
    dim_index = num_dim - 3;
    /*  Do a quick check on the descriptors  */
    if (num_dim != array_desc_out->num_dimensions + 1)
    {
	fprintf (stderr, "Array descriptors number of dimensions: %u %u\n",
		 array_desc_inp->num_dimensions,
		 array_desc_out->num_dimensions);
	a_prog_bug (function_name);
    }
    /*  Determine number of iterations above cube to be rendered  */
    iter_above = 1;
    for (dim_count = 0; dim_count < dim_index; ++dim_count)
    {
	iter_above *= array_desc_inp->dimensions[dim_count]->length;
    }
    pack_desc_inp = array_desc_inp->packet;
    pack_desc_out = array_desc_out->packet;
    /*  Determine strides  */
    pack_size_inp = ds_get_packet_size (pack_desc_inp);
    pack_size_out = ds_get_packet_size (pack_desc_out);
    inp_stride = pack_size_inp;
    for (dim_count = dim_index; dim_count < array_desc_inp->num_dimensions;
	 ++dim_count)
    {
	inp_stride *= array_desc_inp->dimensions[dim_count]->length;
    }
    out_stride = pack_size_out;
    for (dim_count = dim_index; dim_count < array_desc_out->num_dimensions;
	 ++dim_count)
    {
	out_stride *= array_desc_out->dimensions[dim_count]->length;
    }
    /*  Get/ compute cube co-ordinate arrays (time saver)  */
    for (dim_count = dim_index; dim_count < array_desc_inp->num_dimensions;
	 ++dim_count)
    {
	if ( ( coordinates[dim_count - dim_index] =
	       ds_get_coordinate_array (array_desc_inp->dimensions[dim_count])
	       ) == NULL )
	{
	    num_dim = dim_count;
	    for (dim_count = 0; dim_count < num_dim; ++dim_count)
	    {
		m_free (coordinates[dim_count]);
	    }
	    fprintf (stderr,"Error getting array of dimension co-ordinates\n");
	    return (FALSE);
	}
    }
    /*  Copy cube dimension lengths  */
    lengths[0] = array_desc_inp->dimensions[dim_index]->length;
    lengths[1] = array_desc_inp->dimensions[dim_index + 1]->length;
    lengths[2] = array_desc_inp->dimensions[dim_index + 2]->length;
    /*  Iterate through upper dimensions  */
    for (iter_above_count = 0; iter_above_count < iter_above;
	 ++iter_above_count)
    {
	/*  Iterate through elements  */
	for (elem_count = 0; elem_count < pack_desc_inp->num_elements;
	     ++elem_count)
	{
	    elem_type = pack_desc_inp->element_types[elem_count];
	    if (render_cube (inp_data + ds_get_element_offset (pack_desc_inp,
							       elem_count),
			     lengths, coordinates,
			     elem_type, pack_size_inp,
			     (float *) (out_data +
					ds_get_element_offset (pack_desc_out,
							       elem_count) ),
			     array_desc_out->dimensions + dim_index,
			     ds_element_is_complex
			     (pack_desc_out->element_types[elem_count]),
			     pack_size_out / sizeof (float),
			     focus, position, vertical,
			     lower_threshold, upper_threshold)
		!= TRUE)
	    {
		a_func_abort (function_name, "error rendering a cube");
		return (FALSE);
	    }
	}
	inp_data += inp_stride;
	out_data += out_stride;
    }
    /*  Deallocate cube co-ordinate arrays  */
    m_free (coordinates[0]);
    m_free (coordinates[1]);
    m_free (coordinates[2]);
    return (TRUE);
}   /*  End Function process_occurrence  */

flag reorder_one_array (encls_desc, type, data, index)
/*  This routine will re-order one instance of an array.
    The array descriptor must be pointed to by  encls_desc  .
    The type of the descriptor must be given by  type  and must be equal to the
    value K_ARRAY.
    The array data must be pointed to by  data  .
    The  index  parameter is ignored.
    This routine is meant to be called from the  ds_foreach_occurrence
    family of routines.
    The routine returns TRUE on success, else it returns FALSE.
*/
char *encls_desc;
unsigned int type;
char *data;
unsigned int index;
{
    static char function_name[] = "reorder_one_array";

    if ( (encls_desc == NULL) || (data == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    if (type != K_ARRAY)
    {
	fprintf (stderr, "Descriptor type must be: K_ARRAY\n");
	a_prog_bug (function_name);
    }
    if ( !ds_reorder_array ( (array_desc *) encls_desc, reorder_list, data,
			     FALSE) )
    {
	a_func_abort (function_name, "could not re-order array descriptor");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function reorder_one_array  */
