/*  kmkgalaxy.c

    Source file for  kmkgalaxy  (generate simple spiral galaxy in Karma format)

    Copyright (C) 2003-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will generate the projection of a simple 3-dimensional
    spiral galaxy into RA-DEC-vel space and will write the data to a Karma data
    file.


    Written by      Richard Gooch   13-AUG-2003

    Last updated by Richard Gooch   2-APR-2004


*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <karma.h>
#include <k_version.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_dsxfr.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_m.h>
#include <karma_s.h>


#define VERSION "0.2"

#define THICKNESS_IN_FWHM 3.0
#define OVERSAMPLE_RATIO  5


struct dim_struct
{
    unsigned long length;
    unsigned long index;
    float scale;
    float offset;
};


/*  Private functions  */
STATIC_FUNCTION (flag kmkgalaxy, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (void generate_file, (CONST char *outfile) );


/*  Private data  */
static float bulge_density = 0.1;  /*  Proportional density in bulge        */
static float csize = 2e3;          /*  Channel size in m/s                  */
static char *densitycube_f = "densitycube";
static float dfwhm = 230.0;        /*  FWHM vertical density off disc (pc)  */
static float dmax = 0.57e-6;       /*  Maximum atomic density/cubic metre   */
static float inclination = 45.0;   /*  Inclination angle in degrees         */
static float psize = 25.0;         /*  Pixel size in parsecs                */
static float rhole = 600.0;        /*  Radius of central hole in parsecs    */
static float rturnover = 2000.0;   /*  Radius of turnover point in parsecs  */
static float rconstant = 5000.0;   /*  Radius of constant linear velocity   */
static char *tvelimage_f = "tvelimage";
static float vmax = 220e3;         /*  Maximum linear velocity in m/s       */


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    im_register_lib_version (KARMA_VERSION);
    if ( ( panel = panel_create (FALSE) ) == NULL )
	m_abort (function_name, "control panel");
    panel_add_item (panel, "vmax", "max. tangential rotation velocity (m/s)",
		    K_FLOAT, &vmax,
		    PIA_END);
    panel_add_item (panel, "tvelimage",
		    "file to write tangential velocities to",
		    K_VSTRING, &tvelimage_f,
		    PIA_END);
    panel_add_item (panel, "rturnover", "radius of turnover (pc)",
		    K_FLOAT, &rturnover,
		    PIA_END);
    panel_add_item (panel, "rhole", "radius of hole (pc)", K_FLOAT, &rhole,
		    PIA_END);
    panel_add_item (panel, "rconstant",
		    "radius of constant linear velocity (pc)",
		    K_FLOAT,&rconstant,
		    PIA_END);
    panel_add_item (panel, "psize", "pixel size in parsecs",
		    K_FLOAT, &psize,
		    PIA_END);
    panel_add_item (panel, "inclination", "inclination angle in degrees",
		    K_FLOAT, &inclination,
		    PIA_END);
    panel_add_item (panel, "dmax",
		    "maximum atomic density per cubic metre",
		    K_FLOAT, &dmax,
		    PIA_END);
    panel_add_item (panel, "dfwhm", "FWHM vertical density off disc (pc)",
		    K_FLOAT, &dfwhm,
		    PIA_END);
    panel_add_item (panel, "densitycube", "file to write density cube to",
		    K_VSTRING, &densitycube_f,
		    PIA_END);
    panel_add_item (panel, "csize", "channel size in m/s", K_FLOAT, &csize,
		    PIA_END);
    panel_add_item (panel, "bulge_density", "proportional density in bulge",
		    K_FLOAT, &bulge_density,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: outfile");
    module_run (argc, argv, "kmkgalaxy", VERSION, kmkgalaxy, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */


/*  Private functions follow  */

static flag kmkgalaxy (CONST char *command, FILE *fp)
{
    char *arrayfile;

    for ( ; command; command = ex_command_skip (command) )
    {
	if ( ( arrayfile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting arrayfile name\n");
	    return (TRUE);
	}
	generate_file (arrayfile);
	m_free (arrayfile);
    }
    return (TRUE);
}   /*  End Function kmkgalaxy  */

static void generate_file (CONST char *outfile)
/*  [SUMMARY] Generate a cube.
    [PURPOSE] This routine will generate a Karma arrayfile with a
    3-dimensional array of a single atomic element.
    <outfile> The name of the Karma arrayfile that will be generated.
    [RETURNS] Nothing.
*/
{
    iarray cube, densitycube, tvelimage, iweightimage;
    struct dim_struct xdim, ydim, zdim, vdim;
    float subpixel_weight = 1.0 / (OVERSAMPLE_RATIO * OVERSAMPLE_RATIO);
    float rfactor;
    double xcentre, ycentre, ystep, zstep;
    double sini = sin (inclination * PION180);
    double cosi = cos (inclination * PION180);
    double tani = tan (inclination * PION180);
    double *zdensities, *spectrum;
    static char function_name[] = "generate_file";

    xdim.length = rconstant * 2.0 / psize * OVERSAMPLE_RATIO;
    xdim.scale = psize / (float) OVERSAMPLE_RATIO;
    xdim.offset = -rconstant;
    ydim = xdim;
    zdim.length = THICKNESS_IN_FWHM * dfwhm / psize * OVERSAMPLE_RATIO;
    zdim.scale = psize / (float) OVERSAMPLE_RATIO;
    zdim.offset = -THICKNESS_IN_FWHM * dfwhm * 0.5;
    vdim.length = vmax / csize + 1;
    vdim.scale = csize;
    vdim.offset = 0.0;
    fprintf (stderr, "Creating %lu*%lu*%lu density cube\n",
	     xdim.length / OVERSAMPLE_RATIO, ydim.length / OVERSAMPLE_RATIO,
	     zdim.length / OVERSAMPLE_RATIO);
    if ( ( densitycube = iarray_create_3D (zdim.length / OVERSAMPLE_RATIO,
					   ydim.length / OVERSAMPLE_RATIO,
					   xdim.length / OVERSAMPLE_RATIO,
					   K_FLOAT) ) == NULL )
	return;
    panel_put_history_with_stack (iarray_multi_array (densitycube), TRUE);
    iarray_set_world_coords (densitycube, 2, xdim.offset, -xdim.offset);
    iarray_set_world_coords (densitycube, 1, ydim.offset, -ydim.offset);
    iarray_set_world_coords (densitycube, 0, zdim.offset, -zdim.offset);
    iarray_set_dim_name (densitycube, 2, "x (pc)", TRUE);
    iarray_set_dim_name (densitycube, 1, "y (pc)", TRUE);
    iarray_set_dim_name (densitycube, 0, "z (pc)", TRUE);
    iarray_set_value_name (densitycube, "density (atoms/cubic metre)", TRUE);
    fprintf (stderr, "Creating %lu*%lu tangential velocity image\n",
	     xdim.length, ydim.length);
    if ( ( tvelimage = iarray_create_2D (ydim.length, xdim.length, K_FLOAT) )
	 == NULL )
    {
	iarray_dealloc (densitycube);
	return;
    }
    panel_put_history_with_stack (iarray_multi_array (tvelimage), TRUE);
    iarray_set_world_coords (tvelimage, 1, xdim.offset, -xdim.offset);
    iarray_set_world_coords (tvelimage, 0, ydim.offset, -ydim.offset);
    iarray_set_dim_name (tvelimage, 1, "x (pc)", TRUE);
    iarray_set_dim_name (tvelimage, 0, "y (pc)", TRUE);
    iarray_set_value_name (tvelimage, "M/S", TRUE);
    if ( ( iweightimage = iarray_create_2D (ydim.length, xdim.length,K_FLOAT) )
	 == NULL )
    {
	iarray_dealloc (tvelimage);
	iarray_dealloc (densitycube);
	return;
    }
    panel_put_history_with_stack (iarray_multi_array (iweightimage), TRUE);
    if ( ( zdensities = m_alloc (sizeof *zdensities * zdim.length) ) == NULL )
	m_abort (function_name, "zdensities array");
    ds_gausscurve (zdensities, NULL, zdim.length, zdim.length / 2, dmax,
		   dfwhm / zdim.scale, 0.0, FALSE);
    /*  Compute an image containing tangential (rotational) velocities  */
    fprintf (stderr, "Computing velocity image...\n");
    rfactor = rturnover / atanh (0.9);
    for (ydim.index = 0; ydim.index < ydim.length; ++ydim.index)
    {
	float y = ydim.offset + (float) ydim.index * ydim.scale;

	for (xdim.index = 0; xdim.index < xdim.length; ++xdim.index)
	{
	    float x = xdim.offset + (float) xdim.index * xdim.scale;
	    float radius = sqrt (x * x + y * y);
	    float velocity, intensity;

	    if (radius < rconstant) velocity = vmax * tanh (radius / rfactor);
	    else
	    {
		F2 (tvelimage, ydim.index, xdim.index) = TOOBIG;
		continue;
	    }
	    F2 (tvelimage, ydim.index, xdim.index) = velocity;
	    if (radius < rhole) intensity = bulge_density;
	    else intensity = 1.0;
	    F2 (iweightimage, ydim.index, xdim.index) = intensity;
	}
    }
    if (tvelimage_f && *tvelimage_f)
    {
	fprintf (stderr, "Writing velocity image...\n");
	iarray_write (tvelimage, tvelimage_f);
    }
    /*  Compute a mass density cube with three spatial dimensions  */
    fprintf (stderr, "Computing density cube...\n");
    for (zdim.index = OVERSAMPLE_RATIO / 2 + 1; zdim.index < zdim.length;
	 zdim.index += OVERSAMPLE_RATIO)
    {
	float density = zdensities[zdim.index];
	unsigned int yout;
	unsigned int zout = zdim.index / OVERSAMPLE_RATIO;

	for (ydim.index = 0, yout = 0; ydim.index < ydim.length;
	     ydim.index += OVERSAMPLE_RATIO, ++yout)
	{
	    unsigned int xout;

	    for (xdim.index = 0, xout = 0; xdim.index < xdim.length;
		 xdim.index += OVERSAMPLE_RATIO, ++xout)
	    {
		float intensity = F2 (iweightimage, ydim.index, xdim.index);

		F3 (densitycube, zout, yout, xout) = intensity * density;
	    }
	}
    }
    if (densitycube_f && *densitycube_f)
    {
	fprintf (stderr, "Writing density cube...\n");
	iarray_write (densitycube, densitycube_f);
    }
    iarray_dealloc (densitycube);
    fprintf (stderr, "Creating %lu*%lu*%lu tangential velocity cube\n",
	     xdim.length, ydim.length, vdim.length);
    if ( ( cube = iarray_create_3D (vdim.length,
				    ydim.length / OVERSAMPLE_RATIO,
				    xdim.length / OVERSAMPLE_RATIO,
				    K_FLOAT) ) == NULL )
    {
	m_free (zdensities);
	iarray_dealloc (iweightimage);
	iarray_dealloc (tvelimage);
	return;
    }
    panel_put_history_with_stack (iarray_multi_array (cube), TRUE);
    iarray_set_world_coords (cube, 2, xdim.offset, -xdim.offset);
    iarray_set_world_coords (cube, 1, ydim.offset, -ydim.offset);
    iarray_set_world_coords (cube, 0, 0.0, vmax);
    iarray_set_dim_name (cube, 2, "x (pc)", TRUE);
    iarray_set_dim_name (cube, 1, "y (pc)", TRUE);
    iarray_set_dim_name (cube, 0, "tangential velocity (m/s)", TRUE);
    iarray_set_value_name (cube, "intensity", TRUE);
#if 1
    fprintf (stderr, "Computing spinplane cube...\n");
    for (vdim.index = 0; vdim.index < vdim.length; ++vdim.index)
    {
	float v = vdim.offset + (float) vdim.index * vdim.scale;
	float vlow, vhigh;

	vlow = v - vdim.scale * 0.5;
	vhigh = v + vdim.scale * 0.5;
	fprintf (stderr, "V: %-6g  low: %-6g  high: %-6g\r",
		 v * 1e-3, vlow * 1e-3, vhigh * 1e-3);
	for (ydim.index = 0; ydim.index < ydim.length; ++ydim.index)
	{
	    unsigned int yout = ydim.index / OVERSAMPLE_RATIO;

	    for (xdim.index = 0; xdim.index < xdim.length; ++xdim.index)
	    {
		unsigned int xout = xdim.index / OVERSAMPLE_RATIO;
		float val;

		val = F2 (tvelimage, ydim.index, xdim.index);
		if ( (val < vlow) || (val > vhigh) ) val = 0.0;
		else val = F2 (iweightimage, ydim.index, xdim.index);
		F3 (cube, vdim.index, yout, xout) += val * subpixel_weight;
	    }
	}
    }
    fprintf (stderr, "Writing spinplane cube...\n");
    iarray_write (cube, outfile);
    iarray_clear (cube);
#endif
    /*  Compute a "sky plane cube" (a conventional spectral line cube), as it
	would appear viewed through an ideal telescope  */
    /*  Re-use the same cube for the velocity cube  */
    fprintf (stderr, "Computing velocity cube...\n");
    xcentre = (int) (-xdim.offset / xdim.scale + 0.5);/*Centre in oversampled*/
    ycentre = (int) (-ydim.offset / ydim.scale + 0.5);/*Centre in oversampled*/
    ystep = sini * 0.5;
    zstep = cosi * 0.5;
    vdim.scale *= 2.0;
    vdim.offset = -vmax;
    iarray_set_world_coords (cube, 0, vdim.offset, -vdim.offset);
    iarray_set_dim_name (cube, 0, "velocity (m/s)", TRUE);
    iarray_set_value_name (cube, "column density", TRUE);
    if ( ( spectrum = m_alloc (sizeof *spectrum * vdim.length) ) == NULL )
	m_abort (function_name, "spectrum array");
    for (ydim.index = 0; ydim.index < ydim.length; ++ydim.index)
    {
	flag nonzero_spectrum = FALSE;
	unsigned int yout = ydim.index / OVERSAMPLE_RATIO;
	double yin;

	yin = (ydim.index - ycentre) / cosi;   /* Distance from centre in SP */
	yin += ycentre;                        /* Position in spinplane      */
	yin -= zdim.length * 0.5 * tani;
	fprintf (stderr, "Y: %-6lu of %-6lu ip: %-10g\r",
		 ydim.index, ydim.length, yin);
	if (yin + zdim.length * tani < 0.0) continue;
	if (yin >= ydim.length) continue;
	for (xdim.index = 0; xdim.index < xdim.length; ++xdim.index)
	{
	    int old_iy = -1;
	    int old_iz = -1;
	    unsigned int xout = xdim.index / OVERSAMPLE_RATIO;
	    double y, z;

	    if (xdim.index % OVERSAMPLE_RATIO == 0)
	    {
		m_clear (spectrum, sizeof *spectrum * vdim.length);
		nonzero_spectrum = FALSE;
	    }
	    for (z = 0.0, y = yin; TRUE; z += zstep, y += ystep)
	    {
		int iy = y + 0.5;
		int iz = z + 0.5;
		int iv;
		float fval;
		double dval;

		if (iy < 0) continue;
		if ( (iy >= ydim.length) || (iz >= zdim.length) ) break;
		if ( (iy == old_iy) && (iz == old_iz) ) continue;
		old_iy = iy;
		old_iz = iz;
		fval = F2 (tvelimage, iy, xdim.index);
		if ( DOUBLE_IS_BLANK (fval) ) continue;
		/*  Compute component velocity along spinplane Y axis. Use sqrt
		    rather than cos (atan2) because it is faster  */
		dval = xdim.index - xcentre;
		dval = dval * dval + (y - ycentre) * (y - ycentre);
		if (dval < 1e-3) dval = 0.0;
		else dval = fval * (xdim.index - xcentre) / sqrt (dval);
		dval *= sini;  /*  Compute component along line of sight  */
		iv = (dval - vdim.offset) / vdim.scale + 0.5;
		fval = F2 (iweightimage, iy, xdim.index);
		nonzero_spectrum = TRUE;
		spectrum[iv] += fval * zdensities[iz];
	    }
	    if ( nonzero_spectrum
		 && (xdim.index % OVERSAMPLE_RATIO + 1 == OVERSAMPLE_RATIO) )
		for (vdim.index = 0; vdim.index < vdim.length; ++vdim.index)
		    F3 (cube, vdim.index, yout, xout) += spectrum[vdim.index];
	}
    }
    m_free (zdensities);
    m_free (spectrum);
    iarray_write (cube, "skycube");
    iarray_dealloc (cube);
    iarray_dealloc (iweightimage);
    iarray_dealloc (tvelimage);
}   /*  End Function generate_file  */
