#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_module.h>
#include <karma_iarray.h>
#include <karma_panel.h>
#include <karma_wcs.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_im.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <k_version.h>


#define VERSION "0.4"

/* model paramaters provided by user */
static double inclination = 0.0;   /*       inclination of the galaxy        */
static double vsys = 0.0;          /*             Systemic Velocity          */
static double vfirst = 0.0;        /* the first velocity of the desired range*/
static double vlast = 0.0;         /* the last velocity of the desired range */
static double channelwidth = 0.0;  /*        velocity width per channel      */
static flag forward_transform = FALSE;

STATIC_FUNCTION (flag kspinplane, (CONST char *command, FILE *fp) );
STATIC_FUNCTION (flag do_kspinplane,
		 (CONST char *infile, CONST char *outfile) );

double sec (double x) 
{
    return 1.0 / cos (x);
}

/******************************************************************************
There are two ways to transform iarray to deproject the spin plane. They are
forward transformation using projectToSpinPlane and backward transformation
using projectFromSpinPlane.

projectToSpinPlane iterates over the input cube and apply transformation
formula.

projectFromSpinPlane iterates over the output cube and apply the backward
transformation formula.

Only one of them can be used. 

Iterating through the output cube grid assures that each grid is stepped
through. Thus, the output cube will be contiguous. 

So, in this program we are not using the forward transformation. 
We are using projectFromSpinPlane instead.
******************************************************************************/

static void projectToSpinPlane (void* info, unsigned long length, double *xs,
				double *ys, double *zs)
/*  [SUMMARY] Perform forward transformation: sky to spin plane.
*/
{
    unsigned int i;
    double sinInc = sin (inclination * PION180);
    double secInc = sec (inclination * PION180);

#if 0
    for (i = 0; i < length; i++)
    {
	double X = xs[i];
	double Y = ys[i] * secInc;

	ys[i] = Y;
	if (X == 0.0) zs[i] = TOOBIG;
	else zs[i] = (zs[i] - vsys) / (cos (atan2 (Y, X) ) * sinInc);
    }
#else
    for (i = 0; i < length; i++)
    {
	double X = xs[i];
	double Y = ys[i] * secInc;
	double cos_phi;

	ys[i] = Y;
	cos_phi = X * X + Y * Y;
	if (cos_phi == 0.0) zs[i] = TOOBIG;
	else
	{
	    cos_phi = X / sqrt (cos_phi);
	    zs[i] = (zs[i] - vsys) / (cos_phi * sinInc);
	}
    }
#endif
}   /*  End Function projectToSpinPlane  */

static void projectFromSpinPlane (void *info, unsigned long length, double *xs,
				  double *ys, double *zs)
/*  [SUMMARY] Perform backward transformation: spin to sky plane.
*/
{
    double secInc = sec (inclination * PION180);
    double sinInc = sin (inclination * PION180);
    unsigned int i;

#if 0
    for (i = 0; i < length; i++)
    {
	double X = xs[i];
	double Y = ys[i];
	double Z = zs[i];

	ys[i] = Y / secInc;
	if (X == 0.0) zs[i] = TOOBIG;  /*  Avoid dividing by 0  */
	else zs[i] = Z * cos ( atan2 (Y, X) ) * sinInc + vsys;
   
    }
#else
    for (i = 0; i < length; i++)
    {
	double X = xs[i];
	double Y = ys[i];
	double Z = zs[i];
	double cos_phi;

	ys[i] = Y / secInc;
	cos_phi = X * X + Y * Y;
	if (cos_phi == 0.0) zs[i] = TOOBIG;
	else
	{
	    cos_phi = X / sqrt (cos_phi);
	    zs[i] = Z * cos_phi * sinInc + vsys;
	}
    }
#endif
}   /*  End Function projectFromSpinPlane  */

static void findExtremum (iarray in_arr, KwcsAstro in_ap, double *minX,
			  double *maxX, double *minY, double *maxY,
			  double *minZ, double *maxZ)
{
    unsigned long arrayOffset = 0;
    int i = 0;
    /* dimension lengths of the input cube  */
    int stopX = iarray_dim_length (in_arr, 2);
    int stopY = iarray_dim_length (in_arr, 1);
    int stopZ = iarray_dim_length (in_arr, 0);  
    unsigned long arrayLength = 4 * (stopX + stopY);
    double *xs = m_alloc (arrayLength * sizeof *xs);
    double *ys = m_alloc (arrayLength * sizeof *ys);
    double *zs = m_alloc (arrayLength * sizeof *zs);

    if (xs == NULL || ys == NULL || zs == NULL)
    {
	fprintf (stderr, "unable to allocate memory! abandon ship!");
	exit (1);
    }
    /* populate the array with values from the cube */
    for (i = 0; i < stopX; i++)
    {
	xs[arrayOffset] = i;
	ys[arrayOffset] = 0;
	zs[arrayOffset] = 0;
	arrayOffset++;
    }
    for (i = 0; i < stopX; i++)
    {
	xs[arrayOffset] = i;
	ys[arrayOffset] = 0;
	zs[arrayOffset] = stopZ - 1;
	arrayOffset++;
    }
    for (i = 0; i < stopX; i++)
    {
	xs[arrayOffset] = i;
	ys[arrayOffset] = stopY - 1;
	zs[arrayOffset] = 0;
	arrayOffset++;
    }
    for (i = 0; i < stopX; i++)
    {
	xs[arrayOffset] = i;
	ys[arrayOffset] = stopY - 1;
	zs[arrayOffset] = stopZ - 1;
	arrayOffset++;
    }
    for (i = 0; i < stopY; i++)
    {
	xs[arrayOffset] = 0;
	ys[arrayOffset] = i;
	zs[arrayOffset] = 0;
	arrayOffset++;
    }
    for (i = 0; i < stopY; i++)
    {
	xs[arrayOffset] = 0;
	ys[arrayOffset] = i;
	zs[arrayOffset] = stopZ - 1;
	arrayOffset++;
    }
    for (i = 0; i < stopY; i++)
    {
	xs[arrayOffset] = stopX - 1;
	ys[arrayOffset] = i;
	zs[arrayOffset] = 0;
	arrayOffset++;
    }
    for (i = 0; i < stopY; i++)
    {
	xs[arrayOffset] = stopX - 1;
	ys[arrayOffset] = i;
	zs[arrayOffset] = stopZ - 1;
	arrayOffset++;
    }
    /* DEBUG TEST */
    if (arrayOffset != arrayLength)
    {
	fprintf (stderr, "arrayOffset != arrayLength\n");
	exit (1);
    }
    /*  Convert pixel co-ordinates to linear co-ordinates  */
    ds_convert_coordinates (iarray_get_dim_desc (in_arr, 2), arrayLength, xs,
			    FALSE, FALSE, FALSE);
    ds_convert_coordinates (iarray_get_dim_desc (in_arr, 1), arrayLength, ys,
			    FALSE, FALSE, FALSE);
    ds_convert_coordinates (iarray_get_dim_desc (in_arr, 0), arrayLength, zs,
			    FALSE, FALSE, FALSE);
    /*  Convert linear co-ordinates to world co-ordinates  */
    if (in_ap)
	wcs_astro_transform3 (in_ap, FALSE, arrayLength,
			      iarray_dim_name (in_arr, 2), xs, FALSE,
			      iarray_dim_name (in_arr, 1), ys, FALSE,
			      iarray_dim_name (in_arr, 0), zs, FALSE,
			      0, NULL, NULL);
    /* transform the array */
    projectToSpinPlane (NULL, arrayLength, xs, ys, zs);
    /* look for extremum */
    for(i = 0; i < arrayLength; i++)
    {
	if (xs[i] < *minX) *minX = xs[i];
	if (xs[i] > *maxX) *maxX = xs[i];
	if (ys[i] < *minY) *minY = ys[i];
	if (ys[i] > *maxY) *maxY = ys[i];
	if (zs[i] < *minZ) *minZ = zs[i];
	if (zs[i] > *maxZ) *maxZ = zs[i];
    }
    m_free (xs);
    m_free (ys);
    m_free (zs);
}   /*  End Function findExtremum  */

static void findExtremumDEBUG (iarray in_arr, KwcsAstro in_ap)
{
    double minX, maxX, minY, maxY, minZ, maxZ;
    int i, j, k;
    int stopX = iarray_dim_length (in_arr, 2);
    int stopY = iarray_dim_length (in_arr, 1);
    int stopZ = iarray_dim_length (in_arr, 0);  
    double x;
    double y;
    double z;

    minX = minY = minZ = TOOBIG;
    maxX = maxY = maxZ = -TOOBIG;
    for (i = 0; i < stopX; i++)
    {
	for (j = 0; j < stopY; j++)
	{
	    for (k = 0; k < stopZ; k++)
	    {
		x = i;
		y = j;
		z = k;
		/*  Convert pixel co-ordinates to linear co-ordinates  */
		ds_convert_coordinates (iarray_get_dim_desc (in_arr, 2), 1, &x,
					FALSE, FALSE, FALSE);
		ds_convert_coordinates (iarray_get_dim_desc (in_arr, 1), 1, &y,
					FALSE, FALSE, FALSE);
		ds_convert_coordinates (iarray_get_dim_desc (in_arr, 0), 1, &z,
					FALSE, FALSE, FALSE);
		/*  Convert linear co-ordinates to world co-ordinates  */
		if (in_ap)
		    wcs_astro_transform3 (in_ap, FALSE, 1,
					  iarray_dim_name (in_arr,2), &x,FALSE,
					  iarray_dim_name (in_arr,1), &y,FALSE,
					  iarray_dim_name (in_arr,0), &z,FALSE,
					  0, NULL, NULL);
		projectToSpinPlane (NULL,1,&x,&y,&z);
		if (x < minX) minX = x;
		if (x > maxX) maxX = x;
		if (y < minY) minY = y;
		if (y > maxY) maxY = y;
		if (z < minZ) minZ = z;
		if (z > maxZ) maxZ = z;
	    }
	}
    }
    fprintf (stderr, "\nDEBUG EXTREMUM INFO\n");
    fprintf (stderr, "minX:%g\n", minX);
    fprintf (stderr, "maxX:%g\n", maxX);
    fprintf (stderr, "minY:%g\n", minY);
    fprintf (stderr, "maxY:%g\n", maxY);
    fprintf (stderr, "minZ:%g\n", minZ);
    fprintf (stderr, "maxZ:%g\n", maxZ);
}   /*  End Function findExtremumDEBUG  */


int main (int argc, char **argv)
{
    KControlPanel panel;
    static char function_name[] = "main";

    if ( ( panel = panel_create (FALSE) ) == NULL )
	m_abort (function_name, "control panel");
    panel_add_item (panel, "i", "inclination angle (degrees)", K_DOUBLE,
		    &inclination,
		    PIA_END);
    panel_add_item (panel, "vsys", "systemic velocity (m/s)", K_DOUBLE, &vsys,
		    PIA_END);
    panel_add_item (panel, "vfirst", "vsysout(=0) - desired range (m/s)",
		    K_DOUBLE, &vfirst,
		    PIA_END);
    panel_add_item (panel, "vlast", "vsysout(=0) + desired range (m/s)",
		    K_DOUBLE, &vlast,
		    PIA_END);
    panel_add_item (panel, "forward_transform",
		    "use forward transform (prone to aliasing artefacts)",
		    PIT_FLAG, &forward_transform,
		    PIA_END);
    panel_add_item (panel, "channelwidth",
		    "Velocity Width per channel for the output cube (m/s)",
		    K_DOUBLE, &channelwidth,
		    PIA_END);
    panel_push_onto_stack (panel);
    panel_set_extra_comment_string (panel, "\nGeneral usage: infile outfile");
    im_register_lib_version (KARMA_VERSION);
    module_run (argc, argv, "kspinplane", VERSION, kspinplane, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main   */

static flag kspinplane (CONST char *command, FILE *fp)
{
    for ( ; command != NULL; command = ex_word_skip (command) )
    {
	flag ok;
	char *infile, *outfile;

	if ( ( infile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting input filename\n");
	    return (TRUE);
	}
	if ( ( outfile = ex_str (command, &command) ) == NULL )
	{
	    fprintf (fp, "Error extracting output filename\n");
	    return (TRUE);
	}
	ok = do_kspinplane (infile, outfile);
	m_free (infile);
	m_free (outfile);
	if (!ok) continue;
    }
    return (TRUE);
}   /*  End Function kspinplane   */

static flag do_kspinplane (CONST char *infile, CONST char *outfile)
{
    KwcsAstro in_ap = NULL;
    KwcsAstro out_ap;
    iarray in_arr = NULL;
    iarray out_arr;
    flag ok;
    unsigned num_dim;
    double minval = -TOOBIG;
    double maxval = TOOBIG;
    CONST dim_desc *in_dim_x, *in_dim_y, *in_dim_z;
    unsigned long dim_lengths[3];
    /*CONST char *dim_names[3];*/
    char txt[STRING_LENGTH];
    double minX,maxX,minY,maxY,minZ,maxZ;
    unsigned int ftype;
    double num_channels;
    extern CONST char module_name[];
    static char *dim_names[3] = {"tangential velocity (m/s)",
				 "y (pc)", "x (pc)"};

    /*  Read input array  */
    if ( !foreign_read_and_setup (infile, K_CH_MAP_IF_AVAILABLE, FALSE,
				  &ftype, TRUE, 0, K_FLOAT, TRUE,
				  &in_arr, NULL, NULL,
				  TRUE, &in_ap) )
	return (FALSE);
    if (in_ap == NULL)
	fprintf (stderr, "File: \"%s\" has no astronomical projection\n",
		 infile);
    num_dim = iarray_num_dim (in_arr);
    if (num_dim != 3)
    {
	fprintf (stderr, "Unsupported number of dimensions: %u\n", num_dim);
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    in_dim_x = iarray_get_dim_desc(in_arr, 2);
    in_dim_y = iarray_get_dim_desc(in_arr, 1);
    in_dim_z = iarray_get_dim_desc(in_arr, 0);
    dim_lengths[2] = iarray_dim_length(in_arr, 2);
    dim_lengths[1] = iarray_dim_length(in_arr, 1) * sec (inclination* PION180);
    /* These are to be used to find min and max for header info. */
    minX = TOOBIG;
    maxX = -TOOBIG;
    minY = TOOBIG;
    maxY = -TOOBIG;
    minZ = TOOBIG;
    maxZ = -TOOBIG;
    /* findExtremumDEBUG (in_arr, in_ap); */
    fprintf (stderr, "\nabout to search for extremum\n");
    findExtremum (in_arr, in_ap,  &minX, &maxX, &minY, &maxY, &minZ, &maxZ);
    fprintf (stderr, "minX:%g\n", minX);
    fprintf (stderr, "maxX:%g\n", maxX);
    fprintf (stderr, "minY:%g\n", minY);
    fprintf (stderr, "maxY:%g\n", maxY);
    fprintf (stderr, "minZ:%g\n", minZ);
    fprintf (stderr, "maxZ:%g\n", maxZ);
    if (minZ < vfirst) minZ = vfirst;
    if (maxZ > vlast) maxZ = vlast;
    /* Calculating the deprojected dimension of output using user provided
       inputs for velocity range -- in the future spatial range to be used.  */
    num_channels = (maxZ - minZ) / channelwidth;
    dim_lengths[0] = ceil (num_channels);
    /* find the extent of the output cube */
    /* create output cube */
    out_arr = iarray_create2 (K_FLOAT, 3, NULL, dim_lengths, NULL, NULL,FALSE);
    if (out_arr == NULL)
    {
	fprintf (stderr, "Failed to allocate output array\n");
	iarray_dealloc (in_arr);
	return (FALSE);
    }
    /* set crval1 */
    if ( !iarray_put_double (out_arr, "CRVAL1", (minX+maxX)*0.5) )
    {
	fprintf (stderr, "Failed to set CRVAL1\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set crval2 */
    if ( !iarray_put_double (out_arr, "CRVAL2", (minY+maxY)*0.5) )
    {
	fprintf (stderr, "Failed to set CRVAL2\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set crval3 */
    if ( !iarray_put_double (out_arr, "CRVAL3", (minZ+maxZ)/2) )
    {
	fprintf (stderr, "Failed to set CRVAL3\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set cdelt1 */
    if ( !iarray_put_double (out_arr, "CDELT1", (maxX - minX)/dim_lengths[2]) )
    {
	fprintf (stderr, "Failed to set CDELT1\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set cdelt2 */
    if ( !iarray_put_double (out_arr, "CDELT2", (maxY - minY)/dim_lengths[1]) )
    {
	fprintf (stderr, "Failed to set CDELT2\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    fprintf (stderr, "cdelt3 is = %f\n", channelwidth);
    /* set cdelt3 */
    if ( !iarray_put_double (out_arr, "CDELT3", channelwidth) )
    {
	fprintf (stderr, "Failed to set CDELT3\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    iarray_set_dim_name (out_arr, 2, dim_names[2], TRUE);
    iarray_set_dim_name (out_arr, 1, dim_names[1], TRUE);
    iarray_set_dim_name (out_arr, 0, dim_names[0], TRUE);
    /* set ctype1 */
    if ( !iarray_put_named_string (out_arr, "CTYPE1", dim_names[2]) )
    {
	fprintf (stderr, "Failed to set CTYPE1\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set ctype2 */
    if ( !iarray_put_named_string (out_arr, "CTYPE2", dim_names[1]) )
    {
	fprintf (stderr, "Failed to set CTYPE2\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set ctype3 */
    if ( !iarray_put_named_string (out_arr, "CTYPE3", dim_names[0]) )
    {
	fprintf (stderr, "Failed to set CTYPE3\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set crpix1 */
    if ( !iarray_put_double (out_arr, "CRPIX1",
			     -(minX/((maxX - minX)/dim_lengths[2])) + 1) )
    {
	fprintf (stderr, "Failed to set CRPIX1\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set crpix2 */
    if ( !iarray_put_double (out_arr, "CRPIX2",
			     -(minY/((maxY - minY)/dim_lengths[1])) + 1) )
    {
	fprintf (stderr, "Failed to set CRPIX2\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    /* set crpix3 */
    if ( !iarray_put_double (out_arr, "CRPIX3", dim_lengths[0]/2) )
    {
	fprintf (stderr, "Failed to set CRPIX3\n");
	iarray_dealloc (in_arr);
	iarray_dealloc (out_arr);
	return (FALSE);
    }
    out_ap = iarray_get_associated_wcs_astro (out_arr);
    /* transform the cube using our deprojecting formula. */
    if (forward_transform)
    {
	iarray_fill_float (out_arr, TOOBIG);
	iarray_transform_3D (out_arr, out_ap, in_arr, in_ap,
			     projectToSpinPlane, NULL);
    }
    else iarray_untransform_3D (out_arr, out_ap, in_arr, in_ap,
				projectFromSpinPlane, NULL);
    iarray_min_max (out_arr, CONV_CtoR_REAL, &minval, &maxval);
    fprintf (stderr, "Output min: %g  max: %g\n", minval, maxval);
    ok = iarray_put_double (out_arr, "DATAMIN", minval);
    if (ok) ok = iarray_put_double (out_arr, "DATAMAX", maxval);
    /* write out the output cube */
    if (ok) 
	ok = ds_history_copy ( iarray_multi_array (out_arr),
			       iarray_multi_array (in_arr) );
    if (ok) 
	ok = panel_put_history_with_stack (iarray_multi_array (out_arr), TRUE);
    if (ok)
    {
	sprintf (txt, "%s: infile: %s outfile: %s",
		 module_name, infile, outfile);
	ok = ds_history_append_string (iarray_multi_array (out_arr), txt,TRUE);
    }
    if (ok) ok = iarray_write (out_arr, outfile);
    iarray_dealloc (in_arr);
    iarray_dealloc (out_arr);
    return (ok);
}   /*  End Function do_kspinplane  */
