/*  This module controls the persp_3d rendering module.


    Written by      Richard Gooch   21-DEC-1992

    Updated by      Richard Gooch   25-DEC-1992

    Updated by      Richard Gooch   4-JAN-1993: Added event blocking (wait for
  rendering module to finish).

    Updated by      Richard Gooch   25-JAN-1993: Took account of change to
  array_desc  structure.

    Updated by      Richard Gooch   27-MAR-1993: Took account of change to
  iarray_read_nD  function.

    Updated by      Richard Gooch   4-APR-1993: Took account of change to
  conn_register_client_protocol  .

    Updated by      Richard Gooch   16-MAY-1993: Took account of replacement of
  dsrw_write_multi_desc  and  dsrw_write_multi_data  with  dsrw_write_multi  .

    Updated by      Richard Gooch   20-MAY-1994: Moved some #include's around.

    Last updated by Richard Gooch   27-JUN-1999: Cosmetic changes and fixes for
  pedantic compilers. No longer send leading '/' to processing module.

    $Id: xpersp_3d.c,v 2.1 2011/07/27 02:40:59 cal103 Exp $

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/param.h>
#include <dirent.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/panel.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_dsrw.h>
#include <karma_notify_chm.h>
#include <karma_ch.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_r.h>
#include <karma_a.h>
#include <karma_m.h>


#define DEFAULT_FRAME_WIDTH 650
#define DEFAULT_FRAME_HEIGHT 600
#define DEFAULT_ICON_WIDTH 64
#define DEFAULT_ICON_HEIGHT 64

#define ANGLE_MIN -180
#define ANGLE_MAX 180

#define RADIUS_RATIO_MIN 10
#define RADIUS_RATIO_MAX 200
#define RADIUS_RATIO_DEFAULT 150

/*  Local functions  */
void xpersp_3d_init (/* display */);
int quit (/* item, event */);
void scan_directory (/* item, event */);
int load_cube (/* item, event */);
void transmit_command (/* command */);
void compute ();
void rotate_vector (/* vector, roll, pitch, yaw */);
void scatter_cube (/* array  */);
flag render_module_finished (/* connection, info */);


/*  Private functions  */
static void register_server_exit ();
static void slider_roll (/* item, value, event */);
static void slider_pitch (/* item, value, event */);
static void slider_yaw (/* item, value, event */);
static void slider_rho (/* item, value, event */);
    

/*  Global variables for XView  */

/*  Main window info  */
static Frame base_frame;
static Panel_item directory_list;

/*  Other global variables  */
Display *my_display = NULL;

/*  Cube globals  */
/*  x, y, z  */
float cube_centre[3];
float cube_size[3];
flag loaded = FALSE;
char dim_names[3][STRING_LENGTH];

/*  View specification globals  */
float roll = 0.0;
float pitch = 0.0;
float yaw = 0.0;
float rho = 1.5;


int main (int argc, char **argv)
{
    Panel panel;
    Xv_Server server;
    char hostname[MAXHOSTNAMELEN + 1];
    char header_name[STRING_LENGTH];
    char curr_dir_name[STRING_LENGTH];
    extern Frame base_frame;
    extern Panel_item directory_list;
    extern Display *my_display;
    extern char module_name[STRING_LENGTH + 1];

    /*  Initialise module  */
    im_register_module_name ("xpersp_3d_control");

    if (gethostname (hostname, MAXHOSTNAMELEN + 1) != 0)
    {
	fprintf (stderr, "Error getting hostname\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    if (getcwd (curr_dir_name, STRING_LENGTH) == NULL)
    {
	fprintf (stderr, "Error getting current working directory\t%s\n",
		 ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    sprintf (header_name, "%s @ %s        cwd: %s",
		    module_name, hostname, curr_dir_name);
    /*  Start up X server connection, initialise XView  */
    server = xv_init (XV_INIT_ARGC_PTR_ARGV, &argc, argv, NULL);

    my_display = (Display *) xv_get (server, XV_DISPLAY);
    /*  Initialise rest of module  */
    xpersp_3d_init (my_display);
    /*  Create XView objects  */
    base_frame = (Frame) xv_create ( (Xv_object) NULL, FRAME,
				    FRAME_LABEL, header_name,
/*
				    XV_WIDTH, DEFAULT_FRAME_WIDTH,
				    XV_HEIGHT, DEFAULT_FRAME_HEIGHT,
*/
				    NULL );
    /*  Install a panel  */
    panel = (Panel) xv_create (base_frame, PANEL,
			       PANEL_LAYOUT, PANEL_VERTICAL,
			       NULL);
    /*  Install buttons on the panel  */
    xv_create (panel, PANEL_BUTTON,
	       PANEL_LABEL_STRING, "Scan Directory",
	       PANEL_NOTIFY_PROC, scan_directory,
	       NULL);
    xv_create (panel, PANEL_BUTTON,
	       PANEL_LABEL_STRING, "Load Cube",
	       PANEL_NOTIFY_PROC, load_cube,
	       NULL);
    xv_create (panel, PANEL_BUTTON,
	       PANEL_LABEL_STRING, "Quit",
	       PANEL_NOTIFY_PROC, quit,
	       NULL);
    xv_create (panel, PANEL_SLIDER,
	       PANEL_LABEL_STRING, "Roll",
	       PANEL_VALUE, 0,
	       PANEL_MIN_VALUE, ANGLE_MIN,
	       PANEL_MAX_VALUE, ANGLE_MAX,
	       PANEL_NOTIFY_PROC, slider_roll,
	       PANEL_NOTIFY_LEVEL, PANEL_DONE,
	       NULL);
    xv_create (panel, PANEL_SLIDER,
	       PANEL_LABEL_STRING, "Pitch",
	       PANEL_VALUE, 0,
	       PANEL_MIN_VALUE, ANGLE_MIN,
	       PANEL_MAX_VALUE, ANGLE_MAX,
	       PANEL_NOTIFY_PROC, slider_pitch,
	       PANEL_NOTIFY_LEVEL, PANEL_DONE,
	       NULL);
    xv_create (panel, PANEL_SLIDER,
	       PANEL_LABEL_STRING, "Yaw",
	       PANEL_VALUE, 0,
	       PANEL_MIN_VALUE, ANGLE_MIN,
	       PANEL_MAX_VALUE, ANGLE_MAX,
	       PANEL_NOTIFY_PROC, slider_yaw,
	       PANEL_NOTIFY_LEVEL, PANEL_DONE,
	       NULL);
    xv_create (panel, PANEL_SLIDER,
	       PANEL_LABEL_STRING, "Distance Ratio %",
	       PANEL_VALUE, RADIUS_RATIO_DEFAULT,
	       PANEL_MIN_VALUE, RADIUS_RATIO_MIN,
	       PANEL_MAX_VALUE, RADIUS_RATIO_MAX,
	       PANEL_NOTIFY_PROC, slider_rho,
	       PANEL_NOTIFY_LEVEL, PANEL_DONE,
	       NULL);
    /*  Create the scrolling list of files  */
    directory_list = (Panel_item) xv_create (panel, PANEL_LIST,
					     PANEL_CHOOSE_ONE, TRUE,
					     PANEL_LIST_DISPLAY_ROWS, 10,
					     PANEL_LIST_WIDTH, 200,
					     PANEL_NEXT_COL, 50,
					     NULL);
    scan_directory (NULL, NULL);
    window_fit (panel);
    window_fit (base_frame);

    /*  Enter the main loop (Notifier)  */
    xv_main_loop (base_frame);
    return (RV_OK);
}   /*  End Function main   */

void xpersp_3d_init (display)
/*  This routine will initialise the xpersp_3d module.
    The display the module is connected to must be pointed to by  display  .
    The routine returns nothing.
*/
Display *display;
{
    /*  Register channel manager (Notifier version)  */
    conn_register_managers (notify_chm_manage, notify_chm_unmanage,
			    register_server_exit);
    /*  Register the protocols  */
    dsxfr_register_connection_limits (-1, 0);
    conn_register_client_protocol ("command_line", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("processing_finished", 0, 1,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   render_module_finished,
				   ( void (*) () ) NULL);
}   /*  End Function xpersp_3d_init  */

static void register_server_exit ()
/*  This routine will register a message from a Karma client to exit.
    The routine returns nothing.
*/
{
    extern int quit ();

    quit (NULL, NULL);
}   /*  End Function register_server_exit  */

int quit (item, event)
/*  This routine will save the size of the base frame in external variables and
    will then destroy any XView objects and cause the notifier to stop.
    Control is then returned to the  main  function.
*/
Panel_item item;
Event *event;
{
    extern Frame base_frame;
    static char function_name[] = "quit";

    if (xv_destroy_safe (base_frame) != XV_OK)
    {
	fprintf (stderr, "Error destroying base frame\n");
	a_prog_bug (function_name);
    }
    notify_stop ();
    return (XV_OK);
}   /*  End Function quit  */

void scan_directory (item, event)
/*  This routine will show the command frame
*/
Panel_item item;
Event *event;
{
    char *ch;
    DIR *directory;
    struct dirent *entry;
    extern Panel_item directory_list;
    static int entry_count = 0;

    if (entry_count > 0)
    {
	/*  Remove old entries  */
	while (--entry_count > -1)
	{
	    xv_set (directory_list, PANEL_LIST_DELETE, entry_count, NULL);
	}
    }
    entry_count = 0;
    if ( ( directory = opendir (".") ) == NULL )
    {
	fprintf (stderr, "Error reading directory\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    while ( (entry = readdir (directory) ) != NULL )
    {
	if ( ( ch = strrchr ( (*entry).d_name, '.' ) ) != NULL )
	{
	    if (strcmp (ch, ".kf") == 0)
	    {
		/*  Karma data file  */
		xv_set (directory_list,
			PANEL_LIST_INSERT, entry_count,
			PANEL_LIST_STRING, entry_count, (*entry).d_name,
			NULL);
		++entry_count;
	    }
	}
    }
    closedir (directory);
}   /*  End Function scan_directory  */

static void slider_roll (item, value, event)
/*  This routine will register a change in the slider value for  roll  .
    The value is written to a global variable in radians.
    The routine returns nothing.
*/
Panel_item item;
int value;
Event *event;
{
    extern float roll;

    roll = PION180 * (float) value;
    compute ();
}   /*  End Function slider_roll  */

static void slider_pitch (item, value, event)
/*  This routine will register a change in the slider value for  pitch  .
    The value is written to a global variable in radians.
    The routine returns nothing.
*/
Panel_item item;
int value;
Event *event;
{
    extern float pitch;

    pitch = PION180 * (float) value;
    compute ();
}   /*  End Function slider_pitch  */

static void slider_yaw (item, value, event)
/*  This routine will register a change in the slider value for  yaw  .
    The value is written to a global variable in radians.
    The routine returns nothing.
*/
Panel_item item;
int value;
Event *event;
{
    extern float yaw;

    yaw = PION180 * (float) value;
    compute ();
}   /*  End Function slider_yaw  */

static void slider_rho (item, value, event)
/*  This routine will register a change in the slider value for  rho  .
    The value is written to a global variable in raw ratio.
    The routine returns nothing.
*/
Panel_item item;
int value;
Event *event;
{
    extern float rho;

    rho = 0.01 * (float) value;
    compute ();
}   /*  End Function slider_rho  */

int load_cube (item, event)
/*  This routine will load the cube.
*/
Panel_item item;
Event *event;
{
    int index;
    iarray array;
    char *arrayfile;
    dim_desc *dim;
    extern flag loaded;
    extern Panel_item directory_list;
    extern float cube_centre[3];
    extern float cube_size[3];
    extern char dim_names[3][STRING_LENGTH];
    /*static char function_name[] = "load_cube";*/

    index = (int) xv_get (directory_list, PANEL_LIST_FIRST_SELECTED);
    arrayfile = (char *) xv_get (directory_list, PANEL_LIST_STRING, index);
    if ( ( array = iarray_read_nD (arrayfile, FALSE, NULL, 3, NULL, NULL,
				   K_CH_MAP_LOCAL) )
	== NULL )
    {
	fprintf (stderr, "Could not read data set: \"%s\"\n", arrayfile);
	return (XV_OK);
    }
    if (conn_get_num_client_connections ("multi_array") < 2)
    {
	/*  Only one connection  */
	iarray_write (array, "connection");
    }
    else
    {
	/*  Chop up cube and transmit a bit to each server  */
	scatter_cube (array);
    }
    /*  Extract cube information  */
    for (index = 0; index < 3; ++index)
    {
	dim = (* (*array).arr_desc ).dimensions[2 - index];
	cube_size[index] = (*dim).maximum - (*dim).minimum;
	cube_centre[index] = ( (*dim).maximum - (*dim).minimum ) / 2.0;
	strcpy (dim_names[index], (*dim).name);
    }
    iarray_dealloc (array);
    loaded = TRUE;
    compute ();
    return (XV_OK);
}   /*  End Function load_cube  */

void compute ()
/*  This routine will compute.
*/
{
    float distance;
    float vision_vector[3];
    float vertical_vector[3];
    float position[3];
    char txt[STRING_LENGTH];
    extern Frame base_frame;
    extern flag loaded;
    extern float roll, pitch, yaw, rho;
    extern float cube_centre[3];
    extern float cube_size[3];
    extern char dim_names[3][STRING_LENGTH];
    /*static char function_name[] = "compute";*/

    if (!loaded) return;
    /*  Set the frame busy attribute  */
    xv_set (base_frame, FRAME_BUSY, TRUE, NULL);
    distance = cube_size[0] * cube_size[0] + cube_size[1] * cube_size[1];
    distance += cube_size[2] * cube_size[2];
    distance = sqrt (distance);
    vision_vector[0] = 0.0;
    vision_vector[1] = 0.0;
    vision_vector[2] = -1.0;
    rotate_vector (vision_vector, pitch, yaw, roll);
    vertical_vector[0] = 0.0;
    vertical_vector[1] = 1.0;
    vertical_vector[2] = 0.0;
    rotate_vector (vertical_vector, pitch, yaw, roll);
    position[0] = cube_centre[0] - distance * rho * vision_vector[0];
    position[1] = cube_centre[1] - distance * rho * vision_vector[1];
    position[2] = cube_centre[2] - distance * rho * vision_vector[2];
    sprintf (txt, "focus  %e %e %e",
	     cube_centre[0], cube_centre[1], cube_centre[2]);
    transmit_command (txt);
    sprintf (txt, "position  %e %e %e",
	     position[0], position[1], position[2]);
    transmit_command (txt);
    sprintf (txt, "vertical  %e %e %e",
	     vertical_vector[0], vertical_vector[1],vertical_vector[2]);
    transmit_command (txt);
    sprintf (txt, "connection \"%s\" \"%s\" \"%s\"",
	     dim_names[0], dim_names[1], dim_names[2]);
    transmit_command (txt);
}   /*  End Function compute  */

void rotate_vector (vector, roll, pitch, yaw)
/*  This routine will rotate a vector through roll, then pitch and then yaw.
    Roll rotates about +x axis.
    Pitch rotates about -y axis.
    Yaw rotates about +z axis.
    In this application: roll should rotate about z axis
                         pitch should rotate about x axis
			 yaw should rotate about y axis
    Hence, call with (vector, pitch, yaw, roll).
    The vector must be pointed to by  vector  .
    The roll, pitch and yaw angles (in radians) must be given by  roll  ,
    pitch  and  yaw  ,respectively.
    The routine returns nothing.
*/
float vector[3];
float roll;
float pitch;
float yaw;
{
    float x, y, z;
    float sin_roll, cos_roll;
    float sin_pitch, cos_pitch;
    float sin_yaw, cos_yaw;

    x = vector[0];
    y = vector[1];
    z = vector[2];
    sin_roll = sin (roll);
    cos_roll = cos (roll);
    sin_pitch = sin (pitch);
    cos_pitch = cos (pitch);
    sin_yaw = sin (yaw);
    cos_yaw = cos (yaw);
    vector[0] = (x * cos_pitch + (z * cos_roll + y * sin_roll) *
		 sin_pitch) * cos_yaw - (y * cos_roll - z *
					 sin_roll) * sin_yaw;
    vector[1] = (y * cos_roll - z * sin_roll) * cos_yaw + (x * cos_pitch +
							   (z * cos_roll + y *
							    sin_roll) *
							   sin_pitch) *sin_yaw;
    vector[2] = (z * cos_roll + y * sin_roll) * cos_pitch - x * sin_pitch;
}   /*  End Function rotate_vector  */

void transmit_command (command)
/*  This routine will transmit a command to all outgoing "command_line"
    protocol connections. A newline character will be appended.
    The command must be pointed to by  command  .
    The routine returns nothing.
*/
char *command;
{
    unsigned int count;
    unsigned int num_connections;
    Channel channel;
    Connection connection;
    static char function_name[] = "transmit_command";

    num_connections = conn_get_num_client_connections ("command_line");
    for (count = 0; count < num_connections; ++count)
    {
	if ( ( connection = conn_get_client_connection ("command_line",
							count) )
	    == NULL )
	{
	    fprintf (stderr, "Error getting connection: %u\n", count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_channel (connection);
	if (ch_puts (channel, command, TRUE) != TRUE)
	{
	    fprintf (stderr, "Error writing to channel\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (ch_flush (channel) != TRUE)
	{
	    fprintf (stderr, "Error flushing channel\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
}   /*  End Function transmit_command  */

void scatter_cube (cube)
/*  This routine will chop up a cube and will transmit each part to a
    "multi_array" connection.
    The cube must be given by  cube  .
    The routine returns nothing.
*/
iarray cube;
{
    unsigned int conn_count;
    unsigned int slice_length;
    unsigned int last_slice_length;
    unsigned int num_conn;
    unsigned int index;
    Channel channel;
    Connection connection;
    char **array;
    multi_array *inp_multi_desc;
    multi_array *out_multi_desc;
    packet_desc *top_pack_desc;
    array_desc *out_array_desc;
    dim_desc *inp_top_dim;
    dim_desc *out_top_dim;
    static char function_name[] = "scatter_cube";

    fprintf (stderr, "Scattering cube...\n");
    num_conn = conn_get_num_client_connections ("multi_array");
    fprintf (stderr, "num connections: %u\n", num_conn);
    inp_multi_desc = (*cube).multi_desc;
    inp_top_dim = (* (*cube).arr_desc ).dimensions[0];
    if (num_conn > (*inp_top_dim).length)
    {
	fprintf (stderr, "Too many connections\n");
	return;
    }
    slice_length = (*inp_top_dim).length / num_conn;
    last_slice_length = (*inp_top_dim).length - slice_length * (num_conn - 1);
    /*  Create dummy data structure  */
    if ( ( out_multi_desc = ds_alloc_multi (1) ) == NULL )
    {
	m_abort (function_name, "multi_array");
    }
    if ( ( top_pack_desc = ds_copy_desc_until ( (*cube).top_pack_desc, NULL ) )
	== NULL )
    {
	m_abort (function_name, "descriptors");
    }
    (*out_multi_desc).headers[0] = top_pack_desc;
    out_array_desc = (array_desc *) (*top_pack_desc).element_desc[0];
    out_top_dim = (*out_array_desc).dimensions[0];
    /*  Allocate data  */
    if ( ( (*out_multi_desc).data[0] = ds_alloc_packet (top_pack_desc) )
	== NULL )
    {
	m_abort (function_name, "top level packet");
    }
    array = (char **) (*out_multi_desc).data[0];
    /*  Alias sub-cubes and transmit  */
    for (conn_count = 0, index = 0; conn_count < num_conn; ++conn_count)
    {
	/*  Size top dimension  */
	if (conn_count == num_conn - 1)
	{
	    (*out_top_dim).length = last_slice_length;
	}
	else
	{
	    (*out_top_dim).length = slice_length;
	}
	(*out_top_dim).minimum = ds_get_coordinate (inp_top_dim, index);
	(*out_top_dim).maximum = ds_get_coordinate (inp_top_dim, index +
						    (*out_top_dim).length - 1);
	(*out_array_desc).lengths[0] = slice_length;
	*array = (*cube).data + (*cube).offsets[0][index];
	if ( ( connection = conn_get_client_connection ("multi_array",
							conn_count) )
	    == NULL )
	{
	    fprintf (stderr, "Error getting \"multi_array\" connection: %u\n",
		     conn_count);
	    a_prog_bug (function_name);
	}
	channel = conn_get_channel (connection);
	dsrw_write_multi (channel, out_multi_desc);
	ch_flush (channel);
	fprintf (stderr, "Sent to connection: %u  %lu %u\n",
		 conn_count, (*out_top_dim).length, index);
	index += (*out_top_dim).length;
    }
    *array = NULL;
    ds_dealloc_multi (out_multi_desc);
}   /*  End Function scatter_cube  */

flag render_module_finished (connection, info)
/*  This routine will read in data from the connection given by  connection
    and will write any appropriate information to the pointer pointed to by
    info  .
    The routine returns TRUE on successful reading,
    else it returns FALSE (indicating the connection should be closed).
    Note that the  close_func  will be called if this routine returns FALSE
*/
Connection connection;
void **info;
{
    Channel channel;
    char buffer;
    extern Frame base_frame;

    channel = conn_get_channel (connection);
    if (ch_read (channel, &buffer, 1) < 1)
    {
	fprintf (stderr, "Error reading from render module\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Clear the frame busy attribute  */
    xv_set (base_frame, FRAME_BUSY, FALSE, NULL);
    return (TRUE);
}   /*  End Function render_module_finished  */
