/*  active_2d_cursor.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the 2-D active cursor code.


    Written by      Richard Gooch   19-JUN-1995

    Updated by      Richard Gooch   24-JUN-1995

    Updated by      Richard Gooch   4-MAR-1996: Improved depth estimation for
  opacity based shaders.

    Updated by      Richard Gooch   24-MAR-1996: Fixed bug converting plane
  co-ordinate of 3-D point to depth down a ray.

    Updated by      Richard Gooch   19-APR-1996: Minor adjustment of hot gas
  depth estimation (added 0.5).

    Updated by      Richard Gooch   6-MAY-1996: Added support for
  "hot_gas_three_colour" shader.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   28-NOV-1996: Took account of substance
  changes.

    Updated by      Richard Gooch   21-JAN-1997: Removed threshold test for
  opacity in Hot Gas Substances shader.

    Updated by      Richard Gooch   27-NOV-1997: Minor cosmetic changes.

    Updated by      Richard Gooch   22-JUN-1999: Minor cosmetic changes.

    Updated by      Richard Gooch   11-JUL-2000: Removed unnecessary #includes.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <string.h>


*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <sys/time.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_vrender.h>
#include <karma_a.h>
#include "xraycast_3d.h"

#define STEP_X 0
#define STEP_Y 1
#define STEP_Z 2


/*  Private routines  */
STATIC_FUNCTION (float depth_to_t, (Kcoord_3d direction, float depth) );
STATIC_FUNCTION (unsigned int shader_max_voxel,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );
STATIC_FUNCTION (unsigned int shader_min_voxel,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );
STATIC_FUNCTION (unsigned int shader_front_voxel,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );
STATIC_FUNCTION (unsigned int shader_sum_voxels,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );
STATIC_FUNCTION (unsigned int shader_hot_gas_mono,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );
STATIC_FUNCTION (unsigned int shader_hot_gas_subs,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );
STATIC_FUNCTION (unsigned int shader_hot_gas_fullcolour,
		 (CONST signed char *ray, unsigned int ray_length,
		  void *pixel_ptr, float *opacities) );


/*  Public routines follow  */

CONST signed char *get_ray_from_image_plane (Kcoord_3d *coord, Kcoord_2d pos,
					     unsigned int *ray_length,
					     CONST char *shader_name,
					     unsigned int *depth,
					     float *opacities)
/*  [PURPOSE] This routine will get a ray through the volume projected from the
    image plane and will estimate the position of a feature the user sees.
    <coord> The 3-D co-ordinate (in co-ordinate indices) of the feature the
    user sees will be written here.
    <pos> The 2-dimensional position in the image plane to project from.
    <ray_length> The length of the returned ray.
    <shader_name> The name of the current shader.
    <depth> The depth along the ray corresponding to what the user sees.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] A pointer to a statically allocated array containing the ray
    data. This data is overwritten on each call.
*/
{
    Kcoord_3d direction, ray_start;
    float t, t_enter, t_leave;
    CONST signed char *ray;
    extern KVolumeRenderContext main_vrend_ctx;
    static char function_name[] = "get_ray_from_image_plane";

    if ( ( ray = vrender_collect_ray (main_vrend_ctx,
				      VRENDER_EYE_CHOICE_CYCLOPS, pos,
				      &ray_start, &direction,
				      &t_enter, &t_leave, ray_length) )
	== NULL ) return (NULL);
    /*  Shift ray_start to the actual start of the ray: i.e. the first voxel */
    ray_start.x += direction.x * t_enter;
    ray_start.y += direction.y * t_enter;
    ray_start.z += direction.z * t_enter;
    if (*ray_length > 4096)
    {
	fprintf (stderr, "Opacities buffer too small\n");
	a_prog_bug (function_name);
    }
    *depth = get_feature_depth (ray, *ray_length, NULL, shader_name,
				opacities);
    /*  Note here I am using the ray direction vector as the basis for
	determining which axis to step through. This is incorrect: the normal
	to the raster plane should be used (which reflects the internals of the
	<vrender_> package. I'll fix this sometime. This can lead to incorrect
	results when substantial perspective is used.  */
    t = depth_to_t (direction, (float) *depth);
    coord->x = ray_start.x + t * direction.x;
    coord->y = ray_start.y + t * direction.y;
    coord->z = ray_start.z + t * direction.z;
    return (ray);
}   /*  End Function get_ray_from_image_plane  */

unsigned int get_feature_depth (CONST signed char *ray,unsigned int ray_length,
				void *pixel_ptr, CONST char *shader_name,
				float *opacities)
/*  [PURPOSE] This routine will approximate the depth of the feature the user
    sees down a ray.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <shader_name> The name of the current shader.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    unsigned int (*shader) (CONST signed char *ray, unsigned int ray_length,
			    void *pixel_ptr, float *opacities) = NULL;
    static char function_name[] = "get_feature_depth";

    if (opacities != NULL) *opacities = TOOBIG;
    if (strcmp (shader_name, "max_voxel") == 0)
	shader = shader_max_voxel;
    else if (strcmp (shader_name, "min_voxel") == 0)
	shader = shader_min_voxel;
    else if (strcmp (shader_name, "front_voxel") == 0)
	shader = shader_front_voxel;
    else if (strcmp (shader_name, "sum_voxels") == 0)
	shader = shader_sum_voxels;
    else if (strcmp (shader_name, "hot_gas_mono") == 0)
	shader = shader_hot_gas_mono;
    else if (strcmp (shader_name, "hot_gas_substances") == 0)
	shader = shader_hot_gas_subs;
    else if (strcmp (shader_name, "hot_gas_continuous") == 0)
	shader = shader_hot_gas_fullcolour;
    else if (strcmp (shader_name, "hot_gas_three_colour") == 0)
	shader = shader_hot_gas_fullcolour;
    else
    {
	fprintf (stderr, "Illegal shader: \"%s\"\n", shader_name);
	a_prog_bug (function_name);
    }
    return (*shader) (ray, ray_length, pixel_ptr, opacities);
}   /*  End Function get_feature_depth  */


/*  Private routines follow  */

static float depth_to_t (Kcoord_3d direction, float depth)
/*  [PURPOSE] This routine will convert depth (plane coord) along a ray into a
    't' value (distance along a ray).
    <direction> The ray direction.
    <depth> The depth (plane coord) along the ray.
    [RETURNS] The 't' value.
*/
{
    float t;

    direction.x = fabs (direction.x);
    direction.y = fabs (direction.y);
    direction.z = fabs (direction.z);
    if ( (direction.z >= direction.y) && (direction.z >= direction.x) )
    {
	t = depth / direction.z;
    }
    else if (direction.y >= direction.x)
    {
	t = depth / direction.y;
    }
    else
    {
	t = depth / direction.x;
    }
    return (t);
}   /*  End Function depth_to_t  */

static unsigned int shader_max_voxel (CONST signed char *ray,
				      unsigned int ray_length, void *pixel_ptr,
				      float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    int depth = -1;
    int voxel;
    int pixel = -128;
    unsigned int count;

    for (count = 0; count < ray_length; ++count)
    {
	if ( (voxel = ray[count]) < -127 ) continue;
	if (voxel > pixel)
	{
	    depth = count;
	    pixel = voxel;
	}
    }
    if (pixel_ptr != NULL) *(signed char *) pixel_ptr = pixel;
    return (depth);
}   /*  End Function shader_max_voxel  */

static unsigned int shader_min_voxel (CONST signed char *ray,
				      unsigned int ray_length, void *pixel_ptr,
				      float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    int depth = -1;
    int voxel;
    int pixel = 128;
    unsigned int count;

    for (count = 0; count < ray_length; ++count)
    {
	if ( (voxel = ray[count]) < -127 ) continue;
	if (voxel < pixel)
	{
	    depth = count;
	    pixel = voxel;
	}
    }
    if (pixel_ptr != NULL) *(signed char *) pixel_ptr = pixel;
    return (depth);
}   /*  End Function shader_min_voxel  */

static unsigned int shader_front_voxel (CONST signed char *ray,
					unsigned int ray_length,
					void *pixel_ptr, float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    unsigned int count;

    for (count = 0; count < ray_length; ++count)
    {
	if (ray[count] < -127) continue;
	if (pixel_ptr != NULL) *(signed char *) pixel_ptr = ray[count];
	return (count);
    }
    if (pixel_ptr != NULL) *(signed char *) pixel_ptr = ray[count];
    return (count);
}   /*  End Function shader_front_voxel  */

static unsigned int shader_sum_voxels (CONST signed char *ray,
				       unsigned int ray_length,
				       void *pixel_ptr, float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    int depth = -1;
    int voxel;
    int pixel = 0;
    int max_voxel = -128;
    unsigned int count;

    /*  HACK: same as max voxel algorithm at the moment  */
    for (count = 0; count < ray_length; ++count)
    {
	if ( (voxel = ray[count]) < -127 ) continue;
	if (voxel > max_voxel)
	{
	    depth = count;
	    max_voxel = voxel;
	}
	pixel += voxel;
    }
    if (pixel_ptr != NULL) *(float *) pixel_ptr = pixel;
    return (depth);
}   /*  End Function shader_sum_voxels  */

static unsigned int shader_hot_gas_mono (CONST signed char *ray,
					 unsigned int ray_length,
					 void *pixel_ptr, float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    int voxel;
    int index;
    unsigned int count;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float intensity = 0.0;
    float cumulative_transparency = 1.0;
    float opacity_depth_sum = 0.0;
    float opacity_sum = 0.0;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];

    for (count = 0; count < ray_length; ++count)
    {
	if ( (voxel = ray[count]) < -127 )
	{
	    if (opacities != NULL) opacities[count] = 0.0;
	    continue;
	}
	index = voxel + 127;
	opacity = mono_hg_opacities[index];
	if (opacities != NULL) opacities[count] = opacity;
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	intensity += mono_hg_intensities[index] * emissivity;
	cumulative_transparency *= transparency;
	/*  Calculate sum of (opacity * depth) and sum of opacity  */
	opacity_depth_sum += opacity * (float) count;
	opacity_sum += opacity;
	if (cumulative_transparency < visible_threshold)
	{
	    if (opacities != NULL)
	    {
		while (++count < ray_length) opacities[count] = 0.0;
	    }
	    else count = ray_length;
	}
    }
    if (pixel_ptr != NULL) *(float *) pixel_ptr = intensity;
    if (opacity_sum < 0.1) return (ray_length - 1);
    return (unsigned int) (opacity_depth_sum / opacity_sum + 0.5);
}   /*  End Function shader_hot_gas_mono  */

static unsigned int shader_hot_gas_subs (CONST signed char *ray,
					 unsigned int ray_length,
					 void *pixel_ptr, float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    int voxel;
    int index;
    unsigned int count;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float old_opacity = 0.0;
    float old_red = 0.0;
    float old_green = 0.0;
    float old_blue = 0.0;
    int old_index = -1;
    float red = 0.0;
    float green = 0.0;
    float blue = 0.0;
    float cumulative_transparency = 1.0;
    float opacity_depth_sum = 0.0;
    float opacity_sum = 0.0;
    extern int num_substances;
    extern unsigned int bad_substance_count;
    extern char voxel_to_substance_indices[256];
    extern substance substances[MAX_SUBSTANCES];

    for (count = 0; count < ray_length; ++count)
    {
	if ( (voxel = ray[count]) < -127 )
	{
	    if (opacities != NULL) opacities[count] = 0.0;
	    continue;
	}
	if ( (index = voxel_to_substance_indices[voxel + 128])
	     >= num_substances )
	{
	    /*  Outside substance table: skip  */
	    ++bad_substance_count;
	    if (opacities != NULL) opacities[count] = 0.0;
	    continue;
	}
	if (old_index != index)
	{
	    old_opacity = substances[index].opacity;
	    old_red = substances[index].red;
	    old_green = substances[index].green;
	    old_blue = substances[index].blue;
	    old_index = index;
	}
	if (opacities != NULL) opacities[count] = old_opacity;
	transparency = one - old_opacity;
	emissivity = old_opacity * cumulative_transparency;
	red += old_red * emissivity;
	green += old_green * emissivity;
	blue += old_blue * emissivity;
	cumulative_transparency *= transparency;
	/*  Calculate sum of (opacity * depth) and sum of opacity  */
	opacity_depth_sum += old_opacity * (float) count;
	opacity_sum += old_opacity;
	if (cumulative_transparency < visible_threshold)
	{
	    if (opacities != NULL)
	    {
		while (++count < ray_length) opacities[count] = 0.0;
	    }
	    else count = ray_length;
	}
    }
    emissivity = 255.0;
    if (pixel_ptr != NULL)
    {
	*(unsigned char *) pixel_ptr = (int) (red * emissivity);
	*( (unsigned char *) pixel_ptr + 1 ) = (int) (green * emissivity);
	*( (unsigned char *) pixel_ptr + 2 ) = (int) (blue * emissivity);
    }
    if (opacity_sum < 0.1) return (ray_length - 1);
    return (unsigned int) (opacity_depth_sum / opacity_sum + 0.5);
}   /*  End Function shader_hot_gas_subs  */

static unsigned int shader_hot_gas_fullcolour (CONST signed char *ray,
					       unsigned int ray_length,
					       void *pixel_ptr,
					       float *opacities)
/*  [PURPOSE] This routine will determine the depth along a ray which the user
    will see.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <pixel_ptr> The pixel data for the entire ray will be written here. If this
    is NULL, nothing is written here.
    <opacities> The opacity values for the voxels is written here. If the
    shader does not define opacities, the value TOOBIG will be written to the
    first element. If this is NULL, nothing is written here.
    [RETURNS] The depth.
*/
{
    int voxel;
    int index;
    unsigned int count;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float red = 0.0;
    float green = 0.0;
    float blue = 0.0;
    float cumulative_transparency = 1.0;
    float opacity_depth_sum = 0.0;
    float opacity_sum = 0.0;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];

    for (count = 0; count < ray_length; ++count)
    {
	if ( (voxel = ray[count]) < -127 )
	{
	    if (opacities != NULL) opacities[count] = 0.0;
	    continue;
	}
	index = voxel + 127;
	opacity = fullcolour_hg_opacities[index];
	if (opacities != NULL) opacities[count] = opacity;
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	red += fullcolour_hg_reds[index] * emissivity;
	green += fullcolour_hg_greens[index] * emissivity;
	blue += fullcolour_hg_blues[index] * emissivity;
	cumulative_transparency *= transparency;
	/*  Calculate sum of (opacity * depth) and sum of opacity  */
	opacity_depth_sum += opacity * (float) count;
	opacity_sum += opacity;
	if (cumulative_transparency < visible_threshold)
	{
	    if (opacities != NULL)
	    {
		while (++count < ray_length) opacities[count] = 0.0;
	    }
	    else count = ray_length;
	}
    }
    emissivity = 255.0;
    if (pixel_ptr != NULL)
    {
	*(unsigned char *) pixel_ptr = (int) (red * emissivity);
	*( (unsigned char *) pixel_ptr + 1 ) = (int) (green * emissivity);
	*( (unsigned char *) pixel_ptr + 2 ) = (int) (blue * emissivity);
    }
    if (opacity_sum < 0.1) return (ray_length - 1);
    return (unsigned int) (opacity_depth_sum / opacity_sum + 0.5);
}   /*  End Function shader_hot_gas_fullcolour  */
