/*  client.c

    Source file for  render_client  (Karma volume rendering tool client).

    Copyright (C) 1996-2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the main function.


    Written by      Richard Gooch   15-DEC-1996

    Updated by      Richard Gooch   25-MAY-1997: Created <message_notify_func>.

    Updated by      Richard Gooch   9-DEC-1997: Took account of changes to
  initialisation of <vrender> package in setup_vrender.c

    Updated by      Richard Gooch   27-OCT-1998: Took account of change to
  <create_vrender_context>. Added update command.

    Updated by      Richard Gooch   5-JAN-1999: Now update global variable
  <substances_all_transparent>.

    Updated by      Richard Gooch   8-JAN-1999: Make use of new function
  <panel_item_register_callback>. Removed update command.

    Updated by      Richard Gooch   11-JAN-1999: Added bg_processing flag.

    Updated by      Richard Gooch   19-MAR-1999: Took account of change to
  <create_vrender_context>.

    Last updated by Richard Gooch   21-DEC-2001: Included k_version.h directly.


*/
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <sys/time.h>
#include <karma.h>
#include <k_version.h>
#include <karma_vrender.h>
#include <karma_module.h>
#include <karma_panel.h>
#include <karma_conn.h>
#include <karma_im.h>
#include <karma_m.h>
#include <karma_w.h>
#include <karma_p.h>
#include <karma_a.h>
#include "render.h"


#define VERSION "1.2.0"


/*  Private functions  */
STATIC_FUNCTION (void cache_computed_func,
		 (KVolumeRenderContext context, void **info, uaddr eyes) );
STATIC_FUNCTION (void message_notify_func,
		 (KVolumeRenderContext context,
		  void **info, KVolumeRenderMessage *msg) );
STATIC_FUNCTION (void update, (KPanelItem item, void **info) );
STATIC_FUNCTION (void change_w_support, (KPanelItem item, void **info) );


/*  Private data  */
static KVolumeRenderContext main_vrend_ctx = NULL;
static flag show_timings = FALSE;
static flag support_background_processing = TRUE;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KControlPanel panel;
    KPanelItem item;
    extern flag verbose;
    static char function_name[] = "main";

    conn_register_client_protocol ("hog_request", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    if ( ( panel = panel_create (FALSE) ) == NULL )
    {
	m_abort (function_name, "control panel");
    }
    item = panel_add_item (panel, "bg_processing",
			   "support background processing", PIT_FLAG,
			   &support_background_processing, PIA_END);
    panel_item_register_callback (item, change_w_support, NULL);
    item = panel_add_item (panel, "show_timings", "flag", PIT_FLAG,
			   &show_timings, PIA_END);
    panel_item_register_callback (item, update, NULL);
    item = panel_add_item (panel, "verbose", "flag", PIT_FLAG, &verbose,
			   PIA_END);
    panel_push_onto_stack (panel);
    im_register_lib_version (KARMA_VERSION);
    initialise_vrender_package ();
    main_vrend_ctx = create_vrender_context (verbose, show_timings);
    vrender_cache_notify_func (main_vrend_ctx, cache_computed_func, NULL);
    vrender_message_notify_func (main_vrend_ctx, message_notify_func, NULL);
    /*  Make this conditional in case the internal default changes  */
    if (support_background_processing) w_register_support ();
    module_run (argc, argv, "render_client", VERSION, NULL, -1, 0, FALSE);
    return (RV_OK);
}   /*  End Function main  */


/*  Private routines follow  */

static void cache_computed_func (KVolumeRenderContext context, void **info,
				 uaddr eyes)
/* [PURPOSE] This routine is called when the cache for one or more eye
   views is computed.
   <context> The volume render context.
   <info> A pointer to the arbitrary information pointer.
   <eyes> A bitmask specifying which eye views to compute. The bitmask may
   be constructed as the OR of the following values:
       VRENDER_EYE_MASK_CYCLOPS     VRENDER_EYE_MASK_LEFT
       VRENDER_EYE_MASK_RIGHT
   [RETURNS] Nothing.
*/
{
    extern flag verbose;
    static char flash_ch = '-';

    if (verbose)
    {
	fputs ("Cache computed for:", stderr);
	if (eyes & VRENDER_EYE_MASK_CYCLOPS) fputs (" cyclops", stderr);
	if (eyes & VRENDER_EYE_MASK_LEFT) fputs (" left eye", stderr);
	if (eyes & VRENDER_EYE_MASK_RIGHT) fputs (" right eye", stderr);
	putc ('\t', stderr);
	putc (flash_ch, stderr);
	putc ('\n', stderr);
	flash_ch = (flash_ch == '-') ? '|' : '-';
    }
}   /*  End Function cache_computed_func  */

static void message_notify_func (KVolumeRenderContext context,
				 void **info, KVolumeRenderMessage *msg)
/*  [SUMMARY] Message received callback.
    [PURPOSE] This routine is called when a message is received from the master
    <context> The volume render context.
    <info> A pointer to the arbitrary information pointer.
    <msg> The message.
    [RETURNS] Nothing.
*/
{
    int count;
    CONST char *ptr;
    extern flag substances_all_transparent;
    extern int voxel_sum_position;
    extern int num_substances;
    extern char voxel_to_substance_indices[256];
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern substance substances[MAX_SUBSTANCES];
    static char function_name[] = "message_notify_func";

    switch (msg->command)
    {
      case MtoS_VOXEL_OFFSET:
	voxel_sum_position = *(CONST signed char *) msg->message;
	break;
      case MtoS_MONO_HOT_GAS_INFO:
	if (msg->length != MONO_HOT_GAS_CMAP_SIZE * 8)
	{
	    fprintf (stderr,"%s: mono_hot_gas message length: %lu is not %u\n",
		     function_name, msg->length, MONO_HOT_GAS_CMAP_SIZE * 8);
	    return;
	}
	if ( !p_read_buf_floats (msg->message, MONO_HOT_GAS_CMAP_SIZE,
				 mono_hg_intensities, NULL) )
	{
	    fprintf (stderr, "Error converting mono_hot_gas intensities\n");
	    a_prog_bug (function_name);
	}
	if ( !p_read_buf_floats (msg->message + MONO_HOT_GAS_CMAP_SIZE * 4,
				 MONO_HOT_GAS_CMAP_SIZE, mono_hg_opacities,
				 NULL) )
	{
	    fprintf (stderr, "Error converting mono_hot_gas opacities\n");
	    a_prog_bug (function_name);
	}
	break;
      case MtoS_FULLCOLOUR_HOT_GAS_INFO:
	if (msg->length != FULLCOLOUR_HOT_GAS_CMAP_SIZE * 16)
	{
	    fprintf (stderr,
		     "%s: fullcolour_hot_gas message length: %lu is not %u\n",
		     function_name, msg->length,
		     FULLCOLOUR_HOT_GAS_CMAP_SIZE * 16);
	    return;
	}
	if ( !p_read_buf_floats (msg->message, FULLCOLOUR_HOT_GAS_CMAP_SIZE,
				 fullcolour_hg_reds, NULL) )
	{
	    fprintf (stderr,
		     "Error converting fullcolour_hot_gas red intensities\n");
	    a_prog_bug (function_name);
	}
	if ( !p_read_buf_floats (msg->message +
				 FULLCOLOUR_HOT_GAS_CMAP_SIZE * 4,
				 FULLCOLOUR_HOT_GAS_CMAP_SIZE,
				 fullcolour_hg_greens,
				 NULL) )
	{
	    fprintf(stderr,
		    "Error converting fullcolour_hot_gas green intensities\n");
	    a_prog_bug (function_name);
	}
	if ( !p_read_buf_floats (msg->message +
				 FULLCOLOUR_HOT_GAS_CMAP_SIZE * 4 * 2,
				 FULLCOLOUR_HOT_GAS_CMAP_SIZE,
				 fullcolour_hg_blues,
				 NULL) )
	{
	    fprintf (stderr,
		     "Error converting fullcolour_hot_gas blue intensities\n");
	    a_prog_bug (function_name);
	}
	if ( !p_read_buf_floats (msg->message +
				 FULLCOLOUR_HOT_GAS_CMAP_SIZE * 4 * 3,
				 FULLCOLOUR_HOT_GAS_CMAP_SIZE,
				 fullcolour_hg_opacities,
				 NULL) )
	{
	    fprintf (stderr,"Error converting fullcolour_hot_gas opacities\n");
	    a_prog_bug (function_name);
	}
	break;
      case MtoS_SUBSTANCES_INFO:
	if ( (msg->length - 256) % 17 != 0 )
	{
	    fprintf (stderr,
		     "%s: hot_gas_substances message length: %lu is bad\n",
		     function_name, msg->length);
	    return;
	}
	m_copy (voxel_to_substance_indices, msg->message, 256);
	num_substances = (msg->length - 256) / 17;
	substances_all_transparent = TRUE;
	for (count = 0, ptr = msg->message + 256; count < num_substances;
	     ++count, ptr += 17)
	{
	    substances[count].max_voxel = *(CONST signed char *) ptr;
	    p_read_buf_float (ptr + 1, &substances[count].red);
	    p_read_buf_float (ptr + 5, &substances[count].green);
	    p_read_buf_float (ptr + 9, &substances[count].blue);
	    p_read_buf_float (ptr + 13, &substances[count].opacity);
	    if (substances[count].opacity > 0.0)
		substances_all_transparent = FALSE;
	}
	break;
      default:
	fprintf (stderr, "%s: illegal command: %lu\n",
		 function_name, msg->command);
	break;
    }
}   /*  End Function message_notify_func  */

static void update (KPanelItem item, void **info)
/*  [SUMMARY] Update the volume rendering parameters.
    <p> A string which is ignored.
    [RETURNS] Nothing.
*/
{
    extern flag verbose;

    vrender_set_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_VERBOSE, verbose,
				    VRENDER_CONTEXT_ATT_SHOW_TIMINGS,
				    show_timings,
				    VRENDER_CONTEXT_ATT_END);
}   /*  End Function update  */

static void change_w_support (KPanelItem item, void **info)
/*  [SUMMARY] Item change callback.
    [PURPOSE] This function is called whenever an items value is changed.
    <item> The panel item which changed.
    <info> The pointer to the arbitrary information pointer.
    [RETURNS] Nothing.
*/
{
    if (support_background_processing)
    {
	if ( !w_test_supported () )
	{
	    if ( w_work_to_be_done () )
		fprintf (stderr, "Previous work will resume\n");
	    w_register_support ();
	}
    }
    else
    {
	if ( w_test_supported () )
	{
	    if ( w_work_to_be_done () )
		fprintf (stderr, "Work still remains\n");
	    w_unregister_support ();
	}
    }
}   /*  End Function change_w_support  */
