/*  cursor_3d.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the 3-dimensional cursor.


    Written by      Richard Gooch   23-FEB-1996

    Updated by      Richard Gooch   4-MAR-1996: Added 2D "shaded" cursor which
  can blink out when behind a feature.

    Updated by      Richard Gooch   16-MAR-1996: Added 3D "shaded" cursor which
  dissolves away when moved behind a feature.

    Updated by      Richard Gooch   24-MAR-1996: Fixed bugs converting plane
  co-ordinate of 3-D point to depth down a ray.

    Updated by      Richard Gooch   6-MAY-1996: Added support for
  "hot_gas_three_colour" shader.

    Updated by      Richard Gooch   19-MAY-1996: Created <get_colour> function
  to deal with all colourcells allocated for DirectColour canvases.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   4-JUN-1996: Took account of new fields in
  dimension descriptor for first and last co-ordinate.

    Updated by      Richard Gooch   21-JUL-1996: Changed to
  <canvas_convert_from_canvas_coords>.

    Updated by      Richard Gooch   25-JUL-1996: Removed redundant cursor
  parameter for <move_3d_cursor> and created <cursor_3d_position_func>.

    Updated by      Richard Gooch   26-JUL-1996: Use hot gas algorithm for
  3D shaded cursor with "hot_gas_three_colour" shader and made <get_colour>
  public.

    Updated by      Richard Gooch   23-AUG-1996: Added "shaded 2-hair in 3-D"
  cursor and renamed existing cursors. Changed <move_3d_cursor> to take new
  cursor parameter and dispatch cursor move events.

    Updated by      Richard Gooch   29-AUG-1996: Moved to <change_3d_cursor>

    Updated by      Richard Gooch   30-AUG-1996: Implemented sheded sphere
  cursor.

    Updated by      Richard Gooch   28-NOV-1996: Took account of substance
  changes. Trapped problems due to floating point roundoff in
  <show_cursor_shaded_sphere>.

    Updated by      Richard Gooch   6-DEC-1996: Created <draw_shaded_segment>
  which produces a nicer cursor. Added quick hack for loading file with 3D
  segments.

    Updated by      Richard Gooch   10-DEC-1996: Created
  <register_cursor_func> now that wireframe for rendered images is done using
  overlay lists. This separates the management of the wireframe and cursor.

    Updated by      Richard Gooch   11-DEC-1996: Switched to overlay list for
  drawing file segments.

    Updated by      Richard Gooch   12-DEC-1996: Disposed of file segment
  reading code: now taken care of in segments_xt.c

    Updated by      Richard Gooch   19-DEC-1996: Convert segment from world
  co-ordinates when no KwcsAstro object available.

    Updated by      Richard Gooch   21-JAN-1997: Removed threshold test for
  opacity in Hot Gas Substances shader.

    Updated by      Richard Gooch   4-FEB-1997: Made filled rectangles for
  thick cursor points width 1.0 rather than 0.5.

    Updated by      Richard Gooch   16-JUN-1998: Made use of new function
  <register_3d_cursor_change>. No longer move cursor on mouse release.

    Updated by      Richard Gooch   17-JUN-1998: Created <subvolume_prerender>.

    Updated by      Richard Gooch   28-JUL-1998: Added ratio limit on subcube
  size in <subvolume_prerender>.

    Updated by      Richard Gooch   4-AUG-1998: Created first working cut of
  3D region display.

    Updated by      Richard Gooch   19-MAR-1999: Fixed name of hot gas
  continuous depth shader.

    Updated by      Richard Gooch   13-APR-1999: Fixed bug in function
  <draw_shaded_segment>: was not calling <get_colour> for overlay list case.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Last updated by Richard Gooch   22-DEC-2002: Compiler warning fix.


*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_vrender.h>
#include <karma_viewimg.h>
#include <karma_canvas.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_dsra.h>
#include <karma_wcs.h>
#include <karma_chs.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_m.h>
#include "xraycast_3d.h"

#define SEGMENT_BUF_SIZE 64
#define POINT_BUF_SIZE 1024
#define HALF_2HAIR_SIZE 10
#define NUM_AREAS 3
#define MAX_INTENSITY 65535.0
#define SUBCUBE_SIZE  64
#define SUBCUBE_RATIO 0.25

#define FACE_MASK_MIN_X 0x01
#define FACE_MASK_MAX_X 0x02
#define FACE_MASK_MIN_Y 0x04
#define FACE_MASK_MAX_Y 0x08
#define FACE_MASK_MIN_Z 0x10
#define FACE_MASK_MAX_Z 0x20

#define THICK_CURSOR
#define USE_SEGMENTS


#define get_worldcanvas_info(c) get_pixcanvas_info(canvas_get_pixcanvas((c)))

/*  Structure declarations follow  */

typedef struct
{
    flag valid;
    unsigned int num_areas;
    KPixCanvasRefreshArea areas[NUM_AREAS];
} pixcanvas_info;

typedef struct
{
    KVolumeRenderContext context;
    unsigned int eye_view;
    flag has_image;
} cursor_info;

typedef struct
{
    char *x0y0z0_x1y0z0;
    char *x1y0z0_x1y1z0;
    char *x1y1z0_x0y1z0;
    char *x0y1z0_x0y0z0;
    char *x0y0z1_x1y0z1;
    char *x1y0z1_x1y1z1;
    char *x1y1z1_x0y1z1;
    char *x0y1z1_x0y0z1;
    char *x0y0z0_x0y0z1;
    char *x1y0z0_x1y0z1;
    char *x1y1z0_x1y1z1;
    char *x0y1z0_x0y1z1;
} region_type;

typedef void (*draw_line_func) (KWorldCanvas canvas,
				KVolumeRenderContext context,
				unsigned int eye_view, CONST char *colourname,
				Kcoord_3d pos1, Kcoord_3d pos2,
				double *xmin, double *ymin,
				double *xmax, double *ymax);


/*  External functions  */
EXTERN_FUNCTION (void shader_DEPTH_max_voxel, () );
EXTERN_FUNCTION (void shader_DEPTH_min_voxel, () );
EXTERN_FUNCTION (void shader_DEPTH_front_voxel, () );
EXTERN_FUNCTION (void shader_DEPTH_sum_voxels, () );
EXTERN_FUNCTION (void shader_DEPTH_hot_gas_mono, () );
EXTERN_FUNCTION (void shader_DEPTH_hot_gas_subs, () );
EXTERN_FUNCTION (void shader_DEPTH_hot_gas_fullcolour, () );


/*  Private functions  */
STATIC_FUNCTION (pixcanvas_info *get_pixcanvas_info, (KPixCanvas canvas) );
STATIC_FUNCTION (void show_3d_cursor,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view, flag allow_shading) );
STATIC_FUNCTION (void show_cursor_overlay_2hair_in_2d,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view) );
STATIC_FUNCTION (void show_cursor_overlay_3hair,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view) );
STATIC_FUNCTION (void show_cursor_shaded_2hair_in_2d,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view) );
STATIC_FUNCTION (void show_cursor_shaded_3hair,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view) );
STATIC_FUNCTION (void show_cursor_shaded_2hair_in_3d,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view) );
STATIC_FUNCTION (void show_cursor_shaded_sphere,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  unsigned int eye_view) );
STATIC_FUNCTION (void draw_region,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view, region_type control,
		  draw_line_func draw_line,
		  Kcoord_3d cursor_pos, Kcoord_3d corner3d,
		  double *xmin, double *ymin, double *xmax,double *ymax) );
STATIC_FUNCTION (void draw_unshaded_line,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view, CONST char *colourname,
		  Kcoord_3d pos1, Kcoord_3d pos2,
		  double *xmin, double *ymin, double *xmax, double *ymax) );
STATIC_FUNCTION (void draw_shaded_line,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view, CONST char *colourname,
		  Kcoord_3d pos1, Kcoord_3d pos2,
		  double *xmin, double *ymin, double *xmax, double *ymax) );
STATIC_FUNCTION (void draw_shaded_segment,
		 (KWorldCanvas canvas, KOverlayList olist,
		  KVolumeRenderContext context, unsigned int eye_view,
		  CONST char *colourname, Kcoord_3d point1, Kcoord_3d point2,
		  double *xmin, double *ymin, double *xmax, double *ymax) );
STATIC_FUNCTION (void draw_shaded_point,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view, CONST char *colourname,
		  Kcoord_3d point,
		  double *xmin, double *ymin, double *xmax, double *ymax) );
STATIC_FUNCTION (void worldcanvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize,
		  void **info, PostScriptPage pspage) );
STATIC_FUNCTION (float shader_hot_gas_mono,
		 (CONST signed char *ray, unsigned int ray_length,
		  unsigned int cursor_depth, double cursor_intensity) );
STATIC_FUNCTION (unsigned long shader_hot_gas_subs,
		 (CONST signed char *ray, unsigned int ray_length,
		  unsigned int cursor_depth, RGBcolour *colour,
		  unsigned long red_mask, unsigned long green_mask,
		  unsigned long blue_mask) );
STATIC_FUNCTION (unsigned long shader_hot_gas_fullcolour,
		 (CONST signed char *ray, unsigned int ray_length,
		  unsigned int cursor_depth, RGBcolour *colour,
		  unsigned long red_mask, unsigned long green_mask,
		  unsigned long blue_mask) );
STATIC_FUNCTION (unsigned int get_offset,
		 (packet_desc *pack_desc, CONST char *name) );
STATIC_FUNCTION (flag move_cursor_in_face,
		 (KVolumeRenderContext main_vrend_ctx, uaddr eye,
		  Kcoord_3d cursor, double x, double y) );
STATIC_FUNCTION (flag move_cursor_with_keys,
		 (KVolumeRenderContext main_vrend_ctx, uaddr eye,
		  Kcoord_3d cursor, double x, double y, char key) );
STATIC_FUNCTION (void initialise_region_control, (region_type *control) );
#ifdef not_needed
STATIC_FUNCTION (int rasterise_line,
		 (int *x_arr, int *y_arr, int x0, int y0, int x1, int y1) );
#endif


/*  Private data  */
static char *str_depth_desc[] =
{
    "PACKET",
    "  8",
    "END",
    "  ELEMENT",
    "    FLOAT",
    "    Pixel 1",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    Pixel 2",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    Pixel 3",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    Depth",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    Cumulative Opacity",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    Opacity Rolloff",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    depth_offset",
    "  END",
    "  ELEMENT",
    "    FLOAT",
    "    depth_factor",
    "  END",
    NULL
};


/*  Public functions follow  */

void register_cursor_func (KWorldCanvas canvas, KVolumeRenderContext context,
			   unsigned int eye_view, flag has_image)
/*  [PURPOSE] This routine will register an internal refresh function on a
    worldcanvas which will display the 3D cursor.
    <canvas> The world canvas. If this is NULL the routine does nothing.
    <context> The volume render context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CHOICE_CYCLOPS    VRENDER_EYE_CHOICE_LEFT
	VRENDER_EYE_CHOICE_RIGHT
    <has_image> If TRUE, the canvas has images displayed.
    [RETURNS] Nothing.
*/
{
    cursor_info *info;
    static char function_name[] = "register_cursor_func";

    if (canvas == NULL) return;
    if ( ( info = m_alloc (sizeof *info) ) == NULL )
    {
	m_abort (function_name, "cursor_info structure");
    }
    info->context = context;
    info->eye_view = eye_view;
    info->has_image = has_image;
    canvas_register_refresh_func (canvas, worldcanvas_refresh_func, info);
}   /*  End Function register_cursor_func  */

flag segments_post_render_callback (void *object, void *client1_data,
				    void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    unsigned int count;
    double xmin = 0.0;
    double xmax = 0.0;
    double ymin = 0.0;
    double ymax = 0.0;
    Kcoord_3d p1, p2;
    double xscale, yscale, zscale;
    post_render_info *info = (post_render_info *) call_data;
    segment_list *list;
    dim_desc *xdim, *ydim, *zdim;
    double x[2], y[2], z[2];
    extern KwcsAstro astro_projection;
    extern KWorldCanvas visible_worldcanvas;
    extern KVolumeRenderContext main_vrend_ctx;
    extern KOverlayList cyclops_olist;
    extern KOverlayList left_olist;
    extern KOverlayList right_olist;
    extern iarray byte_cube;
    extern segment_list *first_segment_list;

    xdim = iarray_get_dim_desc (byte_cube, 2);
    ydim = iarray_get_dim_desc (byte_cube, 1);
    zdim = iarray_get_dim_desc (byte_cube, 0);
    xscale = (double) (xdim->length-1) / (xdim->last_coord -xdim->first_coord);
    yscale = (double) (ydim->length-1) / (ydim->last_coord -ydim->first_coord);
    zscale = (double) (zdim->length-1) / (zdim->last_coord -zdim->first_coord);
    for (list = first_segment_list; list != NULL; list = list->next)
    {
	if (list->num_segments < 1) continue;
	/*  First process file segments  */
	for (count = 0; count < list->num_segments; ++count)
	{
	    if (astro_projection == NULL)
	    {
		p1 = list->p1[count];
		p2 = list->p2[count];
		p1.x = (p1.x - xdim->first_coord) * xscale;
		p1.y = (p1.y - ydim->first_coord) * yscale;
		p1.z = (p1.z - zdim->first_coord) * zscale;
		p2.x = (p2.x - xdim->first_coord) * xscale;
		p2.y = (p2.y - ydim->first_coord) * yscale;
		p2.z = (p2.z - zdim->first_coord) * zscale;
	    }
	    else
	    {
		x[0] = list->p1[count].x;
		y[0] = list->p1[count].y;
		z[0] = list->p1[count].z;
		x[1] = list->p2[count].x;
		y[1] = list->p2[count].y;
		z[1] = list->p2[count].z;
		wcs_astro_transform3 (astro_projection, FALSE, 2,
				      iarray_dim_name (byte_cube, 2), x, TRUE,
				      iarray_dim_name (byte_cube, 1), y, TRUE,
				      iarray_dim_name (byte_cube, 0), z, TRUE,
				      0, NULL, NULL);
		p1.x = x[0];
		p1.y = y[0];
		p1.z = z[0];
		p2.x = x[1];
		p2.y = y[1];
		p2.z = z[1];
	    }
	    if (info->did_stereo)
	    {
		draw_shaded_segment (visible_worldcanvas, left_olist,
				     main_vrend_ctx, VRENDER_EYE_CHOICE_LEFT,
				     list->colourname, p1, p2,
				     &xmin, &ymin, &xmax, &ymax);
		draw_shaded_segment (visible_worldcanvas, right_olist,
				     main_vrend_ctx, VRENDER_EYE_CHOICE_RIGHT,
				     list->colourname, p1, p2,
				     &xmin, &ymin, &xmax, &ymax);
	    }
	    else draw_shaded_segment (visible_worldcanvas, cyclops_olist,
				      main_vrend_ctx,VRENDER_EYE_CHOICE_CYCLOPS,
				      list->colourname, p1, p2,
				      &xmin, &ymin, &xmax, &ymax);
	}
    }
    return (FALSE);
}   /*  End Function segments_post_render_callback  */

flag change_3d_cursor (Kcoord_3d new_cursor, flag move)
/*  [PURPOSE] This routine will update the new position or shape of the
    3-dimensional cursor. The cursor move event is dispatched to the
    <<cursor3d_event_list>> callback list.
    <new_cursor> The cursor position. If this is outside the subcube boundaries
    it is ignored.
    <move> If TRUE, a new cursor position has been supplied and the cursor
    position should be changed ([<register_3d_cursor_change>] is then called).
    This should be FALSE if only the cursor shape has changed.
    [RETURNS] TRUE if the new cursor is within the subcube boundaries,
    else FALSE.
*/
{
    flag visible;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    pixcanvas_info *info;
    extern KPixCanvas orientation_pixcanvas;
    extern KWorldCanvas visible_worldcanvas;
    extern KWorldCanvas left_image_pc_worldcanvas, right_image_pc_worldcanvas;
    extern KWorldCanvas left_image_rgb_worldcanvas,right_image_rgb_worldcanvas;
    extern KVolumeRenderContext main_vrend_ctx;
    extern Kcoord_3d cursor3d;

    vrender_get_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    if (move)
    {
	/*  Ignore if new position lies outside the subcube  */
	if (new_cursor.x < x_s) return (FALSE);
	if (new_cursor.x > x_e) return (FALSE);
	if (new_cursor.y < y_s) return (FALSE);
	if (new_cursor.y > y_e) return (FALSE);
	if (new_cursor.z < z_s) return (FALSE);
	if (new_cursor.z > z_e) return (FALSE);
	/*  Cursor is OK: update  */
	cursor3d = new_cursor;
    }
    /*  Refresh the image display canvas  */
    if ( (visible_worldcanvas == left_image_pc_worldcanvas) ||
	 (visible_worldcanvas == right_image_pc_worldcanvas) )
    {
	/*  Refresh the left PseudoColour canvas  */
	info = get_worldcanvas_info (left_image_pc_worldcanvas);
	if (info->valid)
	{
	    viewimg_partial_refresh (left_image_pc_worldcanvas,
				     info->num_areas, info->areas);
	}
	else canvas_resize (left_image_pc_worldcanvas, NULL, FALSE);
	/*  Refresh the right PseudoColour canvas  */
	info = get_worldcanvas_info (right_image_pc_worldcanvas);
	if (info->valid) viewimg_partial_refresh (right_image_pc_worldcanvas,
						  info->num_areas,info->areas);
	else canvas_resize (right_image_pc_worldcanvas, NULL, FALSE);
    }
    else if ( (visible_worldcanvas == left_image_rgb_worldcanvas) ||
	      (visible_worldcanvas == right_image_rgb_worldcanvas) )
    {
	/*  Refresh the left RGBColour canvas  */
	info = get_worldcanvas_info (left_image_rgb_worldcanvas);
	if (info->valid) viewimg_partial_refresh (left_image_rgb_worldcanvas,
						  info->num_areas,info->areas);
	else canvas_resize (left_image_rgb_worldcanvas, NULL, FALSE);
	/*  Refresh the right PseudoColour canvas  */
	info = get_worldcanvas_info (right_image_rgb_worldcanvas);
	if (info->valid) viewimg_partial_refresh (right_image_rgb_worldcanvas,
						  info->num_areas,info->areas);
	else canvas_resize (right_image_rgb_worldcanvas, NULL, FALSE);
    }
    else
    {
	/*  Refresh the canvas  */
	info = get_worldcanvas_info (visible_worldcanvas);
	if (info->valid) viewimg_partial_refresh (visible_worldcanvas,
						  info->num_areas,info->areas);
	else canvas_resize (visible_worldcanvas, NULL, FALSE);
    }
    kwin_get_attributes (orientation_pixcanvas,
			 KWIN_ATT_VISIBLE, &visible,
			 KWIN_ATT_END);
    if (visible)
    {
	/*  Refresh the orientation canvas  */
	info = get_pixcanvas_info (orientation_pixcanvas);
	if (info->valid) kwin_partial_refresh (orientation_pixcanvas,
					       info->num_areas,info->areas,
					       FALSE);
	else kwin_resize (orientation_pixcanvas, TRUE, 0, 0, 0, 0);
    }
    if (move)
    {
	/*  Dispatch events  */
	register_3d_cursor_change (new_cursor);
    }
    return (TRUE);
}   /*  End Function change_3d_cursor  */

void init_depth_shaders ()
/*  [PURPOSE] This routine will initialise the depth shaders.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    char *blank_packet;
    packet_desc *pack_desc;
    extern uaddr offset_pixel1;
    extern uaddr offset_pixel2;
    extern uaddr offset_pixel3;
    extern uaddr offset_depth;
    extern uaddr offset_cumulative_opacity;
    extern uaddr offset_opacity_rolloff;
    extern uaddr offset_depth_offset;
    extern uaddr offset_depth_factor;
    static char function_name[] = "init_depth_shaders";

    /*  Create descriptor  */
    if ( ( channel = ch_open_and_fill_memory (str_depth_desc) ) == NULL )
    {
	m_abort (function_name, "memory channel");
    }
    if ( ( pack_desc = dsra_packet_desc (channel) ) == NULL )
    {
	ch_close (channel);
	m_abort (function_name, "overlay object list descriptor");
    }
    ch_close (channel);
    if ( ( blank_packet = ds_alloc_packet (pack_desc) ) == NULL )
    {
	m_abort (function_name, "blank packet");
    }
    offset_pixel1 = get_offset (pack_desc, "Pixel 1");
    offset_pixel2 = get_offset (pack_desc, "Pixel 2");
    offset_pixel3 = get_offset (pack_desc, "Pixel 3");
    offset_depth = get_offset (pack_desc, "Depth");
    offset_cumulative_opacity = get_offset (pack_desc, "Cumulative Opacity");
    offset_opacity_rolloff = get_offset (pack_desc, "Opacity Rolloff");
    offset_depth_offset = get_offset (pack_desc, "depth_offset");
    offset_depth_factor = get_offset (pack_desc, "depth_factor");
    *(signed short *) (blank_packet + offset_depth) = -1.0;
    /*  Register depth shaders  */
    vrender_register_shader (shader_DEPTH_max_voxel, NULL,
			     "DEPTH_max_voxel", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_min_voxel, NULL,
			     "DEPTH_min_voxel", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_front_voxel, NULL,
			     "DEPTH_front_voxel", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_sum_voxels, NULL,
			     "DEPTH_sum_voxels", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_hot_gas_mono, NULL,
			     "DEPTH_hot_gas_mono", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_hot_gas_fullcolour, NULL,
			     "DEPTH_hot_gas_continuous", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_hot_gas_fullcolour, NULL,
			     "DEPTH_hot_gas_three_colour", pack_desc,
			     blank_packet, NULL, FALSE);
    vrender_register_shader (shader_DEPTH_hot_gas_subs, NULL,
			     "DEPTH_hot_gas_subs", pack_desc,
			     blank_packet, NULL, FALSE);
    ds_dealloc_packet (pack_desc, NULL);
}   /*  End Function init_depth_shaders  */

void draw_cut_plane (float cut_plane_depth)
/*  [PURPOSE] This routine will draw the cut plane.
    <cut_plane_depth> The depth of the cut plane.
    [RETURNS] Nothing.
*/
{
    int x, y;
    int pixel, threshold, solid_pixel;
    unsigned int size;
    float alpha, beta, feature_depth;
    double min, max;
    char *ptr;
    array_desc *arr_desc;
    char new_shader[STRING_LENGTH];
    extern KVolumeRenderContext main_vrend_ctx;
    extern flag depth_plane_valid;
    extern uaddr offset_pixel1;
    extern uaddr offset_depth;
    extern uaddr offset_depth_offset;
    extern uaddr offset_depth_factor;
    extern double movie_min;
    extern double movie_max;
    extern CONST char *current_shader_name;
    static ViewableImage vimage = NULL;
    static iarray image = NULL;
    static int xlen, ylen;
    static unsigned int buf_size = 0;
    static char *buffer = NULL;
    static packet_desc *pack_desc = NULL;
    static char function_name[] = "draw_cut_plane";

    if (!depth_plane_valid)
    {
	/*  First do sanity checks  */
	vrender_get_context_attributes (main_vrend_ctx,
					VRENDER_CONTEXT_ATT_IMAGE_DESC,
					&arr_desc,
					VRENDER_CONTEXT_ATT_END);
	if (arr_desc->packet->num_elements != 1)
	{
	    fprintf (stderr, "%s: unsupported number of elements: %u\n",
		     function_name, arr_desc->packet->num_elements);
	    return;
	}
	if (arr_desc->packet->element_types[0] != K_BYTE)
	{
	    fprintf (stderr, "%s: unsupported element type: %u\n",
		     function_name, arr_desc->packet->element_types[0]);
	    return;
	}
	/*  Set depth shader  */
	strcpy (new_shader, "DEPTH_");
	strcat (new_shader, current_shader_name);
	vrender_set_context_attributes (main_vrend_ctx,
					VRENDER_CONTEXT_ATT_SHADER, new_shader,
					VRENDER_CONTEXT_ATT_END);
	vrender_get_context_attributes (main_vrend_ctx,
					VRENDER_CONTEXT_ATT_IMAGE_DESC,
					&arr_desc,
					VRENDER_CONTEXT_ATT_END);
	/*  Have to cache this information since the contents of the array
	    descriptor will change later.  */
	pack_desc = arr_desc->packet;
	xlen = arr_desc->dimensions[1]->length;
	ylen = arr_desc->dimensions[0]->length;
	/*  Ensure buffer is big enough  */
	size = ds_get_array_size (arr_desc) * ds_get_packet_size (pack_desc);
	if (buf_size < size)
	{
	    if (buffer != NULL) m_free (buffer);
	    if ( ( buffer = m_alloc (size) ) == NULL )
	    {
		m_abort (function_name, "depth buffer");
	    }
	    buf_size = size;
	    depth_plane_valid = FALSE;
	}
	fprintf (stderr, "depth rendering...");
	if ( !vrender_to_buffer (main_vrend_ctx, buffer, NULL, &min, &max,
				 ( void (*) (void *) ) NULL, NULL) )
	{
	    fprintf (stderr, "Error rendering\n");
	    a_prog_bug (function_name);
	}
	fprintf (stderr, "rendered\tmin: %g  max: %g\n", min, max);
	depth_plane_valid = TRUE;
	/*  Have valid depth plane: now ensure image array is big enough  */
	if ( (image != NULL) &&
	     ( (iarray_dim_length (image, 0) != ylen) ||
	       (iarray_dim_length (image, 1) != xlen) ) )
	{
	    iarray_dealloc (image);
	    if (vimage != NULL) viewimg_destroy (vimage);
	    image = NULL;
	}
	if (image == NULL)
	{
	    fprintf (stderr, "allocating image...\n");
	    if ( ( image = iarray_create_2D (ylen, xlen, K_BYTE) ) == NULL )
	    {
		m_abort (function_name, "image");
	    }
	    iarray_set_world_coords (image, 0,
				     arr_desc->dimensions[0]->first_coord,
				     arr_desc->dimensions[0]->last_coord);
	    iarray_set_world_coords (image, 1,
				     arr_desc->dimensions[1]->first_coord,
				     arr_desc->dimensions[1]->last_coord);
	}
	/*  Restore all as it was  */
	vrender_set_context_attributes (main_vrend_ctx,
					VRENDER_CONTEXT_ATT_SHADER,
					current_shader_name,
					VRENDER_CONTEXT_ATT_END);
    }
    size = ds_get_packet_size (pack_desc);
    /*  Pixel must be at least 25% of the image(canvas) range  */
    threshold = movie_min + (movie_max - movie_min) / 4.0;
    solid_pixel = movie_min + (movie_max - movie_min) / 2.0;
/*
    cut_plane_depth = -60.0 + cut_plane_depth * 1.92;
*/
    for (y = 0, ptr = buffer; y < ylen; ++y) for (x = 0; x < xlen; ++x)
    {
	pixel = *(signed char *) (ptr + offset_pixel1);
	if (pixel < threshold)
	{
	    B2 (image, y, x) = solid_pixel;
	    ptr += size;
	    continue;
	}
	feature_depth = *(float *) (ptr + offset_depth);
	if (feature_depth < 0.0)
	{
	    B2 (image, y, x) = pixel;
	    ptr += size;
	    continue;
	}
	alpha = *(float *) (ptr + offset_depth_offset);
	beta = *(float *) (ptr + offset_depth_factor);
/*
	fprintf (stderr, "%g  %g\n",
			alpha + cut_plane_depth * beta + 57.0,
			(float) feature_depth);
			*/
	if (alpha + cut_plane_depth * beta < (float) feature_depth)
	{
	    /*  Cut plane is in front of feature  */
	    pixel = solid_pixel /*movie_max*/;
	}
	B2 (image, y, x) = pixel;
	ptr += size;
    }
    if (vimage == NULL)
    {
	KWorldCanvas canvas;
	extern KWorldCanvas cyclops_image_pc_worldcanvas;
	extern KWorldCanvas cyclops_image_rgb_worldcanvas;

	canvas = cyclops_image_pc_worldcanvas;
	if (canvas == NULL) canvas = cyclops_image_rgb_worldcanvas;
	vimage = viewimg_create_from_iarray (canvas, image, FALSE);
    }
    else viewimg_register_data_change (vimage);
    viewimg_make_active (vimage);
}   /*  End Function draw_cut_plane  */

flag cursor_3d_position_func (KWorldCanvas canvas, double x, double y,
			      unsigned int event_code, void *e_info,
			      void **f_info, double x_lin, double y_lin)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    The middle mouse position is used to move the 3D cursor around one face of
    the cube, and the 'm', 'f' and 'c' keys are used to move the 3D cursor so
    that it lies behind the mouse and moves forwards and backwards,
    respectively.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    [NOTE] These values will have been transformed by the registered
    transform function (see [<canvas_register_transform_func>]).
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    uaddr eye = (uaddr) *f_info;
    extern KVolumeRenderContext main_vrend_ctx;
    extern Kcoord_3d cursor3d;
    extern int cursor_3d_mode;

    if (main_vrend_ctx == NULL) return (FALSE);
    if (cursor_3d_mode == CURSOR_3D_NONE) return (FALSE);
    switch (event_code)
    {
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
	return move_cursor_in_face (main_vrend_ctx, eye, cursor3d, x, y);
	/*break;*/
      case K_CANVAS_EVENT_PLAIN_KEY_PRESS:
	return move_cursor_with_keys (main_vrend_ctx, eye, cursor3d, x, y,
				      *(char *) e_info);
	/*break;*/
      default:
	break;
    }
    return (FALSE);
}   /*  End Function cursor_3d_position_func  */

void compute_cursor_3d_strings (Kcoord_3d cursor,
				char pixel_string[STRING_LENGTH],
				char world_string[STRING_LENGTH])
/*  [SUMMARY] Compute display strings for 3D cursor.
    <cursor> The 3D cursor position in voxel indices.
    <pixel_string> The pixel co-ordinate string is written here. This may be
    NULL.
    <world_string> The world co-ordinate string is written here. This may be
    NULL.
    [RETURNS] Nothing.
*/
{
    double wx, wy, wz;
    char ra_str[STRING_LENGTH], dec_str[STRING_LENGTH], vel_str[STRING_LENGTH];
    extern KwcsAstro astro_projection;
    extern iarray byte_cube;

    if (pixel_string != NULL)
    {
	sprintf (pixel_string, "px: %.1f  py: %.1f  pz: %.1f",
		 cursor.x, cursor.y, cursor.z);
    }
    if (world_string == NULL) return;
    /*  Convert voxel co-ordinates to linear world co-ordinates  */
    cursor.x = iarray_get_coordinate (byte_cube, 2, cursor.x);
    cursor.y = iarray_get_coordinate (byte_cube, 1, cursor.y);
    cursor.z = iarray_get_coordinate (byte_cube, 0, cursor.z);
    if (astro_projection == NULL)
    {
	sprintf (world_string, "wx: %6.2e  wy: %6.2e  wz: %6.2e",
		 cursor.x, cursor.y, cursor.z);
    }
    else
    {
	wx = cursor.x;
	wy = cursor.y;
	wz = cursor.z;
	wcs_astro_transform3 (astro_projection, FALSE, 1,
			      iarray_dim_name (byte_cube, 2), &wx, FALSE,
			      iarray_dim_name (byte_cube, 1), &wy, FALSE,
			      iarray_dim_name (byte_cube, 0), &wz, FALSE,
			      0, NULL, NULL);
	cursor.x = wx;
	cursor.y = wy;
	cursor.z = wz;
	if ( wcs_astro_test_radec (astro_projection) &&
	     wcs_astro_test_velocity (astro_projection) )
	{
	    wcs_astro_format_hms (ra_str, cursor.x / 15.0);
	    wcs_astro_format_dms (dec_str, cursor.y);
	    wcs_astro_format_vel (astro_projection, vel_str, cursor.z);
	    sprintf (world_string, "Ra %s  Dec %s  %s",
		     ra_str, dec_str, vel_str);
	}
	else sprintf (world_string, "wx: %6.2e  wy: %6.2e  wz: %6.2e",
		      cursor.x, cursor.y, cursor.z);
    }
}   /*  End Function compute_cursor_3d_strings  */

flag get_colour (KWorldCanvas canvas, CONST char *name,
		 unsigned long *pixel_value, RGBcolour *colour)
/*  [SUMMARY] Get a colour value.
    <canvas> The canvas to get the colour for.
    <name> The name of the colour.
    <pixel_value> The pixel value will be written here. If this is NULL nothing
    is written here.
    <colour> The RGB colour will be written here. If this is is NULL nothing is
    written here.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    KPixCanvas pixcanvas;
    RGBcolour clr;
    unsigned short red, green, blue;
    unsigned int visual;
    unsigned long pix, red_mask, green_mask, blue_mask;

    pixcanvas = canvas_get_pixcanvas (canvas);
    if (strcmp (name, "turquoise1") == 0)
    {
	clr.red = 0.0;
	clr.green = 0.96;
	clr.blue = 1.0;
    }
    else if (strcmp (name, "yellow") == 0)
    {
	clr.red = 1.0;
	clr.green = 1.0;
	clr.blue = 0.0;
    }
    else if (strcmp (name, "red") == 0)
    {
	clr.red = 1.0;
	clr.green = 0.0;
	clr.blue = 0.0;
    }
    else if (strcmp (name, "green") == 0)
    {
	clr.red = 0.0;
	clr.green = 1.0;
	clr.blue = 0.0;
    }
    else if (strcmp (name, "white") == 0)
    {
	clr.red = 1.0;
	clr.green = 1.0;
	clr.blue = 1.0;
    }
    else if (strcmp (name, "black") == 0)
    {
	clr.red = 0.0;
	clr.green = 0.0;
	clr.blue = 0.0;
    }
    else
    {
	/*  Have to use the window system to get the colour  */
	if ( !kwin_get_colour (pixcanvas, name, &pix,
			       &red, &green, &blue) )
	{
	    fprintf (stderr, "Unable to allocate colour: \"%s\"\n", name);
	    return (FALSE);
	}
	clr.red = (float) red / MAX_INTENSITY;
	clr.green = (float) green / MAX_INTENSITY;
	clr.blue = (float) blue / MAX_INTENSITY;
	if (colour != NULL) *colour = clr;
	if (pixel_value != NULL) *pixel_value = pix;
	return (TRUE);
    }
    if (colour != NULL) *colour = clr;
    if (pixel_value == NULL) return (TRUE);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_VISUAL, &visual,
			 KWIN_ATT_END);
    if (visual == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	return kwin_get_colour (pixcanvas, name, pixel_value, NULL, NULL,NULL);
    }
    /*  TrueColour or DirectColour canvas: compute pixel value because with
	DirectColour canvas all colourcells are already allocated  */
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_PIX_RED_MASK, &red_mask,
			 KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			 KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			 KWIN_ATT_END);
    *pixel_value = (unsigned long) (red_mask * clr.red) & red_mask;
    *pixel_value |= (unsigned long) (green_mask * clr.green) & green_mask;
    *pixel_value |= (unsigned long) (blue_mask * clr.blue) & blue_mask;
    return (TRUE);
}   /*  End Function get_colour  */

flag subvolume_prerender (KVolumeRenderContext vrend_ctx)
/*  [SUMMARY] Compute information prior to rendering for the subvolume display.
    <vrend_ctx> The KVolumeRenderContext.
    [RETURNS] TRUE on success, else FALSE indicating rendering should be
    aborted.
*/
{
    iarray cube;
    signed long xs, ys, zs, xe, ye, ze, subcube_size;
    extern Kcoord_3d cursor3d;
    extern rotation_specification rotation;

    vrender_get_context_attributes (vrend_ctx,
				    VRENDER_CONTEXT_ATT_CUBE, &cube,
				    VRENDER_CONTEXT_ATT_END);
    subcube_size = iarray_dim_length (cube, 2);
    if (iarray_dim_length (cube, 1) > subcube_size)
	subcube_size = iarray_dim_length (cube, 1);
    if (iarray_dim_length (cube, 0) > subcube_size)
	subcube_size = iarray_dim_length (cube, 0);
    subcube_size *= SUBCUBE_RATIO;
    if (subcube_size > SUBCUBE_SIZE) subcube_size = SUBCUBE_SIZE;
    xs = cursor3d.x - subcube_size / 2;
    if (xs < 0) xs = 0;
    xe = xs + subcube_size;
    if (iarray_dim_length (cube, 2) < xe) xe = iarray_dim_length (cube, 2);
    --xe;
    ys = cursor3d.y - subcube_size / 2;
    if (ys < 0) ys = 0;
    ye = ys + subcube_size;
    if (iarray_dim_length (cube, 1) < ye) ye = iarray_dim_length (cube, 1);
    --ye;
    zs = cursor3d.z - subcube_size / 2;
    if (zs < 0) zs = 0;
    ze = zs + subcube_size;
    if (iarray_dim_length (cube, 0) < ze) ze = iarray_dim_length (cube, 0);
    --ze;
    vrender_set_context_attributes (vrend_ctx,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, xs,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, ys,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, zs,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, xe,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, ye,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, ze,
				    VRENDER_CONTEXT_ATT_END);
    /*  Since the view depends on the subcube, must update view here  */
    generic_rotation_callback (vrend_ctx, NULL, &rotation, NULL);
    return (TRUE);
}   /*  End Function subvolume_prerender  */


/*  Private routines follow  */

static pixcanvas_info *get_pixcanvas_info (KPixCanvas canvas)
/*  [PURPOSE] This routine will get the pixcanvas_info structure associated
    with a canvas. If no previous information was available, it is allocated.
    <canvas> The canvas.
    [RETURNS] A pointer to the pixcanvas_info structure.
*/
{
    pixcanvas_info *info;
    static char function_name[] = "get_pixcanvas_info";

    kwin_get_attributes (canvas,
			 KWIN_ATT_USER_PTR, &info,
			 KWIN_ATT_END);
    if (info != NULL) return (info);
    if ( ( info = m_alloc (sizeof *info) ) == NULL )
    {
	m_abort (function_name, "pixcanvas_info structure");
    }
    info->valid = FALSE;
    kwin_set_attributes (canvas,
			 KWIN_ATT_USER_PTR, info,
			 KWIN_ATT_END);
    return (info);
}   /*  End Function get_pixcanvas_info  */

static void show_3d_cursor (KWorldCanvas canvas, KVolumeRenderContext context,
			    Kcoord_3d cursor_pos, Kcoord_3d corner3d,
			    unsigned int eye_view, flag allow_shading)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    <allow_shading> If TRUE, shading of cursors is allowed.
    [RETURNS] Nothing.
*/
{
    pixcanvas_info *info;
    extern int cursor_3d_mode;

    info = get_worldcanvas_info (canvas);
    info->valid = FALSE;
    switch (cursor_3d_mode)
    {
      case CURSOR_3D_NONE:
	break;
      case CURSOR_3D_OVERLAY_2HAIR_IN_2D:
	show_cursor_overlay_2hair_in_2d (canvas, context, cursor_pos, corner3d,
					 eye_view);
	break;
      case CURSOR_3D_OVERLAY_3HAIR:
	show_cursor_overlay_3hair (canvas, context, cursor_pos, corner3d,
				   eye_view);
	break;
      case CURSOR_3D_SHADED_2HAIR_IN_2D:
	if (allow_shading) show_cursor_shaded_2hair_in_2d (canvas, context,
							   cursor_pos,corner3d,
							   eye_view);
	else show_cursor_overlay_2hair_in_2d (canvas, context, cursor_pos,
					      corner3d, eye_view);
	break;
      case CURSOR_3D_SHADED_3HAIR:
	if (allow_shading) show_cursor_shaded_3hair (canvas, context,
						     cursor_pos, corner3d,
						     eye_view);
	else show_cursor_overlay_3hair (canvas, context, cursor_pos, corner3d,
					eye_view);
	break;
      case CURSOR_3D_SHADED_2HAIR_IN_3D:
	if (allow_shading)
	    show_cursor_shaded_2hair_in_3d (canvas, context, cursor_pos,
					    corner3d, eye_view);
	else show_cursor_shaded_2hair_in_2d (canvas, context, cursor_pos,
					     corner3d, eye_view);
	break;
      case CURSOR_3D_SHADED_SPHERE:
	if (allow_shading)
	    show_cursor_shaded_sphere (canvas, context, cursor_pos, corner3d,
				       eye_view);
	else show_cursor_shaded_2hair_in_2d (canvas, context, cursor_pos,
					     corner3d, eye_view);
	break;
      default:
	fprintf (stderr, "Illegal 3-D cursor mode: %u\n",cursor_3d_mode);
	break;
    }
}   /*  End Function show_3d_cursor  */

static void show_cursor_overlay_2hair_in_2d (KWorldCanvas canvas,
					     KVolumeRenderContext context,
					     Kcoord_3d cursor_pos,
					     Kcoord_3d corner3d,
					     unsigned int eye_view)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    Kcoord_2d p_2d;
    region_type region_control;
    double px, py;
    unsigned long pixel_value;
    double xrange[2], yrange[2];
    pixcanvas_info *info;

    pixcanvas = canvas_get_pixcanvas (canvas);
    info = get_pixcanvas_info (pixcanvas);
    vrender_project_3d (context, eye_view, cursor_pos, &p_2d, FALSE);
    if ( !get_colour (canvas, "green", &pixel_value, NULL) ) return;
    px = p_2d.x;
    py = p_2d.y;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &px, &py, &px, &py);
    /*  Initialise the range values  */
    xrange[0] = px - HALF_2HAIR_SIZE;
    xrange[1] = px + HALF_2HAIR_SIZE;
    yrange[0] = py + HALF_2HAIR_SIZE;
    yrange[1] = py - HALF_2HAIR_SIZE;
    canvas_convert_to_canvas_coords (canvas, FALSE, 2, xrange, yrange,
				     NULL, NULL, xrange, yrange);
    kwin_draw_line (pixcanvas,
		    px - HALF_2HAIR_SIZE, py, px + HALF_2HAIR_SIZE, py,
		    pixel_value);
    kwin_draw_line (pixcanvas,
		    px, py - HALF_2HAIR_SIZE, px, py + HALF_2HAIR_SIZE,
		    pixel_value);
    /*  Also draw the region  */
    initialise_region_control (&region_control);
    draw_region (canvas, context, eye_view, region_control,
		 draw_unshaded_line, cursor_pos, corner3d,
		 xrange, yrange, xrange + 1, yrange + 1);
    /*  Set up areas for future partial refresh  */
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, xrange, yrange,
				       xrange, yrange);
    m_clear (info->areas, sizeof info->areas);
    info->areas[0].startx = xrange[0] - 1.0;
    info->areas[0].endx = xrange[1] + 1.0;
    info->areas[0].starty = yrange[1] - 1.0;
    info->areas[0].endy = yrange[0] + 1.0;
    info->areas[0].clear = TRUE;
    info->num_areas = 1;
    info->valid = TRUE;
}   /*  End Function show_cursor_overlay_2hair_in_2d  */

static void show_cursor_overlay_3hair (KWorldCanvas canvas,
				       KVolumeRenderContext context,
				       Kcoord_3d cursor_pos,
				       Kcoord_3d corner3d,
				       unsigned int eye_view)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    Kcoord_3d p1_3d, p2_3d;
    Kcoord_3d cube_min, cube_max;
    region_type region_control;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    double xrange[2], yrange[2];
    pixcanvas_info *info;

    info = get_worldcanvas_info (canvas);
    vrender_get_context_attributes (context,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    cube_min.x = x_s;
    cube_min.y = y_s;
    cube_min.z = z_s;
    cube_max.x = x_e;
    cube_max.y = y_e;
    cube_max.z = z_e;
    /*  Initialise the range values  */
    xrange[0] = TOOBIG;
    xrange[1] = -TOOBIG;
    yrange[0] = TOOBIG;
    yrange[1] = -TOOBIG;
    /*  Display parallel to  x  axis  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cursor_pos.y;
    p1_3d.z = cursor_pos.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cursor_pos.y;
    p2_3d.z = cursor_pos.z;
    draw_unshaded_line (canvas, context, eye_view, "turquoise", p1_3d, p2_3d,
			xrange, yrange, xrange + 1, yrange + 1);
    /*  Display parallel to  y  axis  */
    p1_3d.x = cursor_pos.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cursor_pos.z;
    p2_3d.x = cursor_pos.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cursor_pos.z;
    draw_unshaded_line (canvas, context, eye_view, "yellow", p1_3d, p2_3d,
			xrange, yrange, xrange + 1, yrange + 1);
    /*  Display parallel to  z  axis  */
    p1_3d.x = cursor_pos.x;
    p1_3d.y = cursor_pos.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cursor_pos.x;
    p2_3d.y = cursor_pos.y;
    p2_3d.z = cube_max.z;
    draw_unshaded_line (canvas, context, eye_view, "red", p1_3d, p2_3d,
			xrange, yrange, xrange + 1, yrange + 1);
    /*  Also draw the region  */
    initialise_region_control (&region_control);
    region_control.x0y0z0_x1y0z0 = NULL;
    region_control.x0y1z0_x0y0z0 = NULL;
    region_control.x0y0z0_x0y0z1 = NULL;
    draw_region (canvas, context, eye_view, region_control,
		 draw_unshaded_line, cursor_pos, corner3d,
		 xrange, yrange, xrange + 1, yrange + 1);
    /*  Set up areas for future partial refresh  */
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, xrange, yrange,
				       xrange, yrange);
    m_clear (info->areas, sizeof info->areas);
    info->areas[0].startx = xrange[0] - 1.0;
    info->areas[0].endx = xrange[1] + 1.0;
    info->areas[0].starty = yrange[1] - 1.0;
    info->areas[0].endy = yrange[0] + 1.0;
    info->areas[0].clear = TRUE;
    info->num_areas = 1;
    info->valid = TRUE;
}   /*  End Function show_cursor_overlay_3hair  */

static void show_cursor_shaded_2hair_in_2d (KWorldCanvas canvas,
					    KVolumeRenderContext context,
					    Kcoord_3d cursor_pos,
					    Kcoord_3d corner3d,
					    unsigned int eye_view)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    Kcoord_2d p_2d;
    Kcoord_3d direction, ray_start;
    flag visible = FALSE;
    unsigned int ray_length, feature_depth;
    float cursor_depth, t_enter, t_leave;
    double v_min, v_max;
    pixcanvas_info *info;
    CONST signed char *ray;
    char pixel[sizeof (double)];
    extern CONST char *current_shader_name;

    canvas_get_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, &v_min,
			   CANVAS_ATT_VALUE_MAX, &v_max,
			   CANVAS_ATT_END);
    info = get_worldcanvas_info (canvas);
    /*  Project cursor "hotspot" to image plane  */
    vrender_project_3d (context, eye_view, cursor_pos, &p_2d, FALSE);
    /*  Get the ray from the projected point  */
    if ( ( ray = vrender_collect_ray (context, eye_view, p_2d,
				      &ray_start, &direction,
				      &t_enter, &t_leave, &ray_length) )
	 == NULL ) return;
    /*  Shift ray_start to the actual start of the ray: i.e. the first voxel */
    ray_start.x += direction.x * t_enter;
    ray_start.y += direction.y * t_enter;
    ray_start.z += direction.z * t_enter;
    /*  Convert 3-D cursor position to depth along ray (= plane coord)  */
    if ( ( fabs (direction.z) >= fabs (direction.y) ) &&
	 ( fabs (direction.z) >= fabs (direction.x) ) )
    {
	if (direction.z > 0.0) cursor_depth = cursor_pos.z - ray_start.z;
	else cursor_depth = ray_start.z - cursor_pos.z;
    }
    else if ( fabs (direction.y) >= fabs (direction.x) )
    {
	if (direction.y > 0.0) cursor_depth = cursor_pos.y - ray_start.y;
	else cursor_depth = ray_start.y - cursor_pos.y;
    }
    else
    {
	if (direction.x > 0.0) cursor_depth = cursor_pos.x - ray_start.x;
	else cursor_depth = ray_start.x - cursor_pos.x;
    }
    /*  Check if hotspot is visible  */
    /*  First get depth of feature user sees  */
    feature_depth = get_feature_depth (ray, ray_length, pixel,
				       current_shader_name, NULL);
    if (feature_depth > (unsigned int) cursor_depth) visible = TRUE;
    /*  Check if pixel value is considered sufficient to obscure cursor  */
    if (strcmp (current_shader_name, "max_voxel") == 0)
    {
	/*  Pixel must be at least 25% of the image(canvas) range  */
	if ( *(signed char *) pixel < (int) (v_min + (v_max - v_min) / 4.0) )
	{
	    visible = TRUE;
	}
    }
    else if (strcmp (current_shader_name, "min_voxel") == 0)
    {
	/*  Pixel must be no more than 75% of the image(canvas) range  */
	if ( *(signed char *) pixel > (int) (v_min + (v_max - v_min) * 0.75) )
	{
	    visible = TRUE;
	}
    }
    if (visible)
    {
	/*  Cursor is indeed visible  */
	show_cursor_overlay_2hair_in_2d (canvas, context, cursor_pos, corner3d,
					 eye_view);
	return;
    }
    m_clear (info->areas, sizeof info->areas);
    info->areas[0].startx = 0;
    info->areas[0].endx = 0;
    info->areas[0].starty = 0;
    info->areas[0].endy = 0;
    info->areas[0].clear = FALSE;
    info->num_areas = 1;
    info->valid = TRUE;
}   /*  End Function show_cursor_shaded_2hair_in_2d  */

static void show_cursor_shaded_3hair (KWorldCanvas canvas,
				      KVolumeRenderContext context,
				      Kcoord_3d cursor_pos, Kcoord_3d corner3d,
				      unsigned int eye_view)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    Kcoord_3d p1_3d, p2_3d;
    Kcoord_3d cube_min, cube_max;
    region_type region_control;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    double xrange[2], yrange[2];
    pixcanvas_info *info;

    info = get_worldcanvas_info (canvas);
    vrender_get_context_attributes (context,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    cube_min.x = x_s;
    cube_min.y = y_s;
    cube_min.z = z_s;
    cube_max.x = x_e;
    cube_max.y = y_e;
    cube_max.z = z_e;
    /*  Initialise the range values  */
    xrange[0] = TOOBIG;
    xrange[1] = -TOOBIG;
    yrange[0] = TOOBIG;
    yrange[1] = -TOOBIG;
    /*  Display parallel to  x  axis  */
    p1_3d.x = cube_min.x;
    p1_3d.y = cursor_pos.y;
    p1_3d.z = cursor_pos.z;
    p2_3d.x = cube_max.x;
    p2_3d.y = cursor_pos.y;
    p2_3d.z = cursor_pos.z;
    draw_shaded_line (canvas, context, eye_view, "turquoise1",
		      p1_3d, p2_3d, xrange, yrange, xrange + 1, yrange + 1);
    /*  Display parallel to  y  axis  */
    p1_3d.x = cursor_pos.x;
    p1_3d.y = cube_min.y;
    p1_3d.z = cursor_pos.z;
    p2_3d.x = cursor_pos.x;
    p2_3d.y = cube_max.y;
    p2_3d.z = cursor_pos.z;
    draw_shaded_line (canvas, context, eye_view, "yellow",
		      p1_3d, p2_3d, xrange, yrange, xrange + 1, yrange + 1);
    /*  Display parallel to  z  axis  */
    p1_3d.x = cursor_pos.x;
    p1_3d.y = cursor_pos.y;
    p1_3d.z = cube_min.z;
    p2_3d.x = cursor_pos.x;
    p2_3d.y = cursor_pos.y;
    p2_3d.z = cube_max.z;
    draw_shaded_line (canvas, context, eye_view, "red",
		      p1_3d, p2_3d, xrange, yrange, xrange + 1, yrange + 1);
    /*  Also draw the region  */
    initialise_region_control (&region_control);
    region_control.x0y0z0_x1y0z0 = NULL;
    region_control.x0y1z0_x0y0z0 = NULL;
    region_control.x0y0z0_x0y0z1 = NULL;
    draw_region (canvas, context, eye_view, region_control,
		 draw_shaded_line, cursor_pos, corner3d,
		 xrange, yrange, xrange + 1, yrange + 1);
    /*  Set up areas for future partial refresh  */
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, xrange, yrange,
				       xrange, yrange);
    m_clear (info->areas, sizeof info->areas);
    info->areas[0].startx = xrange[0] - 1.0;
    info->areas[0].endx = xrange[1] + 1.0;
    info->areas[0].starty = yrange[1] - 1.0;
    info->areas[0].endy = yrange[0] + 1.0;
    info->areas[0].clear = TRUE;
    info->num_areas = 1;
    info->valid = TRUE;
}   /*  End Function show_cursor_shaded_3hair  */

static void show_cursor_shaded_2hair_in_3d (KWorldCanvas canvas,
					    KVolumeRenderContext context,
					    Kcoord_3d cursor_pos,
					    Kcoord_3d corner3d,
					    unsigned int eye_view)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    Kcoord_3d ras_centre, direction, horizontal;
    view_specification view;
    region_type region_control;
    unsigned int count;
    double dummy = 0.0;
    Kcoord_3d p3d[4];
    Kcoord_2d p2d[4];
    double xrange[2], yrange[2];
    double px[4], py[4];
    pixcanvas_info *info;

    info = get_worldcanvas_info (canvas);
    /*  Initialise the range values  */
    xrange[0] = TOOBIG;
    xrange[1] = -TOOBIG;
    yrange[0] = TOOBIG;
    yrange[1] = -TOOBIG;
    /*  Get eye information  */
    vrender_get_eye_info (context, eye_view, &ras_centre, &direction,
			  &horizontal);
    vrender_get_context_attributes (context,
				    VRENDER_CONTEXT_ATT_VIEW, &view,
				    VRENDER_CONTEXT_ATT_END);
    /*  Construct end-points  */
    p3d[0].x = cursor_pos.x - horizontal.x * HALF_2HAIR_SIZE;
    p3d[0].y = cursor_pos.y - horizontal.y * HALF_2HAIR_SIZE;
    p3d[0].z = cursor_pos.z - horizontal.z * HALF_2HAIR_SIZE;
    p3d[1].x = cursor_pos.x + horizontal.x * HALF_2HAIR_SIZE;
    p3d[1].y = cursor_pos.y + horizontal.y * HALF_2HAIR_SIZE;
    p3d[1].z = cursor_pos.z + horizontal.z * HALF_2HAIR_SIZE;
    p3d[2].x = cursor_pos.x - view.vertical.x * HALF_2HAIR_SIZE;
    p3d[2].y = cursor_pos.y - view.vertical.y * HALF_2HAIR_SIZE;
    p3d[2].z = cursor_pos.z - view.vertical.z * HALF_2HAIR_SIZE;
    p3d[3].x = cursor_pos.x + view.vertical.x * HALF_2HAIR_SIZE;
    p3d[3].y = cursor_pos.y + view.vertical.y * HALF_2HAIR_SIZE;
    p3d[3].z = cursor_pos.z + view.vertical.z * HALF_2HAIR_SIZE;
    draw_shaded_line (canvas, context, eye_view, "green",
		      p3d[0], p3d[1], &dummy, &dummy, &dummy, &dummy);
    draw_shaded_line (canvas, context, eye_view, "green",
		      p3d[2], p3d[3], &dummy, &dummy, &dummy, &dummy);
    for (count = 0; count < 4; ++count)
    {
	vrender_project_3d (context, eye_view, p3d[count], p2d + count, FALSE);
	px[count] = p2d[count].x /*- 0.55 + (float) (count % 2) * 1.1 */;
	py[count] = p2d[count].y /*- 0.55 + (float) (count % 2) * 1.1 */;
    }
    /*  Also draw the region  */
    initialise_region_control (&region_control);
    draw_region (canvas, context, eye_view, region_control,
		 draw_shaded_line, cursor_pos, corner3d,
		 xrange, yrange, xrange + 1, yrange + 1);
    /*  Set up areas for future partial refresh  */
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 4, px, py, px,py);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, xrange, yrange,
				       xrange, yrange);
    m_clear (info->areas, sizeof info->areas);
    info->areas[0].startx = px[0] - 1;
    info->areas[0].endx = px[1] + 1;
    info->areas[0].starty = py[1] - 1;
    info->areas[0].endy = py[0] + 1;
    info->areas[0].clear = TRUE;
    info->areas[1].startx = px[2] - 1;
    info->areas[1].endx = px[3] + 1;
    info->areas[1].starty = py[3] - 1;
    info->areas[1].endy = py[2] + 1;
    info->areas[1].clear = TRUE;
    info->areas[2].startx = xrange[0] - 1.0;
    info->areas[2].endx = xrange[1] + 1.0;
    info->areas[2].starty = yrange[1] - 1.0;
    info->areas[2].endy = yrange[0] + 1.0;
    info->areas[2].clear = TRUE;
    info->num_areas = (corner3d.x < 0.0) ? 2 : 3;
    info->valid = TRUE;
}   /*  End Function show_cursor_shaded_2hair_in_3d  */

static void show_cursor_shaded_sphere (KWorldCanvas canvas,
				       KVolumeRenderContext context,
				       Kcoord_3d cursor_pos,
				       Kcoord_3d corner3d,
				       unsigned int eye_view)
/*  [PURPOSE] This routine will show the position of the 3 dimensional cursor.
    <canvas> The canvas.
    <context> The volume rendering context.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    [RETURNS] Nothing.
*/
{
    Kcoord_3d ras_centre, direction, horizontal, p3d;
    view_specification view;
    float lon, lon_step, lat, lat_step, radius, h, v, d, rad, x, y;
    float sin_lat, cos_lat, sin_lon, cos_lon;
    double xrange[2], yrange[2];
    pixcanvas_info *info;
    extern int sphere_radius;

    info = get_worldcanvas_info (canvas);
    radius = sphere_radius;
    /*  Get eye information  */
    vrender_get_eye_info (context, eye_view, &ras_centre, &direction,
			  &horizontal);
    vrender_get_context_attributes (context,
				    VRENDER_CONTEXT_ATT_VIEW, &view,
				    VRENDER_CONTEXT_ATT_END);
    /*  Initialise the range values  */
    xrange[0] = TOOBIG;
    xrange[1] = -TOOBIG;
    yrange[0] = TOOBIG;
    yrange[1] = -TOOBIG;
    /*  Draw further part of sphere  */
    /*  First draw lines of latitude  */
    lat_step = PI / 10.0;
    for (lat = -PI_ON_2; lat < PI_ON_2; lat += lat_step)
    {
	sin_lat = sin (lat);
	cos_lat = cos (lat);
	rad = cos_lat * radius;
	for (x = -rad; x < rad; ++x)
	{
	    h = x;
	    v = sin_lat * radius;
	    d = rad * rad - x * x;
	    if (d < 0.0) continue;  /*  Required for IRIX6_mips4  */
	    d = sqrt (d);
	    p3d.x = cursor_pos.x + horizontal.x * h + view.vertical.x * v +
		direction.x * d;
	    p3d.y = cursor_pos.y + horizontal.y * h + view.vertical.y * v +
		direction.y * d;
	    p3d.z = cursor_pos.z + horizontal.z * h + view.vertical.z * v +
		direction.z * d;
	    draw_shaded_point (canvas, context, eye_view, "green",
			       p3d, xrange, yrange, xrange + 1, yrange + 1);
	}
    }
    /*  Now draw lines of longitude  */
    lon_step = PI / 10.0;
    for (lon = 0.0; lon < PI; lon += lon_step)
    {
	sin_lon = sin (lon);
	cos_lon = cos (lon);
	for (y = -radius; y < radius; ++y)
	{
	    rad = sqrt (radius * radius - y * y);
	    h = cos_lon * rad;
	    v = y;
	    d = sin_lon * rad;
	    p3d.x = cursor_pos.x + horizontal.x * h + view.vertical.x * v +
		direction.x * d;
	    p3d.y = cursor_pos.y + horizontal.y * h + view.vertical.y * v +
		direction.y * d;
	    p3d.z = cursor_pos.z + horizontal.z * h + view.vertical.z * v +
		direction.z * d;
	    draw_shaded_point (canvas, context, eye_view, "green",
			       p3d, xrange, yrange, xrange + 1, yrange + 1);
	}
    }
    /*  Draw closer part of sphere  */
    rad = radius * sqrt (2.0) / 2.0;
    for (h = -rad; h < rad; ++h)
    {
	v = h;
	d = sqrt (radius * radius - h * h - v * v);
	p3d.x = cursor_pos.x + horizontal.x * h + view.vertical.x * v -
	    direction.x * d;
	p3d.y = cursor_pos.y + horizontal.y * h + view.vertical.y * v -
	    direction.y * d;
	p3d.z = cursor_pos.z + horizontal.z * h + view.vertical.z * v -
	    direction.z * d;
	draw_shaded_point (canvas, context, eye_view, "red",
			   p3d, xrange, yrange, xrange + 1, yrange + 1);
	p3d.x = cursor_pos.x + horizontal.x * h - view.vertical.x * v -
	    direction.x * d;
	p3d.y = cursor_pos.y + horizontal.y * h - view.vertical.y * v -
	    direction.y * d;
	p3d.z = cursor_pos.z + horizontal.z * h - view.vertical.z * v -
	    direction.z * d;
	draw_shaded_point (canvas, context, eye_view, "green",
			   p3d, xrange, yrange, xrange + 1, yrange + 1);
    }
    /*  Set up areas for future partial refresh  */
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 2, xrange, yrange,
				       xrange, yrange);
    m_clear (info->areas, sizeof info->areas);
    info->areas[0].startx = xrange[0] - 1.0;
    info->areas[0].endx = xrange[1] + 1.0;
    info->areas[0].starty = yrange[1] - 1.0;
    info->areas[0].endy = yrange[0] + 1.0;
    info->areas[0].clear = TRUE;
    info->num_areas = 1;
    info->valid = TRUE;
}   /*  End Function show_cursor_shaded_sphere  */

static void draw_region (KWorldCanvas canvas, KVolumeRenderContext context,
			 unsigned int eye_view, region_type control,
			 draw_line_func draw_line,
			 Kcoord_3d cursor_pos, Kcoord_3d corner3d,
			 double *xmin, double *ymin, double *xmax,double *ymax)
/*  [SUMMARY] Draw a region over a volume rendered image.
    <canvas> The canvas.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    <control> The region control.
    <cursor_pos> The cursor position.
    <corner3d> The position of the region other corner.
    <xmin> The minimum 2D horizontal co-ordinate. This is modified.
    <ymin> The minimum 2D vertical co-ordinate. This is modified.
    <xmax> The maximum 2D horizontal co-ordinate. This is modified.
    <ymax> The maximum 2D vertical co-ordinate. This is modified.
    [RETURNS] Nothing.
*/
{
    Kcoord_3d p1, p2;

    if ( (corner3d.x < 0.0) || (corner3d.y < 0.0) || (corner3d.z < 0.0) )
	return;
    /*  Bottom box in z0  */
    if (control.x0y0z0_x1y0z0 != NULL)
    {
	p1 = cursor_pos;
	p2 = p1;
	p2.x = corner3d.x;
	(*draw_line) (canvas, context, eye_view, control.x0y0z0_x1y0z0, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x1y0z0_x1y1z0 != NULL)
    {
	p1 = cursor_pos;
	p1.x = corner3d.x;
	p2 = p1;
	p2.y = corner3d.y;
	(*draw_line) (canvas, context, eye_view, control.x1y0z0_x1y1z0, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x1y1z0_x0y1z0 != NULL)
    {
	p2 = cursor_pos;
	p2.y = corner3d.y;
	p1 = p2;
	p1.x = corner3d.x;
	(*draw_line) (canvas, context, eye_view, control.x1y1z0_x0y1z0, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x0y1z0_x0y0z0 != NULL)
    {
	p2 = cursor_pos;
	p1 = p2;
	p1.y = corner3d.y;
	(*draw_line) (canvas, context, eye_view, control.x0y1z0_x0y0z0, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    /*  Top box in z1  */
    if (control.x0y0z1_x1y0z1 != NULL)
    {
	p1 = cursor_pos;
	p1.z = corner3d.z;
	p2 = p1;
	p2.x = corner3d.x;
	(*draw_line) (canvas, context, eye_view, control.x0y0z1_x1y0z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x1y0z1_x1y1z1 != NULL)
    {
	p1 = cursor_pos;
	p1.x = corner3d.x;
	p1.z = corner3d.z;
	p2 = p1;
	p2.y = corner3d.y;
	(*draw_line) (canvas, context, eye_view, control.x1y0z1_x1y1z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x1y1z1_x0y1z1 != NULL)
    {
	p2 = cursor_pos;
	p2.y = corner3d.y;
	p2.z = corner3d.z;
	p1 = p2;
	p1.x = corner3d.x;
	(*draw_line) (canvas, context, eye_view, control.x1y1z1_x0y1z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x0y1z1_x0y0z1 != NULL)
    {
	p2 = cursor_pos;
	p2.z = corner3d.z;
	p1 = p2;
	p1.y = corner3d.y;
	(*draw_line) (canvas, context, eye_view, control.x0y1z1_x0y0z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    /*  Join the dots^H^H^H^Hboxes  */
    if (control.x0y0z0_x0y0z1 != NULL)
    {
	p1 = cursor_pos;
	p2 = p1;
	p2.z = corner3d.z;
	(*draw_line) (canvas, context, eye_view, control.x0y0z0_x0y0z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x1y0z0_x1y0z1 != NULL)
    {
	p1 = cursor_pos;
	p1.x = corner3d.x;
	p2 = p1;
	p2.z = corner3d.z;
	(*draw_line) (canvas, context, eye_view, control.x1y0z0_x1y0z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x1y1z0_x1y1z1 != NULL)
    {
	p2 = corner3d;
	p1 = p2;
	p1.z = cursor_pos.z;
	(*draw_line) (canvas, context, eye_view, control.x1y1z0_x1y1z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
    if (control.x0y1z0_x0y1z1 != NULL)
    {
	p1 = cursor_pos;
	p1.y = corner3d.y;
	p2 = p1;
	p2.z = corner3d.z;
	(*draw_line) (canvas, context, eye_view, control.x0y1z0_x0y1z1, p1, p2,
		      xmin, ymin, xmax, ymax);
    }
}   /*  End Function draw_region  */

static void draw_unshaded_line (KWorldCanvas canvas,
				KVolumeRenderContext context,
				unsigned int eye_view, CONST char *colourname,
				Kcoord_3d pos1, Kcoord_3d pos2,
				double *xmin, double *ymin,
				double *xmax, double *ymax)
/*  [PURPOSE] This routine will draw an unshaded line over a volume rendered
    image.
    <canvas> The canvas.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    <colourname> The name of the colour to draw in.
    <pos1> The co-ordinate of one of the line endpoints.
    <pos2> The co-ordinate of the other of the line endpoints.
    <xmin> The minimum 2D horizontal co-ordinate. This is modified.
    <ymin> The minimum 2D vertical co-ordinate. This is modified.
    <xmax> The maximum 2D horizontal co-ordinate. This is modified.
    <ymax> The maximum 2D vertical co-ordinate. This is modified.
    [RETURNS] Nothing.
*/
{
    Kcoord_2d p1_2d, p2_2d;
    unsigned long pixel_value;

    if ( !get_colour (canvas, colourname, &pixel_value, NULL) ) return;
    vrender_project_3d (context, eye_view, pos1, &p1_2d, FALSE);
    vrender_project_3d (context, eye_view, pos2, &p2_2d, FALSE);
    canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y, p2_2d.x, p2_2d.y,
			pixel_value);
    /*  Expand range as required  */
    if (p1_2d.x < *xmin) *xmin = p1_2d.x;
    if (p1_2d.x > *xmax) *xmax = p1_2d.x;
    if (p2_2d.x < *xmin) *xmin = p2_2d.x;
    if (p2_2d.x > *xmax) *xmax = p2_2d.x;
    if (p1_2d.y < *ymin) *ymin = p1_2d.y;
    if (p1_2d.y > *ymax) *ymax = p1_2d.y;
    if (p2_2d.y < *ymin) *ymin = p2_2d.y;
    if (p2_2d.y > *ymax) *ymax = p2_2d.y;
}   /*  End Function draw_unshaded_line  */

static void draw_shaded_line (KWorldCanvas canvas,KVolumeRenderContext context,
			      unsigned int eye_view, CONST char *colourname,
			      Kcoord_3d pos1, Kcoord_3d pos2,
			      double *xmin, double *ymin,
			      double *xmax, double *ymax)
/*  [PURPOSE] This routine will draw a shaded line over a volume rendered
    image.
    <canvas> The canvas.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    <colourname> The name of the colour to draw in.
    <pos1> The co-ordinate of one of the line endpoints.
    <pos2> The co-ordinate of the other of the line endpoints.
    <xmin> The minimum 2D horizontal co-ordinate. This is modified.
    <ymin> The minimum 2D vertical co-ordinate. This is modified.
    <xmax> The maximum 2D horizontal co-ordinate. This is modified.
    <ymax> The maximum 2D vertical co-ordinate. This is modified.
    [RETURNS] Nothing.
*/
{
    Kcoord_2d p1_2d, p2_2d;
    Kcoord_3d point, offset;
#ifdef USE_SEGMENTS
    Kcoord_3d p2;
#endif
    float num_steps, t;

    /*  Process points along line in raster fashion. Want to minimise the
	number of points generated, yet must have sufficient points to
	represent line properly. Projecting the line onto the view plane and
	then rasterising would require computing the nearest point on the 3-D
	line to the ray projected from each point on the view plane: not good.
	The alternative is to rasterise the 3-D line directly, but it is hard
	to know how finely to subdivide the line: it is easy to undersample or
	oversample. So, the endpoints of the 3-D line are projected and the
	degree of subdivision is determined from these, but it is the 3-D line
	which is rasterised.
	*/
    vrender_project_3d (context, eye_view, pos1, &p1_2d, FALSE);
    vrender_project_3d (context, eye_view, pos2, &p2_2d, FALSE);
    /*  Determine number of steps  */
    num_steps = fabs (p2_2d.x - p1_2d.x);
    if (fabs (p2_2d.y - p1_2d.y) > num_steps)
    {
	num_steps = fabs (p2_2d.y - p1_2d.y);
    }
    num_steps = ceil (num_steps);
    offset.x = (pos2.x - pos1.x) / num_steps;
    offset.y = (pos2.y - pos1.y) / num_steps;
    offset.z = (pos2.z - pos1.z) / num_steps;
    /*  Now step along line  */
    for (t = 0.0; t <= num_steps; ++t)
    {
	point.x = pos1.x + t * offset.x;
	point.y = pos1.y + t * offset.y;
	point.z = pos1.z + t * offset.z;
#ifdef USE_SEGMENTS
	p2.x = point.x + offset.x;
	p2.y = point.y + offset.y;
	p2.z = point.z + offset.z;
	draw_shaded_segment (canvas, NULL, context, eye_view, colourname,
			     point, p2, xmin, ymin, xmax, ymax);
#else
	draw_shaded_point (canvas, context, eye_view, colourname,
			   point, xmin, ymin, xmax, ymax);
#endif
    }
}   /*  End Function draw_shaded_line  */

static void draw_shaded_segment (KWorldCanvas canvas, KOverlayList olist,
				 KVolumeRenderContext context,
				 unsigned int eye_view, CONST char *colourname,
				 Kcoord_3d point1, Kcoord_3d point2,
				 double *xmin, double *ymin,
				 double *xmax, double *ymax)
/*  [PURPOSE] This routine will draw a shaded segment over a volume rendered
    image.
    <canvas> The canvas.
    <olist> The KOverlayList to draw to. This may be NULL.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    <colourname> The name of the colour to draw in.
    <point1> The co-ordinate of the one end of the segment.
    <point2> The co-ordinate of the other end of the segment.
    <xmin> The minimum 2D horizontal co-ordinate. This is modified.
    <ymin> The minimum 2D vertical co-ordinate. This is modified.
    <xmax> The maximum 2D horizontal co-ordinate. This is modified.
    <ymax> The maximum 2D vertical co-ordinate. This is modified.
    [NOTE] The segment must be short. Shading/visibility is computed from the
    midpoint of the segment.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    Kcoord_2d midpoint_2d, p1_2d, p2_2d;
    Kcoord_3d midpoint_3d, ray_start, direction;
    flag visible = FALSE;
    unsigned int ray_length, feature_depth;
    unsigned int canvas_depth;
    unsigned long red_mask, green_mask, blue_mask;
    unsigned long pixel_value;
    RGBcolour colour;
    float cursor_depth, t_enter, t_leave;
    double v_min, v_max;
    char txt[STRING_LENGTH];
    double value[2];
    CONST signed char *ray;
    char pixel[sizeof (double)];
    extern CONST char *current_shader_name;

    if ( !get_colour (canvas, colourname, &pixel_value, &colour) ) return;
    midpoint_3d.x = (point1.x + point2.x) * 0.5;
    midpoint_3d.y = (point1.y + point2.y) * 0.5;
    midpoint_3d.z = (point1.z + point2.z) * 0.5;
    /*  Project 3-D midpoint to image plane  */
    vrender_project_3d (context, eye_view, midpoint_3d, &midpoint_2d, FALSE);
    /*  Get the ray from the projected midpoint  */
    if ( ( ray = vrender_collect_ray (context, eye_view, midpoint_2d,
				      &ray_start, &direction,
				      &t_enter, &t_leave, &ray_length) )
	 == NULL )
    {
	return;
    }
    /*  Project 3-D endpoints to image plane  */
    vrender_project_3d (context, eye_view, point1, &p1_2d, FALSE);
    vrender_project_3d (context, eye_view, point2, &p2_2d, FALSE);
    /*  Expand range as required  */
    if (p1_2d.x < *xmin) *xmin = p1_2d.x;
    if (p1_2d.x > *xmax) *xmax = p1_2d.x;
    if (p1_2d.y < *ymin) *ymin = p1_2d.y;
    if (p1_2d.y > *ymax) *ymax = p1_2d.y;
    if (p2_2d.x < *xmin) *xmin = p2_2d.x;
    if (p2_2d.x > *xmax) *xmax = p2_2d.x;
    if (p2_2d.y < *ymin) *ymin = p2_2d.y;
    if (p2_2d.y > *ymax) *ymax = p2_2d.y;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, &v_min,
			   CANVAS_ATT_VALUE_MAX, &v_max,
			   CANVAS_ATT_END);
    /*  Shift ray_start to the actual start of the ray: i.e. the first voxel */
    ray_start.x += direction.x * t_enter;
    ray_start.y += direction.y * t_enter;
    ray_start.z += direction.z * t_enter;
    /*  Convert 3-D cursor position to depth along ray (= plane coord)  */
    if ( ( fabs (direction.z) >= fabs (direction.y) ) &&
	 ( fabs (direction.z) >= fabs (direction.x) ) )
    {
	if (direction.z > 0.0) cursor_depth = midpoint_3d.z - ray_start.z;
	else cursor_depth = ray_start.z - midpoint_3d.z;
    }
    else if ( fabs (direction.y) >= fabs (direction.x) )
    {
	if (direction.y > 0.0) cursor_depth = midpoint_3d.y - ray_start.y;
	else cursor_depth = ray_start.y - midpoint_3d.y;
    }
    else
    {
	if (direction.x > 0.0) cursor_depth = midpoint_3d.x - ray_start.x;
	else cursor_depth = ray_start.x - midpoint_3d.x;
    }
    /*  Get pixel canvas information  */
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_DEPTH, &canvas_depth,
			 KWIN_ATT_END);
    if (canvas_depth == 24)
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_RED_MASK, &red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			     KWIN_ATT_END);
    }
    /*  Now have to draw the point using an appropriate shading algorithm  */
    if (strcmp (current_shader_name, "hot_gas_mono") == 0)
    {
	/*  It is simple to process the ray and generate a new pixel  */
	value[0] = shader_hot_gas_mono (ray, ray_length, cursor_depth, v_max);
	value[1] = 0.0;
	if (olist == NULL) canvas_draw_line (canvas, p1_2d.x, p1_2d.y, p2_2d.x,
					     p2_2d.y, value);
	else
	{
	    sprintf (txt, "!%g %g", value[0], value[1]);
	    overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
			  OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, txt);
	}
	return;
    }
    else if ( (strcmp (current_shader_name, "hot_gas_substances") == 0) &&
	      (canvas_depth == 24) )
    {
	/*  It is simple to process the ray and generate a new pixel  */
	pixel_value = shader_hot_gas_subs (ray, ray_length, cursor_depth,
					   &colour,
					   red_mask, green_mask, blue_mask);
	if (olist == NULL) canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y,
					       p2_2d.x, p2_2d.y, pixel_value);
	else
	{
	    sprintf ( txt, "#%04x%04x%04x",
		      (int) (colour.red * MAX_INTENSITY),
		      (int) (colour.green * MAX_INTENSITY),
		      (int) (colour.blue * MAX_INTENSITY) );
	    overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
			  OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, txt);
	}
	return;
    }
    else if ( (strcmp (current_shader_name, "hot_gas_continuous") == 0) &&
	      (canvas_depth == 24) )
    {
	/*  It is simple to process the ray and generate a new pixel  */
	pixel_value = shader_hot_gas_fullcolour (ray, ray_length, cursor_depth,
						 &colour,
						 red_mask, green_mask,
						 blue_mask);
	if (olist == NULL) canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y,
					       p2_2d.x, p2_2d.y, pixel_value);
	else
	{
	    sprintf ( txt, "#%04x%04x%04x",
		      (int) (colour.red * MAX_INTENSITY),
		      (int) (colour.green * MAX_INTENSITY),
		      (int) (colour.blue * MAX_INTENSITY) );
	    overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
			  OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, txt);
	}
	return;
    }
    else if ( (strcmp (current_shader_name, "hot_gas_three_colour") == 0) &&
	      (canvas_depth == 24) )
    {
	/*  It is simple to process the ray and generate a new pixel  */
	pixel_value = shader_hot_gas_fullcolour (ray, ray_length, cursor_depth,
						 &colour,
						 red_mask, green_mask,
						 blue_mask);
	if (olist == NULL) canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y,
					       p2_2d.x, p2_2d.y, pixel_value);
	else
	{
	    sprintf ( txt, "#%04x%04x%04x",
		      (int) (colour.red * MAX_INTENSITY),
		      (int) (colour.green * MAX_INTENSITY),
		      (int) (colour.blue * MAX_INTENSITY) );
	    overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
			  OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, txt);
	}
	return;
    }
    /*  It is not possible to merge the cursor into the data: we have to settle
	for a simple visible/not visible test.  */
    /*  Check if point is visible  */
    /*  First get depth of feature user sees  */
    feature_depth = get_feature_depth (ray, ray_length, pixel,
				       current_shader_name, NULL);
    if (feature_depth > (unsigned int) cursor_depth) visible = TRUE;
    /*  Check if pixel value is considered sufficient to obscure cursor  */
    if (strcmp (current_shader_name, "max_voxel") == 0)
    {
	/*  Pixel must be at least 25% of the image(canvas) range  */
	if ( *(signed char *) pixel < (int) (v_min + (v_max - v_min) / 4.0) )
	{
	    visible = TRUE;
	}
    }
    else if (strcmp (current_shader_name, "min_voxel") == 0)
    {
	/*  Pixel must be no more than 75% of the image(canvas) range  */
	if ( *(signed char *) pixel > (int) (v_min + (v_max - v_min) * 0.75) )
	{
	    visible = TRUE;
	}
    }
    if (visible)
    {
	/*  Cursor is indeed visible  */
	if (olist == NULL) canvas_draw_line_p (canvas, p1_2d.x, p1_2d.y,
					       p2_2d.x, p2_2d.y, pixel_value);
	else overlay_line (olist, OVERLAY_COORD_WORLD, p1_2d.x, p1_2d.y,
			   OVERLAY_COORD_WORLD, p2_2d.x, p2_2d.y, colourname);
	return;
    }
}   /*  End Function draw_shaded_segment  */

static void draw_shaded_point (KWorldCanvas canvas,
			       KVolumeRenderContext context,
			       unsigned int eye_view, CONST char *colourname,
			       Kcoord_3d point,
			       double *xmin, double *ymin,
			       double *xmax, double *ymax)
/*  [PURPOSE] This routine will draw a shaded point over a volume rendered
    image.
    <canvas> The canvas.
    <context> The volume rendering context.
    <eye_view> The eye which will see the point. Legal values are:
        VRENDER_EYE_CYCLOPS    VRENDER_EYE_LEFT    VRENDER_EYE_RIGHT
    <pixel_value> The pixel value to draw in. Used for some shaders.
    <colourname> The name of the colour to draw in.
    <xmin> The minimum 2D horizontal co-ordinate. This is modified.
    <ymin> The minimum 2D vertical co-ordinate. This is modified.
    <xmax> The maximum 2D horizontal co-ordinate. This is modified.
    <ymax> The maximum 2D vertical co-ordinate. This is modified.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    Kcoord_2d p_2d;
    Kcoord_3d ray_start, direction;
    flag visible = FALSE;
    unsigned int ray_length, feature_depth;
    unsigned int canvas_depth;
    unsigned long red_mask, green_mask, blue_mask;
    unsigned long pixel_value;
    RGBcolour colour;
    float cursor_depth, t_enter, t_leave;
    double v_min, v_max;
    double value[2];
    CONST signed char *ray;
    char pixel[sizeof (double)];
    extern CONST char *current_shader_name;

    if ( !get_colour (canvas, colourname, &pixel_value, &colour) ) return;
    /*  Project 3-D point to image plane  */
    vrender_project_3d (context, eye_view, point, &p_2d, FALSE);
    /*  Get the ray from the projected point  */
    if ( ( ray = vrender_collect_ray (context, eye_view, p_2d,
				      &ray_start, &direction,
				      &t_enter, &t_leave, &ray_length) )
	 == NULL )
    {
/*
	fprintf (stderr, "Error collecting ray: %f  %f\n",
			p_2d.x, p_2d.y);
*/
	return;
    }
    /*  Expand range as required  */
    if (p_2d.x < *xmin) *xmin = p_2d.x;
    if (p_2d.x > *xmax) *xmax = p_2d.x;
    if (p_2d.y < *ymin) *ymin = p_2d.y;
    if (p_2d.y > *ymax) *ymax = p_2d.y;
    canvas_get_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, &v_min,
			   CANVAS_ATT_VALUE_MAX, &v_max,
			   CANVAS_ATT_END);
    /*  Shift ray_start to the actual start of the ray: i.e. the first voxel */
    ray_start.x += direction.x * t_enter;
    ray_start.y += direction.y * t_enter;
    ray_start.z += direction.z * t_enter;
    /*  Convert 3-D cursor position to depth along ray (= plane coord)  */
    if ( ( fabs (direction.z) >= fabs (direction.y) ) &&
	 ( fabs (direction.z) >= fabs (direction.x) ) )
    {
	if (direction.z > 0.0) cursor_depth = point.z - ray_start.z;
	else cursor_depth = ray_start.z - point.z;
    }
    else if ( fabs (direction.y) >= fabs (direction.x) )
    {
	if (direction.y > 0.0) cursor_depth = point.y - ray_start.y;
	else cursor_depth = ray_start.y - point.y;
    }
    else
    {
	if (direction.x > 0.0) cursor_depth = point.x - ray_start.x;
	else cursor_depth = ray_start.x - point.x;
    }
    /*  Get pixel canvas information  */
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_get_attributes (pixcanvas,
			 KWIN_ATT_DEPTH, &canvas_depth,
			 KWIN_ATT_END);
    if (canvas_depth == 24)
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_RED_MASK, &red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			     KWIN_ATT_END);
    }
    /*  Now have to draw the point using an appropriate shading algorithm  */
    if (strcmp (current_shader_name, "hot_gas_mono") == 0)
    {
	/*  It is simple to process the ray and generate a new pixel  */
	value[0] = shader_hot_gas_mono (ray, ray_length, cursor_depth, v_max);
	value[1] = 0.0;
#ifdef THICK_CURSOR
	canvas_fill_rectangle (canvas, p_2d.x - 0.5, p_2d.y - 0.5, 1.0, 1.0,
			       value);
#else
	canvas_draw_point (canvas, p_2d.x, p_2d.y, value);
#endif
	return;
    }
    else if ( (strcmp (current_shader_name, "hot_gas_substances") == 0) &&
	      (canvas_depth == 24) )
    {
	/*  It is simple to process the ray and generate a new pixel  */
	pixel_value = shader_hot_gas_subs (ray, ray_length, cursor_depth,
					   &colour,
					   red_mask, green_mask, blue_mask);
#ifdef THICK_CURSOR
	canvas_fill_rectangle_p (canvas,p_2d.x - 0.5, p_2d.y - 0.5, 1.0, 1.0,
				 pixel_value);
#else
	canvas_draw_point_p (canvas, p_2d.x, p_2d.y, pixel_value);
#endif
	return;
    }
    else if ( (strcmp (current_shader_name, "hot_gas_continuous") == 0) &&
	      (canvas_depth == 24) )
    {
	/*  It is simple to process the ray and generate a new pixel  */
	pixel_value = shader_hot_gas_fullcolour (ray, ray_length, cursor_depth,
						 &colour,
						 red_mask, green_mask,
						 blue_mask);
#ifdef THICK_CURSOR
	canvas_fill_rectangle_p (canvas,p_2d.x - 0.5, p_2d.y - 0.5, 1.0, 1.0,
				 pixel_value);
#else
	canvas_draw_point_p (canvas, p_2d.x, p_2d.y, pixel_value);
#endif
	return;
    }
    else if ( (strcmp (current_shader_name, "hot_gas_three_colour") == 0) &&
	      (canvas_depth == 24) )
    {
	/*  It is simple to process the ray and generate a new pixel  */
	pixel_value = shader_hot_gas_fullcolour (ray, ray_length, cursor_depth,
						 &colour,
						 red_mask, green_mask,
						 blue_mask);
#ifdef THICK_CURSOR
	canvas_fill_rectangle_p (canvas,p_2d.x - 0.5, p_2d.y - 0.5, 1.0, 1.0,
				 pixel_value);
#else
	canvas_draw_point_p (canvas, p_2d.x, p_2d.y, pixel_value);
#endif
	return;
    }
    /*  It is not possible to merge the cursor into the data: we have to settle
	for a simple visible/not visible test.  */
    /*  Check if point is visible  */
    /*  First get depth of feature user sees  */
    feature_depth = get_feature_depth (ray, ray_length, pixel,
				       current_shader_name, NULL);
    if (feature_depth > (unsigned int) cursor_depth) visible = TRUE;
    /*  Check if pixel value is considered sufficient to obscure cursor  */
    if (strcmp (current_shader_name, "max_voxel") == 0)
    {
	/*  Pixel must be at least 25% of the image(canvas) range  */
	if ( *(signed char *) pixel < (int) (v_min + (v_max - v_min) / 4.0) )
	{
	    visible = TRUE;
	}
    }
    else if (strcmp (current_shader_name, "min_voxel") == 0)
    {
	/*  Pixel must be no more than 75% of the image(canvas) range  */
	if ( *(signed char *) pixel > (int) (v_min + (v_max - v_min) * 0.75) )
	{
	    visible = TRUE;
	}
    }
    if (visible)
    {
	/*  Cursor is indeed visible  */
#ifdef THICK_CURSOR
	canvas_fill_rectangle_p (canvas,p_2d.x - 0.5, p_2d.y - 0.5, 1.0, 1.0,
				 pixel_value);
#else
	canvas_draw_point_p (canvas, p_2d.x, p_2d.y, pixel_value);
#endif
	return;
    }
}   /*  End Function draw_shaded_point  */

static void worldcanvas_refresh_func (KWorldCanvas canvas,
				      int width, int height,
				      struct win_scale_type *win_scale,
				      Kcolourmap cmap, flag cmap_resize,
				      void **info, PostScriptPage pspage)
/*  [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    [RETURNS] Nothing.
*/
{
    cursor_info *c_info;
    extern Kcoord_3d cursor3d, corner3d;
    extern flag byte_cube_valid;
    /*static char function_name[] = "cursor_canvas_refresh_func";*/

    if (!byte_cube_valid) return;
    c_info = (cursor_info *) *info;
    show_3d_cursor (canvas, c_info->context, cursor3d, corner3d,
		    c_info->eye_view, c_info->has_image);
}   /*  End Function worldcanvas_refresh_func  */

static float shader_hot_gas_mono (CONST signed char *ray,
				  unsigned int ray_length,
				  unsigned int cursor_depth,
				  double cursor_intensity)
/*  [PURPOSE] This routine will process a ray, merging in the cursor.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <cursor_depth> The depth down the ray the cursor is located.
    <cursor_intensity> The intensity of the cursor.
    [RETURNS] The pixel.
*/
{
    int voxel;
    int index;
    unsigned int count;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity, intensity;
    float pixel = 0.0;
    float cumulative_transparency = 1.0;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];

    for (count = 0; count < ray_length; ++count)
    {
	if (count == cursor_depth)
	{
	    opacity = 1.0;
	    intensity = cursor_intensity;
	}
	else
	{
	    if ( (voxel = ray[count]) < -127 ) continue;
	    index = voxel + 127;
	    opacity = mono_hg_opacities[index];
	    intensity = mono_hg_intensities[index];
	}
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	pixel += intensity * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold) break;
    }
    return (pixel);
}   /*  End Function shader_hot_gas_mono  */

static unsigned long shader_hot_gas_subs (CONST signed char *ray,
					  unsigned int ray_length,
					  unsigned int cursor_depth,
					  RGBcolour *colour,
					  unsigned long red_mask,
					  unsigned long green_mask,
					  unsigned long blue_mask)
/*  [PURPOSE] This routine will process a ray, merging in the cursor.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <cursor_depth> The depth down the ray the cursor is located.
    <colour> The colour to draw in. This is updated with the merged pixel.
    <red_mask> The red pixel mask.
    <green_mask> The green pixel mask.
    <blue_mask> The blue pixel mask.
    [RETURNS] The pixel.
*/
{
    int voxel;
    int index;
    unsigned int count;
    float transparency;
    float visible_threshold = 0.02;
    float one = 1.0;
    float old_opacity = 0.0;
    float old_red = 0.0;
    float old_green = 0.0;
    float old_blue = 0.0;
    int old_index = -1;
    unsigned long pixel_value;
    float emissivity;
    float red = 0.0;
    float green = 0.0;
    float blue = 0.0;
    float cumulative_transparency = 1.0;
    extern int num_substances;
    extern unsigned int bad_substance_count;
    extern char voxel_to_substance_indices[256];
    extern substance substances[MAX_SUBSTANCES];

    for (count = 0; count < ray_length; ++count)
    {
	if (count == cursor_depth)
	{
	    old_opacity = one;
	    old_red = colour->red;
	    old_green = colour->green;
	    old_blue = colour->blue;
	    old_index = -1;
	}
	else
	{
	    if ( (voxel = ray[count]) < -127 ) continue;
	    if ( (index = voxel_to_substance_indices[voxel + 128])
		 >= num_substances )
	    {
		/*  Outside substance table: skip  */
		++bad_substance_count;
		continue;
	    }
	    if (old_index != index)
	    {
		old_opacity = substances[index].opacity;
		old_red = substances[index].red;
		old_green = substances[index].green;
		old_blue = substances[index].blue;
		old_index = index;
	    }
	}
	transparency = one - old_opacity;
	emissivity = old_opacity * cumulative_transparency;
	red += old_red * emissivity;
	green += old_green * emissivity;
	blue += old_blue * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold) break;
    }
    colour->red = red;
    colour->green = green;
    colour->blue = blue;
    pixel_value = (long) ( (float) red_mask * red ) & red_mask;
    pixel_value |= (long) ( (float) green_mask * green ) & green_mask;
    pixel_value |= (long) ( (float) blue_mask * blue ) & blue_mask;
    return (pixel_value);
}   /*  End Function shader_hot_gas_subs  */

static unsigned long shader_hot_gas_fullcolour (CONST signed char *ray,
						unsigned int ray_length,
						unsigned int cursor_depth,
						RGBcolour *colour,
						unsigned long red_mask,
						unsigned long green_mask,
						unsigned long blue_mask)
/*  [PURPOSE] This routine will process a ray, merging in the cursor.
    <ray> A pointer to the ray data.
    <ray_length> The length of the ray.
    <cursor_depth> The depth down the ray the cursor is located.
    <colour> The colour to draw in. This is updated with the merged pixel.
    <red_mask> The red pixel mask.
    <green_mask> The green pixel mask.
    <blue_mask> The blue pixel mask.
    [RETURNS] The pixel.
*/
{
    int voxel;
    int index = 0;  /*  Initialised to keep compiler happy  */
    unsigned int count;
    unsigned long pixel_value;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float red = 0.0;
    float green = 0.0;
    float blue = 0.0;
    float cumulative_transparency = 1.0;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];

    for (count = 0; count < ray_length; ++count)
    {
	if (count == cursor_depth)
	{
	    opacity = 1.0;
	}
	else
	{
	    if ( (voxel = ray[count]) < -127 ) continue;
	    index = voxel + 127;
	    opacity = fullcolour_hg_opacities[index];
	}
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	if (count == cursor_depth)
	{
	    red += colour->red * emissivity;
	    green += colour->green * emissivity;
	    blue += colour->blue * emissivity;
	}
	else
	{
	    red += fullcolour_hg_reds[index] * emissivity;
	    green += fullcolour_hg_greens[index] * emissivity;
	    blue += fullcolour_hg_blues[index] * emissivity;
	}
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold) break;
    }
    colour->red = red;
    colour->green = green;
    colour->blue = blue;
    pixel_value = (long) ( (float) red_mask * red ) & red_mask;
    pixel_value |= (long) ( (float) green_mask * green ) & green_mask;
    pixel_value |= (long) ( (float) blue_mask * blue ) & blue_mask;
    return (pixel_value);
}   /*  End Function shader_hot_gas_fullcolour  */

static unsigned int get_offset (packet_desc *pack_desc, CONST char *name)
/*  [PURPOSE] This routine will get the offset of an element in a packet.
    <pack_desc> The packet descriptor.
    <name> The name of the element.
    [RETURNS] The offset of the element. On failure the process aborts.
*/
{
    unsigned int index;
    static char function_name[] = "get_offset";

    if ( ( index = ds_f_elem_in_packet (pack_desc, name) ) >=
	 pack_desc->num_elements )
    {
	fprintf (stderr, "Element: \"%s\" not found\n", name);
	a_prog_bug (function_name);
    }
    return ds_get_element_offset (pack_desc, index);
}   /*  End Function get_offset  */

static flag move_cursor_in_face (KVolumeRenderContext main_vrend_ctx, uaddr eye,
				 Kcoord_3d cursor, double x, double y)
/*  [SUMMARY] Move the 3D cursor around one face of the cube.
    <main_vrend_ctx> The volume rendering context.
    <eye> The eye.
    <cursor> The current cursor position.
    <x> The horizontal canvas co-ordinate.
    <y> The vertical canvas co-ordinate.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Kcoord_2d pos_2d;
    Kcoord_3d ray_start, ray_direction, intersection;
    unsigned int ray_length;
    unsigned int face_mask = 0;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    float t_enter, t_leave;

    /*  Project ray from screen co-ordinate to volume  */
    pos_2d.x = x;
    pos_2d.y = y;
    if (vrender_collect_ray (main_vrend_ctx, eye,
			     pos_2d, &ray_start, &ray_direction,
			     &t_enter, &t_leave,
			     &ray_length) == NULL) return (FALSE);
    /*  Compute 3D co-ordinate of the intersection with volume  */
    intersection.x = ray_start.x + t_enter * ray_direction.x;
    intersection.y = ray_start.y + t_enter * ray_direction.y;
    intersection.z = ray_start.z + t_enter * ray_direction.z;
    /*  Determine which face(s) the point lies on  */
    vrender_get_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    if (fabs (intersection.x - (float)x_s) < 0.5) face_mask |= FACE_MASK_MIN_X;
    if (fabs (intersection.x - (float)x_e) < 0.5) face_mask |= FACE_MASK_MAX_X;
    if (fabs (intersection.y - (float)y_s) < 0.5) face_mask |= FACE_MASK_MIN_Y;
    if (fabs (intersection.y - (float)y_e) < 0.5) face_mask |= FACE_MASK_MAX_Y;
    if (fabs (intersection.z - (float)z_s) < 0.5) face_mask |= FACE_MASK_MIN_Z;
    if (fabs (intersection.z - (float)z_e) < 0.5) face_mask |= FACE_MASK_MAX_Z;
    /*  Move the 3D cursor if point is in only one face  */
    switch (face_mask)
    {
      case FACE_MASK_MIN_X:
      case FACE_MASK_MAX_X:
	cursor.y = intersection.y;
	cursor.z = intersection.z;
	change_3d_cursor (cursor, TRUE);
	break;
      case FACE_MASK_MIN_Y:
      case FACE_MASK_MAX_Y:
	cursor.x = intersection.x;
	cursor.z = intersection.z;
	change_3d_cursor (cursor, TRUE);
	break;
      case FACE_MASK_MIN_Z:
      case FACE_MASK_MAX_Z:
	cursor.x = intersection.x;
	cursor.y = intersection.y;
	change_3d_cursor (cursor, TRUE);
	break;
    }
    return (TRUE);
}   /*  End Function move_cursor_in_face  */

static flag move_cursor_with_keys (KVolumeRenderContext main_vrend_ctx,
				   uaddr eye, Kcoord_3d cursor,
				   double x, double y, char key)
/*  [SUMMARY] Move the 3D cursor using keys.
    <main_vrend_ctx> The volume rendering context.
    <eye> The eye.
    <cursor> The current cursor position.
    <x> The horizontal canvas co-ordinate.
    <y> The vertical canvas co-ordinate.
    <key> The key pressed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Kcoord_2d pos_2d;
    Kcoord_3d tmp3d, ray_start, ray_direction;
    Kcoord_3d ras_centre, direction, horizontal;
    int move;
    unsigned int ray_length;
    float t_enter, t_leave, depth;

    switch (key)
    {
      case 'f':
      case 'F':
	move = 1;
	break;
      case 'c':
      case 'C':
	move = -1;
	break;
      case 'm':
      case 'M':
	move = 0;
	break;
      default:
	return (FALSE);
/*
	break;
*/
    }
    /*  Get eye information  */
    vrender_get_eye_info (main_vrend_ctx, eye, &ras_centre, &direction,
			  &horizontal);
    if (move != 0)
    {
	/*  Move 3D cursor forwards or backwards  */
	cursor.x += (float) move * direction.x;
	cursor.y += (float) move * direction.y;
	cursor.z += (float) move * direction.z;
	change_3d_cursor (cursor, TRUE);
	return (TRUE);
    }
    /*  Move 3D cursor to lie behind the mouse cursor, at approximately the
	same depth (if possible). First find distance from 3D cursor to raster
	plane  */
    tmp3d.x = cursor.x - ras_centre.x;
    tmp3d.y = cursor.y - ras_centre.y;
    tmp3d.z = cursor.z - ras_centre.z;
    depth = geom_vector_dot_product (tmp3d, direction);
    /*  Project ray from screen co-ordinate to volume  */
    pos_2d.x = x;
    pos_2d.y = y;
    if (vrender_collect_ray (main_vrend_ctx, eye,
			     pos_2d, &ray_start, &ray_direction,
			     &t_enter, &t_leave,
			     &ray_length) == NULL) return (TRUE);
    /*  Compute new depth down ray for new cursor position. This is done by
	scaling the existing distance from the 3D cursor to the raster plane
	by the inverse of the dot product between the ray direction vector and
	the direction vector to infinity  */
    depth /= geom_vector_dot_product (ray_direction, direction);
    /*  Limit depth down ray so that it lies within the boundaries  */
    if (depth < t_enter) depth = t_enter;
    if (depth > t_leave) depth = t_leave;
    /*  Now can generate new 3D cursor position  */
    cursor.x = ray_start.x + depth * ray_direction.x;
    cursor.y = ray_start.y + depth * ray_direction.y;
    cursor.z = ray_start.z + depth * ray_direction.z;
    change_3d_cursor (cursor, TRUE);
    return (TRUE);
}   /*  End Function move_cursor_with_keys  */

static void initialise_region_control (region_type *control)
/*  [SUMMARY] Initialise a region control structure with some defaults.
    <control> The structure to initialise.
    [RETURNS] Nothing.
*/
{
    static char *turquoise = "turquoise";
    static char *yellow = "yellow";
    static char *red = "red";
    static char *white = "white";

    /*  Bottom box in z0  */
    control->x0y0z0_x1y0z0 = turquoise;
    control->x1y0z0_x1y1z0 = white;
    control->x1y1z0_x0y1z0 = white;
    control->x0y1z0_x0y0z0 = yellow;
    /*  Top box in z1  */
    control->x0y0z1_x1y0z1 = white;
    control->x1y0z1_x1y1z1 = white;
    control->x1y1z1_x0y1z1 = white;
    control->x0y1z1_x0y0z1 = white;
    /*  Join the dots^H^H^H^Hboxes  */
    control->x0y0z0_x0y0z1 = red;
    control->x1y0z0_x1y0z1 = white;
    control->x1y1z0_x1y1z1 = white;
    control->x0y1z0_x0y1z1 = white;
}   /*  End Function initialise_region_control  */

#ifdef not_needed
static int rasterise_line (int *x_arr, int *y_arr,
			   int x0, int y0, int x1, int y1)
/*  [PURPOSE] This routine will rasterise a line into its constituent pixels.
    <x_arr> The array of x co-ordinates will be written here. The length of
    this array must be atleast equal to the sum of the range of x and y values
    plus 2.
    <y_arr> The array of y co-ordinates will be written here. The length of
    this array must be atleast equal to the sum of the range of x and y values
    plus 2.
    <x0> The horizontal co-ordinate of the first point.
    <y0> The vertical co-ordinate of the first point.
    <x1> The horizontal co-ordinate of the second point.
    <y1> The vertical co-ordinate of the second point.
    [RETURNS] The number of points in the line.
*/
{
    int num_points = 0;
    int tmp;
    int sum, inc;
    int x, y, delta_x, delta_y;

    if ( (x1 == x0) && (y1 == y0) )
    {
	x_arr[0] = x0;
	y_arr[0] = y0;
	return (1);
    }
    if (x1 != x0)
    {
	/*  Draw horizontal line  */
	if (x1 < x0)
	{
	    swap (x0, x1, tmp);
	    swap (y0, y1, tmp);
	}
	delta_x = x1 - x0;
	if (y1 < y0)
	{
	    /*  Line goes up  */
	    delta_y = y0 - y1;
	    inc = -1;
	}
	else
	{
	    /*  Line goes down  */
	    delta_y = y1 - y0;
	    inc = 1;
	}
	sum = delta_x / 2;
	for (x = x0, y = y0; x <= x1; ++x, ++num_points)
	{
	    x_arr[num_points] = x;
	    y_arr[num_points] = y;
	    sum += delta_y;
	    while (sum >= delta_x)
	    {
		y += inc;
		sum -= delta_x;
	    }
	}
    }
    if (y1 != y0)
    {
	/*  Draw vertical line  */
	if (y1 < y0)
	{
	    swap (x0, x1, tmp);
	    swap (y0, y1, tmp);
	}
	delta_y = y1 - y0;
	if (x1 < x0)
	{
	    /*  Line goes left  */
	    delta_x = x0 - x1;
	    inc = -1;
	}
	else
	{
	    /*  Line goes right  */
	    delta_x = x1 - x0;
	    inc = 1;
	}
	sum = delta_y / 2;
	for (x = x0, y = y0; y <= y1; ++y, ++num_points)
	{
	    x_arr[num_points] = x;
	    y_arr[num_points] = y;
	    sum += delta_x;
	    while (sum >= delta_y)
	    {
		x += inc;
		sum -= delta_y;
	    }
	}
    }
    return (num_points);
}   /*  End Function rasterise_line  */
#endif  /*  not_needed  */
