/*  cursor_3d_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for controlling the 3D cursor.


    Written by      Richard Gooch   29-AUG-1996: Copied filter menu code from
  main_xt.c

    Updated by      Richard Gooch   2-SEP-1996: Added depth slider.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   16-JUN-1998: Popup cursor and region render
  popups.

    Updated by      Richard Gooch   31-AUG-1998: Added dynamic creation of
  render popup.

    Updated by      Richard Gooch   27-OCT-1998: Took account of change to
  <create_vrender_context>.

    Updated by      Richard Gooch   19-MAR-1999: Took account of change to
  <create_vrender_context>.

    Last updated by Richard Gooch   11-DEC-1999: Removed unnecessary externs.


*/
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <karma_vrender.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_wcs.h>
#include <karma_c.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/Twodpos.h>
/*#include <Xkw/RenderDisplayPopup.h>*/
#include "RenderDisplayPopup.h"
#define NEED_CURSOR_3D_ALTERNATIVES
#include "xraycast_3d.h"


#define MAX_SPHERE_RADIUS 100
#define MAX_DEPTH 1000

/*  Private functions  */
STATIC_FUNCTION (void cursor_3d_cbk, (Widget w, XtPointer client_data,
				      XtPointer call_data) );
STATIC_FUNCTION (void slider_radius_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag cursor_callback,
		 (void *object, void *client1_data, void *call_data,
		  void *client2_data) );
STATIC_FUNCTION (void compute_sphere_sum,
		 (Widget label, iarray cube, Kcoord_3d centre, int radius) );
#if 0
STATIC_FUNCTION (void twodthing_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
#endif
STATIC_FUNCTION (void compute_depth_range,
		 (KVolumeRenderContext main_vrend_ctx,
		  float *min, float *max) );
STATIC_FUNCTION (void view_change_callback,
		 (KVolumeRenderContext context, void **info) );
STATIC_FUNCTION (void slider_depth_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */
static Widget sum_label = NULL;
static double radii_sums[MAX_SPHERE_RADIUS];
static Widget depth_sld = NULL;
/*static Widget twodpos = NULL;*/
static float min_depth = 0.0;
static float max_depth = 1.0;


Widget create_cursor_3d_menu (Widget form, Widget vert, Widget horiz)
/*  [PURPOSE] This routine will create the 3D cursor mode menu.
    <form> The form to place the widget in.
    <vert> The widget to place the menu under.
    <horiz> The widget to place the menu to the right of.
    [RETURNS] The cursor menu.
*/
{
    int count;
    Widget w, menu_btn, popup, radius_sld;
    extern KCallbackList cursor3d_event_list;
    extern KVolumeRenderContext main_vrend_ctx;
    extern int sphere_radius;

    /*  Invalidate sphere sums  */
    for (count = 0; count < MAX_SPHERE_RADIUS; ++count)
	radii_sums[count] = TOOBIG;
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, form,
				 XtNmenuName, "3DcursorMenu",
				 XkwNchoiceName, "3-D CURSOR",
				 XtNfromVert, vert,
				 XtNfromHoriz, horiz,
				 XkwNnumItems, NUM_CURSOR_3D_ALTERNATIVES,
				 XkwNitemStrings, cursor_3d_alternatives,
				 NULL);
    menu_btn = w;
    popup = XtVaCreatePopupShell ("cursor3dwinpopup",
				  topLevelShellWidgetClass, form,
				  XtNtitle, "3D Cursor Control Window",
				  NULL);
    XtAddCallback (menu_btn, XkwNselectCallback, cursor_3d_cbk, popup);
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, popup);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Sphere Radius",
				 XtNfromVert, w,
				 XtNorientation, XtorientVertical,
				 XkwNminimum, 1,
				 XkwNmaximum, MAX_SPHERE_RADIUS,
				 XkwNmodifier, 1,
				 XtNvalue, sphere_radius,
				 XkwNvaluePtr, &sphere_radius,
				 NULL);
    radius_sld = w;
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, "Region sum:                  ",
				 XtNfromVert, w,
				 XtNborderWidth, 0,
				 XtNresizable, TRUE,
				 NULL);
    sum_label = w;
    XtAddCallback (radius_sld, XkwNvalueChangeCallback, slider_radius_cbk,
		   sum_label);
    c_register_callback (&cursor3d_event_list, cursor_callback, sum_label,
			 NULL, FALSE, NULL, FALSE, FALSE);
#ifdef dummy
    w = XtVaCreateManagedWidget ("twodthing", twodposWidgetClass, form,
				 XtNfromVert, w,
				 XtNwidth, 155,
				 XtNheight, 155,
				 NULL);
    twodpos = w;
    XtAddCallback (w, XkwNvalueChangeCallback, twodthing_cbk, NULL);
#endif
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Cursor Depth",
				 XtNfromVert, w,
				 XtNorientation, XtorientVertical,
				 XkwNminimum, 0,
				 XkwNmaximum, MAX_DEPTH,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNshowValue, False,
				 NULL);
    depth_sld = w;
    XtAddCallback (depth_sld, XkwNvalueChangeCallback, slider_depth_cbk,
		   main_vrend_ctx);
    vrender_view_notify_func (main_vrend_ctx, view_change_callback, NULL);
    return (menu_btn);
}   /*  End Function create_cursor_3d_menu  */


/*  Private functions follow  */

static void cursor_3d_cbk (Widget w, XtPointer client_data,XtPointer call_data)
/*  [PURPOSE] This routine is called when the user changes the cursor type.
*/
{
    KVolumeRenderContext cursor_vrend_ctx;
    Widget popup = (Widget) client_data;
    extern KPixCanvas orientation_pixcanvas;
    extern flag verbose;
    extern unsigned int cursor_3d_mode;
    extern Widget image_display;
    static Widget cursor_render_popup = NULL;
    /*static char function_name[] = "cursor_3d_cbk";*/

    cursor_3d_mode = *(int *) call_data;  /*  Must come first  */
    if (cursor_render_popup == NULL)
    {
	cursor_vrend_ctx = create_vrender_context (verbose, FALSE);
	cursor_render_popup =
	    XtVaCreatePopupShell ("cursorSubvolume",
				  renderDisplayPopupWidgetClass,
				  image_display,
				  XkwNvolumeRenderContext, cursor_vrend_ctx,
				  XkwNpreRenderFunction, subvolume_prerender,
				  NULL);
	XtAddCallback (w, XkwNselectCallback, cursor_3d_cbk,
		       cursor_render_popup);
	if (cursor_3d_mode != CURSOR_3D_NONE)
	    XtPopup (cursor_render_popup, XtGrabNone);
    }
    if (cursor_3d_mode == CURSOR_3D_NONE) XtPopdown (popup);
    else XtPopup (popup, XtGrabNone);
    kwin_resize (orientation_pixcanvas, TRUE, 0, 0, 0, 0);
    refresh_display_window ();
}   /*  End Function cursor_3d_cbk  */

static void slider_radius_cbk (Widget w, XtPointer client_data,
			       XtPointer call_data)
/*  This is the slider sphere radius callback.
*/
{
    int radius = *(int *) call_data;
    Widget sum_label = (Widget) client_data;
    extern iarray byte_cube, unfiltered_cube;
    extern Kcoord_3d cursor3d;
    extern unsigned int cursor_3d_mode;

    if (cursor_3d_mode == CURSOR_3D_SHADED_SPHERE)
	change_3d_cursor (cursor3d, FALSE);
    compute_sphere_sum (sum_label,
			(unfiltered_cube == NULL) ? byte_cube :unfiltered_cube,
			cursor3d, radius);
}   /*  End Function slider_radius_cbk   */

static flag cursor_callback (void *object, void *client1_data, void *call_data,
			     void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    Kcoord_3d tmp3d;
    Kcoord_3d ras_centre, direction, horizontal;
    int count;
    float depth;
    Widget sum_label = (Widget) object;
    Kcoord_3d *cursor = (Kcoord_3d *) call_data;
    extern KVolumeRenderContext main_vrend_ctx;
    extern iarray byte_cube, unfiltered_cube;
    extern int sphere_radius;

    /*  Update depth slider and 2D position widget */
    vrender_get_eye_info (main_vrend_ctx, VRENDER_EYE_CHOICE_CYCLOPS,
			  &ras_centre, &direction, &horizontal);
    tmp3d.x = cursor->x - ras_centre.x;
    tmp3d.y = cursor->y - ras_centre.y;
    tmp3d.z = cursor->z - ras_centre.z;
    depth = geom_vector_dot_product (tmp3d, direction) - min_depth;
    depth = depth / (max_depth - min_depth) * (float) MAX_DEPTH;
    XtVaSetValues (depth_sld,
		   XtNvalue, (int) depth,
		   NULL);
    /*  Invalidate sphere sums  */
    for (count = 0; count < MAX_SPHERE_RADIUS; ++count)
	radii_sums[count] = TOOBIG;
    compute_sphere_sum (sum_label,
			(unfiltered_cube == NULL) ? byte_cube :unfiltered_cube,
			*cursor, sphere_radius);
    return (FALSE);
}   /*  End Function cursor_callback  */

static void compute_sphere_sum (Widget label, iarray cube, Kcoord_3d centre,
				int radius)
/*  [SUMMARY] Compute and display sum of data in a sphere.
    <label> The label widget to display to.
    <cube> The cube.
    <centre> The centre of the cube.
    <radius> The radius of the cube.
    [RETURNS] Nothing.
*/
{
    int x, y, z, plane_sq_radius, plane_radius, xrange, xp, yp, zp;
    int xlen, ylen, zlen;
    unsigned int type;
    double value, val, scale, offset, units_scale;
    char txt[STRING_LENGTH], units_str[STRING_LENGTH];
    char format_str[STRING_LENGTH], all_format[STRING_LENGTH];

    iarray_get_data_scaling (cube, &scale, &offset);
    if ( DOUBLE_IS_BLANK (value = radii_sums[radius]) )
    {
	type = iarray_type (cube);
	/*  Compute volume sum  */
	value = 0.0;
	xlen = iarray_dim_length (cube, 2);
	ylen = iarray_dim_length (cube, 1);
	zlen = iarray_dim_length (cube, 0);
	for (z = -radius; z <= radius; ++z)
	{
	    plane_sq_radius = radius * radius - z * z;
	    plane_radius = sqrt (plane_sq_radius);
	    plane_sq_radius = plane_radius * plane_radius;
	    for (y = -plane_radius; y <= plane_radius; ++y)
	    {
		xrange = sqrt (plane_sq_radius - y * y);
		for (x = -xrange; x <= xrange; ++x)
		{
		    xp = x + centre.x;
		    yp = y + centre.y;
		    zp = z + centre.z;
		    if ( (xp < 0) || (xp >= xlen) ) continue;
		    if ( (yp < 0) || (yp >= ylen) ) continue;
		    if ( (zp < 0) || (zp >= zlen) ) continue;
		    if (type == K_FLOAT) val = F3 (cube, zp, yp, xp);
		    else val = B3 (cube, zp, yp, xp);
		    value += val * scale + offset;
		}
	    }
	}
	radii_sums[radius] = value;
    }
    if ( wcs_astro_get_preferred_units ( units_str, format_str, &units_scale,
					 iarray_value_name (cube) ) )
    {
	sprintf (all_format, "Region sum: %s %s", format_str, units_str);
	sprintf (txt, all_format, value * units_scale);
    }
    else sprintf ( txt, "Region sum: %e %s", value, iarray_value_name (cube) );
    XtVaSetValues (label,
		   XtNlabel, txt,
		   NULL);
}   /*  End Function compute_sphere_sum  */

#if 0
static void twodthing_cbk (Widget w, XtPointer client_data,XtPointer call_data)
{
    TwodposCallbackPtr data = (TwodposCallbackPtr) call_data;

    fprintf (stderr, "pos: %e %e\n", data->value_x, data->value_y);
}   /*  End Function twodthing_cbk  */
#endif

static void compute_depth_range (KVolumeRenderContext main_vrend_ctx,
				 float *min, float *max)
/*  [SUMMARY] Compute range of depths from raster plane.
    <main_vrend_ctx> The volume rendering context.
    <min> The minimum depth is written here.
    <max> The maximum depth is written here.
    [RETURNS] Nothing.
*/
{
    Kcoord_3d ras_centre, direction, horizontal, tmp3d;
    unsigned int count;
    unsigned long x_s, y_s, z_s, x_e, y_e, z_e;
    float depth;
    Kcoord_3d points[8];

    vrender_get_eye_info (main_vrend_ctx, VRENDER_EYE_CHOICE_CYCLOPS,
			  &ras_centre, &direction, &horizontal);
    vrender_get_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_START, &x_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_START, &y_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_START, &z_s,
				    VRENDER_CONTEXT_ATT_SUBCUBE_X_END, &x_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Y_END, &y_e,
				    VRENDER_CONTEXT_ATT_SUBCUBE_Z_END, &z_e,
				    VRENDER_CONTEXT_ATT_END);
    points[0].x = x_s;
    points[0].y = y_s;
    points[0].z = z_s;
    points[1].x = x_e;
    points[1].y = y_s;
    points[1].z = z_s;
    points[2].x = x_e;
    points[2].y = y_e;
    points[2].z = z_s;
    points[3].x = x_s;
    points[3].y = y_s;
    points[3].z = z_s;
    points[4].x = x_s;
    points[4].y = y_s;
    points[4].z = z_e;
    points[5].x = x_e;
    points[5].y = y_s;
    points[5].z = z_e;
    points[6].x = x_e;
    points[6].y = y_e;
    points[6].z = z_e;
    points[7].x = x_s;
    points[7].y = y_e;
    points[7].z = z_e;
    *min = TOOBIG;
    *max = -TOOBIG;
    for (count = 0; count < 8; ++count)
    {
	tmp3d.x = points[count].x - ras_centre.x;
	tmp3d.y = points[count].y - ras_centre.y;
	tmp3d.z = points[count].z - ras_centre.z;
	depth = geom_vector_dot_product (tmp3d, direction);
	if (depth < *min) *min = depth;
	if (depth > *max) *max = depth;
    }
}   /*  End Function compute_depth_range  */

static void view_change_callback (KVolumeRenderContext context, void **info)
/*  [SUMMARY] View changed callback.
    [PURPOSE] This routine is called when the view is changed.
    <context> The volume render context.
    <info> A pointer to the arbitrary information pointer.
    [RETURNS] Nothing.
*/
{
    compute_depth_range (context, &min_depth, &max_depth);
}   /*  End Function view_change_callback  */

static void slider_depth_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the slider cursor depth callback.
*/
{
    KVolumeRenderContext main_vrend_ctx = (KVolumeRenderContext) client_data;
    Kcoord_3d ras_centre, direction, horizontal, pos, tmp3d;
    float depth = (float) *(int *) call_data;
    float old_depth;
    extern Kcoord_3d cursor3d;

    /*  Change the depth of the 3D cursor. To make things simple, find the
	existing depth and move cursor from that old position  */
    vrender_get_eye_info (main_vrend_ctx, VRENDER_EYE_CHOICE_CYCLOPS,
			  &ras_centre, &direction, &horizontal);
    tmp3d.x = cursor3d.x - ras_centre.x;
    tmp3d.y = cursor3d.y - ras_centre.y;
    tmp3d.z = cursor3d.z - ras_centre.z;
    old_depth = geom_vector_dot_product (tmp3d, direction);
    depth = min_depth + depth * (max_depth - min_depth) / (float) MAX_DEPTH;
    pos.x = cursor3d.x + (depth - old_depth) * direction.x;
    pos.y = cursor3d.y + (depth - old_depth) * direction.y;
    pos.z = cursor3d.z + (depth - old_depth) * direction.z;
    change_3d_cursor (pos, TRUE);
}   /*  End Function slider_depth_cbk   */
