/*  display.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all image display code.


    Written by      Richard Gooch   26-DEC-1994

    Updated by      Richard Gooch   27-DEC-1994

    Updated by      Richard Gooch   7-JAN-1995: Added timing information.

    Updated by      Richard Gooch   15-JAN-1995: Added test for
  XRAYCAST_WRITE_SUNRAS environment variable.

    Updated by      Richard Gooch   9-JUL-1995: Added movie support.

    Updated by      Richard Gooch   30-AUG-1995: Changed to
  <canvas_set_attributes> to rescale canvas.

    Updated by      Richard Gooch   21-FEB-1996: Supported <inhibit_images>
  flag.

    Updated by      Richard Gooch   3-MAR-1996: Made raw image buffers global.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   8-DEC-1996: Added restriction information
  to ViewableImages. Required for proper overlay support.

    Updated by      Richard Gooch   10-DEC-1996: Called <overlay_orientation>
  inside <render>, as this is currently the best place to do so. May create
  a callback list for post-rendering operations (with information given about
  whether stereo was displayed or not).

    Updated by      Richard Gooch   11-DEC-1996: Created the post-render event
  list <post_render_event_list> now that v1.6 is released.

    Updated by      Richard Gooch   29-MAY-1997: Made use of show_timings flag.

    Updated by      Richard Gooch   3-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   7-MAY-1998: No longer allocate right movie
  buffer until needed.

    Updated by      Richard Gooch   8-OCT-1998: Comment fixes.

    Updated by      Richard Gooch   27-OCT-1998: No longer use show_timings
  flag.

    Updated by      Richard Gooch   20-FEB-1999: Switched to global variable
  <compute_time>.

    Updated by      Richard Gooch   19-MAY-1999: Support RGB visuals for
  PseudoColour data.

    Updated by      Richard Gooch   10-SEP-1999: Fixed bug in <post_render>:
  was passing static restriction data rather than allocated data.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <karma.h>
#include <karma_vrender.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_m.h>
#include <karma_a.h>
#include <karma_c.h>
#include "xraycast_3d.h"
#include "interpol.h"


#define IARRAY_MAGIC_NUMBER 939032982

/*  Private structures  */
typedef struct
{
    flag pseudocolour, did_stereo;
    double min, max;
    char *left_ptr, *right_ptr;
    unsigned int num_restr;
    unsigned long exp_image_size;
    array_desc *raw_arr_desc;
    char **restr_names;
    double *restr_values;
    
} render_type;


/*  Private functions  */
STATIC_FUNCTION (void post_render, (render_type *myinfo) );
STATIC_FUNCTION (void do_expansion,
		 (char *exp_image, array_desc *exp_arr_desc,
		  char *raw_image, array_desc *raw_arr_desc,
		  int image_expand_factor) );


/*  Private data  */
static char *movie_right_buffer = NULL;
static struct timeval start_time;


/*  Public routines follow  */

flag render ()
/*  [PURPOSE] This routine renders the cube.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int pack_size, n_frames;
    unsigned int raw_image_width, raw_image_height;
    unsigned int exp_image_width, exp_image_height;
    unsigned long raw_image_size;
    extern KWorldCanvas left_image_pc_worldcanvas;
    extern KWorldCanvas left_image_rgb_worldcanvas;
    extern KVolumeRenderContext main_vrend_ctx;
    extern flag stereo;
    extern flag running_movie;
    extern flag verbose, show_timings;
    extern int num_frames_to_compute;
    extern int image_expand_factor;
    extern unsigned int num_frames;
    extern unsigned int movie_buf_len;
    extern unsigned int bad_substance_count;
    extern unsigned int raycast_mode;
    extern double movie_min, movie_max;
    extern char *raw_left_buffer, *raw_right_buffer;
    extern char *movie_left_buffer;
    extern array_desc *exp_arr_desc;
    static unsigned int raw_image_buf_size = 0;  /*  In bytes  */
    static render_type myinfo;
    static char function_name[] = "render";

    if (verbose) fprintf (stderr, "num_frames: %d\n", num_frames);
    bad_substance_count = 0;
    /*  Prepare for computing extent of cube  */
    vrender_get_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_IMAGE_DESC,
				    &myinfo.raw_arr_desc,
				    VRENDER_CONTEXT_ATT_END);
    if (myinfo.raw_arr_desc == NULL)
    {
	fprintf (stderr, "No image descriptor!\n");
	a_prog_bug (function_name);
    }
    raw_image_width = myinfo.raw_arr_desc->dimensions[1]->length;
    raw_image_height = myinfo.raw_arr_desc->dimensions[0]->length;
    exp_image_width = raw_image_width * image_expand_factor;
    exp_image_height = raw_image_height * image_expand_factor;
    /*  Assume if <exp_arr_desc> is defined it is valid. Any changes which
	could change the array descriptor should invalidate <exp_arr_desc>.
	Need to be careful that this is done. Increasing the image expand
	factor is a special case, though.
	*/
    if (exp_arr_desc != NULL)
    {
	/*  Check this just in case. Should never happen.  */
	if ( !ds_compare_packet_desc (myinfo.raw_arr_desc->packet,
				      exp_arr_desc->packet, TRUE) )
	{
	    fprintf (stderr, "Packet descriptors do not match!\n");
	    a_prog_bug (function_name);
	}
	if ( (exp_arr_desc->dimensions[0]->length != exp_image_height) ||
	     (exp_arr_desc->dimensions[1]->length != exp_image_width) )
	{
	    destroy_all_vimages ();
	    ds_dealloc_array_desc (exp_arr_desc);
	    exp_arr_desc = NULL;
	}
    }
    if (exp_arr_desc == NULL)
    {
	if ( ( exp_arr_desc = ds_copy_array_desc_until (myinfo.raw_arr_desc,
							NULL) ) == NULL )
	{
	    m_abort (function_name, "expanded array descriptor");
	}
	exp_arr_desc->dimensions[0]->length = exp_image_height;
	exp_arr_desc->dimensions[1]->length = exp_image_width;
	exp_arr_desc->lengths[0] = exp_image_height;
	exp_arr_desc->lengths[1] = exp_image_width;
    }
    /*  Ensure final output buffer is large enough  */
    if (running_movie)
    {
	n_frames = (num_frames_to_compute < 1) ? 1 : num_frames_to_compute;
	if ( ( myinfo.restr_names = m_alloc (sizeof *myinfo.restr_names) )
	     == NULL )
	{
	    m_abort (function_name, "array of restriction names");
	}
	if ( ( myinfo.restr_names[0] = st_dup ("Frame Number") ) == NULL )
	{
	    m_abort (function_name, "restriction name");
	}
	if ( ( myinfo.restr_values = m_alloc (sizeof *myinfo.restr_values) )
	     == NULL )
	{
	    m_abort (function_name, "array of restriction values");
	}
	*myinfo.restr_values = num_frames;
	myinfo.num_restr = 1;
    }
    else
    {
	n_frames = 1;
	destroy_all_vimages ();
	movie_min = TOOBIG;
	movie_max = -TOOBIG;
	myinfo.num_restr = 0;
	myinfo.restr_names = NULL;
	myinfo.restr_values = NULL;
    }
    pack_size = ds_get_packet_size (myinfo.raw_arr_desc->packet);
    /*  Ensure raw image buffer is large enough  */
    raw_image_size = pack_size * raw_image_width * raw_image_height;
    myinfo.exp_image_size = pack_size * exp_image_width * exp_image_height;
    if (raw_image_buf_size < raw_image_size)
    {
	if (raw_left_buffer != NULL) m_free (raw_left_buffer);
	if ( ( raw_left_buffer = m_alloc (raw_image_size) ) == NULL )
	{
	    m_abort (function_name, "raw left image buffer");
	}
	m_clear (raw_left_buffer, raw_image_buf_size);
	if ( ( raw_right_buffer = m_alloc (raw_image_size) ) == NULL )
	{
	    m_abort (function_name, "raw right image buffer");
	}
	m_clear (raw_right_buffer, raw_image_buf_size);
	raw_image_buf_size = raw_image_size;
    }
    /*  Ensure expanded (final) output buffer is large enough  */
    if (movie_buf_len < myinfo.exp_image_size * n_frames)
    {
	if (movie_left_buffer != NULL) m_free (movie_left_buffer);
	if (movie_right_buffer != NULL)
	{
	    m_free (movie_right_buffer);
	    movie_right_buffer = NULL;
	}
	movie_buf_len = myinfo.exp_image_size * n_frames;
	if ( ( movie_left_buffer = m_alloc (movie_buf_len) ) == NULL )
	{
	    m_abort (function_name, "left image/movie buffer");
	}
	if (stereo)
	{
	    if ( ( movie_right_buffer = m_alloc (movie_buf_len) ) == NULL )
	    {
		m_abort (function_name, "right image/movie buffer");
	    }
	}
	if (verbose) fprintf (stderr, "Allocated %smovie buffer: %u bytes%s\n",
			      stereo ? "stereo " : "", movie_buf_len,
			      stereo ? " each" : "");
    }
    /*  Do the actual rendering  */
    if (image_expand_factor > 1)
    {
	/*  Image will be expanded: write rendered image to a temporary buffer
	    and the expand into the final buffer  */
	myinfo.left_ptr = raw_left_buffer;
	myinfo.right_ptr = raw_right_buffer;
    }
    else
    {
	/*  Image is 1:1 so render directly into the final buffer  */
	myinfo.left_ptr = movie_left_buffer +
	    myinfo.exp_image_size * num_frames;
	myinfo.right_ptr = movie_right_buffer +
	    myinfo.exp_image_size * num_frames;
    }
    if (!stereo) myinfo.right_ptr = NULL;
    myinfo.did_stereo = stereo;
    if (myinfo.raw_arr_desc->packet->num_elements == 1)
    {
	myinfo.pseudocolour = TRUE;
	if (left_image_pc_worldcanvas == NULL)
	{
	    myinfo.right_ptr = NULL;
	    myinfo.did_stereo = FALSE;
	}
    }
    else
    {
	myinfo.pseudocolour = FALSE;
	if ( (left_image_rgb_worldcanvas == NULL) &&
	     (left_image_pc_worldcanvas == NULL) )
	{
	    myinfo.right_ptr = NULL;
	    myinfo.did_stereo = FALSE;
	}
    }
    /*  Have to do this locally in order to include time to compute smooth
	cache  */
    if (show_timings)
    {
	if (gettimeofday (&start_time, NULL) != 0)
	{
	    fprintf (stderr, "%s: error getting time of day\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
    if (raycast_mode == MODE_SMOOTH_ALWAYS)
    {
	if (myinfo.right_ptr == NULL)
	{
	    vrender_compute_caches (main_vrend_ctx, VRENDER_EYE_MASK_CYCLOPS,
				    FALSE);
	}
	else vrender_compute_caches (main_vrend_ctx,
				     VRENDER_EYE_MASK_LEFT |
				     VRENDER_EYE_MASK_RIGHT,
				     FALSE);
    }
    if ( !vrender_to_buffer (main_vrend_ctx, myinfo.left_ptr, myinfo.right_ptr,
			     &myinfo.min, &myinfo.max,
			     ( void (*) () ) post_render, &myinfo) )
    {
	fprintf (stderr, "Error rendering\n");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function render  */

static void post_render (render_type *myinfo)
/*  [PURPOSE] This routine is called after the cube is rendered.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    ViewableImage vimage;
    KWorldCanvas left_wc, right_wc;
    flag ok;
    post_render_info render_info;
    struct timeval finish_time;
    extern KWorldCanvas cyclops_image_pc_worldcanvas;
    extern KWorldCanvas cyclops_image_rgb_worldcanvas;
    extern KWorldCanvas left_image_pc_worldcanvas, right_image_pc_worldcanvas;
    extern KWorldCanvas left_image_rgb_worldcanvas,right_image_rgb_worldcanvas;
    extern KCallbackList post_render_event_list;
    extern flag running_movie;
    extern flag verbose, inhibit_images, show_timings;
    extern ViewableImage vimage_left, vimage_right;
    extern int image_expand_factor;
    extern unsigned int num_frames;
    extern long compute_time;
    extern double movie_min, movie_max;
    extern ViewableImage *vimages_left, *vimages_right;
    extern char *raw_left_buffer, *raw_right_buffer;
    extern char *movie_left_buffer;
    extern array_desc *exp_arr_desc;
    static char function_name[] = "post_render";

    if (show_timings)
    {
	if (gettimeofday (&finish_time, NULL) != 0)
	{
	    fprintf (stderr, "%s: error getting time of day\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	compute_time = (finish_time.tv_sec - start_time.tv_sec) * 1000;
	compute_time += (finish_time.tv_usec - start_time.tv_usec) / 1000;
	fprintf (stderr, "Compute time: %ld ms\n", compute_time);
    }
    if (myinfo->min < movie_min) movie_min = myinfo->min;
    if (myinfo->max > movie_max) movie_max = myinfo->max;
    if (myinfo->pseudocolour)
    {
	if (movie_min >= movie_max)
	{
	    fprintf (stderr, "WARNING: min: %e not less than max: %e\n",
		     movie_min, movie_max);
	    fprintf (stderr, "Setting to 0.0 and 1.0 respectively\n");
	    movie_min = 0.0;
	    movie_max = 1.0;
	}
	else if (verbose) fprintf (stderr, "min: %e\tmax: %e\n",
				   movie_min, movie_max);
    }
    /*  Make new viewable image  */
    if (image_expand_factor > 1)
    {
	/*  Need to use Tom's expansion routines  */
	myinfo->left_ptr = movie_left_buffer +
	    myinfo->exp_image_size * num_frames;
	myinfo->right_ptr = movie_right_buffer +
	    myinfo->exp_image_size * num_frames;
	m_clear (myinfo->left_ptr, myinfo->exp_image_size);
	if (myinfo->did_stereo) m_clear (myinfo->right_ptr,
					 myinfo->exp_image_size);
	do_expansion (myinfo->left_ptr, exp_arr_desc, raw_left_buffer,
		      myinfo->raw_arr_desc, image_expand_factor);
	if (myinfo->did_stereo)
	{
	    do_expansion (myinfo->right_ptr, exp_arr_desc, raw_right_buffer,
			  myinfo->raw_arr_desc, image_expand_factor);
	}
	else myinfo->right_ptr = NULL;
   }
    /*  At this point myinfo->left_ptr and myinfo->right_ptr point into their
	respective movie buffers  */
#ifdef DUMMY
    /*  HACK to brighten green component of 24 bit images and movies. Did this
	for Bob Sault and the Jupiter data  */
    if ( !myinfo->pseudocolour &&
	 ( !running_movie || ( running_movie &&
			       (num_frames == num_frames_to_compute - 1) ) ) )
    {
	int i_max, val;
	unsigned int count, num_values, inc;
	uaddr off;
	unsigned char *base_ptr, *ptr;

	if (running_movie)
	{
	    base_ptr = (unsigned char *) movie_left_buffer;
	    num_values = exp_image_width * exp_image_height * (num_frames + 1);
	    inc = 3;
	}
	else
	{
	    base_ptr = (unsigned char *) myinfo->left_ptr;
	    num_values = exp_image_width * exp_image_height;
	    inc = 3;
	}
#  ifdef USE_MAX
	/*  Determine maximum green component  */
	ptr = base_ptr + 1;
	for (count = 0, i_max = 0; count < num_values; ++count, ptr += inc)
	{
	    val = *ptr;
	    if (val > i_max) i_max = val;
	}
	fprintf (stderr, "max component: %d\n", i_max);
#  else
	i_max = 85;
#  endif
	/*  Now brighten green component of image  */
	ptr = base_ptr + 1;
	for (count = 0; count < num_values; ++count, ptr += inc)
	{
	    val = (int) *ptr * 255 / i_max;
	    if (val > 255) val = 255;
	    *ptr = val;
	}
    }
#endif
    /*  Determine canvas to use and select visual  */
    if (myinfo->did_stereo)
    {
	if (myinfo->pseudocolour)
	{
	    if (left_image_pc_worldcanvas == NULL)
	    {
		select_visual (left_image_rgb_worldcanvas);
		left_wc = left_image_rgb_worldcanvas;
		right_wc = right_image_rgb_worldcanvas;
	    }
	    else
	    {
		select_visual (left_image_pc_worldcanvas);
		left_wc = left_image_pc_worldcanvas;
		right_wc = right_image_pc_worldcanvas;
	    }
	}
	else
	{
	    if (left_image_rgb_worldcanvas == NULL)
	    {
		select_visual (left_image_pc_worldcanvas);
		left_wc = left_image_pc_worldcanvas;
		right_wc = right_image_pc_worldcanvas;
	    }
	    else
	    {
		select_visual (left_image_rgb_worldcanvas);
		left_wc = left_image_rgb_worldcanvas;
		right_wc = right_image_rgb_worldcanvas;
	    }
	}
    }
    else
    {
	/*  Monoscopic  */
	right_wc = NULL;
	if (myinfo->pseudocolour)
	{
	    if (cyclops_image_pc_worldcanvas == NULL)
	    {
		select_visual (cyclops_image_rgb_worldcanvas);
		left_wc = cyclops_image_rgb_worldcanvas;
	    }
	    else
	    {
		select_visual (cyclops_image_pc_worldcanvas);
		left_wc = cyclops_image_pc_worldcanvas;
	    }
	}
	else
	{
	    if (cyclops_image_rgb_worldcanvas == NULL)
	    {
		select_visual (cyclops_image_pc_worldcanvas);
		left_wc = cyclops_image_pc_worldcanvas;
	    }
	    else
	    {
		select_visual (cyclops_image_rgb_worldcanvas);
		left_wc = cyclops_image_rgb_worldcanvas;
	    }
	}
    }
    /*  First draw the right image if required  */
    if (myinfo->did_stereo)
    {
	/*  Create the ViewableImage  */
	if (myinfo->pseudocolour)
	{
	    if ( ( vimage =
		   viewimg_create_restr (right_wc, NULL, exp_arr_desc,
					 myinfo->right_ptr, 1, 0, 0,
					 myinfo->num_restr,
					 (CONST char **) myinfo->restr_names,
					 myinfo->restr_values) ) == NULL )
	    {
		m_abort (function_name, "ViewableImage");
	    }
	    canvas_set_attributes (right_wc,
				   CANVAS_ATT_VALUE_MIN, movie_min,
				   CANVAS_ATT_VALUE_MAX, movie_max,
				   CANVAS_ATT_END);
	}
	else
	{
	    /*  TrueColour image  */
	    if ( ( vimage =
		   viewimg_create_rgb (right_wc, NULL, exp_arr_desc,
				       myinfo->right_ptr, 1, 0,
				       0, 1, 2,
				       myinfo->num_restr,
				       (CONST char **) myinfo->restr_names,
				       myinfo->restr_values) ) == NULL )
	    {
		m_abort (function_name, "ViewableImage");
	    }
	}
	if (!inhibit_images) viewimg_set_active (vimage, TRUE);
	/*  Remove old ViewableImage and save new ViewableImage  */
	if (running_movie)
	{
	    if (vimages_right[num_frames] != NULL)
	    {
		viewimg_destroy (vimages_right[num_frames]);
	    }
	    vimages_right[num_frames] = vimage;
	}
	else
	{
	    if (vimage_right != NULL) viewimg_destroy (vimage_right);
	    vimage_right = vimage;
	}
    }
    /*  Now draw the left image  */
    if (myinfo->pseudocolour)
    {
	if ( ( vimage =
	       viewimg_create_restr (left_wc, NULL, exp_arr_desc,
				     myinfo->left_ptr, 1, 0, 0,
				     myinfo->num_restr,
				     (CONST char **) myinfo->restr_names,
				     myinfo->restr_values) ) == NULL )
	{
	    m_abort (function_name, "ViewableImage");
	}
	canvas_set_attributes (left_wc,
			       CANVAS_ATT_VALUE_MIN, movie_min,
			       CANVAS_ATT_VALUE_MAX, movie_max,
			       CANVAS_ATT_END);
    }
    else
    {
	/*  TrueColour image  */
	if ( ( vimage =
	       viewimg_create_rgb (left_wc, NULL, exp_arr_desc,
				   myinfo->left_ptr, 1, 0,
				   0, 1, 2,
				   myinfo->num_restr,
				   (CONST char **) myinfo->restr_names,
				   myinfo->restr_values) ) == NULL )
	{
	    m_abort (function_name, "ViewableImage");
	}
    }
    if (running_movie)
    {
	if (vimages_left[num_frames] != NULL)
	{
	    viewimg_destroy (vimages_left[num_frames]);
	}
	vimages_left[num_frames] = vimage;
    }
    else
    {
	if (vimage_left != NULL) viewimg_destroy (vimage_left);
	vimage_left = vimage;
    }
    freeze_controls (FALSE);
    ok = TRUE;
    if (!inhibit_images) ok = viewimg_set_active (vimage, TRUE);
    if (!inhibit_images && ok)
    {
	render_info.did_stereo = myinfo->did_stereo;
	c_call_callbacks (post_render_event_list, &render_info);
    }
}   /*  End Function post_render  */

void destroy_all_vimages ()
/*  [PURPOSE] This routine will destroy all ViewableImages and clear the
    appropriate pointers.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    extern ViewableImage vimage_left, vimage_right;
    extern unsigned int num_frames;
    extern ViewableImage *vimages_left, *vimages_right;

    for (count = 0; count < num_frames; ++count)
    {
	if ( (vimages_left != NULL) && (vimages_left[count] != NULL) )
	{
	    viewimg_destroy (vimages_left[count]);
	    vimages_left[count] = NULL;
	}
	if ( (vimages_right != NULL) && (vimages_right[count] != NULL) )
	{
	    viewimg_destroy (vimages_right[count]);
	    vimages_right[count] = NULL;
	}
    }
    if (vimage_left != NULL) viewimg_destroy (vimage_left);
    vimage_left = NULL;
    if (vimage_right != NULL) viewimg_destroy (vimage_right);
    vimage_right = NULL;
    num_frames = 0;
}   /*  End Function destroy_all_vimages  */


/*  Private routines follow  */

static void do_expansion (char *exp_image, array_desc *exp_arr_desc,
			  char *raw_image, array_desc *raw_arr_desc,
			  int image_expand_factor)
/*  [PURPOSE] This routine will expand an image using anti-aliasing.
    <exp_image> A pointer to storage for the expanded image.
    <exp_arr_desc> The expanded array descriptor.
    <raw_image> A pointer to the input image.
    <raw_arr_desc> The original array descriptor.
    <image_expand_factor> The image expand factor.
    [RETURNS] Nothing.
*/
{
    unsigned int elem_count;
    flag contiguous[2];
    extern flag verbose;
    static iarray inp = NULL;
    static iarray out = NULL;
    static char function_name[] = "do_expansion";

    if (raw_arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (raw_arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    if (exp_arr_desc->offsets == NULL)
    {
	if (verbose) fprintf (stderr, "computing exp offsets...\n");
	if ( !ds_compute_array_offsets (exp_arr_desc) )
	{
	    fprintf (stderr, "Error computing address offsets\n");
	    a_prog_bug (function_name);
	}
    }
    /*  What follows are immoral hacks  */
    if (inp == NULL)
    {
	if ( ( inp = (iarray) m_alloc (sizeof *inp) ) == NULL )
	{
	    m_abort (function_name, "input iarray");
	}
    }
    if (out == NULL)
    {
	if ( ( out = (iarray) m_alloc (sizeof *out) ) == NULL )
	{
	    m_abort (function_name, "output iarray");
	}
    }
    contiguous[0] = TRUE;
    contiguous[1] = TRUE;
    /*  Fill input structure  */
    inp->offsets = raw_arr_desc->offsets;
    inp->lengths = raw_arr_desc->lengths;  /*  Wrong if tiled!  */
    inp->contiguous = contiguous;
    inp->top_pack_desc = NULL;
    inp->top_packet = NULL;
    inp->multi_desc = NULL;
    inp->arr_desc = raw_arr_desc;
    inp->array_num = 0;
    inp->boundary_width = 0;
    inp->num_dim = 2;
    inp->orig_dim_indices = NULL;
    inp->restrictions = NULL;
    inp->magic_number = IARRAY_MAGIC_NUMBER;
    inp->destroy_callbacks = NULL;
    /*  Fill output structure  */
    out->offsets = exp_arr_desc->offsets;
    out->lengths = exp_arr_desc->lengths;  /*  Wrong if tiled!  */
    out->contiguous = contiguous;
    out->top_pack_desc = NULL;
    out->top_packet = NULL;
    out->multi_desc = NULL;
    out->arr_desc = exp_arr_desc;
    out->array_num = 0;
    out->boundary_width = 0;
    out->num_dim = 2;
    out->orig_dim_indices = NULL;
    out->restrictions = NULL;
    out->magic_number = IARRAY_MAGIC_NUMBER;
    out->destroy_callbacks = NULL;
    for (elem_count = 0; elem_count < raw_arr_desc->packet->num_elements;
	 ++elem_count)
    {
	inp->data = raw_image + ds_get_element_offset (raw_arr_desc->packet,
						       elem_count);
	inp->elem_index = elem_count;
	out->data = exp_image + ds_get_element_offset (exp_arr_desc->packet,
						       elem_count);
	out->elem_index = elem_count;
	switch ( iarray_type (inp) )
	{
	  case K_FLOAT:
	    if ( !iarray_expand_float (inp, out, image_expand_factor,
				       INTER_EDGE_SAME) ) return;
	    break;
	  case K_BYTE:
	    if ( !iarray_expand_byte (inp, out, image_expand_factor,
				      INTER_EDGE_SAME) ) return;
	    break;
	  case K_UBYTE:
	    if ( !iarray_expand_ubyte (inp, out, image_expand_factor,
				       INTER_EDGE_SAME) ) return;
	    break;
	  default:
	    fprintf ( stderr, "Data type: %u not supported\n",
		      iarray_type (inp) );
	    a_prog_bug (function_name);
	    break;
	}
    }
}   /*  End Function do_expansion  */
