/*  geom.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-1996  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file provides some simple geometry functions.


    Written by      Richard Gooch   9-FEB-1994

    Updated by      Richard Gooch   9-FEB-1994

    Updated by      Richard Gooch   21-APR-1995: Moved function declarations to
  xraycast_3d.h

    Updated by      Richard Gooch   17-JUL-1995: Moved over to coord_3d type.

    Updated by      Richard Gooch   27-JUL-1995: Created <geom_vector_multiply>
  with left-hand rule.

    Last updated by Richard Gooch   17-DEC-1995: Moved over to Kcoord_3d type.


*/

#include <stdio.h>
#include <math.h>
#include "xraycast_3d.h"


void geom_rotate_3d_r (Kcoord_3d *vector, float roll)
/*  This routine will rotate a vector through roll (+x axis).
    The vector must be pointed to by  vector  .
    The roll angles (in radians) must be given by  roll  .
    The routine returns nothing.
*/
{
    float y, z;
    float sin_roll, cos_roll;

    y = vector->y;
    z = vector->z;
    sin_roll = sin (roll);
    cos_roll = cos (roll);
    vector->y = y * cos_roll - z * sin_roll;
    vector->z = z * cos_roll + y * sin_roll;
}   /*  End Function geom_rotate_3d_r  */

void geom_rotate_3d_p (Kcoord_3d *vector, float pitch)
/*  This routine will rotate a vector through pitch (+y axis).
    The vector must be pointed to by  vector  .
    The pitch angles (in radians) must be given by  pitch  .
    The routine returns nothing.
*/
{
    float x, z;
    float sin_pitch, cos_pitch;

    x = vector->x;
    z = vector->z;
    sin_pitch = sin (pitch);
    cos_pitch = cos (pitch);
    vector->x = x * cos_pitch + z * sin_pitch;
    vector->z = z * cos_pitch - x * sin_pitch;
}   /*  End Function geom_rotate_3d_p  */

void geom_rotate_3d_y (Kcoord_3d *vector, float yaw)
/*  This routine will rotate a vector through yaw (+z axis).
    The vector must be pointed to by  vector  .
    The yaw angles (in radians) must be given by  yaw  .
    The routine returns nothing.
*/
{
    float x, y;
    float sin_yaw, cos_yaw;

    x = vector->x;
    y = vector->y;
    sin_yaw = sin (yaw);
    cos_yaw = cos (yaw);
    vector->x = x * cos_yaw - y * sin_yaw;
    vector->y = y * cos_yaw + x * sin_yaw;
}   /*  End Function geom_rotate_3d_y  */

void geom_rotate_3d_rpy (Kcoord_3d *vector, float roll, float pitch, float yaw)
/*  This routine will rotate a vector through roll, then pitch and then yaw.
    Roll rotates about +x axis.
    Pitch rotates about -y axis.
    Yaw rotates about +z axis.
    In this application: roll should rotate about z axis
                         pitch should rotate about x axis
			 yaw should rotate about y axis
    Hence, call with (vector, pitch, yaw, roll).
    The vector must be pointed to by  vector  .
    The roll, pitch and yaw angles (in radians) must be given by  roll  ,
    pitch  and  yaw  ,respectively.
    The routine returns nothing.
*/
{
    float x, y, z;
    float sin_roll, cos_roll;
    float sin_pitch, cos_pitch;
    float sin_yaw, cos_yaw;

    x = vector->x;
    y = vector->y;
    z = vector->z;
    sin_roll = sin (roll);
    cos_roll = cos (roll);
    sin_pitch = sin (pitch);
    cos_pitch = cos (pitch);
    sin_yaw = sin (yaw);
    cos_yaw = cos (yaw);
/*
    geom_rotate_3d_r (vector, roll);
    geom_rotate_3d_p (vector, pitch);
    geom_rotate_3d_y (vector, yaw);
    return;
*/
    vector->x = (x * cos_pitch + (z * cos_roll + y * sin_roll) *
		 sin_pitch) * cos_yaw - (y * cos_roll - z *
					 sin_roll) * sin_yaw;
    vector->y = (y * cos_roll - z * sin_roll) * cos_yaw + (x * cos_pitch +
							   (z * cos_roll + y *
							    sin_roll) *
							   sin_pitch) *sin_yaw;
    vector->z = (z * cos_roll + y * sin_roll) * cos_pitch - x * sin_pitch;
}   /*  End Function geom_rotate_3d_rpy  */

void geom_project_3d_ortho_2d (Kcoord_3d eye_pos, Kcoord_3d vision,
			       Kcoord_3d vertical, Kcoord_3d point_3d,
			       Kcoord_2d *point_2d)
/*  This routine will perform an orthogonal projection of a point in 3
    dimensional space to a view plane.
    The position of the eye (which must lie in the view plane) must be pointed
    to by  eye_pos  .
    The direction the eye is looking at must be pointed to by  vision  .
    The vector defining the vertical of the view plane must be pointed to by
    vertical  .
    The 3 dimensional point to be projected must be pointed to by  point_3d  .
    The projected 2 dimensional point will be written to the storage pointed
    to by  point_2d  .
    The routine returns nothing.
*/
{
    Kcoord_3d horizontal;
    Kcoord_3d direction;

    geom_vector_multiply (&horizontal, vision, vertical);
    direction.x = point_3d.x - eye_pos.x;
    direction.y = point_3d.y - eye_pos.y;
    direction.z = point_3d.z - eye_pos.z;
    point_2d->x = (horizontal.x * direction.x +
		   horizontal.y * direction.y +
		   horizontal.z * direction.z);
    point_2d->x /= sqrt (horizontal.x * horizontal.x +
			 horizontal.y * horizontal.y +
			 horizontal.z * horizontal.z);
    point_2d->y = (vertical.x * direction.x +
		   vertical.y * direction.y +
		   vertical.z * direction.z);
    point_2d->y /= sqrt (vertical.x * vertical.x +
			 vertical.y * vertical.y +
			 vertical.z * vertical.z);
}   /*  End Function geom_project_3d_ortho_2d  */

void geom_rect_to_polar (float x, float y, float z,
			 float *rho, float *theta, float *phi)
/*  This routine will convert rectangular to polar co-ordinates.
*/
{
    *rho = sqrt (x * x + y * y + z * z);
    *theta = atan2 (y, x);
    *phi = asin (z / *rho);
}   /*  End Function geom_rect_to_polar  */

void geom_polar_to_rect (float rho, float theta, float phi,
			 float *x, float *y, float *z)
/*  This routine will convert polar to rectangular co-ordinates.
*/
{
    float k;

    k = rho * cos (phi);
    *z = rho * sin (phi);
    *x = k * cos (theta);
    *y = k * sin (theta);
}   /*  End Function geom_polar_to_rect  */

void geom_rotate_in_plane (Kcoord_3d *vector, float angle, Kcoord_3d normal)
/*  This routine will rotate a vector in a plane.
    The vector must be pointed to by  vector  .This value is modified.
    The angle to rotate by must be given by  angle  .
    The normal to the plane must be pointed to by  normal  .
    The routine returns nothing.
*/
{ 
    float n_rho, n_theta, n_phi;

    geom_rect_to_polar (normal.x, normal.y, normal.z,
			&n_rho, &n_theta, &n_phi);
    /*  Rotate vector so that normal would become +x axis  */
    geom_rotate_3d_y (vector, -n_theta);
    geom_rotate_3d_p (vector, n_phi);
    if ( (vector->x < -1e-2) || (vector->x > 1e-2) )
    {
	(void) fprintf (stderr, "bad x: %e\n", vector->x);
    }
    vector->x = 0.0;
    geom_rotate_3d_r (vector, angle);
    /*  Rotate new vector back  */
    geom_rotate_3d_p (vector, -n_phi);
    geom_rotate_3d_y (vector, n_theta);
}   /*  End Function geom_rotate_in_plane  */

void geom_vector_multiply (Kcoord_3d *out, Kcoord_3d vec1, Kcoord_3d vec2)
/*  [PURPOSE] This routine will multiply two vectors together, using the left
    hand rule, to produce another mutually perpendicular vector.
    <out> The new vector will be written here.
    <vec1> One of the input vectors.
    <vec2> The other input vector.
    [RETURNS] Nothing.
*/
{
    out->x = vec1.z * vec2.y - vec1.y * vec2.z;
    out->y = vec1.x * vec2.z - vec1.z * vec2.x;
    out->z = vec1.y * vec2.x - vec1.x * vec2.y;
}   /*  End Function geom_vector_multiply  */

float geom_vector_dot_product (Kcoord_3d vec1, Kcoord_3d vec2)
/*  [PURPOSE] This routine will compute the dot product of two vectors.
    <vec1> One of the input vectors.
    <vec2> The other input vector.
    [RETURNS] The dot product.
*/
{
    return (vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z);
}   /*  End Function geom_vector_dot_product  */
