/*  initialise.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the initialisation code.


    Written by      Richard Gooch   22-DEC-1995

    Updated by      Richard Gooch   3-JAN-1996

    Updated by      Richard Gooch   17-MAR-1996: Fixed initial eye seperation
  value by scaling appropriately.

    Updated by      Richard Gooch   6-MAY-1996: Added support for
  "hot_gas_three_colour" shader.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   28-NOV-1996: Took account of substance
  changes.

    Updated by      Richard Gooch   8-DEC-1996: Initialised orientation overlay
  lists.

    Updated by      Richard Gooch   11-DEC-1996: Made use of new post-render
  callback list for orientation (wireframe) display and segments display.

    Updated by      Richard Gooch   14-DEC-1996: Added cute "flasher" display
  whenever cache(s) computed.

    Updated by      Richard Gooch   15-DEC-1996: Moved <vrender> support to
  vrender_setup.c

    Updated by      Richard Gooch   27-DEC-1996: Removed read event handler for
  <dsxfr> package.

    Updated by      Richard Gooch   25-MAY-1997: Manipulate global
  <voxel_sum_offset> variable in <register_voxel_offset> function. Created
  <client_notify_func> function. Made use of <vrender_send_message>.

    Updated by      Richard Gooch   9-DEC-1997: Create volume render contexts
  for 3D cursor and 3D region.

    Updated by      Richard Gooch   1-FEB-1998: Made use of new switching
  fabric.

    Updated by      Richard Gooch   11-JUN-1998: No longer update "front_voxel"
  blank packet in <register_voxel_offset>.

    Updated by      Richard Gooch   28-AUG-1998: Moved creation of region
  vrender context to region_xt.c

    Updated by      Richard Gooch   31-AUG-1998: Moved creation of cursor
  vrender context to region_xt.c

    Updated by      Richard Gooch   7-SEP-1998: Remove registration of hog
  request protocol as it's now done in <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   27-OCT-1998: Took account of change to
  <create_vrender_context>.

    Updated by      Richard Gooch   3-DEC-1998: Pass FALSE for <<show_timings>>
  flag passed to <create_vrender_context>.

    Updated by      Richard Gooch   19-MAR-1999: Took account of change to
  <create_vrender_context>.

    Updated by      Richard Gooch   10-SEP-1999: Removed code to become server
  and set up title name.

    Updated by      Richard Gooch   26-JAN-2000: Removed unnecessary includes.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kcmap.h>


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_vrender.h>
#include <karma_overlay.h>
#include <karma_kcmap.h>
#include <karma_dsxfr.h>
#include <karma_conn.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_c.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void image_desc_change_func,
		 (KVolumeRenderContext context, void **info) );
STATIC_FUNCTION (void cache_computed_func,
		 (KVolumeRenderContext context, void **info, uaddr eyes) );
STATIC_FUNCTION (void client_notify_func,
		 (KVolumeRenderContext context, void **info) );


/*  Public routines follow  */

void initialise ()
/*  This routine will initialise the xraycast_3d module.
    The routine returns nothing.
*/
{
    extern KOverlayList histogram_olist;
    extern KOverlayList cyclops_olist;
    extern KOverlayList left_olist;
    extern KOverlayList right_olist;
    extern KCallbackList new_byte_cube_event_list;
    extern KCallbackList post_render_event_list, subcube_event_list;
    extern KCallbackList rotation_event_list, shader_event_list;
    extern KCallbackList projection_event_list, eye_separation_event_list;
    extern flag verbose;
    extern KVolumeRenderContext main_vrend_ctx;
    extern array_desc *exp_arr_desc;
    static char function_name[] = "xray_initialise";

    /*  Register the protocols  */
    /*  Support a shitload of protocols  */
    dsxfr_register_connection_limits (1, 0);
    conn_register_server_protocol ("3D_cursor_position", 0, 0,
				   ( flag (*) () ) NULL, ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("cont_hot_gas_info", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    conn_register_client_protocol ("mono_hot_gas_info", 0, 0,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL,
				   ( void (*) () ) NULL);
    if ( ( histogram_olist =
	   overlay_va_create_list (NULL, "histogramOverlayList",
				   OVERLAY_ATT_END) ) == NULL )
    {
	m_abort (function_name, "overlay list");
    }
    if ( ( cyclops_olist = overlay_va_create_list (NULL, "cyclopsOverlayList",
						   OVERLAY_ATT_END) ) == NULL )
    {
	m_abort (function_name, "overlay list");
    }
    if ( ( left_olist = overlay_va_create_list (NULL, "leftOverlayList",
						OVERLAY_ATT_END) ) == NULL )
    {
	m_abort (function_name, "overlay list");
    }
    if ( ( right_olist = overlay_va_create_list (NULL, "rightOverlayList",
						 OVERLAY_ATT_END) ) == NULL )
    {
	m_abort (function_name, "overlay list");
    }
    c_register_callback (&post_render_event_list,
			 orientation_post_render_callback, NULL,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&post_render_event_list,
			 segments_post_render_callback, NULL,
			 NULL, FALSE, NULL, FALSE, FALSE);
    initialise_vrender_package ();
    main_vrend_ctx = create_vrender_context (verbose, FALSE);
    vrender_image_desc_notify_func (main_vrend_ctx,
				    image_desc_change_func,
				    &exp_arr_desc);
    vrender_cache_notify_func (main_vrend_ctx, cache_computed_func, NULL);
    vrender_client_notify_func (main_vrend_ctx, client_notify_func, NULL);
    c_register_callback (&new_byte_cube_event_list, generic_new_cube_callback,
			 main_vrend_ctx, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&subcube_event_list,
			 generic_subcube_callback, main_vrend_ctx,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&rotation_event_list,
			 generic_rotation_callback, main_vrend_ctx,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&shader_event_list,
			 generic_shader_callback, main_vrend_ctx,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&shader_event_list,
			 ( flag (*) () ) compute, NULL,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&projection_event_list,
			 generic_projection_callback, main_vrend_ctx,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&eye_separation_event_list,
			 generic_eye_separation_callback, main_vrend_ctx,
			 NULL, FALSE, NULL, FALSE, FALSE);
    kcmap_initialise ();
}   /*  End Function initialise  */

void register_voxel_offset (signed char offset)
/*  [PURPOSE] This routine will register a new voxel offset to use for blank
    values.
    <offset> The voxel offset.
    [RETURNS] Nothing.
*/
{
    extern KVolumeRenderContext main_vrend_ctx;
    extern int voxel_sum_position;

    voxel_sum_position = offset;
    vrender_change_shader_blank_packet ("max_voxel", (char *) &offset);
    vrender_change_shader_blank_packet ("min_voxel", (char *) &offset);
    vrender_send_message (main_vrend_ctx, MtoS_VOXEL_OFFSET,
			  (CONST char *) &offset, 1, FALSE);
}   /*  End Function register_voxel_offset  */


/*  Private routines follow  */

static void image_desc_change_func (KVolumeRenderContext context, void **info)
/*  [PURPOSE] This routine is called when the output image descriptor is
    changed.
    <context> The volume render context.
    <info> A pointer to the arbitrary information pointer.
    [RETURNS] Nothing.
*/
{
    array_desc **arr_desc = (array_desc **) *info;

    if (*arr_desc == NULL) return;
    destroy_all_vimages ();
    ds_dealloc_array_desc (*arr_desc);
    *arr_desc = NULL;
}   /*  End Function image_desc_change_func  */

static void cache_computed_func (KVolumeRenderContext context, void **info,
				 uaddr eyes)
/* [PURPOSE] This routine is called when the cache for one or more eye
   views is computed.
   <context> The volume render context.
   <info> A pointer to the arbitrary information pointer.
   <eyes> A bitmask specifying which eye views to compute. The bitmask may
   be constructed as the OR of the following values:
       VRENDER_EYE_MASK_CYCLOPS     VRENDER_EYE_MASK_LEFT
       VRENDER_EYE_MASK_RIGHT
   [RETURNS] Nothing.
*/
{
    extern flag verbose;
    extern unsigned int raycast_mode;
    static char flash_ch = '-';

    if (verbose)
    {
	fputs ("Cache computed for:", stderr);
	if (eyes & VRENDER_EYE_MASK_CYCLOPS) fputs (" cyclops", stderr);
	if (eyes & VRENDER_EYE_MASK_LEFT) fputs (" left eye", stderr);
	if (eyes & VRENDER_EYE_MASK_RIGHT) fputs (" right eye", stderr);
	putc ('\t', stderr);
	putc (flash_ch, stderr);
	putc ('\n', stderr);
	flash_ch = (flash_ch == '-') ? '|' : '-';
    }
    if (raycast_mode == MODE_SMOOTH_DELAYED)
    {
	if (verbose) fputs ("Delayed smooth render...\n", stderr);
	render ();
    }
}   /*  End Function cache_computed_func  */

static void client_notify_func (KVolumeRenderContext context, void **info)
/*  [SUMMARY] Client connected callback.
    [PURPOSE] This routine is called when a new client connects to the master.
    <context> The volume render context.
    <info> A pointer to the arbitrary information pointer.
    [RETURNS] Nothing.
*/
{
    extern int voxel_sum_position;

    vrender_send_message (context, MtoS_VOXEL_OFFSET,
			  (CONST char *) &voxel_sum_position, 1, FALSE);
    transmit_fullcolour_info (context);
    transmit_mono_hg_info (context);
    transmit_substances (context);
}   /*  End Function client_notify_func  */
