/*  load_disp.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains code to load images and movies.


    Written by      Richard Gooch   10-JUL-1995

    Updated by      Richard Gooch   30-AUG-1995: Changed to
  <canvas_set_attributes> to refresh canvas.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kcmap.h>


*/
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_kcmap.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include "xraycast_3d.h"


/*  Private data  */


/*  Public functions follow  */

flag display_file (CONST char *inp_filename, KWorldCanvas pseudo_canvas,
		   KWorldCanvas rgb_canvas)
/*  [PURPOSE] This routine will load a file and display it.
    <inp_filename> The name of the file to load.
    <pseudo_canvas> The PseudoColour canvas.
    <rgb_canvas> The TrueColour or DirectColour canvas. May be NULL.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    Kcolourmap cmap;
    iarray image_pseudo, image_red, image_green, image_blue;
    iarray movie_pseudo, movie_red, movie_green, movie_blue;
    unsigned int cmap_index, ftype;
    double i_min, i_max;
    multi_array *multi_desc;
    extern ViewableImage vimage_left;
    extern unsigned int num_frames, vimage_buf_len;
    extern ViewableImage *vimages_left, *vimages_right;
    static char function_name[] = "display_file";

    if ( ( multi_desc = foreign_guess_and_read (inp_filename, K_CH_MAP_LOCAL,
						FALSE, &ftype,
						FA_GUESS_READ_END) ) == NULL )
    {
	(void) fprintf (stderr, "Error reading file: \"%s\"\n", inp_filename);
	return (FALSE);
    }
    if (rgb_canvas == NULL) rgb_canvas = pseudo_canvas;
    /*  Try to get 2-D image  */
    if ( iarray_get_image_from_multi (multi_desc, &image_pseudo,
				      &image_red, &image_green, &image_blue,
				      &cmap_index) )
    {
	destroy_all_vimages ();
	if (image_pseudo == NULL)
	{
	    if ( ( vimage_left = viewimg_create_rgb (rgb_canvas, multi_desc,
						     image_red->arr_desc,
						     image_red->data, 1, 0,
						     image_red->elem_index,
						     image_green->elem_index,
						     image_blue->elem_index,
						     0, (CONST char **) NULL,
						     (double *) NULL) )
		== NULL )
	    {
		(void) fprintf (stderr,
				"Error getting ViewableImage from Iarray\n");
		ds_dealloc_multi (multi_desc);
		iarray_dealloc (image_red);
		iarray_dealloc (image_green);
		iarray_dealloc (image_blue);
		return (FALSE);
	    }
	    iarray_dealloc (image_red);
	    iarray_dealloc (image_green);
	    iarray_dealloc (image_blue);
	    if ( !viewimg_make_active (vimage_left) )
	    {
		(void) fprintf (stderr, "Error making ViewableImage active\n");
		ds_dealloc_multi (multi_desc);
		viewimg_destroy (vimage_left);
		vimage_left = NULL;
		return (FALSE);
	    }
	}
	else
	{
	    if ( ( vimage_left = viewimg_create_from_iarray (pseudo_canvas,
							     image_pseudo,
							     FALSE) ) == NULL )
	    {
		(void) fprintf (stderr,
				"Error getting ViewableImage from Iarray\n");
		ds_dealloc_multi (multi_desc);
		iarray_dealloc (image_pseudo);
		return (FALSE);
	    }
	    cmap = canvas_get_cmap (pseudo_canvas);
	    if ( (cmap != NULL) && (cmap_index < multi_desc->num_arrays) )
	    {
		if ( !kcmap_copy_from_struct (cmap,
					      multi_desc->headers[cmap_index],
					      multi_desc->data[cmap_index]) )
		{
		    (void) fprintf (stderr, "Error changing colourmap\n");
		}
	    }
	    if ( !iarray_min_max (image_pseudo, CONV1_REAL, &i_min, &i_max) )
	    {
		(void) fprintf (stderr, "Error computing min-max\n");
	    }
	    iarray_dealloc (image_pseudo);
	    canvas_set_attributes (pseudo_canvas,
				   CANVAS_ATT_VALUE_MIN, i_min,
				   CANVAS_ATT_VALUE_MAX, i_max,
				   CANVAS_ATT_END);
	    if ( !viewimg_make_active (vimage_left) )
	    {
		(void) fprintf (stderr, "Error making ViewableImage active\n");
		ds_dealloc_multi (multi_desc);
		viewimg_destroy (vimage_left);
		vimage_left = NULL;
		return (FALSE);
	    }
	}
	if (image_pseudo == NULL) select_visual (rgb_canvas);
	else select_visual (pseudo_canvas);
	ds_dealloc_multi (multi_desc);
	return (TRUE);
    }
    /*  Try to get movie  */
    if ( !iarray_get_movie_from_multi (multi_desc, &movie_pseudo,
				       &movie_red, &movie_green, &movie_blue,
				       &cmap_index) )
    {
	(void) fprintf (stderr, "Error getting movie\n");
	ds_dealloc_multi (multi_desc);
	return (FALSE);
    }
    destroy_all_vimages ();
    if (vimages_left != NULL) m_free ( (char *) vimages_left );
    vimages_left = NULL;
    if (vimages_right != NULL) m_free ( (char *) vimages_right );
    vimages_right = NULL;
    vimage_buf_len = 0;
    if (movie_pseudo == NULL)
    {
	if ( ( vimages_left =
	      viewimg_create_rgb_sequence (rgb_canvas, multi_desc,
					   movie_red->arr_desc,
					   movie_red->data,
					   2, 1, 0,
					   movie_red->elem_index,
					   movie_green->elem_index,
					   movie_blue->elem_index,
					   0, (CONST char **) NULL,
					   (CONST double *) NULL) )
	    == NULL )
	{
	    (void) fprintf (stderr, "Error creating ViewableImage sequence\n");
	    ds_dealloc_multi (multi_desc);
	    iarray_dealloc (movie_red);
	    iarray_dealloc (movie_green);
	    iarray_dealloc (movie_blue);

	    return (FALSE);
	}
	num_frames = iarray_dim_length (movie_red, 0);
	iarray_dealloc (movie_red);
	iarray_dealloc (movie_green);
	iarray_dealloc (movie_blue);
    }
    else
    {
	if ( ( vimages_left =
	      viewimg_create_sequence (pseudo_canvas, multi_desc,
				       movie_pseudo->arr_desc,
				       movie_pseudo->data,
				       2, 1, 0, movie_pseudo->elem_index) )
	    == NULL )
	{
	    (void) fprintf (stderr, "Error creating ViewableImage sequence\n");
	    ds_dealloc_multi (multi_desc);
	    iarray_dealloc (movie_pseudo);
	    return (FALSE);
	}
	num_frames = iarray_dim_length (movie_pseudo, 0);
	cmap = canvas_get_cmap (pseudo_canvas);
	if ( (cmap != NULL) && (cmap_index < multi_desc->num_arrays) )
	{
	    if ( !kcmap_copy_from_struct (cmap,
					  multi_desc->headers[cmap_index],
					  multi_desc->data[cmap_index]) )
	    {
		(void) fprintf (stderr, "Error changing colourmap\n");
	    }
	    i_min = 0;
	    i_max = kcmap_get_pixels (cmap, (unsigned long **) NULL) - 1;
	}
	else
	{
	    (void)fprintf(stderr,"Computing minimum and maximum of cube...\n");
	    if ( !iarray_min_max (movie_pseudo, CONV1_REAL, &i_min, &i_max) )
	    {
		(void) fprintf (stderr, "Error computing min-max\n");
	    }
	    (void) fprintf (stderr, "Minimum: %e  maximum: %e\n",i_min, i_max);
	}
	iarray_dealloc (movie_pseudo);
	/*  Set intensity scale (since autoscaling is not on)  */
	canvas_set_attributes (pseudo_canvas,
			       CANVAS_ATT_VALUE_MIN, i_min,
			       CANVAS_ATT_VALUE_MAX, i_max,
			       CANVAS_ATT_END);
    }
    ds_dealloc_multi (multi_desc);
    if ( ( vimages_right = (ViewableImage *) m_alloc
	  (sizeof *vimages_right * num_frames) ) == NULL )
    {
	m_abort (function_name, "right ViewableImage array");
    }
    m_clear ( (char *) vimages_right, sizeof *vimages_right * num_frames );
    vimage_buf_len = num_frames;
    if ( !viewimg_make_active (vimages_left[0]) )
    {
        (void) fprintf (stderr, "Error making ViewableImage active\n");
        return (FALSE);
    }
    if (movie_pseudo == NULL) select_visual (rgb_canvas);
    else select_visual (pseudo_canvas);
    return (TRUE);
}   /*  End Function display_file  */
