/*  movie_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 2000-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the code necessary for controlling movie production.


    Written by      Richard Gooch   12-JUN-2000: Moved from main_xt.c

    Updated by      Richard Gooch   12-JUN-2000: Use post-render callback to
  prevent computing new frame before previous one is finished computing.

    Last updated by Richard Gooch   14-FEB-2004: Added <save_ppm_movie_cbk>.


*/
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <errno.h>
#include <sys/time.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_m.h>
#include <karma_c.h>
#include <Xkw/Value.h>
#include <Xkw/AnimateControl.h>
#include <Xkw/ExportMenu.h>
#include "xraycast_3d.h"


#define ANGLE_RATE_MIN -180
#define ANGLE_RATE_MAX 180


/*  Private functions  */
STATIC_FUNCTION (void start_movie_cbk, (Widget w, XtPointer client_data,
					XtPointer call_data) );
STATIC_FUNCTION (void stop_movie_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (void save_movie_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (void freeze_movie_controls, (flag freeze) );
STATIC_FUNCTION (Boolean next_movie_frame, (XtPointer client_data) );
STATIC_FUNCTION (flag movie_post_render_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (void save_ppm_movie_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );


/*  Private data  */
static int roll_rate = 0;
static int pitch_rate = 0;
static int yaw_rate = 0;
static Widget movie_close_btn = NULL;
static Widget movie_save_btn = NULL;
static Widget roll_rate_sld = NULL;
static Widget pitch_rate_sld = NULL;
static Widget yaw_rate_sld = NULL;
static Widget num_frames_sld = NULL;
static XtAppContext app_context = NULL;


/*  Public functions follow  */

void create_movie_win (Widget main_shell)
{
    Widget form, start_btn, stop_btn;
    Widget dialog;
    extern KCallbackList post_render_event_list;
    extern int num_frames_to_compute;
    extern Widget movie_winpopup;
    /*static char function_name[] = "create_movie_win";*/

    app_context = XtWidgetToApplicationContext (main_shell);
    movie_winpopup = XtVaCreatePopupShell ("moviewinpopup",
					   topLevelShellWidgetClass,
					   main_shell,
					   XtNtitle, "Movie Window",
					   NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass,
				    movie_winpopup,
				    XtNborderWidth, 0,
				    NULL);
    movie_close_btn = XtVaCreateManagedWidget ("closeButton",
					       commandWidgetClass, form,
					       XtNlabel, "Close",
					       NULL);
    XtAddCallback (movie_close_btn, XtNcallback, xtmisc_popdown_cbk,
		   movie_winpopup);
    start_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					 XtNlabel, "Start",
					 XtNfromHoriz, movie_close_btn,
					 NULL);
    XtAddCallback (start_btn, XtNcallback, start_movie_cbk, NULL);
    stop_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					XtNlabel, "Stop",
					XtNfromHoriz, start_btn,
					NULL);
    XtAddCallback (stop_btn, XtNcallback, stop_movie_cbk, NULL);
    movie_save_btn = XtVaCreateManagedWidget ("button",
					      commandWidgetClass, form,
					      XtNlabel, "Save",
					      XtNfromHoriz, stop_btn,
					      NULL);
    dialog = XtVaCreateManagedWidget ("filename", dialogWidgetClass, form,
				      XtNlabel, "Movie Arrayfile name:",
				      XtNborderWidth, 0,
				      XtNfromVert, movie_close_btn,
				      XtNvalue, "movie",
				      XtNresizable, TRUE,
				      NULL);
    XtAddCallback (movie_save_btn, XtNcallback, save_movie_cbk, dialog);
    roll_rate_sld = XtVaCreateManagedWidget ("slider", valueWidgetClass, form,
					     XtNlabel, "Roll Increment",
					     XtNfromVert, dialog,
					     XkwNminimum, ANGLE_RATE_MIN,
					     XkwNmaximum, ANGLE_RATE_MAX,
					     XkwNmodifier, 1,
					     XtNvalue, 0,
					     XkwNvaluePtr, &roll_rate,
					     NULL);
    pitch_rate_sld = XtVaCreateManagedWidget ("slider", valueWidgetClass, form,
					      XtNlabel, "Pitch Increment",
					      XtNfromVert, roll_rate_sld,
					      XkwNminimum, ANGLE_RATE_MIN,
					      XkwNmaximum, ANGLE_RATE_MAX,
					      XkwNmodifier, 1,
					      XtNvalue, 0,
					      XkwNvaluePtr, &pitch_rate,
					      NULL);
    yaw_rate_sld = XtVaCreateManagedWidget ("slider", valueWidgetClass, form,
					    XtNlabel, "Yaw Increment",
					    XtNfromVert, pitch_rate_sld,
					    XkwNminimum, ANGLE_RATE_MIN,
					    XkwNmaximum, ANGLE_RATE_MAX,
					    XkwNmodifier, 1,
					    XtNvalue, 0,
					    XkwNvaluePtr, &yaw_rate,
					    NULL);
    num_frames_sld = XtVaCreateManagedWidget ("slider", valueWidgetClass, form,
					      XtNlabel, "Num of Frames",
					      XtNfromVert, yaw_rate_sld,
					      XkwNminimum, 0,
					      XkwNmaximum, ANGLE_RANGE,
					      XkwNmodifier, 1,
					      XtNvalue, 0,
					      XkwNvaluePtr,
					      &num_frames_to_compute,
					      NULL);
    freeze_save_movie_button (TRUE);
    XtRealizeWidget (movie_winpopup);
    c_register_callback (&post_render_event_list,
			 movie_post_render_callback, app_context,
			 NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function create_movie_win  */

void freeze_save_movie_button (flag freeze)
/*  [PURPOSE] This routine will set the state of the "Save Movie" button.
    <freeze> If TRUE, the button is frozen, else it is unfrozen.
    [RETURNS] Nothing.
*/
{
    Boolean sensitive;
    Widget w;
    extern Widget image_display;
    static flag frozen = FALSE;

    if (freeze == frozen) return;
    sensitive = freeze ? False : True;
    XtSetSensitive (movie_save_btn, sensitive);
    XSync (XtDisplay (movie_save_btn), False);
    frozen = freeze;
    if (!image_display) return;
    w = XtNameToWidget (image_display, "exportMenu");
    if (freeze) XtRemoveCallback (w, XkwNPPMWindowMovieCallback,
				  save_ppm_movie_cbk, NULL);
    else XtAddCallback (w, XkwNPPMWindowMovieCallback, save_ppm_movie_cbk,
			NULL);
}   /*  End Function freeze_save_movie_button  */


/*  Private routines follow  */

static void start_movie_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
{
    extern flag running_movie;
    extern int num_frames_to_compute;
    extern unsigned int num_frames;
    extern unsigned int vimage_buf_len;
    extern double movie_min, movie_max;
    extern ViewableImage *vimages_left, *vimages_right;
    static char function_name[] = "start_movie_cbk";

    if (running_movie) return;
    num_frames = 0;
    freeze_movie_controls (TRUE);
    freeze_save_movie_button (TRUE);
    destroy_all_vimages ();
    running_movie = TRUE;
    movie_min = TOOBIG;
    movie_max = -TOOBIG;
    if (vimage_buf_len < num_frames_to_compute)
    {
	vimage_buf_len = num_frames_to_compute;
	if (vimages_left) m_free (vimages_left);
	if (vimages_right) m_free (vimages_right);
	if ( ( vimages_left =
	       m_calloc (sizeof *vimages_left * vimage_buf_len) ) == NULL )
	    m_abort (function_name, "array of left ViewableImage pointers");
	if ( ( vimages_right =
	       m_calloc (sizeof *vimages_right * vimage_buf_len) ) == NULL )
	    m_abort (function_name, "array of right ViewableImage pointers");
    }
    XtAppAddWorkProc (app_context, next_movie_frame, NULL);
}   /*  End Function start_movie_cbk  */

static void stop_movie_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
{
    extern flag running_movie;

    running_movie = FALSE;
    freeze_movie_controls (FALSE);
    freeze_controls (FALSE);
}   /*  End Function stop_movie_cbk  */

static void save_movie_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
{
    char *arrayfile_name;
    Widget dialog = (Widget) client_data;
    extern int num_frames_to_compute;
    extern char *movie_left_buffer;
    extern array_desc *exp_arr_desc;

    arrayfile_name = XawDialogGetValueString (dialog);
    if (arrayfile_name == NULL) arrayfile_name = "movie";
    if (strlen (arrayfile_name) < 1) arrayfile_name = "movie";
    save_movie (arrayfile_name, exp_arr_desc, movie_left_buffer,
		num_frames_to_compute);
}   /*  End Function save_movie_cbk  */

static void freeze_movie_controls (flag freeze)
/*  [PURPOSE] This routine set the state of the movie controls.
    <freeze> If TRUE, the controls are frozen, else they are unfrozen.
    [RETURNS] Nothing.
*/
{
    
    Boolean sensitive;

    sensitive = freeze ? False : True;
    XtSetSensitive (movie_close_btn, sensitive);
    XtSetSensitive (roll_rate_sld, sensitive);
    XtSetSensitive (pitch_rate_sld, sensitive);
    XtSetSensitive (yaw_rate_sld, sensitive);
    XtSetSensitive (num_frames_sld, sensitive);
    XSync (XtDisplay (movie_close_btn), False);
}   /*  End Function freeze_movie_controls  */

static Boolean next_movie_frame (XtPointer client_data)
/*  [SUMMARY] Compute the next movie frame.
    [RETURNS] True if the work procedure is finished, else False to indicate
    that it should be called again.
*/
{
    Widget animate_winpopup;
    extern flag running_movie;
    extern int num_frames_to_compute;
    extern unsigned int num_frames;
    extern Widget image_display;

    if (num_frames >= num_frames_to_compute)
    {
	running_movie = FALSE;
	freeze_movie_controls (FALSE);
	freeze_controls (FALSE);
	freeze_save_movie_button (FALSE);
	animate_winpopup = XtNameToWidget (image_display, "animatepopup");
	XtVaSetValues (animate_winpopup, XkwNnumFrames, num_frames, NULL);
	show_animate_control (TRUE);
	return (True);
    }
    do_rotate (roll_rate, yaw_rate, pitch_rate);
    return (True);
}   /*  End Function next_movie_frame  */

static flag movie_post_render_callback (void *object, void *client1_data,
					void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    extern flag running_movie;
    extern unsigned int num_frames;

    if (!running_movie) return (FALSE);
    ++num_frames;
    XtAppAddWorkProc (object, next_movie_frame, NULL);
    return (FALSE);
}   /*  End Function movie_post_render_callback  */

static void save_ppm_movie_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
{
    int precision;
    unsigned int count;
    CONST char *basename = call_data;
    char filename[STRING_LENGTH];
    extern unsigned int num_frames;
    extern ViewableImage *vimages_left, *vimages_right;

    precision = ceil ( log10 (num_frames) );
    for (count = 0; count < num_frames; ++count)
    {
	viewimg_set_active (vimages_left[count], FALSE);
	if (vimages_right[count])
	    viewimg_set_active (vimages_right[count], FALSE);
	sprintf (filename, "%s.%.*u.ppm", basename, precision, count);
	ExportMenuWritePPMWindow (w, filename);
    }
}   /*  End Function save_ppm_movie_cbk  */
