/*
 * sblibry.c
 *
 * Spaceball Library Routines
 *
 * Version: L4.05a
 * Release Date:	       29 August 1989
 * Latest modification:        30 August 1989
 *
 * Written by John Hilton, Spatial Systems Inc, May 1988
 *
 * Copyright (C) 1988 Spatial Systems Inc. All Rights Reserved.
 * Spaceball is a registered trademark of Spatial Systems Inc.
 *
 * Refer to the Spaceball Application Developer's Reference (Document
 * Number D1005) for more information on the following code.
 *
 * RESTRICTIONS
 *
 * The following code is licensed for use and distribution provided the
 * code is used exclusively in conjunction with a Spaceball.  No portion
 * may be used for other purposes without prior written permission from
 * Spatial Systems Inc.  This notice must accompany any full or partial
 * copies.
 *
 * PURPOSE
 *
 * This file provides the functionality to convert between Spaceball
 * data formats and application data formats. Every effort has been made
 * to allow simple configuration parameters to be set in sbcustom.h so as
 * to meet a wide variety of applications' requirements.
 *
 * The use of the routines contained herein is strongly recommended to
 * save implementation time and effort and to provide for application
 * save implementation time and effort and to ensure portablility.
 */

#include <stdio.h>
#include <string.h>
#ifndef vms
#  include <memory.h>
#endif
#include <ctype.h>
#include <math.h>
#define Sb_ENVIRONMENT_INCLUDES
#include "sblibry.h"

/*======================= CONSTANT DEFINITIONS ===========================*/

#define 	LF			0xa
#define 	CR			0xd
#define 	XON			0x11
#define 	XOFF			0x13
#define 	FFFCu			((SbUint16) 0xfffc)
#define 	FFFCs			((SbInt16) -4)
#define 	REAL_TO_SB_FACTOR	16384.0
#define 	SB_TO_REAL_FACTOR	6.103515625e-5
#define 	REAL_TO_CS_FACTOR	45812.0
#define 	CS_TO_REAL_FACTOR	2.182834192e-5

#define 	APACKET 		0
#define 	FTPACKET		1
#define 	FRPACKET		2
#define 	MPACKET 		3
#define 	NPACKET 		4
#define 	OPACKET 		5
#define 	PPACKET 		6
#define 	RPACKET 		7
#define 	SPACKET 		8
#define 	TPACKET 		9
#define 	XPACKET 		10
#define 	YXZ			0
#define 	ZXY			1
#define 	XYZ			2
#define 	ZYX			3
#define 	XZY			4
#define 	YZX			5

/* define initialisation values for SbSettings */

#define 	DEFAULT_SETTINGS    {\
    "", "", "", "", "", "", "", "", "", "", ""\
}

/* define initialisation values for matrices */

#if Sb_COLUMNS == 4
# if Sb_ROWS == 4
#  define   IDENTITY_MATRIX {\
	{ 1.0, 0.0, 0.0, 0.0 },\
	{ 0.0, 1.0, 0.0, 0.0 },\
	{ 0.0, 0.0, 1.0, 0.0 },\
	{ 0.0, 0.0, 0.0, 1.0 }\
}
# else
#  define   IDENTITY_MATRIX {\
	{ 1.0, 0.0, 0.0, 0.0 },\
	{ 0.0, 1.0, 0.0, 0.0 },\
	{ 0.0, 0.0, 1.0, 0.0 }\
}
# endif
#else
# if Sb_COLUMNS == 3 && Sb_ROWS == 4
#  define   IDENTITY_MATRIX {\
	{ 1.0, 0.0, 0.0 },\
	{ 0.0, 1.0, 0.0 },\
	{ 0.0, 0.0, 1.0 },\
	{ 0.0, 0.0, 0.0 }\
}
# endif
#endif

/* define initialisation for packet handlers */

#define     DEFAULT_FUNCTIONS	    {\
    SbNullFunction,\
    SbNullFunction,\
    SbNullFunction,\
    SbNullFunction,\
    SbNullFunction,\
    SbProcessEcho,\
    SbProcessError,\
    SbProcessInvalid,\
    SbProcessIRotate,\
    SbProcessITranslate,\
    SbProcessOther,\
    SbNullFunction,\
    SbNullFunction,\
    SbProcessReset,\
    SbNullFunction,\
    SbNullFunction\
}

/* define initialisation values for Spaceball variables */

#define     DEFAULT_VALUES	{\
    "",\
    "",\
    DEFAULT_SETTINGS,\
    0,\
    "",\
    0,\
    "",\
    "",\
    "",\
    0,\
    0,\
    0,\
    0,\
    0,\
    0,\
    2,\
    SbGetPacket,\
    TRUE,\
    TRUE,\
    FALSE,\
    FALSE,\
    FALSE,\
    FALSE,\
    FALSE,\
    FALSE,\
    0,\
    0,\
    0,\
    { 0, 0 },\
    { 0, 0, 0, 0, 0, 0 },\
    0,\
    0.0,\
    { 0.0, 0.0, 0.0 },\
    { 0.0, 0.0, 0.0 },\
    { 0.0, 0.0, 0.0 },\
    IDENTITY_MATRIX,\
    IDENTITY_MATRIX,\
    DEFAULT_FUNCTIONS,\
    Sb_DEFAULT_ENVIRONMENT\
}

/* define POSxx and NEGxx for matrix conversions so as to match the usage
 * of rotation matrices by an application. SbBallMode will also need to be
 * correct. Refer to the Programmers Reference Manual for more detail. */

/* if	  negate row		 xor	    negate column      */
#if ((Sb_MATRIX_CONVERSION & 010) ^ (Sb_MATRIX_CONVERSION & 001))
#define 	POS00			-
#define 	NEG00
#else
#define 	POS00
#define 	NEG00			-
#endif
#if ((Sb_MATRIX_CONVERSION & 010) ^ (Sb_MATRIX_CONVERSION & 001))
#define 	POS01			-
#define 	NEG01
#else
#define 	POS01
#define 	NEG01			-
#endif
#if ((Sb_MATRIX_CONVERSION & 010) ^ (Sb_MATRIX_CONVERSION & 001))
#define 	POS02			-
#define 	NEG02
#else
#define 	POS02
#define 	NEG02			-
#endif
#if ((Sb_MATRIX_CONVERSION & 020) ^ (Sb_MATRIX_CONVERSION & 002))
#define 	POS10			-
#define 	NEG10
#else
#define 	POS10
#define 	NEG10			-
#endif
#if ((Sb_MATRIX_CONVERSION & 020) ^ (Sb_MATRIX_CONVERSION & 002))
#define 	POS11			-
#define 	NEG11
#else
#define 	POS11
#define 	NEG11			-
#endif
#if ((Sb_MATRIX_CONVERSION & 020) ^ (Sb_MATRIX_CONVERSION & 002))
#define 	POS12			-
#define 	NEG12
#else
#define 	POS12
#define 	NEG12			-
#endif
#if ((Sb_MATRIX_CONVERSION & 040) ^ (Sb_MATRIX_CONVERSION & 004))
#define 	POS20			-
#define 	NEG20
#else
#define 	POS20
#define 	NEG20			-
#endif
#if ((Sb_MATRIX_CONVERSION & 040) ^ (Sb_MATRIX_CONVERSION & 004))
#define 	POS21			-
#define 	NEG21
#else
#define 	POS21
#define 	NEG21			-
#endif
#if ((Sb_MATRIX_CONVERSION & 040) ^ (Sb_MATRIX_CONVERSION & 004))
#define 	POS22			-
#define 	NEG22
#else
#define 	POS22
#define 	NEG22			-
#endif
/*======================= MACRO DEFINITONS ===============================*/

#define 	min( v1, v2 )		(v1 <= v2 ? v1 : v2)
#define 	max( v1, v2 )		(v1 >= v2 ? v1 : v2)

#if Sb_ODD_ADDR_OK && !Sb_SWAP_BYTES
#   define	SbConvertWord( w )  *((SbInt16 *) ( w ))
#else
#   define	SbConvertWord( w ) \
    ((SbInt16) (*((SbUchar *) w) << 8) | (SbInt16) *((SbUchar *) ( w ) + 1))
#endif

#define 	error_fp		stderr
#define 	output_fp		stdout

/*======================== GLOBAL VARIABLES ==============================*/

#ifdef vms
#define 	GLOBALDEF		globaldef
#else
#define 	GLOBALDEF
#endif

#ifdef Sb_DEBUG
GLOBALDEF int Sb_debug = 0xffff;
#endif

GLOBALDEF SbMatrixType
  SbIdentityMatrix = IDENTITY_MATRIX;

GLOBALDEF struct SbArrayStruct
#if Sb_NUMBER_OF_SPACEBALLS == 1
  Sb[ 1 ] = { DEFAULT_VALUES };
#else
  SbArray[ Sb_NUMBER_OF_SPACEBALLS ] = { DEFAULT_VALUES },
  *Sb = SbArray;    /* Sb-><variable name> is a run time pointer dereference */
#endif

static int
  request_settings_pending = FALSE;

static int
  debuglength;

static char
  debugpacket[ Sb_MAXPACKET ];

/*======================== SENDING ROUTINES ==============================*/

SbPutChar( c )
char c;

/* all output to Spaceball is passed here */

{
    if (c != '\r') {
	SbOutBuffer[ SbOutIndex++ ] = c;    /* buffer output for efficiency */
	if (SbOutIndex >= Sb_MAXPACKET) {
	    fprintf( error_fp, "sblibry: Packet being sent is too big!\n" );
	    SbOutIndex = Sb_MAXPACKET - 1;
	}
    } else {
	if ((SbRS232Mode[ 0 ] | 0x20) == 'p') {     /* printable mode? */
	    if (SbOutBitShift != 2) {	     /* any bits in SbOutBitBuffer? */
		SbOutBuffer[ SbOutIndex++ ] = SbOutBitBuffer;
		if (SbOutIndex >= Sb_MAXPACKET) {
		    fprintf( error_fp,
			"sblibry: Packet being sent is too big!\n"
		    );
		    SbOutIndex = Sb_MAXPACKET - 1;
		}
	    }
	    SbPacking = FALSE;		/* reset packing flag */
	}
	SbOutBuffer[ SbOutIndex++ ] = '\r';
	SbOutput( SbOutBuffer, SbOutIndex );
	SbOutIndex = 0;
	SbOutBitShift = 2;
    }
}

SbPutStr( s )
char *s;
{
    while (*s)
	SbPutChar( *s++ );
}

SbPutByte( c )
char c;
{
    if ((SbRS232Mode[ 0 ] | 0x20) != 'p')
	switch (c) {
	  case CR:
	  case XON:
	  case XOFF:
	  case '^':
	    SbPutChar( '^' );
	    SbPutChar( c | 0x40 );
	    break;
	  default:
	    SbPutChar( c );
	}
    else if ( !SbPacking )	/* printable ascii mode - not packed yet */
	if (isprint( c ) && (c != '^'))
	    SbPutChar( c );
	else {			/* start packing data */
	    SbPacking = TRUE;
	    SbPutChar( '^' );
	    SbOutBitBuffer = 0x40 | c>>2;
	    SbPutChar( SbOutBitBuffer != 0x7f ? SbOutBitBuffer : 0x3f );
	    SbOutBitShift = 4;
	    SbOutBitBuffer = 0x40 | (c & 3) << 4;
	}
    else {			/* printable ascii mode - packed 6 bits/byte */
	if (( SbOutBitBuffer |= c >> SbOutBitShift) != 0x7f)
	    SbPutChar( SbOutBitBuffer );
	else
	    SbPutChar( 0x3f );	   /* instead of DEL */

	if ((SbOutBitShift += 2) >= 8) {
	    if ((SbOutBitBuffer = c | 0x40) != 0x7f)
		SbPutChar( SbOutBitBuffer );
	    else
		SbPutChar( 0x3f );     /* instead of DEL */
	    SbOutBitShift = 2;
	    SbOutBitBuffer = 0x40;
	} else
	    SbOutBitBuffer = (c << (8 - SbOutBitShift)) & 0x3f | 0x40;
    }
}

SbPutBytes( bytes, length )
char *bytes;
int length;
{
    while (length-- > 0)
	SbPutByte( *bytes++ );
}

SbPutWords( words, length )
SbUint16 *words;
int length;
{
    while (length--) {
#if SWAP_BYTE
	SbPutByte( *((char *) words + 1) );
	SbPutByte( *((char *) (words++)) );
#else
	SbPutByte( *((char *) words) );
	SbPutByte( *((char *) (words++) + 1) );
#endif
    }
}

/*======================== UTILITY ROUTINES ==============================*/

SbShowPacket( packet, lth )
char *packet;
int lth;

/* print a hexdump of a packet to output_fp */

{
    int linelth, i;

    for ( ;lth > 0; lth -= 16) {
	linelth = lth > 16 ? 16 : lth;
	for (i=0; i<linelth; i++)
	    fprintf( output_fp, " %2.2x", (SbUchar) *packet++ );
	packet -= linelth;
	for (i=linelth; i<16; i++)
	    fprintf( output_fp, "   " );
	fprintf( output_fp, "    " );
	for (i=0; i<linelth; i++, packet++)
	    if (isprint( *packet ))
		fputc( *packet, output_fp );
	    else
		fputc( '.', output_fp );
	fputc( '\n', output_fp );
    }
}

SbShowMatrix( matrix )
SbMatrixType matrix;

/* print a matrix to error_fp */

{
    int i, j;

    for (i=0; i<Sb_ROWS; i++) {
	for (j=0; j<Sb_COLUMNS; j++)
	    fprintf( output_fp, " %10.5f", matrix[ i ][ j ] );
	fprintf( output_fp, "\n" );
    }
}

SbBad( packet, lth )
char *packet;
int lth;
{
    fprintf( output_fp, "sblibry: Bad %d byte packet from Spaceball\n", lth );
    SbShowPacket( packet, lth );
}

SbPackVector( vector, pvec )
SbReal vector[ 3 ];
SbInt16 pvec[ 3 ];

/* convert a vector to a Spaceball unit vector */

{
    SbReal veclth, scalefactor;

    veclth = sqrt(
	vector[ 0 ] * vector[ 0 ] +
	vector[ 1 ] * vector[ 1 ] +
	vector[ 2 ] * vector[ 2 ]
    );
    scalefactor = REAL_TO_SB_FACTOR / veclth;
    pvec[ 0 ] = vector[ 0 ] * scalefactor;
    pvec[ 1 ] = vector[ 1 ] * scalefactor;
    pvec[ 2 ] = vector[ 2 ] * scalefactor;
}

SbUnpackVector( pvec, vector )
char *pvec;
SbReal vector[ 3 ];

/* unpack a Spaceball vector into an application vector where each
 * element will be within -2.0 <= value < 2.0
 */

{
    *vector++ = ldexp( (double) SbConvertWord( pvec	), -14 );
    *vector++ = ldexp( (double) SbConvertWord( pvec + 2 ), -14 );
    *vector   = ldexp( (double) SbConvertWord( pvec + 4 ), -14 );
}

/*========================= MATRIX ROUTINES ==============================*/

static SbInt16 pack_cs( c, s )
SbReal c, s;

/* pack a cosine and sine into a 16 bit value */

{
    if (fabs( c ) < fabs( s ))
	return ((SbInt16) ( c * REAL_TO_CS_FACTOR )) & FFFCs
		| 2
		| (s < 0 ? 1 : 0);
    return ((SbInt16) ( s * REAL_TO_CS_FACTOR )) & FFFCs
	    | (c < 0 ? 1 : 0);
}

static void unpack_cs( p, c, s )
SbInt16   p;
SbReal *c, *s;

/* unpack a cosine and sine from a 16 bit value */

{
    if (p & 2) {
	*c = (p & FFFCs) * CS_TO_REAL_FACTOR;
	*s = sqrt( 1.0 - *c * *c );
	if (p & 1)
	    *s = -*s;
    } else {
	*s = (p & FFFCs) * CS_TO_REAL_FACTOR;
	*c = sqrt( 1.0 - *s * *s );
	if (p & 1)
	    *c = -*c;
    }
}

static int find_minimum( a, b, c, d, e, f )
SbInt16 a, b, c, d, e, f;

/* find the minimum element of the six rotation matrix diagonals, return the
 * ordering which results in optimal numeric accuracy.
 */

{
    SbInt16 minval = a;
    int retval = YXZ;

    if (minval > b) {
	minval = b;
	retval = ZXY;
    }
    if (minval > c) {
	minval = c;
	retval = XYZ;
    }
    if (minval > d) {
	minval = d;
	retval = ZYX;
    }
    if (minval > e) {
	minval = e;
	retval = XZY;
    }
    if (minval > f)
	return YZX;
    return retval;
}

SbPackMatrix( matrix, pmat )
SbMatrixType matrix;
SbInt16 pmat[ 3 ];

/* pack a rotation matrix into three 16 bits values */

{
    SbReal cx, sx, cy, sy, cz, sz, inv;

    /* find minimum off diagonal element */

    switch(
	find_minimum(
	    abs( (SbInt16) (Sb_INDEX( matrix, 1, 2 ) * REAL_TO_CS_FACTOR) )
		& FFFCs,
	    abs( (SbInt16) (Sb_INDEX( matrix, 2, 1 ) * REAL_TO_CS_FACTOR) )
		& FFFCs,
	    abs( (SbInt16) (Sb_INDEX( matrix, 0, 2 ) * REAL_TO_CS_FACTOR) )
		& FFFCs,
	    abs( (SbInt16) (Sb_INDEX( matrix, 2, 0 ) * REAL_TO_CS_FACTOR) )
		& FFFCs,
	    abs( (SbInt16) (Sb_INDEX( matrix, 0, 1 ) * REAL_TO_CS_FACTOR) )
		& FFFCs,
	    abs( (SbInt16) (Sb_INDEX( matrix, 1, 0 ) * REAL_TO_CS_FACTOR) )
		& FFFCs
	)
    ) {
      case YXZ:
	sx  = POS12  Sb_INDEX( matrix, 1, 2 );
	cx  =  sqrt( 1.0 - sx * sx );
	inv = 1 / cx;
	sy  = POS02 Sb_INDEX( matrix, 0, 2 ) * inv;
	cy  = POS22 Sb_INDEX( matrix, 2, 2 ) * inv;
	sz  = NEG10 Sb_INDEX( matrix, 1, 0 ) * inv;
	cz  = POS11 Sb_INDEX( matrix, 1, 1 ) * inv;
	break;
      case ZXY:
	sx  = NEG21 Sb_INDEX( matrix, 2, 1 );
	cx  = -sqrt( 1.0 - sx * sx );
	inv = 1 / cx;
	sz  = POS01 Sb_INDEX( matrix, 0, 1 ) * inv;
	cz  = POS11 Sb_INDEX( matrix, 1, 1 ) * inv;
	sy  = NEG20 Sb_INDEX( matrix, 2, 0 ) * inv;
	cy  = POS22 Sb_INDEX( matrix, 2, 2 ) * inv;
	break;
      case XYZ:
	sy  = POS02 Sb_INDEX( matrix, 0, 2 );
	cy  =  sqrt( 1.0 -  sy * sy );
	inv = 1 / cy;
	sx  = POS12 Sb_INDEX( matrix, 1, 2 ) * inv;
	cx  = POS22 Sb_INDEX( matrix, 2, 2 ) * inv;
	sz  = POS01 Sb_INDEX( matrix, 0, 1 ) * inv;
	cz  = POS00 Sb_INDEX( matrix, 0, 0 ) * inv;
	break;
      case ZYX:
	sy  = NEG20 Sb_INDEX( matrix, 2, 0 );
	cy  = -sqrt( 1.0 -  sy * sy );
	inv = 1 / cy;
	sz  = NEG10 Sb_INDEX( matrix, 1, 0 ) * inv;
	cz  = POS00 Sb_INDEX( matrix, 0, 0 ) * inv;
	sx  = NEG21 Sb_INDEX( matrix, 2, 1 ) * inv;
	cx  = POS22 Sb_INDEX( matrix, 2, 2 ) * inv;
	break;
      case XZY:
	sz  = POS01 Sb_INDEX( matrix, 0, 1 );
	cz  =  sqrt( 1.0 -  sz * sz );
	inv = 1 / cz;
	sx  = NEG21 Sb_INDEX( matrix, 2, 1 ) * inv;
	cx  = POS11 Sb_INDEX( matrix, 1, 1 ) * inv;
	sy  = POS02 Sb_INDEX( matrix, 0, 2 ) * inv;
	cy  = POS00 Sb_INDEX( matrix, 0, 0 ) * inv;
	break;
      case YZX:
	sz  = NEG10 Sb_INDEX( matrix, 1, 0 );
	cz  = -sqrt( 1.0 -  sz * sz );
	inv = 1 / cz;
	sy  = NEG20 Sb_INDEX( matrix, 2, 0 ) * inv;
	cy  = POS00 Sb_INDEX( matrix, 0, 0 ) * inv;
	sx  = POS12 Sb_INDEX( matrix, 1, 2 ) * inv;
	cx  = POS11 Sb_INDEX( matrix, 1, 1 ) * inv;
    }
    pmat[ 0 ] = pack_cs( cx, sx );
    pmat[ 1 ] = pack_cs( cy, sy );
    pmat[ 2 ] = pack_cs( cz, sz );
}

static int XYZ_select( sx, cx, sy, cy, sz, cz )
SbReal	sx, cx, sy, cy, sz, cz;

/* select the ordering chosen by find_minimum() based on the minimum sine */

{
    /* ensure sines are positive */

    if (sx < 0)
	sx = -sx;
    if (sy < 0)
	sy = -sy;
    if (sz < 0)
	sz = -sz;

    /* find the smallest sine - favour sx then sy then sz if equal */

    if (sx <= sz )
	if (sx <= sy )
	    return cx >= 0 ? YXZ : ZXY;
	else
	    return cy >= 0 ? XYZ : ZYX;
    else
	if (sy <= sz )
	    return cy >= 0 ? XYZ : ZYX;
	else
	    return cz >= 0 ? XZY : YZX;
}

SbUnpackMatrix( pmat, matrix )
char pmat[ 6 ];
SbMatrixType matrix;

/* unpack a rotation matrix from three 16 bit values */

{
    SbReal cx, sx, cy, sy, cz, sz;
    SbReal tmp;

    unpack_cs( SbConvertWord( pmat     ), &cx, &sx );
    unpack_cs( SbConvertWord( pmat + 2 ), &cy, &sy );
    unpack_cs( SbConvertWord( pmat + 4 ), &cz, &sz );

    switch (XYZ_select( sx, cx, sy, cy, sz, cz )) {
      case YXZ:
	Sb_INDEX( matrix, 0, 0 ) = POS00 (cy * cz + (tmp = sy * sx ) * sz);
	Sb_INDEX( matrix, 0, 1 ) = POS01 (cy * sz - tmp * cz);
	Sb_INDEX( matrix, 0, 2 ) = POS02 sy * cx;
	Sb_INDEX( matrix, 1, 0 ) = NEG10 cx * sz;
	Sb_INDEX( matrix, 1, 1 ) = POS11 cx * cz;
	Sb_INDEX( matrix, 1, 2 ) = POS12 sx;
	Sb_INDEX( matrix, 2, 0 ) = NEG20 (sy * cz - (tmp = cy * sx ) * sz);
	Sb_INDEX( matrix, 2, 1 ) = NEG21 (sy * sz + tmp * cz);
	Sb_INDEX( matrix, 2, 2 ) = POS22 cy * cx;
	break;
      case ZXY:
	Sb_INDEX( matrix, 0, 0 ) = POS00 (cz * cy - (tmp = sz * sx ) * sy);
	Sb_INDEX( matrix, 0, 1 ) = POS01 sz * cx;
	Sb_INDEX( matrix, 0, 2 ) = POS02 (cz * sy + tmp * cy);
	Sb_INDEX( matrix, 1, 0 ) = NEG10 (sz * cy + (tmp = cz * sx ) * sy);
	Sb_INDEX( matrix, 1, 1 ) = POS11 cz * cx;
	Sb_INDEX( matrix, 1, 2 ) = NEG12 (sz * sy - tmp * cy);
	Sb_INDEX( matrix, 2, 0 ) = NEG20 cx * sy;
	Sb_INDEX( matrix, 2, 1 ) = NEG21 sx;
	Sb_INDEX( matrix, 2, 2 ) = POS22 cx * cy;
	break;
      case XYZ:
	Sb_INDEX( matrix, 0, 0 ) = POS00 cy * cz;
	Sb_INDEX( matrix, 0, 1 ) = POS01 cy * sz;
	Sb_INDEX( matrix, 0, 2 ) = POS02 sy;
	Sb_INDEX( matrix, 1, 0 ) = NEG10 ((tmp = sx * sy ) * cz + cx * sz);
	Sb_INDEX( matrix, 1, 1 ) = NEG11 (tmp * sz - cx * cz);
	Sb_INDEX( matrix, 1, 2 ) = POS12 sx * cy;
	Sb_INDEX( matrix, 2, 0 ) = NEG20 ((tmp = cx * sy ) * cz - sx * sz);
	Sb_INDEX( matrix, 2, 1 ) = NEG21 (tmp * sz + sx * cz);
	Sb_INDEX( matrix, 2, 2 ) = POS22 cx * cy;
	break;
      case ZYX:
	Sb_INDEX( matrix, 0, 0 ) = POS00 cz * cy;
	Sb_INDEX( matrix, 0, 1 ) = POS01 (sz * cx - (tmp = cz * sy ) * sx);
	Sb_INDEX( matrix, 0, 2 ) = POS02 (sz * sx + tmp * cx);
	Sb_INDEX( matrix, 1, 0 ) = NEG10 sz * cy;
	Sb_INDEX( matrix, 1, 1 ) = POS11 (cz * cx + (tmp = sz * sy ) * sx);
	Sb_INDEX( matrix, 1, 2 ) = POS12 (cz * sx - tmp * cx);
	Sb_INDEX( matrix, 2, 0 ) = NEG20 sy;
	Sb_INDEX( matrix, 2, 1 ) = NEG21 cy * sx;
	Sb_INDEX( matrix, 2, 2 ) = POS22 cy * cx;
	break;
      case XZY:
	Sb_INDEX( matrix, 0, 0 ) = POS00 cz * cy;
	Sb_INDEX( matrix, 0, 1 ) = POS01 sz;
	Sb_INDEX( matrix, 0, 2 ) = POS02 cz * sy;
	Sb_INDEX( matrix, 1, 0 ) = NEG10 ((tmp = cx * sz ) * cy + sx * sy);
	Sb_INDEX( matrix, 1, 1 ) = POS11 cx * cz;
	Sb_INDEX( matrix, 1, 2 ) = NEG12 (tmp * sy - sx * cy);
	Sb_INDEX( matrix, 2, 0 ) = POS20 ((tmp = sx * sz ) * cy - cx * sy);
	Sb_INDEX( matrix, 2, 1 ) = NEG21 sx * cz;
	Sb_INDEX( matrix, 2, 2 ) = POS22 (tmp * sy + cx * cy);
	break;
      case YZX:
	Sb_INDEX( matrix, 0, 0 ) = POS00 cy * cz;
	Sb_INDEX( matrix, 0, 1 ) = POS01 ((tmp = cy * sz ) * cx - sy * sx);
	Sb_INDEX( matrix, 0, 2 ) = POS02 (tmp * sx + sy * cx);
	Sb_INDEX( matrix, 1, 0 ) = NEG10 sz;
	Sb_INDEX( matrix, 1, 1 ) = POS11 cz * cx;
	Sb_INDEX( matrix, 1, 2 ) = POS12 cz * sx;
	Sb_INDEX( matrix, 2, 0 ) = NEG20 sy * cz;
	Sb_INDEX( matrix, 2, 1 ) = NEG21 ((tmp = sy * sz ) * cx + cy * sx);
	Sb_INDEX( matrix, 2, 2 ) = NEG22 (tmp * sx - cy * cx);
	break;
    }
}

SbConvertMatrix( mask, mat )
SbUint16 mask;
SbMatrixType mat;

/* negates rows and columns based on octal setting of mask which takes the
 * form:  00 [row2][row1][row0] [col2][col1][col0]
 * If the bit is set the cooresponding row or column is negated. This allows
 * Spaceball's rotation matrices to match application programs rotation
 * matrices. An even number of bits must be set in mask otherwise the
 * matrices determinant will be negated. See also Sb_MATRIX_CONVERSION.
 */

{
    if ((mask & 010) ^ (mask & 001))
	Sb_INDEX( mat, 0, 0 ) = -Sb_INDEX( mat, 0, 0 );
    if ((mask & 010) ^ (mask & 001))
	Sb_INDEX( mat, 0, 1 ) = -Sb_INDEX( mat, 0, 1 );
    if ((mask & 010) ^ (mask & 001))
	Sb_INDEX( mat, 0, 2 ) = -Sb_INDEX( mat, 0, 2 );
    if ((mask & 020) ^ (mask & 002))
	Sb_INDEX( mat, 1, 0 ) = -Sb_INDEX( mat, 1, 0 );
    if ((mask & 020) ^ (mask & 002))
	Sb_INDEX( mat, 1, 1 ) = -Sb_INDEX( mat, 1, 1 );
    if ((mask & 020) ^ (mask & 002))
	Sb_INDEX( mat, 1, 2 ) = -Sb_INDEX( mat, 1, 2 );
    if ((mask & 040) ^ (mask & 004))
	Sb_INDEX( mat, 2, 0 ) = -Sb_INDEX( mat, 2, 0 );
    if ((mask & 040) ^ (mask & 004))
	Sb_INDEX( mat, 2, 1 ) = -Sb_INDEX( mat, 2, 1 );
    if ((mask & 040) ^ (mask & 004))
	Sb_INDEX( mat, 2, 2 ) = -Sb_INDEX( mat, 2, 2 );
}

/*===================== DEFAULT PACKET HANDLERS ==========================*/

SbNullFunction() {}	/* used to ignore certain packets from Spaceball */

SbProcessReset( packet )
char *packet;
{
    int i;

    if (packet[ 1 ] != '2')     /* respond only to the second reset packet */
	return;
    SbPutChar( '\r' );      /* ensure start of next packet */
    if (sscanf( packet, "@2 Firmware version %d.%d created on %*s",
	    &SbVersionMajor, &SbVersionMinor
	) == 2
    ) {
#if Sb_PRINTABLE_MODE
	if (SbVersionMajor == 1 && SbVersionMinor < 4) {
	    fprintf( error_fp, "\
Spaceball ROM version V%d.%d cannot handle the printable communications mode\n\
Please contact supplier for a ROM upgrade\n", SbVersionMajor, SbVersionMinor
	    );
	} else
	    SbSetRS232Mode( "P" );
#else
	;
#endif
    } else
	fprintf( error_fp, "Invalid reset message from Spaceball\n" );

    if (!request_settings_pending)
	/* initialize Spaceball with present settings */
	for (i=0; i<sizeof( SbSettings )/sizeof( SbSettings[ 0 ]) ; i++) {
	    SbPutChar( SbSettings[i][1] );
	    SbPutBytes( SbSettings[ i ] + 2, (int) *SbSettings[ i ] - 1);
	    SbPutChar( '\r' );
	}
}

SbProcessEcho( s )
char *s;
{
    fprintf( output_fp, "Spaceball echoed: \n\t%s\n", s );
}

SbProcessError( s )
char *s;
{
    fprintf( error_fp, "Spaceball message: \n\t%s\n", s );
}

SbProcessInvalid( packet, lth )
char *packet;
int lth;
{
    fprintf( output_fp,
	"sblibry: Invalid packet returned by Spaceball of %d characters\n",
	lth
    );
    SbShowPacket( packet, lth );
}

SbProcessIRotate( period, ivector )
SbUint16 period;
SbInt16 ivector[ 3 ];
{
    SbReal vector[ 3 ];

    vector[ 0 ] = ldexp( (double) ivector[ 0 ], -14 );
    vector[ 1 ] = ldexp( (double) ivector[ 1 ], -14 );
    vector[ 2 ] = ldexp( (double) ivector[ 2 ], -14 );
    (*SbFunctions.rotate)( period, vector );
}

SbProcessITranslate( period, ivector )
SbUint16 period;
SbInt16 ivector[ 3 ];
{
    SbReal vector[ 3 ];

    vector[ 0 ] = ldexp( (double) ivector[ 0 ], -14 );
    vector[ 1 ] = ldexp( (double) ivector[ 1 ], -14 );
    vector[ 2 ] = ldexp( (double) ivector[ 2 ], -14 );
    (*SbFunctions.translate)( period, vector );
}

SbProcessOther( packet, length )
char *packet;
int length;
{
    fprintf( output_fp, "Unrecognized Packet from Spaceball:\n" );
    SbShowPacket( packet, length );
}

/*===================== HIGH LEVEL ROUTINES ==============================*/

SbProcessPacket( packet, lth )
char *packet;
int lth;

/* validates and decodes the Spaceball packet, then passes the
 * data to functions specified in SbFunctions{} to process the data
 */

{
    char *pp;
    int i;
    SbUint16 oldbuttons, pmat[ 3 ], changed, period, ivector[ 3 ];
    SbReal vector[ 3 ];
    static SbMatrixType matrix = IDENTITY_MATRIX;
    static char nulls[] = { 0, 0, 0, 0, 0, 0 };

    switch (packet[ 0 ]) {
      case '@':
	SbBallDataLength = 0;
	(*SbFunctions.reset)( packet );
	break;
      case 'A':
	if (lth != 7) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ APACKET ][ 0 ] = 7;
	memcpy( SbSettings[ APACKET ] + 1, packet, 7 );
	SbUnpackMatrix( packet + 1, SbAbsolute );
	(*SbFunctions.absolute)( SbAbsolute );
	break;
      case 'B':
	(*SbFunctions.beeper)( packet + 1 );
	break;
      case 'C':
	if (lth == 5)
	    return;	/* should be Chex+ or Chex- */
	if (lth != 3) {
	    SbBad( packet, lth );
	    break;
	}
	SbRS232Mode[ 0 ] = packet[ 1 ];
	SbRS232Mode[ 1 ] = packet[ 2 ];
	SbOutBitShift = 2;
	break;
      case 'D':
	if (lth != SbBallDataLength) {
	    SbPutStr( "m\r" );
	    SbRequestBallData();
	    SbBad( packet, lth );
	    SbShowPacket( debugpacket, debuglength );
	    break;
	}
#if 1
	{ int i;
	for (i=1; i<16; i++)
	    if (packet[i] == '\\' &&
		packet[i+1] == '\\' &&
		packet[i+2] == '\\'
	    ) {
		printf( "HAVE funny packet\n" );
		SbShowPacket( packet, lth );
		break;
	    }
	}
#endif
	SbBallAtRest = TRUE;
	SbBallDataPending = FALSE;
	pp = packet + 1;
	if (SbBallDataHasPeriod) {
	    period = SbConvertWord( pp );
	    pp += 2;

	    /* SbAveragePeriod tracks an average value of period */

	    SbAveragePeriod = (SbAveragePeriod >> 1)	    /*	 1/2 */
			    + (SbAveragePeriod >> 2)	    /* + 1/4 */
			    + (SbAveragePeriod >> 3)	    /* + 1/8 */
			    + (period >> 3);		    /* + 1/8 */
	}
	if (SbSettings[ MPACKET ][ 2 ] != 'N') {
	    SbBallAtRest &= (memcmp( pp, nulls, 6 ) == 0);
	    ivector[ 0 ] = SbConvertWord( pp	 );
	    ivector[ 1 ] = SbConvertWord( pp + 2 );
	    ivector[ 2 ] = SbConvertWord( pp + 4 );
	    pp += 6;
	    (*SbFunctions.itranslate)( period, ivector );
	}
	if (SbSettings[ MPACKET ][ 3 ] != 'N') {
	    switch (SbSettings[ MPACKET ][ 4 ]) {
	      case 'V':
		SbBallAtRest &= (memcmp( pp, nulls, 6 ) == 0);
		ivector[ 0 ] = SbConvertWord( pp     );
		ivector[ 1 ] = SbConvertWord( pp + 2 );
		ivector[ 2 ] = SbConvertWord( pp + 4 );
		(*SbFunctions.irotate)( period, ivector );
		break;
	      case 'D':
		SbBallAtRest &= (memcmp( pp, nulls, 6 ) == 0);
		SbUnpackMatrix( pp, matrix );
		(*SbFunctions.delta)( matrix );
		break;
	      case 'A':
		SbBallAtRest &=
		    (memcmp( pp, SbSettings[ APACKET ] + 2, 6 ) == 0);
		SbSettings[ APACKET ][ 0 ] = 7;
		SbSettings[ APACKET ][ 1 ] = 'A';
		memcpy( SbSettings[ APACKET ] + 2, pp, 6 );
		SbUnpackMatrix( pp, SbAbsolute );
		(*SbFunctions.absolute)( SbAbsolute );
	    }
	}
	if (SbBallAtRest)
	    (*SbFunctions.ballatrest)();
	break;
      case 'E':
	if (lth == 1) {
	    SbBad( packet, lth );
	    break;
	}
	for (i=1; i<lth; i++) {
	    SbPutChar( 'h' );
	    SbPutChar( packet[ i ] );
	    SbPutChar( '\r' );
	}
	break;
      case 'F':
	if (lth != 3) {
	    SbBad( packet, lth );
	    break;
	}
	switch (packet[ 1 ]) {
	  case 'T':
	    SbSettings[ FTPACKET ][ 0 ] = lth;
	    memcpy( SbSettings[ FTPACKET ] + 1, packet, lth );
	    SbTranslationFeel = packet[ 2 ] & 0x3f;
	    break;
	  case 'R':
	    SbSettings[ FRPACKET ][ 0 ] = lth;
	    memcpy( SbSettings[ FRPACKET ] + 1, packet, lth );
	    SbRotationFeel = packet[ 2 ] & 0x3f;
	    break;
	  default:
	    SbBad( packet, lth );
	}
	break;
      case 'H':
	(*SbFunctions.error)( packet + 1 );
	break;
      case 'K':
	if (lth != 3) {
	    SbBad( packet, lth );
	    break;
	}
	oldbuttons = SbKeys;
	SbKeys = ((packet[ 1 ] & 0x1f) << 4) | packet[ 2 ] & 0xf;
	changed = SbKeys & (SbKeys ^ oldbuttons);
	if (changed)
	    (*SbFunctions.pressed)( changed );
	changed = ~SbKeys & (SbKeys ^ oldbuttons);
	if (changed)
	    (*SbFunctions.released)( changed );
	break;
      case 'M':
	SbBallModePending = FALSE;
	if (lth != 6) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ MPACKET ][ 0 ] = 6;
	memcpy( SbSettings[ MPACKET ] + 1, packet, 6 );
	memcpy( SbBallMode, packet + 1, 5 );
	SbBallDataHasPeriod = (packet[ 1 ] != 'N')
	       || (packet[ 2 ] != 'N') && (packet[ 3 ] == 'V');
	SbBallDataLength = 1
		+ (SbBallDataHasPeriod ? 2 : 0)
		+ (packet[ 1 ] != 'N' ? 6 : 0)
		+ (packet[ 2 ] != 'N' ? 6 : 0);
	if (SbBallAtRest = (SbBallDataLength == 1))
	    (*SbFunctions.ballatrest)();
	(*SbFunctions.ballmode)( SbBallMode );
	break;
      case 'N':
	if (lth != 2) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ NPACKET ][ 0 ] = lth;
	memcpy( SbSettings[ NPACKET ] + 1, packet, lth );
	SbNullRegion = packet[ 1 ] & 0x3f;
	break;
      case 'O':
	if (lth != 7) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ OPACKET ][ 0 ] = 7;
	memcpy( SbSettings[ OPACKET ] + 1, packet, 7 );
	SbUnpackMatrix( packet + 1, SbOrientation );
	break;
      case 'P':
	if (lth != 5) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ PPACKET ][ 0 ] = 5;
	memcpy( SbSettings[ PPACKET ] + 1, packet, 5 );
	SbPulse[ 0 ] = ((packet[ 1 ] & 0x3f) << 6) | (packet[ 2 ] & 0x3f);
	SbPulse[ 1 ] = ((packet[ 3 ] & 0x3f) << 6) | (packet[ 4 ] & 0x3f);
	break;
      case 'R':
	if (lth != 10) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ RPACKET ][ 0 ] = lth;
	memcpy( SbSettings[ RPACKET ] + 1, packet, lth );
	memcpy( SbRotationMode, packet + 1, 3 );
	SbUnpackVector( packet + 4, SbRotationFreedom );
	break;
      case 'S':
	if (lth != 3) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ SPACKET ][ 0 ] = 3;
	memcpy( SbSettings[ PPACKET ] + 1, packet, 3 );
	SbSpinRate = ldexp( (double) (packet[ 1 ] & 0x3f),
	    -6 - 2 - (packet[ 2 ] & 0x3f)
	);
	break;
      case 'T':
	if (lth != 10) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ TPACKET ][ 0 ] = lth;
	memcpy( SbSettings[ TPACKET ] + 1, packet, lth );
	memcpy( SbTranslationMode, packet + 1, 3 );
	SbUnpackVector( packet + 4, SbTranslationFreedom );
	break;
      case 'X':
	if (lth != 7) {
	    SbBad( packet, lth );
	    break;
	}
	SbSettings[ XPACKET ][ 0 ] = lth;
	memcpy( SbSettings[ XPACKET ] + 1, packet, lth );
	SbUnpackVector( packet + 4, SbXYZScalings );
	break;
      case '?':
	(*SbFunctions.invalid)( packet + 1, lth - 1 );
	break;
      case '%':
	(*SbFunctions.echo)( packet + 1 );
	break;
      default:
	(*SbFunctions.other)( packet, lth );
    }
}

int SbReset( wait )
int wait;

/* reset Spaceball then obtain the default settings from Spaceball */

{
    int i;
    SbInt16 time_left = 0, timeout = 150;  /* wait 0.15 sec for pending data */
    char dummy[ Sb_MAXPACKET ];

    /* ensure Spaceball has sent any pending characters */
    SbPutStr( "\021\r" );
    timeout += SbMilliseconds();
    do {
	if (time_left > 0)
	    SbWaitForPacket( time_left );
	while ((*SbUserGetPacket)( dummy ) != 0)
	    ;
    } while ((time_left = timeout - SbMilliseconds()) > 0);

    /* reset Spaceball */
    SbPutStr( "\r@RESET\r" );

    /* clear SbSettings */
    for (i=0; i<Sb_NUM_SETTINGS; i++)
	SbSettings[ i ][ 0 ] = '\0';
    memcpy(
	(char *) SbAbsolute,
	(char *) SbIdentityMatrix,
	sizeof( SbMatrixType )
    );

    if (SbWatchingReset = !wait)    /* assign SbWatchingReset */
	return 0;
    return SbWatchReset( 5000 );    /* wait up to 2 seconds */
}

SbWatchReset( timeout )
SbInt16 timeout;
{
    char packet[ Sb_MAXPACKET ];
    SbInt16 time_left = timeout;
    int retval;

    /* care is required here so as to avoid real time bugs as the process
     * may be put to sleep at any time. */

    timeout += SbMilliseconds();	/* calculate absolute time */
    do {
	while ((*SbUserGetPacket)( packet ) > 0)
	    if (memcmp( "@1 Spaceball", packet, 11 ) == 0) {
#ifdef Sb_DEBUG
	    if (Sb_debug & 0x8000)
		fprintf( output_fp, "%s\n", packet );
#endif
#if Sb_PRINTABLE_MODE
		SbSetRS232Mode( "P" );
#endif
		/* request the default SbSettings */
		request_settings_pending = TRUE;
		SbRequestSettings();
		retval = SbSynchronize();
		request_settings_pending = FALSE;
		return retval;
	    }
	SbWaitForPacket( 100 ); /* send following '\r' every 1/10th sec */
	SbPutChar( '\r' );      /* activity on RS232 line after a poweron
				 * reset tells Spaceball to send the reset
				 * message. */
    } while ((time_left = timeout - SbMilliseconds()) > 0);
    return -1;	 /* timeout */
}

SbRequestSettings()
{
    SbPutStr( "\ra\rf\rk\rm\rn\ro\rp\rr\rs\rt\rx\r" );
}

int SbSynchronize()

/* Send a message to Spaceball to be echoed then wait for it.
 * Process any intervening packets.
 * Returns 0 if successfull, -1 if a timeout occurs.
 */

{
    if (SbWatchingReset) {
	SbWatchingReset = FALSE;
	return SbWatchReset( 2000 );	/* will normally call SbSynchronize() */
    }
    SbSendSync();
    return SbWaitForSync( 2000 );  /* two second timeout */
}

SbWaitForSync( timeout )
SbInt16 timeout;
{
    char packet[ Sb_MAXPACKET ];
    int lth;
    SbInt16 time_left = timeout;

    timeout += SbMilliseconds();
    do {
	while ((lth = (*SbUserGetPacket)( packet )) != 0)
	    if (*packet == '?' && SbFirstTime)  /* catch any earlier rubbish */
		SbFirstTime = FALSE;
	    else if (strcmp( packet, " sync" ) == 0)
		return 0;
	    else
		SbProcessPacket( packet, lth );
#if SGI     /* A bug, probably in the OS but possibly in sbcustom.c
	     * requires Spaceball to send more data as a fix.
	     * Without this, characters from Spaceball do not cause
	     * SbInput() to return immediately. Only occurs on 4DGTs. */
	SbPutStr( "c\r" );      /* send comms mode */
#endif
	SbWaitForPacket( 100 );
    } while ((time_left = timeout - SbMilliseconds()) > 0);
    return -1;
}

/*====================== INPUT HANDLING ROUTINES =========================*/

#ifdef vms
static avoid_bug( lth )   /* dummy routine to avoid a vms C compiler
			   * optimizer bug */
int *lth;
{}
#endif

static int SquashPacket( new, old, lth )
char *new, *old;
int lth;

/* copies old[] to new[] removing preceeding LFs and replacing the
 * two byte sequences, ^Q ^S ^M ^^, with a single byte.
 * Returns length of new[], if an error occurs returns a negative
 * length with the error returned in new[]
 */

{
    SbBoolean special_state = FALSE;
    char *newstart = new, c, bitbuffer;
    int bitshift;

#if Sb_DEBUG
    if (Sb_debug & 4) {
	fprintf( output_fp, "Squashing packet of %d characters:\n", lth );
	SbShowPacket( old, lth );
    }
#endif

    if (lth == 0)
	return 0;

    /* skip preceeding LFs */

    while (*old == LF) {
	if (--lth == 0)
	    return 0;
	old++;
    }

    /* process characters */

    if ((SbRS232Mode[ 0 ] | 0x20) != 'p') {
	for (;lth--; old++)
	    if (!special_state)
		if (*old != '^')
		    *new++ = *old;
		else
		    special_state = TRUE;
	    else {
		special_state = FALSE;
		switch (*old) {
		  case 'M':
		  case 'Q':
		  case 'S':
		    *new++ = *old & 0x1f;
		    break;
		  case '^':
		    *new++ = '^';
		    break;
		  default:
		    *new++ = '^';
		    *new++ = *old;
		    *new = '\0';
		    return -(new - newstart);	/* return error */
		}
	    }
    } else {		/* printable ascii mode */
	lth++;
#ifdef vms
	avoid_bug( &lth );  /* avoid a vms C compiler optimizer bug */
#endif
	while (--lth)
	    if ((*old /* &= 0x7f */) != '^')
		*new++ = *old++;
	    else {
		lth--;
		break;
	    }
	bitshift = -2;
	while (lth--) {
	    c = *++old & 0x7f;
	    switch (bitshift) {
	      case -2:
		bitbuffer = c << 2;
		bitshift = 4;
		break;
	      case 4:
		*new++ = bitbuffer | (c & 0x3f) >> 4;
		bitbuffer = c << 4;
		bitshift = 2;
		break;
	      case 2:
		*new++ = bitbuffer | (c & 0x3f) >> 2;
		bitbuffer = c << 6;
		bitshift = 0;
		break;
	      case 0:
		*new++ = bitbuffer | c & 0x3f;
		bitshift = -2;
		break;
	    }
	}
    }
    *new = '\0';        /* ensure null termination */
    return new - newstart;
}

int SbGetPacket( packet )
char *packet;

/* obtains the next packet, return value is the length, 0 indicates
 * no more packets received
 */

{
    int
      retval;			/* return value */

    SbNumchars += SbInput( SbInBuffer + SbNumchars,
	Sb_MAXPACKET - SbNumchars, 0
    );

    do {
	  /* process characters */
	for ( ; SbInIndex < SbNumchars; SbInIndex++)
	    if (SbInBuffer[ SbInIndex ] == CR) {
		if (!SbOverflowed)
		    retval = SquashPacket( packet, SbInBuffer, SbInIndex );
		else {
		    retval = 0;
		    SbOverflowed = FALSE;
		}
		SbNumchars -= SbInIndex + 1;
#if 1
		debuglength = SbInIndex + 1;
		memcpy( debugpacket, SbInBuffer, debuglength );
#endif
		memcpy( SbInBuffer, SbInBuffer + SbInIndex + 1, SbNumchars );
		SbInIndex = 0;
#if Sb_DEBUG
		if ((Sb_debug & 8) && (retval > 0)) {
		    fprintf( output_fp,
			"Packet from Spaceball of %d characters:\n", retval
		    );
		    SbShowPacket( packet, retval );
		}
#endif
		if (retval)
		    return retval;
	    }
	if (SbInIndex == Sb_MAXPACKET) {
	    if (!SbOverflowed) {
		SbOverflowed = TRUE;
		return -Sb_MAXPACKET;
	    }
	    SbNumchars = SbInput( SbInBuffer, Sb_MAXPACKET, 0 );
	    SbInIndex = 0;
	}
    } while (SbInIndex < SbNumchars);
    return 0;
}

int SbPacketReady()

/* checks if a packet is available */

{
    int
      retval;			/* return value */

    SbNumchars += SbInput( SbInBuffer + SbNumchars,
	Sb_MAXPACKET - SbNumchars, 0
    );

    do {
	  /* process characters */
	for ( ; SbInIndex < SbNumchars; SbInIndex++)
	    if (SbInBuffer[ SbInIndex ] == CR) {
		if (!SbOverflowed)
		    return TRUE;
		else
		    SbOverflowed = FALSE;
		SbNumchars -= SbInIndex + 1;
		memcpy( SbInBuffer, SbInBuffer + SbInIndex + 1, SbNumchars );
		SbInIndex = 0;
	    }
	if (SbInIndex == Sb_MAXPACKET) {
	    if (!SbOverflowed) {
		SbOverflowed = TRUE;
		return -1;
	    }
	    SbNumchars = SbInput( SbInBuffer, Sb_MAXPACKET, 0 );
	    SbInIndex = 0;
	}
    } while (SbInIndex < SbNumchars);
    return FALSE;
}

int SbCheckSpaceball()

/* Processes all waiting input from Spaceball.
 * Returns the number of packets received.
 */

{
    char packet[ Sb_MAXPACKET ];
    int lth, retval = FALSE;

    while ((lth = (*SbUserGetPacket)( packet )) != 0) {
	if (lth < 0) {
	    fprintf( output_fp, "sblibry: bad packet received:\n" );
	    SbShowPacket( packet, -lth );
	} else {
	    retval = TRUE;
	    SbProcessPacket( packet, lth );
	}
    }
    return retval;
}

SbWaitForPacket( timeout )

/* Waits for a CR to be received, packet overflow, or period to expire.
 * Maximum period 32767 milliseconds.
 * SbMilliseconds() returns the lowest 16 bits of a millisecond clock,
 * integer overflow is handled correctly.
 * Returns 0 if CR received or packet overflow, -1 if timeout.
 */

SbInt16 timeout;
{
    SbInt16 time_left = timeout;
    int idx = 0;

    timeout += SbMilliseconds();
    do {
	for ( ; idx<SbNumchars; idx++)
	    if (SbInBuffer[ idx ] == '\r')
		return 0;	    /* CR received */
	SbNumchars += SbInput(
	    SbInBuffer + SbNumchars,
	    Sb_MAXPACKET - SbNumchars,
	    time_left
	);
	if (SbNumchars == Sb_MAXPACKET)
	    return 0;		    /* packet overflow or CR received */
	for ( ; idx<SbNumchars; idx++)
	    if (SbInBuffer[ idx ] == '\r')
		return 0;	    /* CR received */
    } while ((time_left = timeout - SbMilliseconds()) > 0);
    return -1;
}

/*========================== SbSet ROUTINES ==============================*/

SbSetAbsolute( matrix )
SbMatrixType matrix;
{
    SbInt16 pmat[ 3 ];

    SbPutChar( 'A' );
    SbPackMatrix( matrix, pmat );
    SbPutWords( (SbUint16 *) pmat, 3 );
    SbPutChar( '\r' );
}

SbSetBallMode( mode )
char *mode;
{
    SbPutChar( 'M' );
    SbPutStr( mode );
    SbPutChar( '\r' );
    SbBallModePending = TRUE;
}

SbSetRS232Mode( mode )
char *mode;
{
    SbPutChar( 'C' );
    SbPutStr( mode );
    SbPutChar( '\r' );
}

#if Sb_NUMBER_OF_SPACEBALLS > 1
SbSetCurrentSpaceball( n )
int n;

/* select the nth Spaceball by setting Sb */

{
    if (n < 0 || n >= Sb_NUMBER_OF_SPACEBALLS) {
	fprintf( error_fp, "Invalid call SbSetSpaceball( %d ) - ignored\n", n );
	return;
    }
    Sb = SbArray + n;		/* change pointer for this file */
}
#endif

SbSetNullRegion( nr )
char nr;
{
    SbPutChar( 'N' );
    SbPutChar( (nr &= 0x3f) == 0x3f ? nr : nr | 0x40 );
    SbPutChar( '\r' );
}

SbSetOrientation( matrix )
SbMatrixType matrix;
{
    SbInt16 pmat[ 3 ];

    SbPutChar( 'O' );
    SbPackMatrix( matrix, pmat );
    SbPutWords( (SbUint16 *) pmat, 3 );
    SbPutChar( '\r' );
}

SbSetPulse( p )
SbUint16 p[ 2 ];
{
    char c;

    SbPutChar( 'P' );
    SbPutChar( (c = (p[ 0 ] >> 6) & 0x3f) == 0x3f ? c : c | 0x40 );
    SbPutChar( (c =  p[ 0 ]	  & 0x3f) == 0x3f ? c : c | 0x40 );
    if (p[ 1 ] != 0) {
	SbPutChar( (c = (p[ 1 ] >> 6) & 0x3f) == 0x3f ? c : c | 0x40 );
	SbPutChar( (c =  p[ 1 ]       & 0x3f) == 0x3f ? c : c | 0x40 );
    }
    SbPutChar( '\r' );
}

SbSetRotationFeel( feel )
char feel;
{
    SbPutStr( "FR" );
    SbPutChar( feel == 0x3f ? feel : feel | 0x40 );
    SbPutChar( '\r' );
}

SbSetRotationFreedom( vector )
SbReal vector[ 3 ];
{
    SbInt16 pvec[ 3 ];

    SbPutStr( "RXXX" );
    SbPackVector( vector, pvec );
    SbPutWords( (SbUint16 *) pvec, 3 );
    SbPutChar( '\r' );
}

SbSetRotationMode( lpsys )
char lpsys[ 3 ];
{
    SbPutChar( 'R' );
    SbPutBytes( lpsys, 3 );	/* Note - these are all printable */
    SbPutChar( '\r' );
}

SbSetSpinRate( spinrate )
SbReal spinrate;
{
    int exp;
    char frac;

    SbPutChar( 'S' );
    frac = frexp( spinrate, &exp ) * 64.0;
    SbPutChar( (frac != 0x3f) ? (0x40 | frac) : frac );
    SbPutChar( 0x40 | (-2 - exp) );
    SbPutChar( '\r' );
}

SbSetTranslationFeel( feel )
char feel;
{
    SbPutStr( "FT" );
    SbPutChar( feel == 0x3f ? feel : feel | 0x40 );
    SbPutChar( '\r' );
}

SbSetTranslationFreedom( vector )
SbReal vector[ 3 ];
{
    SbInt16 pvec[ 3 ];

    SbPutStr( "TXXX" );
    SbPackVector( vector, pvec );
    SbPutWords( (SbUint16 *) pvec, 3 );
    SbPutChar( '\r' );
}

SbSetTranslationMode( lpsys )
char lpsys[ 3 ];
{
    SbPutChar( 'T' );
    SbPutBytes( lpsys, 3 );	/* Note - these are all printable */
    SbPutChar( '\r' );
}

SbSetXYZScalings( xyzscalings )
SbReal xyzscalings[ 3 ];
{
    SbInt16 pvec[ 3 ];

    SbPutChar( 'X' );
    SbPackVector( xyzscalings, pvec );
    SbPutWords( (SbUint16 *) pvec, 3 );
    SbPutChar( '\r' );
}

SbSetZero( zero )
SbUint16 zero[ 6 ];
{
    int i;

    SbPutChar( 'Z' );
    for (i=0; i<6; i++) {
	SbPutChar( 0x40 | (int) (zero[ i ] >> 5) & 0x1f );
	SbPutChar( 0x40 | (int) (zero[ i ]     ) & 0x1f );
    }
    SbPutChar( '\r' );
}

/*========================= UTILITY ROUTINES =============================*/

/* SbDominant: Routine to perform the dominant mode operation which simply
 * zeros the two smaller components of a 3D vector. */

SbDominant( vec )
SbReal vec[3];
{
    if (fabs( vec[0] ) > fabs( vec[1] )) {
	vec[1] = 0.0;
	if (fabs( vec[0] ) > fabs( vec[2] ))
	    vec[2] = 0.0;
	else
	    vec[0] = 0.0;
    } else {
	vec[0] = 0.0;
	if (fabs( vec[1] ) > fabs( vec[2] ))
	    vec[2] = 0.0;
	else
	    vec[1] = 0.0;
    }
}

/* SbMatricize: Generates a delta rotation matrix from a vector and a
 * scaling factor. */

SbMatricize( factor, vec, delta_matrix )
SbReal factor, vec[3];
SbMatrixType delta_matrix;
{
    float length, radians;

    length = sqrt( vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2] );
    if (length == 0.0) {
	memcpy( (char *) delta_matrix, (char *) SbIdentityMatrix,
	    sizeof( SbMatrixType )
	);
	return;
    }
    radians = length * factor;
    vec[0] /= length;	/* convert to a unit vector */
    vec[1] /= length;
    vec[2] /= length;
    SbRotateAbout( radians, vec, delta_matrix );
}

/* SbRotateAbout: Converts a unit vector defining an axis and an angle in
 * radians to a delta rotation matrix. */

SbRotateAbout( radians, unit_vec, delta_matrix )
SbReal radians, unit_vec[3];
SbMatrixType delta_matrix;
{
    SbReal
      sin0, cos0, vers0, kxvers0, kyvers0, kzvers0, kxsin0, kysin0, kzsin0;

    if (SbBallMode[3] == 'R')       /* L4.05 fix */
	radians = -radians;	    /* negate angle for right handed system */
    sin0 = sin( radians );
    cos0 = cos( radians );
    vers0 = 1.0 - cos0;
    kxvers0 = unit_vec[0] * vers0;
    kyvers0 = unit_vec[1] * vers0;
    kzvers0 = unit_vec[2] * vers0;
    kxsin0  = unit_vec[0] * sin0;
    kysin0  = unit_vec[1] * sin0;
    kzsin0  = unit_vec[2] * sin0;
    delta_matrix[0][0] = unit_vec[0] * kxvers0 + cos0;
    delta_matrix[0][1] = unit_vec[1] * kxvers0 - kzsin0;
    delta_matrix[0][2] = unit_vec[2] * kxvers0 + kysin0;
    delta_matrix[1][0] = unit_vec[0] * kyvers0 + kzsin0;
    delta_matrix[1][1] = unit_vec[1] * kyvers0 + cos0;
    delta_matrix[1][2] = unit_vec[2] * kyvers0 - kxsin0;
    delta_matrix[2][0] = unit_vec[0] * kzvers0 - kysin0;
    delta_matrix[2][1] = unit_vec[1] * kzvers0 + kxsin0;
    delta_matrix[2][2] = unit_vec[2] * kzvers0 + cos0;
    /* L4.05 fix */
    delta_matrix[0][3] = delta_matrix[3][0]
	= delta_matrix[1][3] = delta_matrix[3][1]
	= delta_matrix[2][3] = delta_matrix[3][2] = 0.0;
    delta_matrix[3][3] = 1.0;
}

/* SbMatrix3x3Multiply: Multiply the 3x3 parts of two matrices. */

SbMatrix3x3Multiply( mat1, mat2, result )
SbMatrixType mat1, mat2, result;
{
    SbMatrixType tmp;
    int i, j;

    for (i=0; i<3; i++)
	for (j=0; j<3; j++)
	    tmp[i][j] = mat1[i][0] * mat2[0][j]
		      + mat1[i][1] * mat2[1][j]
		      + mat1[i][2] * mat2[2][j];
    for (i=0; i<3; i++)
	for (j=0; j<3; j++)
	    result[i][j] = tmp[i][j];
}

