/*  segments_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for creating the segment controls


    Written by      Richard Gooch   12-DEC-1996

    Updated by      Richard Gooch   11-MAR-1999: Removed unnecessary casts.

    Updated by      Richard Gooch   13-APR-1999: Removed stale code which
  decremented unset values.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <k_event_codes.h>
#include <karma_overlay.h>
#include <karma_chs.h>
#include <karma_ds.h>
#include <karma_ch.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/Filepopup.h>
#include "xraycast_3d.h"

#define MAX_FILE_SEGMENTS 10240


/*  Private functions  */
STATIC_FUNCTION (segment_list *create_segment_list,
		 (Widget main_shell, CONST char *filename) );
STATIC_FUNCTION (void unload_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag file_tester_cbk, (KFileInfo finfo) );
STATIC_FUNCTION (void fileselect_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Public functions follow  */

void load_segments_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Load segments callback.
    [PURPOSE] This routine is called when the "Load Segments" button is pressed
    <w> The widget.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    static Widget filepopup = NULL;

    if (filepopup == NULL)
    {
	filepopup = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
					  client_data,
					  XkwNfilenameTester, file_tester_cbk,
					  XtNtitle, "SegmentFile Selector",
					  XkwNautoPopdown, True,
					  NULL);
	XtAddCallback (filepopup, XkwNfileSelectCallback, fileselect_cbk,
		       client_data);
    }
    XtPopup (filepopup, XtGrabNone);
}   /*  End Function load_segments_cbk  */


/*  Private functions follow  */

static segment_list *create_segment_list (Widget main_shell,
					  CONST char *filename)
/*  [PURPOSE] This routine will create a segment list (complete with popup
    control panel).
    <main_shell> The application main shell.
    [RETURNS] The segment list. On failure the process aborts.
*/
{
    Channel ch;
    unsigned int count;
    unsigned int num_seg = 0;
    Widget popup, form;
    Widget btn, w;
    segment_list *list;
    Kcoord_3d file_segments_p1[MAX_FILE_SEGMENTS];
    Kcoord_3d file_segments_p2[MAX_FILE_SEGMENTS];
    char txt[STRING_LENGTH];
    extern segment_list *first_segment_list;
    static char function_name[] = "create_segment_list";

    /*  First load segments  */
    if ( ( ch = ch_open_file (filename, "r") ) == NULL )
    {
	exit (RV_CANNOT_OPEN);
    }
    while ( chs_get_line (ch, txt, STRING_LENGTH) )
    {
	if (num_seg >= MAX_FILE_SEGMENTS)
	{
	    fprintf (stderr, "Too many segments in file\n");
	    exit (RV_BAD_DATA);
	}
	sscanf (txt, "%e %e %e %e %e %e",
		&file_segments_p1[num_seg].x,
		&file_segments_p1[num_seg].y,
		&file_segments_p1[num_seg].z,
		&file_segments_p2[num_seg].x,
		&file_segments_p2[num_seg].y,
		&file_segments_p2[num_seg].z);
	++num_seg;
    }
    ch_close (ch);
    if ( ( list = m_alloc (sizeof *list) ) == NULL )
    {
	m_abort (function_name, "segment list");
    }
    strcpy (list->colourname, "green");
    if ( ( list->p1 = m_alloc (sizeof *list->p1 * num_seg) ) == NULL )
    {
	m_abort (function_name, "p1 co-ordinates");
    }
    if ( ( list->p2 = m_alloc (sizeof *list->p2 * num_seg) ) == NULL )
    {
	m_abort (function_name, "p2 co-ordinates");
    }
    /*  Copy segment data over  */
    for (count = 0; count < num_seg; ++count)
    {
	list->p1[count] = file_segments_p1[count];
	list->p2[count] = file_segments_p2[count];
    }
    list->num_segments = num_seg;
    list->prev = NULL;
    list->next = first_segment_list;
    if (first_segment_list != NULL) first_segment_list->prev = list;
    first_segment_list = list;
    popup = XtVaCreatePopupShell ("segmentswinpopup",
				  topLevelShellWidgetClass, main_shell,
				  XtNtitle, "Segments Control Window",
				  NULL);
    list->popup_widget = popup;
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				   XtNlabel, "Unload",
				   NULL);
    XtAddCallback (btn, XtNcallback, unload_cbk, list);
    btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				   XtNlabel, "Apply",
				   XtNfromHoriz, btn,
				   NULL);
    XtAddCallback (btn, XtNcallback, apply_cbk, list);
    sprintf (txt, "Segment File: %s", filename);
    w = XtVaCreateManagedWidget ("label", labelWidgetClass, form,
				 XtNlabel, txt,
				 XtNborderWidth, 0,
				 XtNfromVert, btn,
				 NULL);
    w = XtVaCreateManagedWidget ("colourDialog", dialogWidgetClass, form,
				 XtNborderWidth, 0,
				 XtNhorizDistance, 0,
				 XtNvertDistance, 0,
				 XtNlabel, "Colour",
				 XtNvalue, "green",
				 XtNfromVert, w,
				 NULL);
    XtRealizeWidget (popup);
    XtPopup (popup, XtGrabNone);
    return (list);
}   /*  End Function create_segment_list  */

static void unload_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Unload segments callback.
    <w> The widget.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    segment_list *list = client_data;
    extern segment_list *first_segment_list;

    XtDestroyWidget (list->popup_widget);
    if (list->prev != NULL) list->prev->next = list->next;
    if (list->next != NULL) list->next->prev = list->prev;
    m_free (list->p1);
    m_free (list->p2);
    if (list == first_segment_list) first_segment_list = list->next;
    m_free (list);
    compute ();
}   /*  End Function unload_cbk  */

static flag file_tester_cbk (KFileInfo finfo)
/*  [SUMMARY] File tester.
    <finfo> The file information.
    [RETURNS] TRUE if the file is to be displayed, else FALSE.
*/
{
    int len;

    /*  Accept all directories  */
    if (finfo.type == KFILETYPE_DIRECTORY) return (TRUE);
    len = strlen (finfo.filename);
    if (len < 5) return (FALSE);
    if (strcmp (finfo.filename + len - 4, ".seg") == 0) return (TRUE);
    return (FALSE);
}   /*  End Function file_tester_cbk  */

static void fileselect_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  [SUMMARY] Load segments callback.
    <w> The file selector.
    <client_data> Client data.
    <call_data> A pointer to the selected file.
    [RETURNS] Nothing.
*/
{
    create_segment_list (client_data, call_data);
    compute ();
}   /*  End Function fileselect_cbk  */

static void apply_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Apply callback.
    <w> The widget.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    Widget dlg;
    segment_list *list = client_data;
    String colour;

    dlg = XtNameToWidget (list->popup_widget, "topForm.colourDialog");
    XtVaGetValues (dlg, XtNvalue, &colour, NULL);
    strcpy (list->colourname, colour);
    compute ();
}   /*  End Function apply_cbk  */
