/*  setup_vrender.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the <vrender> initialisation code.


    Written by      Richard Gooch   14-DEC-1996: Moved code from initialise.c

    Updated by      Richard Gooch   21-JAN-1997: Removed threshold test for
  opacity in Hot Gas Substances shader.

    Updated by      Richard Gooch   29-MAY-1997: Made use of verbose flag.

    Updated by      Richard Gooch   9-DEC-1997: Split functionality.

    Updated by      Richard Gooch   11-JUN-1998: Use voxel threshold in
  front voxel shader.

    Last updated by Richard Gooch   19-MAR-1999: Removed visual type
  limitations.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <karma_vrender.h>
#include <karma_conn.h>
#include <karma_ds.h>
#include <karma_r.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
#include "render.h"


/*  Slow shaders  */
EXTERN_FUNCTION (void shader_max_voxel, () );
EXTERN_FUNCTION (void shader_min_voxel, () );
EXTERN_FUNCTION (void shader_front_voxel, () );
EXTERN_FUNCTION (void shader_sum_voxels, () );
EXTERN_FUNCTION (void shader_hot_gas_mono, () );
EXTERN_FUNCTION (void shader_hot_gas_subs, () );
EXTERN_FUNCTION (void shader_hot_gas_fullcolour, () );


/*  Private functions  */
STATIC_FUNCTION (void shader_fast_max_voxel,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );
STATIC_FUNCTION (void shader_fast_front_voxel,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );
STATIC_FUNCTION (void shader_fast_min_voxel,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );
STATIC_FUNCTION (void shader_fast_sum_voxels,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );
STATIC_FUNCTION (void shader_fast_hot_gas_mono,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );
STATIC_FUNCTION (void shader_fast_hot_gas_subs,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );
STATIC_FUNCTION (void shader_fast_hot_gas_fullcolour,
		 (signed char *ray, int length,
		  double *minimum_image_value, double *maximum_image_value,
		  void *pixel_ptr) );


/*  Public functions follow  */

void initialise_vrender_package (void)
/*  [PURPOSE] This routine will initialise the xraycast_3d module.
    [RETURNS] Nothing.
*/
{
    packet_desc pack_desc;
    unsigned int elem_types[3];
    char *elem_names[3];
    char blank_value[sizeof (double)];
    /*static char function_name[] = "initialise_vrender_package";*/

    pack_desc.element_types = elem_types;
    pack_desc.element_desc = elem_names;
    /*  Pseudocolour/byte shaders  */
    elem_types[0] = K_BYTE;
    elem_names[0] = "Data Value";
    pack_desc.num_elements = 1;
    *(signed char *) blank_value = -128;
    vrender_register_shader (shader_max_voxel, shader_fast_max_voxel,
			     "max_voxel", &pack_desc,
			     blank_value, NULL, FALSE);
    vrender_register_shader (shader_min_voxel, shader_fast_min_voxel,
			     "min_voxel", &pack_desc,
			     blank_value, NULL, FALSE);
    vrender_register_shader (shader_front_voxel, shader_fast_front_voxel,
			     "front_voxel", &pack_desc,
			     blank_value, NULL, FALSE);
    /*  Pseudocolour/float shaders  */
    elem_types[0] = K_FLOAT;
    *(float *) blank_value = 0.0;
    vrender_register_shader (shader_sum_voxels, shader_fast_sum_voxels,
			     "sum_voxels", &pack_desc,
			     blank_value, NULL, FALSE);
    vrender_register_shader (shader_hot_gas_mono, shader_fast_hot_gas_mono,
			     "hot_gas_mono", &pack_desc,
			     blank_value, NULL, FALSE);
    /*  RGB shaders  */
    elem_types[0] = K_UBYTE;
    elem_types[1] = K_UBYTE;
    elem_types[2] = K_UBYTE;
    elem_names[0] = "Red Intensity";
    elem_names[1] = "Green Intensity";
    elem_names[2] = "Blue Intensity";
    pack_desc.num_elements = 3;
    *(unsigned char *) blank_value = 0;
    *(unsigned char *) (blank_value + 1) = 0;
    *(unsigned char *) (blank_value + 2) = 0;
    vrender_register_shader (shader_hot_gas_fullcolour,
			     shader_fast_hot_gas_fullcolour,
			     "hot_gas_continuous",
			     &pack_desc, blank_value, NULL, FALSE);
    vrender_register_shader (shader_hot_gas_subs, shader_fast_hot_gas_subs,
			     "hot_gas_substances",
			     &pack_desc, blank_value, NULL, FALSE);
    vrender_register_shader (shader_hot_gas_fullcolour,
			     shader_fast_hot_gas_fullcolour,
			     "hot_gas_three_colour",
			     &pack_desc, blank_value, NULL, FALSE);
}   /*  End Function initialise_vrender_package  */

KVolumeRenderContext create_vrender_context (flag verbose, flag show_timings)
/*  [PURPOSE] This routine will create a volume rendering context.
    <verbose> If TRUE verbose output is generated.
    <show_timings> If TRUE timing output is generated.
    [RETURNS] A volume rendering context.
*/
{
    KVolumeRenderContext ctx;
    double sep;
    char *default_shader;
    static char function_name[] = "create_vrender_context";

    default_shader = "max_voxel";
    /*  Compute distance independent of subcube size
	Assume 1 voxel corresponds to 1 pixel (invalid if image expanded)
	Assume screen is 305 mm across (17 inch MAG) with 960 pixels
	Hence 1 mm is 960 / 305 pixels
	If image size is doubled in x and y then 1 voxel corresponds to
	2 pixels
	*/
    sep = (double) DEFAULT_EYE_SEP * 960.0 / 305.0;  /*  Distance in pixels  */
    /*  TODO: sep /= image_expand_factor  */
    if ( ( ctx =
	   vrender_create_context (NULL,
				   VRENDER_CONTEXT_ATT_SHADER, default_shader,
				   VRENDER_CONTEXT_ATT_EYE_SEPARATION, sep,
				   VRENDER_CONTEXT_ATT_VERBOSE, verbose,
				   VRENDER_CONTEXT_ATT_SHOW_TIMINGS,
				   show_timings,
				   VRENDER_CONTEXT_ATT_END) )
	== NULL )
    {
	m_abort (function_name, "volume rendering context");
    }
    return (ctx);
}   /*  End Function create_vrender_context  */


/*  Private functions follow  */

static void shader_fast_max_voxel (signed char *ray, int length,
				   double *minimum_image_value,
				   double *maximum_image_value,
				   void *pixel_ptr)
{
    int pixel = -129;
    int voxel;
    double d_pixel;
    signed char *ray_end;
    extern int voxel_sum_position;

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < -127) continue;
	if (voxel > pixel) pixel = voxel;
    }
    if (pixel < -127) pixel = voxel_sum_position;
    else
    {
	d_pixel = (double) pixel;
	if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
	if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    }
    *(signed char *) pixel_ptr = pixel;
}   /*  End Function shader_fast_max_voxel  */

static void shader_fast_min_voxel (signed char *ray, int length,
				   double *minimum_image_value,
				   double *maximum_image_value,
				   void *pixel_ptr)
{
    int pixel = 128;
    int voxel;
    double d_pixel;
    signed char *ray_end;
    extern int voxel_sum_position;

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < -127) continue;
	if (voxel < pixel) pixel = voxel;
    }
    if (pixel > 127) pixel = voxel_sum_position;
    else
    {
	d_pixel = (double) pixel;
	if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
	if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    }
    *(signed char *) pixel_ptr = pixel;
}   /*  End Function shader_fast_min_voxel  */

static void shader_fast_front_voxel (signed char *ray, int length,
				     double *minimum_image_value,
				     double *maximum_image_value,
				     void *pixel_ptr)
{
    int pixel = -128;
    int voxel;
    double d_pixel;
    signed char *ray_end;
    extern int voxel_threshold;

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < voxel_threshold) continue;
	pixel = voxel;
	ray = ray_end;
    }
    if (pixel > -128)
    {
	d_pixel = (double) pixel;
	if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
	if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    }
    *(signed char *) pixel_ptr = pixel;
}   /*  End Function shader_fast_front_voxel  */

static void shader_fast_sum_voxels (signed char *ray, int length,
				    double *minimum_image_value,
				    double *maximum_image_value,
				    void *pixel_ptr)
{
    int pixel = 0;
    int voxel;
    double d_pixel;
    signed char *ray_end;
    extern int voxel_sum_position;

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < -127) continue;
	pixel += voxel - voxel_sum_position;
    }
    d_pixel = (double) pixel;
    if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
    if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    *(float *) pixel_ptr = pixel;
}   /*  End Function shader_fast_sum_voxels  */

static void shader_fast_hot_gas_mono (signed char *ray, int length,
				      double *minimum_image_value,
				      double *maximum_image_value,
				      void *pixel_ptr)
{
    int voxel;
    int index;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float intensity = 0.0;
    float cumulative_transparency = 1.0;
    double d_pixel;
    signed char *ray_end;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < -127) continue;
	index = voxel + 127;
	opacity = mono_hg_opacities[index];
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	intensity += mono_hg_intensities[index] * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold)
	{
	    ray = ray_end;
	}
    }
    d_pixel = (double) intensity;
    if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
    if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    *(float *) pixel_ptr = intensity;
}   /*  End Function shader_fast_hot_gas_mono  */

static void shader_fast_hot_gas_subs (signed char *ray, int length,
				      double *minimum_image_value,
				      double *maximum_image_value,
				      void *pixel_ptr)
{
    int voxel;
    int index;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float old_opacity = 0.0;
    float old_red = 0.0;
    float old_green = 0.0;
    float old_blue = 0.0;
    int old_index = -1;
    float red = 0.0;
    float green = 0.0;
    float blue = 0.0;
    float cumulative_transparency = 1.0;
    signed char *ray_end;
    extern int num_substances;
    extern unsigned int bad_substance_count;
    extern char voxel_to_substance_indices[256];
    extern substance substances[MAX_SUBSTANCES];

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < -127) continue;
	if ( (index = voxel_to_substance_indices[voxel + 128])
	     >= num_substances )
	{
	    /*  Outside substance table: skip  */
	    ++bad_substance_count;
	    continue;
	}
	if (old_index != index)
	{
	    old_opacity = substances[index].opacity;
	    old_red = substances[index].red;
	    old_green = substances[index].green;
	    old_blue = substances[index].blue;
	    old_index = index;
	}
	transparency = one - old_opacity;
	emissivity = old_opacity * cumulative_transparency;
	red += old_red * emissivity;
	green += old_green * emissivity;
	blue += old_blue * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold)
	{
	    ray = ray_end;
	}
    }
    emissivity = 255.0;
    *(unsigned char *) pixel_ptr = (int) (red * emissivity);
    *( (unsigned char *) pixel_ptr + 1 ) = (int) (green * emissivity);
    *( (unsigned char *) pixel_ptr + 2 ) = (int) (blue * emissivity);
}   /*  End Function shader_fast_hot_gas_subs  */

static void shader_fast_hot_gas_fullcolour (signed char *ray, int length,
					    double *minimum_image_value,
					    double *maximum_image_value,
					    void *pixel_ptr)
{
    int voxel;
    int index;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float red = 0.0;
    float green = 0.0;
    float blue = 0.0;
    float cumulative_transparency = 1.0;
    signed char *ray_end;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];

    ray_end = ray + length;
    while (ray < ray_end)
    {
	voxel = *ray++;
	if (voxel < -127) continue;
	index = voxel + 127;
	opacity = fullcolour_hg_opacities[index];
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	red += fullcolour_hg_reds[index] * emissivity;
	green += fullcolour_hg_greens[index] * emissivity;
	blue += fullcolour_hg_blues[index] * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold)
	{
	    ray = ray_end;
	}
    }
    emissivity = 255.0;
    *(unsigned char *) pixel_ptr = (int) (red * emissivity);
    *( (unsigned char *) pixel_ptr + 1 ) = (int) (green * emissivity);
    *( (unsigned char *) pixel_ptr + 2 ) = (int) (blue * emissivity);
}   /*  End Function shader_fast_hot_gas_fullcolour  */
