/*  subs_hg.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the user-friendly substance controls.


    Written by      Richard Gooch   21-NOV-1993

    Updated by      Richard Gooch   29-NOV-1993

    Updated by      Richard Gooch   2-DEC-1993: Made use of <canvas_get_colour>
  routine.

    Updated by      Richard Gooch   19-MAY-1994: Took account of change to
  <canvas_get_colour> routine.

    Updated by      Richard Gooch   31-JUL-1994: Created
  <setup_substance_canvases> routine.

    Updated by      Richard Gooch   3-OCT-1994: Made use of
  canvas_use_log_scale  .

    Updated by      Richard Gooch   29-AUG-1995: Changed to <kwin_resize> to
  refresh canvas.

    Updated by      Richard Gooch   1-SEP-1995: Changed to <kcmap_va_create>.

    Updated by      Richard Gooch   21-JUL-1996: Changed to
  <canvas_convert_from_canvas_coords>.

    Updated by      Richard Gooch   22-NOV-1996: Moved to
  <offset_position_func>.

    Updated by      Richard Gooch   28-NOV-1996: Took account of substance
  changes.

    Updated by      Richard Gooch   21-JAN-1997: Increased height of opacity
  controls.

    Updated by      Richard Gooch   13-APR-1997: Fixed parent refresh function
  to make child canvases visible prior to refreshing them.

    Updated by      Richard Gooch   25-MAY-1997: Renamed to
  <process_substances> and actually transmit them now.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   29-JUN-1998: Moved to callback mechanism to
  update histogram canvas when the histogram changes.

    Updated by      Richard Gooch   8-SEP-1998: Made use of new function
  <xc_get_location2>.

    Updated by      Richard Gooch   5-JAN-1999: Now update global variable
  <substances_all_transparent>.

    Updated by      Richard Gooch   14-APR-1999: Clip opacity in input events
  and when reading.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kcmap.h>
  and <karma_kwin.h>.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Last updated by Richard Gooch   21-MAR-2003: Switched to function
  <overlay_create_viewable>.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_xc.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_p.h>
#include <karma_c.h>
#include "xraycast_3d.h"


#define MAX_INTENSITY 65535
#define SUBSTANCE_HEIGHT 0.85
#define ACTIVE_BOTTOM    0.9
#define COLOURBAR_HEIGHT 128
#define CURSOR_SIZE 5


/*  Private functions  */
STATIC_FUNCTION (flag get_and_clip_value,
		 (float *value, CONST packet_desc *pack_desc,
		  CONST char *packet, CONST char *name) );
STATIC_FUNCTION (void substance_pixcanvas_refresh_func,
		 (KPixCanvas canvas, int width, int height, void **info) );
STATIC_FUNCTION (void cmapctrl_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (flag cmapctrl_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info) );
STATIC_FUNCTION (void colourbar_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (flag colourbar_canvas_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info, void **f_info,
		  double x_lin, double y_lin, int x_pix, int y_pix) );
STATIC_FUNCTION (void histogram_canvas_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale, Kcolourmap cmap,
		  flag cmap_resize, void **info) );
STATIC_FUNCTION (void update_colours, (substance *substances,
				       flag cmap_only) );

static void hsb_to_rgb (/* hue, saturation, brightness, red, green, blue */);
static void draw_subs_pos_cursor (/* canvas, x, y, pixel_value */);
STATIC_FUNCTION (void draw_hs_pos_cursor,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned long pixel_value) );
static void draw_b_pos_cursor (/* canvas, y, pixel_value */);
STATIC_FUNCTION (flag histogram_change_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );


/*  Public data  */
substance temp_substances[MAX_SUBSTANCES];

/*  Private data  */
static KPixCanvas cmapctrl_pixcanvas = NULL;
static KPixCanvas colourbar_pixcanvas = NULL;
static KPixCanvas histogram_pixcanvas = NULL;
static KWorldCanvas colourbar_worldcanvas = NULL;
/*  Substance information  */
static unsigned int active_substance = 0;
static double subs_pos_old_x = TOOBIG;
static double subs_pos_old_y = TOOBIG;
static double hs_pos_old_x[MAX_SUBSTANCES];
static double hs_pos_old_y[MAX_SUBSTANCES];
static double b_pos_old_y[MAX_SUBSTANCES];


/*  Public routines follow  */

void setup_substance_canvases (KPixCanvas parent, Kdisplay dpy_handle,
			       void *offset_widget,
			       unsigned long foreground_pixel,
			       unsigned long background_pixel,
			       Kcolourmap *subs_cmap)
/*  This routine will setup the substance canvases.
    The parent pixel canvas must be given by  parent  .
    The display handle for the colourmap must be given by  dpy_handle  .
    <offset_widget> The offset track label widget.
    <subs_cmap> The colourmap to display the substances is written here. NULL
    may be written here.
    The routine returns nothing.
*/
{
    KWorldCanvas cmapctrl_worldcanvas, histogram_worldcanvas;
    KViewableOverlayList vlist;
    int width, height;
    double frac;
    struct win_scale_type win_scale;
    extern KOverlayList histogram_olist;
    extern KCallbackList histogram_change_event_list;
    extern int num_substances;
    extern unsigned int histogram_peak;
    extern substance temp_substances[MAX_SUBSTANCES];
    static char function_name[] = "setup_substance_canvases";

    kwin_get_size (parent, &width, &height);
    /*  Setup the colourmap control canvas  */
    if ( ( cmapctrl_pixcanvas =
	   kwin_create_child (parent, 0, 0, width, height - COLOURBAR_HEIGHT,
			      TRUE) ) == NULL )
    {
	m_abort (function_name, "cmapctrl_pixcanvas");
    }
    if (dpy_handle == NULL) *subs_cmap = NULL;
    else
    {
	if ( ( *subs_cmap =
	       kcmap_va_create (DEFAULT_COLOURMAP_NAME, num_substances, FALSE,
				dpy_handle, xc_alloc_colours, xc_free_colours,
				xc_store_colours, xc_get_location2,
				KCMAP_ATT_END) )
	     == NULL )
	{
	    fprintf (stderr, "Error creating main colourmap\n");
	    a_prog_bug (function_name);
	}
    }
    /*  Create the colourmap control world canvas  */
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    win_scale.x_min = 0.0;
    win_scale.x_max = 1.0;
    win_scale.y_min = 0.0;
    win_scale.y_max = 1.0;
    win_scale.z_min = 0.0;
    win_scale.z_max = 7.0;
    win_scale.z_scale = K_INTENSITY_SCALE_LINEAR;
    win_scale.min_sat_pixel = background_pixel;
    win_scale.max_sat_pixel = foreground_pixel;
    if ( ( cmapctrl_worldcanvas = canvas_create (cmapctrl_pixcanvas,
						 *subs_cmap, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (cmapctrl_worldcanvas,
				  cmapctrl_canvas_refresh_func,
				  (void *) foreground_pixel);
    canvas_register_position_event_func (cmapctrl_worldcanvas,
					 cmapctrl_canvas_position_func,
					 (void *) foreground_pixel);
    /*  Setup the colourbar canvas  */
    /*  Create the colourbar pixel canvas  */
    if ( ( colourbar_pixcanvas =
	   kwin_create_child (parent,
			      0, height - COLOURBAR_HEIGHT,
			      width, COLOURBAR_HEIGHT, TRUE) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    /*  Create the colourbar world canvas  */
    win_scale.x_min = -127.0;
    win_scale.x_max = 127.0;
    win_scale.y_min = 0.0;
    win_scale.y_max = 1.0;
    win_scale.z_min = 0.0;
    win_scale.z_max = 7.0;
    win_scale.z_scale = K_INTENSITY_SCALE_LINEAR;
    win_scale.conv_type = KIMAGE_COMPLEX_CONV_REAL;
    win_scale.min_sat_pixel = background_pixel;
    win_scale.max_sat_pixel = foreground_pixel;
    if ( ( colourbar_worldcanvas = canvas_create (colourbar_pixcanvas,
						  *subs_cmap, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_register_refresh_func (colourbar_worldcanvas,
				  colourbar_canvas_refresh_func, NULL);
    canvas_register_position_event_func (colourbar_worldcanvas,
					 colourbar_canvas_position_func,
					 NULL);
    /*  Setup the histogram canvas  */
    /*  Create the histogram pixel canvas  */
    if ( ( histogram_pixcanvas =
	   kwin_create_child (parent, 0, 0, width, height - COLOURBAR_HEIGHT,
			      TRUE) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    /*  Create the histogram world canvas  */
    win_scale.x_min = -127.0;
    win_scale.x_max = 127.0;
    if (histogram_peak < 1)
    {
	/*  No histogram data yet  */
	win_scale.y_min = 0.0;
	win_scale.y_max = 1.0;
    }
    else
    {
	/*  Have histogram data: setup logarithmic scale  */
	win_scale.y_max = histogram_peak;
	frac = log10 (win_scale.y_max);
	if (frac > 6.0) frac = 6.0;
	win_scale.y_min = win_scale.y_max / pow (10.0, frac);
	if (win_scale.y_min < 1.0) win_scale.y_min = 1.0;
	win_scale.y_log = TRUE;
    }
    win_scale.z_min = 0.0;
    win_scale.z_max = 7.0;
    win_scale.z_scale = K_INTENSITY_SCALE_LINEAR;
    win_scale.conv_type = KIMAGE_COMPLEX_CONV_REAL;
    win_scale.min_sat_pixel = background_pixel;
    win_scale.max_sat_pixel = foreground_pixel;
    if ( ( histogram_worldcanvas = canvas_create (histogram_pixcanvas,
						  *subs_cmap, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    canvas_use_log_scale (histogram_worldcanvas, FALSE, TRUE);
    vlist = overlay_create_viewable (histogram_olist, histogram_worldcanvas);
    overlay_set_active (vlist, TRUE, FALSE, TRUE, FALSE, 0);
    canvas_register_refresh_func (histogram_worldcanvas,
				  histogram_canvas_refresh_func, NULL);
    canvas_register_position_event_func (histogram_worldcanvas,
					 offset_position_func,
					 offset_widget);
    kwin_register_refresh_func (parent, substance_pixcanvas_refresh_func,
				NULL);
    c_register_callback (&histogram_change_event_list,
			 generic_histogram_change_callback,
			 histogram_worldcanvas,
			 NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&histogram_change_event_list,
			 histogram_change_callback,
			 cmapctrl_pixcanvas,
			 NULL, FALSE, NULL, FALSE, FALSE);
    update_colours (temp_substances, TRUE);
}   /*  End Function setup_substance_canvases  */

void process_substances (KVolumeRenderContext context)
/*  [SUMMARY] Process substance information.
    <context> The volume rendering context.
    [RETURNS] Nothing.
*/
{
    int count, substance_count;
    extern KPixCanvas substance_pixcanvas;
    extern flag substances_needed, substances_all_transparent;
    extern int num_substances;
    extern substance substances[MAX_SUBSTANCES];
    extern substance temp_substances[MAX_SUBSTANCES];
    extern char voxel_to_substance_indices[256];
    /*static char function_name[] = "process_substances";*/

    /*  Compute voxel to substance array  */
    for (count = -128; count < 128; ++count)
    {
	voxel_to_substance_indices[count + 128] = num_substances;
	for (substance_count = 0; substance_count < num_substances;
	     ++substance_count)
	{
	    if (count <= substances[substance_count].max_voxel)
	    {
		/*  Falls within this substance  */
		voxel_to_substance_indices[count + 128] = substance_count;
		substance_count = num_substances;
	    }
	}
    }
    substances_all_transparent = TRUE;
    for (count = 0; count < num_substances; ++count)
    {
	temp_substances[count].max_voxel = substances[count].max_voxel;
	temp_substances[count].red = substances[count].red;
	temp_substances[count].green = substances[count].green;
	temp_substances[count].blue = substances[count].blue;
	temp_substances[count].opacity = substances[count].opacity;
	if (substances[count].opacity > 0.0) substances_all_transparent =FALSE;
    }
    update_colours (temp_substances, TRUE);
    kwin_resize (substance_pixcanvas, TRUE, 0, 0, 0, 0);
    transmit_substances (context);
    if (substances_needed) compute ();
}   /*  End Function process_substances  */

void transmit_substances (KVolumeRenderContext context)
/*  [SUMMARY] Transmit substance information to slaves.
    <context> The volume rendering context.
    [RETURNS] Nothing.
*/
{
    int count;
    char *message, *ptr;
    extern int num_substances;
    extern substance substances[MAX_SUBSTANCES];
    extern char voxel_to_substance_indices[256];
    static char function_name[] = "transmit_substances";

    message = m_alloc_scratch (num_substances * 17 + 256, function_name);
    m_copy (message, voxel_to_substance_indices, 256);
    for (count = 0, ptr = message + 256; count < num_substances;
	 ++count, ptr += 17)
    {
	*(signed char *) ptr = substances[count].max_voxel;
	p_write_buf_float (ptr + 1, substances[count].red);
	p_write_buf_float (ptr + 5, substances[count].green);
	p_write_buf_float (ptr + 9, substances[count].blue);
	p_write_buf_float (ptr + 13, substances[count].opacity);
    }
    vrender_send_message (context, MtoS_SUBSTANCES_INFO, message,
			  num_substances * 17 + 256, FALSE);
    m_free_scratch ();
}   /*  End Function transmit_substances  */

unsigned int scan_substance_structure (substance *substances,
				       packet_desc *pack_desc, char *packet)
/*  This routine will scan a general data structure for substance information.
    The substance information will be written to the storage pointed to by
    substances  .
    The top level packet descriptor for the data structure must be pointed to
    by  pack_desc  .
    The top level packet data must be pointed to by  packet  .
    [RETURNS] The number of substances on success, else 0.
*/
{
    unsigned int substance_count;
    unsigned int parent_type;
    unsigned int index;
    unsigned int arr_pack_size;
    double value[2];
    char *array;
    array_desc *arr_desc;
    static char function_name[] = "scan_substance_structure";

    switch ( ds_get_handle_in_packet (pack_desc, packet, "Substance Number",
				      (CONST char **) NULL, (double *) NULL, 0,
				      (char **) &arr_desc, &array,
				      &parent_type, &index) )
    {
      case IDENT_NOT_FOUND:
	fprintf (stderr, "Item: \"Substance Number\" not found\n");
	return (0);
        /*break;*/
      case IDENT_MULTIPLE:
	fprintf (stderr,
		 "Multiple occurrences of item: \"Substance Number\"\n");
	return (0);
	/*break;*/
      case IDENT_DIMENSION:
	break;
      case IDENT_ELEMENT:
	fprintf (stderr,
		 "Item: \"Substance Number\" must not be an element\n");
	return (0);
	/*break;*/
      default:
	fprintf (stderr,
		 "Illegal return value from function: ds_get_handle_in_packet");
	a_prog_bug (function_name);
	break;
    }
    if (arr_desc->num_dimensions > 1)
    {
	fprintf (stderr, "Substance array must have only one dimension\n");
	return (0);
    }
    if (arr_desc->dimensions[0]->length >= MAX_SUBSTANCES)
    {
	fprintf (stderr, "Too much substance information\n");
	return (0);
    }
    arr_pack_size = ds_get_packet_size (arr_desc->packet);
    for (substance_count = 0;
	 substance_count < arr_desc->dimensions[0]->length;
	 ++substance_count, array += arr_pack_size)
    {
	if ( !ds_get_unique_named_value (arr_desc->packet, array,
					 "Maximum Voxel Value",
					 NULL, value) )
	{
	    fprintf (stderr, "Error converting value for: \"%s\"\n",
		     "Maximum Voxel Value");
	    return (0);
	}
	substances[substance_count].max_voxel = value[0];
	if ( !get_and_clip_value (&substances[substance_count].red,
				  arr_desc->packet, array, "Red Intensity") )
	    return (0);
	if ( !get_and_clip_value (&substances[substance_count].green,
				  arr_desc->packet, array, "Green Intensity") )
	    return (0);
	if ( !get_and_clip_value (&substances[substance_count].blue,
				  arr_desc->packet, array, "Blue Intensity") )
	    return (0);
	if ( !get_and_clip_value (&substances[substance_count].opacity,
				  arr_desc->packet, array, "Opacity") )
	    return (0);
    }
    return (arr_desc->dimensions[0]->length);
}   /*  End Function scan_substance_structure  */

flag make_substance_structure (substance *substances,
			       unsigned int num_substances,
			       packet_desc **pack_desc, char **packet)
/*  This routine will create a Karma data structure to contain substance
    information.
    The substance information must be pointed to by  substances  .
    The number of substances must be given by  num_substances  .
    The routine will write a pointer to the top level packet descriptor for the
    data structure to the storage pointed to by  pack_desc  .
    The routine will write a pointer to the top level packet data for the data
    structure to the storage pointed to by  packet  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    unsigned int count;
    char *array;
    array_desc *arr_desc;
    packet_desc *arr_pack_desc;
    dim_desc *dim;
    static unsigned int element_types[5] = 
    {
	 K_INT, K_FLOAT, K_FLOAT, K_FLOAT, K_FLOAT
    };
    static char *element_names[5] =
    {
	"Maximum Voxel Value",
	"Red Intensity",
	"Green Intensity",
	"Blue Intensity",
	"Opacity"
    };
    static char function_name[] = "make_substance_structure";

    if ( ( *pack_desc = ds_alloc_packet_desc (1) ) == NULL )
    {
	m_error_notify (function_name, "top level packet descriptor");
	return (FALSE);
    }
    if ( ( arr_desc = ds_alloc_array_desc (1, 0) ) == NULL )
    {
	m_error_notify (function_name, "array descriptor");
	ds_dealloc_packet (*pack_desc, NULL);
	return (FALSE);
    }
    (**pack_desc).element_desc[0] = (char *) arr_desc;
    (**pack_desc).element_types[0] = K_ARRAY;
    if ( ( dim = ds_alloc_dim_desc ("Substance Number", num_substances,
				    0.0, (double) (num_substances - 1), TRUE) )
	== NULL )
    {
	m_error_notify (function_name, "dimension descriptor");
	ds_dealloc_packet (*pack_desc, NULL);
	return (FALSE);
    }
    arr_desc->dimensions[0] = dim;
    arr_desc->lengths[0] = dim->length;
    if ( ( arr_pack_desc = ds_alloc_packet_desc (5) ) == NULL )
    {
	m_error_notify (function_name, "array packet descriptor");
	ds_dealloc_packet (*pack_desc, NULL);
	return (FALSE);
    }
    arr_desc->packet = arr_pack_desc;
    for (count = 0; count < 5; ++count)
    {
	if ( ( arr_pack_desc->element_desc[count] =
	       st_dup (element_names[count]) )
	     == NULL )
	{
	    m_error_notify (function_name, "element name");
	    ds_dealloc_packet (*pack_desc, NULL);
	    return (FALSE);
	}
	arr_pack_desc->element_types[count] = element_types[count];
    }
    if ( ( *packet = ds_alloc_data (*pack_desc, TRUE, TRUE) ) == NULL )
    {
	m_error_notify (function_name, "data space");
	ds_dealloc_packet (*pack_desc, NULL);
	return (FALSE);
    }
    array = *(char **) *packet;
    /*  Write substance information  */
    for (count = 0; count < num_substances; ++count)
    {
	*(int *) array = substances[count].max_voxel;
	array += sizeof (int);
	*(float *) array = substances[count].red;
	array += sizeof (float);
	*(float *) array = substances[count].green;
	array += sizeof (float);
	*(float *) array = substances[count].blue;
	array += sizeof (float);
	*(float *) array = substances[count].opacity;
	array += sizeof (float);
    }
    return (TRUE);
}   /*  End Function make_substance_structure  */


/*  Private functions follow  */

static flag get_and_clip_value (float *value, CONST packet_desc *pack_desc,
				CONST char *packet, CONST char *name)
{
    double val[2];

    if ( !ds_get_unique_named_value (pack_desc, packet, name, NULL, val) )
    {
	fprintf (stderr, "Error converting value for: \"%s\"\n", name);
	return (FALSE);
    }
    if (val[0] < 0.0)
    {
	fprintf (stderr, "%s value: %g is less than zero: clipping\n",
		 name, val[0]);
	*value = 0.0;
	return (TRUE);
    }
    if (val[0] > 1.0)
    {
	fprintf (stderr, "%s value: %g is larger than one: clipping\n",
		 name, val[0]);
	*value = 1.0;
	return (TRUE);
    }
    *value = val[0];
    return (TRUE);
}   /*  End Function get_and_clip_value  */

static void substance_pixcanvas_refresh_func (canvas, width, height, info)
/*  This routine will process a refresh event for a pixel canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
KPixCanvas canvas;
int width;
int height;
void **info;
{
    kwin_set_attributes (cmapctrl_pixcanvas,
			 KWIN_ATT_VISIBLE, TRUE,
			 KWIN_ATT_END);
    kwin_resize (cmapctrl_pixcanvas, FALSE,
		 0, 0, width, height - COLOURBAR_HEIGHT);
    kwin_set_attributes (colourbar_pixcanvas,
			 KWIN_ATT_VISIBLE, TRUE,
			 KWIN_ATT_END);
    kwin_resize (colourbar_pixcanvas, FALSE,
		 0, height - COLOURBAR_HEIGHT,
		 width, COLOURBAR_HEIGHT);
    kwin_set_attributes (histogram_pixcanvas,
			 KWIN_ATT_VISIBLE, TRUE,
			 KWIN_ATT_END);
    kwin_resize (histogram_pixcanvas, FALSE,
		 0, 0, width, height - COLOURBAR_HEIGHT);
}   /*  End Function substance_pixcanvas_refresh_func  */


/*  Colourmap control world canvas callbacks  */

static void cmapctrl_canvas_refresh_func (canvas, width, height, win_scale,
					  cmap, cmap_resize, info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
KWorldCanvas canvas;
int width;
int height;
struct win_scale_type *win_scale;
Kcolourmap cmap;
flag cmap_resize;
void **info;
{
    unsigned int count;
    unsigned long pixel_value;
    extern int num_substances;
    extern unsigned int active_substance;
    static flag first_time = TRUE;

    if (first_time)
    {
	for (count = 0; count < num_substances; ++count)
	{
	    hs_pos_old_x[count] = (double) (count + 0.5) / 8.0;
	    hs_pos_old_y[count] = (double) (count + 0.5) / 8.0;
	    b_pos_old_y[count] = (double) (count + 0.5) / 8.0;
	}
	first_time = FALSE;
    }
    pixel_value = (uaddr) *info;
    draw_subs_pos_cursor (canvas, subs_pos_old_x, subs_pos_old_y, pixel_value);
    draw_hs_pos_cursor (canvas, hs_pos_old_x[active_substance],
			hs_pos_old_y[active_substance], pixel_value);
    draw_b_pos_cursor (canvas, b_pos_old_y[active_substance], pixel_value);
    /*  This will be filled with position indicator drawing  */
}   /*  End Function cmapctrl_canvas_refresh_func  */

static flag cmapctrl_canvas_position_func (canvas, x, y, event_code, e_info,
					   f_info)
/*  This routine is a position event consumer for a world canvas.
    The canvas is given by  canvas  .
    The horizontal world co-ordinate of the event will be given by  x  .
    The vertical world co-ordinate of the event will be given by  y  .
    The arbitrary event code is given by  event_code  .
    The arbitrary event information is pointed to by  e_info  .
    The arbitrary function information pointer is pointed to by  f_info  .
    The routine returns TRUE if the event was consumed, else it returns
    FALSE indicating that the event is still to be processed.
*/
KWorldCanvas canvas;
double x;
double y;
unsigned int event_code;
void *e_info;
void **f_info;
{
    KPixCanvas pixcanvas;
    KPixCanvasRefreshArea area;
    unsigned int count;
    unsigned long pixel_value;
    double shift, slope;
    double max_voxel;
    double red, green, blue;
    double hue;
    double saturation;
    double brightness;
    double px, py;
    extern KPixCanvas substance_pixcanvas;
    extern int num_substances;
    extern substance temp_substances[MAX_SUBSTANCES];

    pixcanvas = canvas_get_pixcanvas (canvas);
    pixel_value = (uaddr) *f_info;
    m_clear ( (char *) &area, sizeof area );
    switch (event_code)
    {
      case K_CANVAS_EVENT_LEFT_MOUSE_CLICK:
      case K_CANVAS_EVENT_LEFT_MOUSE_DRAG:
	subs_pos_old_x = x;
	subs_pos_old_y = y;
	/*  Change substance positions  */
	if ( (y < 0.0) || (y >= 1.0) )
	{
	    slope = 1.0;
	}
	else
	{
	    slope = tan (1.57 * y);
	}
	shift = x;
	for (count = 0; count < num_substances; ++count)
	{
	    max_voxel = ( (double) (count + 1) / 8.0 - 0.5) / slope + shift;
	    if (max_voxel < 0.0) max_voxel = 0.0;
	    if (max_voxel > 1.0) max_voxel = 1.0;
	    temp_substances[count].max_voxel = max_voxel * 254.0 - 127.0;
	}
	temp_substances[num_substances - 1].max_voxel = 127;
	kwin_resize (substance_pixcanvas, TRUE, 0, 0, 0, 0);
	break;
      case K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK:
      case K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG:
	if ( DOUBLE_IS_BLANK (hs_pos_old_x[active_substance]) ||
	     DOUBLE_IS_BLANK (hs_pos_old_y[active_substance]) )
	{
	    /*  Nothing to undraw  */
	    area.clear = FALSE;
	}
	else
	{
	    /*  Setup area to undraw old cursor  */
	    /*  Convert to pixel co-ordinates  */
	    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					       &hs_pos_old_x[active_substance],
					       &hs_pos_old_y[active_substance],
					       &px, &py);
	    area.startx = (int) px - CURSOR_SIZE - 1;
	    area.endx = (int) px + CURSOR_SIZE + 1;
	    area.starty = (int) py - CURSOR_SIZE - 1;
	    area.endy = (int) py + CURSOR_SIZE + 1;
	    area.clear = TRUE;
	}
	hs_pos_old_x[active_substance] = x;
	hs_pos_old_y[active_substance] = y;
	kwin_partial_refresh (pixcanvas, 1, &area, FALSE);
	kwin_resize (histogram_pixcanvas, FALSE, 0, 0, 0, 0);
	hue = x * 6.0;
	saturation = y;
	brightness = b_pos_old_y[active_substance];
	hsb_to_rgb (hue, saturation, brightness, &red, &green, &blue);
	temp_substances[active_substance].red = red;
	temp_substances[active_substance].green = green;
	temp_substances[active_substance].blue = blue;
	update_colours (temp_substances, FALSE);
	break;
      case K_CANVAS_EVENT_RIGHT_MOUSE_CLICK:
      case K_CANVAS_EVENT_RIGHT_MOUSE_DRAG:
	if ( DOUBLE_IS_BLANK (b_pos_old_y[active_substance]) )
	{
	    /*  Nothing to undraw  */
	    area.clear = FALSE;
	}
	else
	{
	    /*  Setup area to undraw old cursor  */
	    /*  Convert to pixel co-ordinates  */
	    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					       &x,
					       &b_pos_old_y[active_substance],
					       &px, &py);
	    area.startx = 0;
	    area.endx = CURSOR_SIZE * 2;
	    area.starty = (int) py - 1;
	    area.endy = (int) py + 1;
	    area.clear = TRUE;
	}
	b_pos_old_y[active_substance] = y;
	kwin_partial_refresh (pixcanvas, 1, &area, FALSE);
	kwin_resize (histogram_pixcanvas, FALSE, 0, 0, 0, 0);
	hue = hs_pos_old_x[active_substance] * 6.0;
	saturation = hs_pos_old_y[active_substance];
	brightness = y;
	hsb_to_rgb (hue, saturation, brightness, &red, &green, &blue);
	temp_substances[active_substance].red = red;
	temp_substances[active_substance].green = green;
	temp_substances[active_substance].blue = blue;
	update_colours (temp_substances, FALSE);
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    return (TRUE);
}   /*  End Function cmapctrl_canvas_position_func  */


/*  Colourbar world canvas callbacks  */

static void colourbar_canvas_refresh_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   Kcolourmap cmap, flag cmap_resize,
					   void **info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned int count;
    int last_max_voxel;
    unsigned long red_mask, green_mask, blue_mask, pixel_value;
    float red, green, blue;
    double x0, y0, x1, y1;
    double value[2];
    extern int num_substances;
    extern substance temp_substances[MAX_SUBSTANCES];

    pixcanvas = canvas_get_pixcanvas (canvas);
    if (cmap == NULL)
    {
	kwin_get_attributes (pixcanvas,
			     KWIN_ATT_PIX_RED_MASK, &red_mask,
			     KWIN_ATT_PIX_GREEN_MASK, &green_mask,
			     KWIN_ATT_PIX_BLUE_MASK, &blue_mask,
			     KWIN_ATT_END);
    }
    value[0] = 0.0;
    value[1] = 0.0;
    for (count = 0, last_max_voxel = -127; count < num_substances; ++count)
    {
	if (temp_substances[count].max_voxel <= last_max_voxel) continue;
	/*  Display substance colour  */
	if (cmap == NULL)
	{
	    /*  Need to construct the pixel value directly  */
	    red = (float) red_mask * temp_substances[count].red;
	    green = (float) green_mask * temp_substances[count].green;
	    blue = (float) blue_mask * temp_substances[count].blue;
	    pixel_value = (red_mask & (unsigned long) red) |
		(green_mask & (unsigned long) green) |
		(blue_mask & (unsigned long) blue);
	    canvas_fill_rectangle_p (canvas,
				     (double) last_max_voxel, 0.0,
				     (double) (temp_substances[count].max_voxel -
					       last_max_voxel), SUBSTANCE_HEIGHT,
				     pixel_value);
	}
	else
	{
	    /*  Can simply index into the colourmap  */
	    value[0] = count;
	    canvas_fill_rectangle (canvas,
				   (double) last_max_voxel, 0.0,
				   (double) (temp_substances[count].max_voxel -
					     last_max_voxel), SUBSTANCE_HEIGHT,
				   value);
	}
	/*  Display substance opacity  */
	x0 = (double) last_max_voxel;
	y0 = SUBSTANCE_HEIGHT * temp_substances[count].opacity;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &x0, &y0, &x0, &y0);
	x1 = (double) temp_substances[count].max_voxel;
	y1 = SUBSTANCE_HEIGHT * temp_substances[count].opacity;
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &x1, &y1, &x1, &y1);
	kwin_draw_line (pixcanvas, x0, y0, x1 - 1, y0,
			win_scale->max_sat_pixel);
	kwin_draw_line (pixcanvas, x0, y0 - 1, x1 - 1, y0 - 1,
			win_scale->min_sat_pixel);
	kwin_draw_line (pixcanvas, x0, y0 + 1, x1 - 1, y0 + 1,
			win_scale->min_sat_pixel);
	if (count == active_substance)
	{
	    canvas_fill_rectangle_p (canvas,
				     (double) last_max_voxel, ACTIVE_BOTTOM,
				     (double) (temp_substances[count].max_voxel -
					       last_max_voxel),
				     1.0 - ACTIVE_BOTTOM,
				     win_scale->max_sat_pixel);
	}
	last_max_voxel = temp_substances[count].max_voxel;
    }
}   /*  End Function colourbar_canvas_refresh_func  */

static flag colourbar_canvas_position_func (KWorldCanvas canvas,
					    double x, double y,
					    unsigned int event_code,
					    void *e_info,
					    void **f_info,
					    double x_lin, double y_lin,
					    int x_pix, int y_pix)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    <x_pix> The horizontal pixel co-ordinate.
    <y_pix> The vertical pixel co-ordinate.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    unsigned int count;
    extern KPixCanvas substance_pixcanvas;
    extern int num_substances;
    extern substance temp_substances[MAX_SUBSTANCES];

    if ( (event_code != K_CANVAS_EVENT_LEFT_MOUSE_CLICK) &&
	 (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_CLICK) &&
	 (event_code != K_CANVAS_EVENT_MIDDLE_MOUSE_DRAG) ) return (TRUE);
    for (count = 0; count < num_substances; ++count)
    {
	if (x <= temp_substances[count].max_voxel)
	{
	    /*  Found the substance: now, change active or change opacity  */
	    if (event_code == K_CANVAS_EVENT_LEFT_MOUSE_CLICK)
	    {
		/*  Change active  */
		if (count != active_substance)
		{
		    active_substance = count;
		    kwin_resize (substance_pixcanvas, TRUE, 0, 0, 0, 0);
		}
		return (TRUE);
	    }
	    /*  Change opacity  */
	    temp_substances[count].opacity = y / SUBSTANCE_HEIGHT;
	    if (temp_substances[count].opacity > 1.0)
		temp_substances[count].opacity = 1.0;
	    kwin_resize (canvas_get_pixcanvas (canvas), TRUE, 0, 0, 0, 0);
	    return (TRUE);
	}
    }
    return (TRUE);
}   /*  End Function colourbar_canvas_position_func  */


/*  Histogram world canvas callbacks  */

static void histogram_canvas_refresh_func (canvas, width, height, win_scale,
					   cmap, cmap_resize, info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    The routine returns nothing.
*/
KWorldCanvas canvas;
int width;
int height;
struct win_scale_type *win_scale;
Kcolourmap cmap;
flag cmap_resize;
void **info;
{
    int count;
    unsigned long pixel_value;
    double x, y;
    extern int num_substances;
    extern unsigned int histogram_peak;
    extern substance temp_substances[MAX_SUBSTANCES];

    if (canvas_get_colour (canvas, "yellow4", &pixel_value,
			   (unsigned short *) NULL, (unsigned short *) NULL,
			   (unsigned short *) NULL) != TRUE) return;
    y = histogram_peak;
    for (count = 0; count < num_substances; ++count)
    {
	x = temp_substances[count].max_voxel;
	canvas_draw_line_p (canvas, x, 0.0, x, y, pixel_value);
    }
    canvas_draw_line_p (canvas, -127.0, 0.0, 127.0, 0.0,
			win_scale->max_sat_pixel);
}   /*  End Function histogram_canvas_refresh_func  */

static void hsb_to_rgb (double hue, double saturation, double brightness,
			double *red, double *green, double *blue)
/*  This routine will convert a colour specified in HSB to RGB.
    The hue must be given by  hue  .The legal range is from 0.0 to 6.0
    The saturation must be given by  saturation  .
    The brightness must be given by  brightness  .
    The red intensity will be written to the storage pointed to by  red  .
    The green intensity will be written to the storage pointed to by  green  .
    The blue intensity will be written to the storage pointed to by  blue  .
    Note: the intensity values will be in the range 0.0 to 1.0
    The routine returns nothing.
*/
{
    int i_hue;
    double p, q, t;
    double f;

    p = ( brightness * (1.0 - saturation) );
    i_hue = hue;
    f = hue - (float) i_hue;
    q = ( brightness * ( 1.0 - (saturation * f) ) );
    t = ( brightness * ( 1.0 - ( saturation * (1.0 - f) ) ) );
    switch (i_hue)
    {
      case 0:
	*red = brightness;
	*green = t;
	*blue = p;
	break;
      case 1:
	*red = q;
	*green = brightness;
	*blue = p;
	break;
      case 2:
	*red = p;
	*green = brightness;
	*blue = t;
	break;
      case 3:
	*red = p;
	*green = q;
	*blue = brightness;
	break;
      case 4:
	*red = t;
	*green = p;
	*blue = brightness;
	break;
      case 5:
	*red = brightness;
	*green = p;
	*blue = q;
	break;
    }
}   /*  End Function hsb_to_rgb  */

static void update_colours (substance *substances, flag cmap_only)
/*  This routine will update the colourmap from supplied substance information.
    The substance information must be given by  substances  .
    <cmap_only> If TRUE, only the colourmap is updated.
    The routine returns nothing.
*/
{
    Kcolourmap cmap;
    unsigned int count;
    extern int num_substances;
    static packet_desc *cmap_pack_desc = NULL;
    static char *cmap_packet = NULL;
    static unsigned short *cmap_array = NULL;
    static char function_name[] = "update_colours";

    cmap = canvas_get_cmap (colourbar_worldcanvas);
    if (cmap == NULL)
    {
	if (cmap_only) return;
	kwin_resize (canvas_get_pixcanvas (colourbar_worldcanvas), FALSE,
		     0, 0, 0, 0);
	return;
    }
    if (cmap_array == NULL)
    {
	if ( ( cmap_array =
	       ds_cmap_alloc_colourmap (num_substances, (multi_array **) NULL,
					&cmap_pack_desc, &cmap_packet) )
	     == NULL )
	{
	    m_abort (function_name, "colourmap data structure");
	}
    }
    for (count = 0; count < num_substances; ++count)
    {
	/*  Copy over colour from substance table  */
	cmap_array[count * 3] = temp_substances[count].red * MAX_INTENSITY;
	cmap_array[count * 3 + 1] = temp_substances[count].green*MAX_INTENSITY;
	cmap_array[count * 3 + 2] = temp_substances[count].blue *MAX_INTENSITY;
    }
    if (kcmap_copy_from_struct (cmap, cmap_pack_desc, cmap_packet) != TRUE)
    {
	m_abort (function_name, "colourmap");
    }
}   /*  End Function update_colours  */

static void draw_subs_pos_cursor (canvas, x, y, pixel_value)
/*  This routine will draw the cursor which indicates the substance positions
    control position.
    The canvas must be given by  canvas  .
    The horizontal co-ordinate must be given by  x  .
    The vertical co-ordinate must be given by  y  .
    The pixel value to draw with must be given by  pixel_value  .
    The routine returns nothing.
*/
KWorldCanvas canvas;
double x;
double y;
unsigned long pixel_value;
{
    KPixCanvas pixcanvas;

    if ( DOUBLE_IS_BLANK (x) || DOUBLE_IS_BLANK (y) ) return;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &x, &y, &x,&y);
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_draw_line (pixcanvas, x - CURSOR_SIZE, y, x + CURSOR_SIZE, y,
		    pixel_value);
    kwin_draw_line (pixcanvas, x, y - CURSOR_SIZE, x, y + CURSOR_SIZE,
		    pixel_value);
}   /*  End Function draw_subs_pos_cursor  */

static void draw_hs_pos_cursor (KWorldCanvas canvas, double x, double y,
				unsigned long pixel_value)
/*  [SUMMARY] Draw the hue-saturation cursor.
    [PURPOSE] This routine will draw the cursor which indicates the
    hue-saturation position control position.
    <canvas> The canvas.
    <x> The horizontal co-ordinate.
    <y> The vertical co-ordinate.
    <pixel_value> The pixel value to draw with.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;

    if ( DOUBLE_IS_BLANK (x) || DOUBLE_IS_BLANK (y) ) return;
    /*  Convert to pixel co-ordinates  */
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &x, &y, &x,&y);
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_draw_ellipse (pixcanvas, x, y, CURSOR_SIZE, CURSOR_SIZE, pixel_value);
}   /*  End Function draw_hs_pos_cursor  */

static void draw_b_pos_cursor (canvas, y, pixel_value)
/*  This routine will draw the cursor which indicates the brightness position
    control position.
    The canvas must be given by  canvas  .
    The vertical co-ordinate must be given by  y  .
    The pixel value to draw with must be given by  pixel_value  .
    The routine returns nothing.
*/
KWorldCanvas canvas;
double y;
unsigned long pixel_value;
{
    KPixCanvas pixcanvas;
    double x = 0.0;

    if ( DOUBLE_IS_BLANK (y) ) return;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &x, &y, &x,&y);
    pixcanvas = canvas_get_pixcanvas (canvas);
    kwin_draw_line (pixcanvas, x, y, x + 2 * CURSOR_SIZE, y, pixel_value);
}   /*  End Function draw_b_pos_cursor  */

static flag histogram_change_callback (void *object, void *client1_data,
				       void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KPixCanvas canvas = object;
    kwin_refresh_if_visible (canvas, FALSE);
    return (FALSE);
}   /*  End Function histogram_change_callback  */
