/*  threecolour_hg_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for creating the three colour Hot Gas
  control window.


    Written by      Richard Gooch   6-MAY-1996

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-NOV-1996: Added opacity peak controls
  and display floating point numbers in sliders.

    Updated by      Richard Gooch   29-JUN-1998: Moved to callback mechanism to
  update histogram canvas when the histogram changes.

    Updated by      Richard Gooch   5-MAR-1999: Fixed <add_widget> to increment
  value rather than pointer.

    Last updated by Richard Gooch   22-DEC-2002: Added #include <stdlib.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_canvas.h>
#include <karma_c.h>
#include <Xkw/Canvas.h>
#include <Xkw/SimpleSlider.h>
#include "xraycast_3d.h"


#define COLOUR_COMPONENT_SLIDER_WIDTH 180
#define ALPHA_SLIDER_WIDTH 180
#define CANVAS_WIDTH 553
#define DIVIDER_HEIGHT 1
#define MAX_WIDGETS 32


/*  Private functions  */
STATIC_FUNCTION (void slider_threecolour_hg_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void threecolour_hg_apply_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag new_cube_func,
		 (void *object, void *client1_data, void *call_data,
		  void *client2_data) );
STATIC_FUNCTION (void add_widget, (Widget widget, Widget widgets[MAX_WIDGETS],
				   unsigned int *num_widgets) );


/*  Private data  */
static Widget blank_start_sld = NULL;
static Widget blank_end_sld = NULL;
static Widget section1_end_sld = NULL;
static Widget section2_end_sld = NULL;


/*  Public functions follow  */

Widget create_threecolour_hg_win (Widget main_shell)
/*  [SUMMARY] Create the continuous Hot Gas control window.
    <main_shell> The application main shell.
    [RETURNS] The control window popup.
*/
{
    KPixCanvas pixcanvas;
    KWorldCanvas worldcanvas;
    unsigned int count;
    unsigned int num_widgets = 0;
    Widget threecolour_hg_winpopup;
    Widget form, hist_cnv, close_btn, apply_btn;
    Widget label, divider, rw, gw, bw, iw, ow, pw, w;
    Pixel label_foreground;
    Widget widgets[MAX_WIDGETS];
    extern KCallbackList new_byte_cube_event_list;
    extern KCallbackList histogram_change_event_list;
    extern int threecolour_hg_section1_red_intensity;
    extern int threecolour_hg_section1_green_intensity;
    extern int threecolour_hg_section1_blue_intensity;
    extern int threecolour_hg_section1_intensity_alpha_tenths;
    extern int threecolour_hg_section1_opacity_alpha_tenths;
    extern int threecolour_hg_section1_opacity_peak;
    extern int threecolour_hg_section1_end;
    extern int threecolour_hg_section2_red_intensity;
    extern int threecolour_hg_section2_green_intensity;
    extern int threecolour_hg_section2_blue_intensity;
    extern int threecolour_hg_section2_intensity_alpha_tenths;
    extern int threecolour_hg_section2_opacity_alpha_tenths;
    extern int threecolour_hg_section2_opacity_peak;
    extern int threecolour_hg_section2_end;
    extern int threecolour_hg_section3_red_intensity;
    extern int threecolour_hg_section3_green_intensity;
    extern int threecolour_hg_section3_blue_intensity;
    extern int threecolour_hg_section3_intensity_alpha_tenths;
    extern int threecolour_hg_section3_opacity_alpha_tenths;
    extern int threecolour_hg_section3_opacity_peak;
    extern int threecolour_hg_blank_min, threecolour_hg_blank_max;
    /*static char function_name[] = "create_threecolour_hg_win";*/

    threecolour_hg_winpopup = XtVaCreatePopupShell ("threecolour_hgwinpopup",
					     topLevelShellWidgetClass,
					     main_shell,
					     XtNtitle,
					     "Three Colour Hot Gas Window",
					     NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass,
				    threecolour_hg_winpopup,
				    XtNborderWidth, 0,
				    NULL);
    close_btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,
					 form,
					 XtNlabel, "Close",
					 NULL);
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk,
		   threecolour_hg_winpopup);
    apply_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					 XtNlabel, "Apply",
					 XtNfromHoriz, close_btn,
					 NULL);
    /*  Create section 1 controls  */
    label = XtVaCreateManagedWidget ("sectionLabel", labelWidgetClass, form,
				     XtNlabel, "Section 1 Controls:",
				     XtNfromVert, apply_btn,
				     XtNborderWidth, 0,
				     NULL);
    XtVaGetValues (label,
		   XtNforeground, &label_foreground,
		   NULL);
    /*  Create colour component sliders  */
    rw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Red Intensity",
				  XtNfromVert, label,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section1_red_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section1_red_intensity,
				  NULL);
    add_widget (rw, widgets, &num_widgets);
    gw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Green Intensity",
				  XtNfromVert, label,
				  XtNfromHoriz, rw,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section1_green_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section1_green_intensity,
				  NULL);
    add_widget (gw, widgets, &num_widgets);
    bw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Blue Intensity",
				  XtNfromVert, label,
				  XtNfromHoriz, gw,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section1_blue_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section1_blue_intensity,
				  NULL);
    add_widget (bw, widgets, &num_widgets);
    /*  Create intensity and opacity alpha sliders  */
    iw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Intensity Alpha",
				  XtNfromVert, rw,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, -50,
				  XkwNmaximum, 50,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section1_intensity_alpha_tenths,
				  XkwNvaluePtr,
				  &threecolour_hg_section1_intensity_alpha_tenths,
				  XkwNscaledFormat, "%.1f",
				  NULL);
    XkwSimpleSliderSetScale (iw, 0.1, 0.0, FALSE, TRUE);
    add_widget (iw, widgets, &num_widgets);
    ow = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Opacity Alpha",
				  XtNfromVert, rw,
				  XtNfromHoriz, iw,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, -50,
				  XkwNmaximum, 50,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section1_opacity_alpha_tenths,
				  XkwNvaluePtr,
				  &threecolour_hg_section1_opacity_alpha_tenths,
				  XkwNscaledFormat, "%.1f",
				  NULL);
    XkwSimpleSliderSetScale (ow, 0.1, 0.0, FALSE, TRUE);
    add_widget (ow, widgets, &num_widgets);
    pw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Opacity Peak",
				  XtNfromVert, rw,
				  XtNfromHoriz, ow,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, 100,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section1_opacity_peak,
				  XkwNvaluePtr,
				  &threecolour_hg_section1_opacity_peak,
				  XkwNscaledFormat, "%.2f",
				  NULL);
    XkwSimpleSliderSetScale (pw, 0.01, 0.0, FALSE, TRUE);
    add_widget (pw, widgets, &num_widgets);
    divider = XtVaCreateManagedWidget ("divider", widgetClass, form,
				       XtNfromVert, iw,
				       XtNwidth, CANVAS_WIDTH,
				       XtNheight, DIVIDER_HEIGHT,
				       XtNbackground, label_foreground,
				       NULL);
    /*  Create section 2 controls  */
    label = XtVaCreateManagedWidget ("sectionLabel", labelWidgetClass, form,
				     XtNlabel, "Section 2 Controls:",
				     XtNfromVert, divider,
				     XtNborderWidth, 0,
				     NULL);
    /*  Create colour component sliders  */
    rw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Red Intensity",
				  XtNfromVert, label,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section2_red_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section2_red_intensity,
				  NULL);
    add_widget (rw, widgets, &num_widgets);
    gw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Green Intensity",
				  XtNfromVert, label,
				  XtNfromHoriz, rw,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section2_green_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section2_green_intensity,
				  NULL);
    add_widget (gw, widgets, &num_widgets);
    bw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Blue Intensity",
				  XtNfromVert, label,
				  XtNfromHoriz, gw,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section2_blue_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section2_blue_intensity,
				  NULL);
    add_widget (bw, widgets, &num_widgets);
    /*  Create intensity and opacity alpha sliders  */
    iw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Intensity Alpha",
				  XtNfromVert, rw,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, -50,
				  XkwNmaximum, 50,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section2_intensity_alpha_tenths,
				  XkwNvaluePtr,
				  &threecolour_hg_section2_intensity_alpha_tenths,
				  XkwNscaledFormat, "%.1f",
				  NULL);
    XkwSimpleSliderSetScale (iw, 0.1, 0.0, FALSE, TRUE);
    add_widget (iw, widgets, &num_widgets);
    ow = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Opacity Alpha",
				  XtNfromVert, rw,
				  XtNfromHoriz, iw,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, -50,
				  XkwNmaximum, 50,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section2_opacity_alpha_tenths,
				  XkwNvaluePtr,
				  &threecolour_hg_section2_opacity_alpha_tenths,
				  XkwNscaledFormat, "%.1f",
				  NULL);
    XkwSimpleSliderSetScale (ow, 0.1, 0.0, FALSE, TRUE);
    add_widget (ow, widgets, &num_widgets);
    pw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Opacity Peak",
				  XtNfromVert, rw,
				  XtNfromHoriz, ow,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, 100,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section2_opacity_peak,
				  XkwNvaluePtr,
				  &threecolour_hg_section2_opacity_peak,
				  XkwNscaledFormat, "%.2f",
				  NULL);
    XkwSimpleSliderSetScale (pw, 0.01, 0.0, FALSE, TRUE);
    add_widget (pw, widgets, &num_widgets);
    divider = XtVaCreateManagedWidget ("divider", widgetClass, form,
				       XtNfromVert, iw,
				       XtNwidth, CANVAS_WIDTH,
				       XtNheight, DIVIDER_HEIGHT,
				       XtNbackground, label_foreground,
				       NULL);
    /*  Create section 3 controls  */
    label = XtVaCreateManagedWidget ("sectionLabel", labelWidgetClass, form,
				     XtNlabel, "Section 3 Controls:",
				     XtNfromVert, divider,
				     XtNborderWidth, 0,
				     NULL);
    /*  Create colour component sliders  */
    rw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Red Intensity",
				  XtNfromVert, label,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section3_red_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section3_red_intensity,
				  NULL);
    add_widget (rw, widgets, &num_widgets);
    gw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Green Intensity",
				  XtNfromVert, label,
				  XtNfromHoriz, rw,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section3_green_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section3_green_intensity,
				  NULL);
    add_widget (gw, widgets, &num_widgets);
    bw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Blue Intensity",
				  XtNfromVert, label,
				  XtNfromHoriz, gw,
				  XtNwidth, COLOUR_COMPONENT_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, INTENSITY_GRADATIONS,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section3_blue_intensity,
				  XkwNvaluePtr,
				  &threecolour_hg_section3_blue_intensity,
				  NULL);
    add_widget (bw, widgets, &num_widgets);
    /*  Create intensity and opacity alpha sliders  */
    iw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Intensity Alpha",
				  XtNfromVert, rw,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, -50,
				  XkwNmaximum, 50,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section3_intensity_alpha_tenths,
				  XkwNvaluePtr,
				  &threecolour_hg_section3_intensity_alpha_tenths,
				  XkwNscaledFormat, "%.1f",
				  NULL);
    XkwSimpleSliderSetScale (iw, 0.1, 0.0, FALSE, TRUE);
    add_widget (iw, widgets, &num_widgets);
    ow = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Opacity Alpha",
				  XtNfromVert, rw,
				  XtNfromHoriz, iw,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, -50,
				  XkwNmaximum, 50,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section3_opacity_alpha_tenths,
				  XkwNvaluePtr,
				  &threecolour_hg_section3_opacity_alpha_tenths,
				  XkwNscaledFormat, "%.1f",
				  NULL);
    XkwSimpleSliderSetScale (ow, 0.1, 0.0, FALSE, TRUE);
    add_widget (ow, widgets, &num_widgets);
    pw = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				  XtNlabel, "Opacity Peak",
				  XtNfromVert, rw,
				  XtNfromHoriz, ow,
				  XtNwidth, ALPHA_SLIDER_WIDTH,
				  XkwNminimum, 0,
				  XkwNmaximum, 100,
				  XkwNmodifier, 1,
				  XtNvalue,
				  threecolour_hg_section3_opacity_peak,
				  XkwNvaluePtr,
				  &threecolour_hg_section3_opacity_peak,
				  XkwNscaledFormat, "%.2f",
				  NULL);
    XkwSimpleSliderSetScale (pw, 0.01, 0.0, FALSE, TRUE);
    add_widget (pw, widgets, &num_widgets);
    divider = XtVaCreateManagedWidget ("divider", widgetClass, form,
				       XtNfromVert, iw,
				       XtNwidth, CANVAS_WIDTH,
				       XtNheight, DIVIDER_HEIGHT,
				       XtNbackground, label_foreground,
				       NULL);
    /*  Create the blank sliders  */
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Blank Start",
				 XtNfromVert, divider,
				 XtNwidth, CANVAS_WIDTH,
				 XkwNminimum, -128,
				 XkwNmaximum, 127,
				 XkwNmodifier, 1,
				 XtNvalue, threecolour_hg_blank_min,
				 XkwNvaluePtr,
				 &threecolour_hg_blank_min,
				 NULL);
    blank_start_sld = w;
    add_widget (w, widgets, &num_widgets);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Blank End",
				 XtNfromVert, w,
				 XtNwidth, CANVAS_WIDTH,
				 XkwNminimum, -128,
				 XkwNmaximum, 127,
				 XkwNmodifier, 1,
				 XtNvalue, threecolour_hg_blank_max,
				 XkwNvaluePtr,
				 &threecolour_hg_blank_max,
				 NULL);
    blank_end_sld = w;
    add_widget (w, widgets, &num_widgets);
    /*  Create the section-end sliders  */
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Section 1 End",
				 XtNfromVert, w,
				 XtNwidth, CANVAS_WIDTH,
				 XkwNminimum, -127,
				 XkwNmaximum, 127,
				 XkwNmodifier, 1,
				 XtNvalue, threecolour_hg_section1_end,
				 XkwNvaluePtr, &threecolour_hg_section1_end,
				 NULL);
    section1_end_sld = w;
    add_widget (w, widgets, &num_widgets);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Section 2 End",
				 XtNfromVert, w,
				 XtNwidth, CANVAS_WIDTH,
				 XkwNminimum, -127,
				 XkwNmaximum, 127,
				 XkwNmodifier, 1,
				 XtNvalue, threecolour_hg_section2_end,
				 XkwNvaluePtr, &threecolour_hg_section2_end,
				 NULL);
    section2_end_sld = w;
    add_widget (w, widgets, &num_widgets);
    /*  Finally create the display canvas  */
    hist_cnv = XtVaCreateManagedWidget ("histCanvas",
					canvasWidgetClass, form,
					XtNfromVert, w,
					XtNwidth, CANVAS_WIDTH,
					XtNheight, 200,
					XkwNsilenceUnconsumed, True,
					NULL);
    XtRealizeWidget (threecolour_hg_winpopup);
    XtVaGetValues (hist_cnv,
		   XkwNmonoPixCanvas, &pixcanvas,
		   NULL);
    XtAddCallback (apply_btn, XtNcallback, threecolour_hg_apply_cbk,pixcanvas);
    for (count = 0; count < num_widgets; ++count)
    {
	XtAddCallback (widgets[count], XkwNvalueChangeCallback,
		       slider_threecolour_hg_cbk, pixcanvas);
    }
    worldcanvas = create_fullcolour_hg_canvas (pixcanvas);
    c_register_callback (&new_byte_cube_event_list, new_cube_func,
			 NULL, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&histogram_change_event_list,
			 generic_histogram_change_callback, worldcanvas,
			 NULL, FALSE, NULL, FALSE, FALSE);
    return (threecolour_hg_winpopup);
}   /*  End Function create_threecolour_hg_win  */


/*  Private functions follow  */

static void slider_threecolour_hg_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  This is the slider continuous Hot Gas callback.
*/
{
    KPixCanvas canvas = client_data;

    compute_threecolour_hg_info ();
    kwin_resize (canvas, TRUE, 0, 0, 0, 0);
}   /*  End Function slider_threecolour_hg_cbk   */

static void threecolour_hg_apply_cbk (Widget w, XtPointer client_data,
				      XtPointer call_data)
/*  This is the continuous Hot Gas apply callback.
*/
{
    KPixCanvas canvas = client_data;

    compute_threecolour_hg_info ();
    kwin_resize (canvas, TRUE, 0, 0, 0, 0);
    compute ();
}   /*  End Function threecolour_hg_apply_cbk   */

static flag new_cube_func (void *object, void *client1_data, void *call_data,
			   void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    extern iarray byte_cube;
    extern double byte_cube_scale, byte_cube_offset;

    XtVaSetValues (blank_start_sld,
		   XkwNscaledUnit, iarray_value_name (byte_cube),
		   NULL);
    XkwSimpleSliderSetScale (blank_start_sld,
			     byte_cube_scale, byte_cube_offset, TRUE, TRUE);
    XtVaSetValues (blank_end_sld,
		   XkwNscaledUnit, iarray_value_name (byte_cube),
		   NULL);
    XkwSimpleSliderSetScale (blank_end_sld, byte_cube_scale, byte_cube_offset,
			     TRUE, TRUE);
    XtVaSetValues (section1_end_sld,
		   XkwNscaledUnit, iarray_value_name (byte_cube),
		   NULL);
    XkwSimpleSliderSetScale (section1_end_sld,
			     byte_cube_scale, byte_cube_offset, TRUE, TRUE);
    XtVaSetValues (section2_end_sld,
		   XkwNscaledUnit, iarray_value_name (byte_cube),
		   NULL);
    XkwSimpleSliderSetScale (section2_end_sld,
			     byte_cube_scale, byte_cube_offset, TRUE, TRUE);
    return (FALSE);
}   /*  End Function new_cube_func  */

static void add_widget (Widget widget, Widget widgets[MAX_WIDGETS],
			unsigned int *num_widgets)
/*  [SUMMARY] Add a widget to a list.
    [RETURNS] Nothing.
*/
{
    if (*num_widgets >= MAX_WIDGETS)
    {
	fprintf (stderr, "Too many widgets for threecolour HG\n");
	exit (RV_MEM_ERROR);
    }
    widgets[(*num_widgets)++] = widget;
}   /*  End Function add_widget  */
